module Test exposing (..)

import Browser
import Html exposing (..)
import Html.Attributes exposing (..)
import Html.Events exposing (onInput, onClick)
import Html.Events.Extra.Wheel as Wheel
import Phosphor exposing (IconWeight(..), toHtml, withSize, withSizeUnit, withClass)



-- MAIN


main =
    Browser.sandbox { init = init, update = update, view = view }



-- MODEL


type alias Model =
    { weight: IconWeight
    , size: Float
    }


init : Model
init =
    Model Thin 64



-- UPDATE


type Msg
    = SetWeight String
    | SetSize String
    | Reset
    | NoOp


update : Msg -> Model -> Model
update msg model =
    case msg of
        SetWeight weight ->
            { model | weight = parseWeight weight }

        SetSize size ->
            { model | size = clamp 4 2000 (size |> String.toFloat |> Maybe.withDefault 128) }

        Reset ->
            init

        _ ->
            model


parseWeight : String -> IconWeight
parseWeight weight =
    case weight of
        "thin" ->
            Thin

        "light" ->
            Light
    
        "bold" ->
            Bold

        "fill" ->
            Fill

        "duotone" ->
            Duotone

        _ ->
            Regular

css =
    """
    body {
        color: white;
        background-color: #35313D;
        font-family: monospace;
    }

    h2 {
        color: darkgrey;
    }

    .row {
        display: flex;
        justify-content: center;
        gap: 32px;
        margin: 32px;
    }
    """


-- VIEW


view : Model -> Html Msg
view model =
    let
        fontSize = String.fromFloat model.size ++ "px"
    in
        div
            [ style "font-size" fontSize ]
            [ Html.node "style" [] [ text css ]
            , div
                [ style "position" "fixed"
                , style "bottom" "0"
                , style "left" "0"
                , style "right" "0"
                , style "z-index" "1"
                , style "display" "flex"
                , style "gap" "16px"
                , style "justify-content" "center"
                , style "align-items" "center"
                , style "padding" "16px"
                ]
                [ select [ onInput SetWeight, style "height" "21px" ]
                    [ option [ value "thin" ] [ text "Thin "] 
                    , option [ value "light" ] [ text "Light "] 
                    , option [ value "regular" ] [ text "Regular "] 
                    , option [ value "bold" ] [ text "Bold "] 
                    , option [ value "fill" ] [ text "Fill "] 
                    , option [ value "duotone" ] [ text "Duotone "] 
                    ]
                , input
                    [ type_ "number"
                    , placeholder "Size"
              
                    , value <| String.fromFloat model.size
                    , onInput SetSize
                    ] []
                , button
                    [ onClick Reset ]
                    [ text "Reset" ]
                ]
           , div [ class "row" ]
                [ Phosphor.addressBook model.weight |> toHtml []
                , Phosphor.addressBook Thin |> toHtml []
                , Phosphor.addressBook Light |> toHtml []
                , Phosphor.addressBook Regular |> toHtml []
                , Phosphor.addressBook Bold |> toHtml []
                , Phosphor.addressBook Fill |> toHtml []
                , Phosphor.addressBook Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.airTrafficControl model.weight |> toHtml []
                , Phosphor.airTrafficControl Thin |> toHtml []
                , Phosphor.airTrafficControl Light |> toHtml []
                , Phosphor.airTrafficControl Regular |> toHtml []
                , Phosphor.airTrafficControl Bold |> toHtml []
                , Phosphor.airTrafficControl Fill |> toHtml []
                , Phosphor.airTrafficControl Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.airplane model.weight |> toHtml []
                , Phosphor.airplane Thin |> toHtml []
                , Phosphor.airplane Light |> toHtml []
                , Phosphor.airplane Regular |> toHtml []
                , Phosphor.airplane Bold |> toHtml []
                , Phosphor.airplane Fill |> toHtml []
                , Phosphor.airplane Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.airplaneInFlight model.weight |> toHtml []
                , Phosphor.airplaneInFlight Thin |> toHtml []
                , Phosphor.airplaneInFlight Light |> toHtml []
                , Phosphor.airplaneInFlight Regular |> toHtml []
                , Phosphor.airplaneInFlight Bold |> toHtml []
                , Phosphor.airplaneInFlight Fill |> toHtml []
                , Phosphor.airplaneInFlight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.airplaneLanding model.weight |> toHtml []
                , Phosphor.airplaneLanding Thin |> toHtml []
                , Phosphor.airplaneLanding Light |> toHtml []
                , Phosphor.airplaneLanding Regular |> toHtml []
                , Phosphor.airplaneLanding Bold |> toHtml []
                , Phosphor.airplaneLanding Fill |> toHtml []
                , Phosphor.airplaneLanding Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.airplaneTakeoff model.weight |> toHtml []
                , Phosphor.airplaneTakeoff Thin |> toHtml []
                , Phosphor.airplaneTakeoff Light |> toHtml []
                , Phosphor.airplaneTakeoff Regular |> toHtml []
                , Phosphor.airplaneTakeoff Bold |> toHtml []
                , Phosphor.airplaneTakeoff Fill |> toHtml []
                , Phosphor.airplaneTakeoff Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.airplaneTilt model.weight |> toHtml []
                , Phosphor.airplaneTilt Thin |> toHtml []
                , Phosphor.airplaneTilt Light |> toHtml []
                , Phosphor.airplaneTilt Regular |> toHtml []
                , Phosphor.airplaneTilt Bold |> toHtml []
                , Phosphor.airplaneTilt Fill |> toHtml []
                , Phosphor.airplaneTilt Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.airplay model.weight |> toHtml []
                , Phosphor.airplay Thin |> toHtml []
                , Phosphor.airplay Light |> toHtml []
                , Phosphor.airplay Regular |> toHtml []
                , Phosphor.airplay Bold |> toHtml []
                , Phosphor.airplay Fill |> toHtml []
                , Phosphor.airplay Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alarm model.weight |> toHtml []
                , Phosphor.alarm Thin |> toHtml []
                , Phosphor.alarm Light |> toHtml []
                , Phosphor.alarm Regular |> toHtml []
                , Phosphor.alarm Bold |> toHtml []
                , Phosphor.alarm Fill |> toHtml []
                , Phosphor.alarm Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alien model.weight |> toHtml []
                , Phosphor.alien Thin |> toHtml []
                , Phosphor.alien Light |> toHtml []
                , Phosphor.alien Regular |> toHtml []
                , Phosphor.alien Bold |> toHtml []
                , Phosphor.alien Fill |> toHtml []
                , Phosphor.alien Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignBottom model.weight |> toHtml []
                , Phosphor.alignBottom Thin |> toHtml []
                , Phosphor.alignBottom Light |> toHtml []
                , Phosphor.alignBottom Regular |> toHtml []
                , Phosphor.alignBottom Bold |> toHtml []
                , Phosphor.alignBottom Fill |> toHtml []
                , Phosphor.alignBottom Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignBottomSimple model.weight |> toHtml []
                , Phosphor.alignBottomSimple Thin |> toHtml []
                , Phosphor.alignBottomSimple Light |> toHtml []
                , Phosphor.alignBottomSimple Regular |> toHtml []
                , Phosphor.alignBottomSimple Bold |> toHtml []
                , Phosphor.alignBottomSimple Fill |> toHtml []
                , Phosphor.alignBottomSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignCenterHorizontal model.weight |> toHtml []
                , Phosphor.alignCenterHorizontal Thin |> toHtml []
                , Phosphor.alignCenterHorizontal Light |> toHtml []
                , Phosphor.alignCenterHorizontal Regular |> toHtml []
                , Phosphor.alignCenterHorizontal Bold |> toHtml []
                , Phosphor.alignCenterHorizontal Fill |> toHtml []
                , Phosphor.alignCenterHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignCenterHorizontalSimple model.weight |> toHtml []
                , Phosphor.alignCenterHorizontalSimple Thin |> toHtml []
                , Phosphor.alignCenterHorizontalSimple Light |> toHtml []
                , Phosphor.alignCenterHorizontalSimple Regular |> toHtml []
                , Phosphor.alignCenterHorizontalSimple Bold |> toHtml []
                , Phosphor.alignCenterHorizontalSimple Fill |> toHtml []
                , Phosphor.alignCenterHorizontalSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignCenterVertical model.weight |> toHtml []
                , Phosphor.alignCenterVertical Thin |> toHtml []
                , Phosphor.alignCenterVertical Light |> toHtml []
                , Phosphor.alignCenterVertical Regular |> toHtml []
                , Phosphor.alignCenterVertical Bold |> toHtml []
                , Phosphor.alignCenterVertical Fill |> toHtml []
                , Phosphor.alignCenterVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignCenterVerticalSimple model.weight |> toHtml []
                , Phosphor.alignCenterVerticalSimple Thin |> toHtml []
                , Phosphor.alignCenterVerticalSimple Light |> toHtml []
                , Phosphor.alignCenterVerticalSimple Regular |> toHtml []
                , Phosphor.alignCenterVerticalSimple Bold |> toHtml []
                , Phosphor.alignCenterVerticalSimple Fill |> toHtml []
                , Phosphor.alignCenterVerticalSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignLeft model.weight |> toHtml []
                , Phosphor.alignLeft Thin |> toHtml []
                , Phosphor.alignLeft Light |> toHtml []
                , Phosphor.alignLeft Regular |> toHtml []
                , Phosphor.alignLeft Bold |> toHtml []
                , Phosphor.alignLeft Fill |> toHtml []
                , Phosphor.alignLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignLeftSimple model.weight |> toHtml []
                , Phosphor.alignLeftSimple Thin |> toHtml []
                , Phosphor.alignLeftSimple Light |> toHtml []
                , Phosphor.alignLeftSimple Regular |> toHtml []
                , Phosphor.alignLeftSimple Bold |> toHtml []
                , Phosphor.alignLeftSimple Fill |> toHtml []
                , Phosphor.alignLeftSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignRight model.weight |> toHtml []
                , Phosphor.alignRight Thin |> toHtml []
                , Phosphor.alignRight Light |> toHtml []
                , Phosphor.alignRight Regular |> toHtml []
                , Phosphor.alignRight Bold |> toHtml []
                , Phosphor.alignRight Fill |> toHtml []
                , Phosphor.alignRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignRightSimple model.weight |> toHtml []
                , Phosphor.alignRightSimple Thin |> toHtml []
                , Phosphor.alignRightSimple Light |> toHtml []
                , Phosphor.alignRightSimple Regular |> toHtml []
                , Phosphor.alignRightSimple Bold |> toHtml []
                , Phosphor.alignRightSimple Fill |> toHtml []
                , Phosphor.alignRightSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignTop model.weight |> toHtml []
                , Phosphor.alignTop Thin |> toHtml []
                , Phosphor.alignTop Light |> toHtml []
                , Phosphor.alignTop Regular |> toHtml []
                , Phosphor.alignTop Bold |> toHtml []
                , Phosphor.alignTop Fill |> toHtml []
                , Phosphor.alignTop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.alignTopSimple model.weight |> toHtml []
                , Phosphor.alignTopSimple Thin |> toHtml []
                , Phosphor.alignTopSimple Light |> toHtml []
                , Phosphor.alignTopSimple Regular |> toHtml []
                , Phosphor.alignTopSimple Bold |> toHtml []
                , Phosphor.alignTopSimple Fill |> toHtml []
                , Phosphor.alignTopSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.amazonLogo model.weight |> toHtml []
                , Phosphor.amazonLogo Thin |> toHtml []
                , Phosphor.amazonLogo Light |> toHtml []
                , Phosphor.amazonLogo Regular |> toHtml []
                , Phosphor.amazonLogo Bold |> toHtml []
                , Phosphor.amazonLogo Fill |> toHtml []
                , Phosphor.amazonLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.anchor model.weight |> toHtml []
                , Phosphor.anchor Thin |> toHtml []
                , Phosphor.anchor Light |> toHtml []
                , Phosphor.anchor Regular |> toHtml []
                , Phosphor.anchor Bold |> toHtml []
                , Phosphor.anchor Fill |> toHtml []
                , Phosphor.anchor Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.anchorSimple model.weight |> toHtml []
                , Phosphor.anchorSimple Thin |> toHtml []
                , Phosphor.anchorSimple Light |> toHtml []
                , Phosphor.anchorSimple Regular |> toHtml []
                , Phosphor.anchorSimple Bold |> toHtml []
                , Phosphor.anchorSimple Fill |> toHtml []
                , Phosphor.anchorSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.androidLogo model.weight |> toHtml []
                , Phosphor.androidLogo Thin |> toHtml []
                , Phosphor.androidLogo Light |> toHtml []
                , Phosphor.androidLogo Regular |> toHtml []
                , Phosphor.androidLogo Bold |> toHtml []
                , Phosphor.androidLogo Fill |> toHtml []
                , Phosphor.androidLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.angularLogo model.weight |> toHtml []
                , Phosphor.angularLogo Thin |> toHtml []
                , Phosphor.angularLogo Light |> toHtml []
                , Phosphor.angularLogo Regular |> toHtml []
                , Phosphor.angularLogo Bold |> toHtml []
                , Phosphor.angularLogo Fill |> toHtml []
                , Phosphor.angularLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.aperture model.weight |> toHtml []
                , Phosphor.aperture Thin |> toHtml []
                , Phosphor.aperture Light |> toHtml []
                , Phosphor.aperture Regular |> toHtml []
                , Phosphor.aperture Bold |> toHtml []
                , Phosphor.aperture Fill |> toHtml []
                , Phosphor.aperture Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.appStoreLogo model.weight |> toHtml []
                , Phosphor.appStoreLogo Thin |> toHtml []
                , Phosphor.appStoreLogo Light |> toHtml []
                , Phosphor.appStoreLogo Regular |> toHtml []
                , Phosphor.appStoreLogo Bold |> toHtml []
                , Phosphor.appStoreLogo Fill |> toHtml []
                , Phosphor.appStoreLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.appWindow model.weight |> toHtml []
                , Phosphor.appWindow Thin |> toHtml []
                , Phosphor.appWindow Light |> toHtml []
                , Phosphor.appWindow Regular |> toHtml []
                , Phosphor.appWindow Bold |> toHtml []
                , Phosphor.appWindow Fill |> toHtml []
                , Phosphor.appWindow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.appleLogo model.weight |> toHtml []
                , Phosphor.appleLogo Thin |> toHtml []
                , Phosphor.appleLogo Light |> toHtml []
                , Phosphor.appleLogo Regular |> toHtml []
                , Phosphor.appleLogo Bold |> toHtml []
                , Phosphor.appleLogo Fill |> toHtml []
                , Phosphor.appleLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.applePodcastsLogo model.weight |> toHtml []
                , Phosphor.applePodcastsLogo Thin |> toHtml []
                , Phosphor.applePodcastsLogo Light |> toHtml []
                , Phosphor.applePodcastsLogo Regular |> toHtml []
                , Phosphor.applePodcastsLogo Bold |> toHtml []
                , Phosphor.applePodcastsLogo Fill |> toHtml []
                , Phosphor.applePodcastsLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.archive model.weight |> toHtml []
                , Phosphor.archive Thin |> toHtml []
                , Phosphor.archive Light |> toHtml []
                , Phosphor.archive Regular |> toHtml []
                , Phosphor.archive Bold |> toHtml []
                , Phosphor.archive Fill |> toHtml []
                , Phosphor.archive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.archiveBox model.weight |> toHtml []
                , Phosphor.archiveBox Thin |> toHtml []
                , Phosphor.archiveBox Light |> toHtml []
                , Phosphor.archiveBox Regular |> toHtml []
                , Phosphor.archiveBox Bold |> toHtml []
                , Phosphor.archiveBox Fill |> toHtml []
                , Phosphor.archiveBox Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.archiveTray model.weight |> toHtml []
                , Phosphor.archiveTray Thin |> toHtml []
                , Phosphor.archiveTray Light |> toHtml []
                , Phosphor.archiveTray Regular |> toHtml []
                , Phosphor.archiveTray Bold |> toHtml []
                , Phosphor.archiveTray Fill |> toHtml []
                , Phosphor.archiveTray Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.armchair model.weight |> toHtml []
                , Phosphor.armchair Thin |> toHtml []
                , Phosphor.armchair Light |> toHtml []
                , Phosphor.armchair Regular |> toHtml []
                , Phosphor.armchair Bold |> toHtml []
                , Phosphor.armchair Fill |> toHtml []
                , Phosphor.armchair Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowArcLeft model.weight |> toHtml []
                , Phosphor.arrowArcLeft Thin |> toHtml []
                , Phosphor.arrowArcLeft Light |> toHtml []
                , Phosphor.arrowArcLeft Regular |> toHtml []
                , Phosphor.arrowArcLeft Bold |> toHtml []
                , Phosphor.arrowArcLeft Fill |> toHtml []
                , Phosphor.arrowArcLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowArcRight model.weight |> toHtml []
                , Phosphor.arrowArcRight Thin |> toHtml []
                , Phosphor.arrowArcRight Light |> toHtml []
                , Phosphor.arrowArcRight Regular |> toHtml []
                , Phosphor.arrowArcRight Bold |> toHtml []
                , Phosphor.arrowArcRight Fill |> toHtml []
                , Phosphor.arrowArcRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendDoubleUpLeft model.weight |> toHtml []
                , Phosphor.arrowBendDoubleUpLeft Thin |> toHtml []
                , Phosphor.arrowBendDoubleUpLeft Light |> toHtml []
                , Phosphor.arrowBendDoubleUpLeft Regular |> toHtml []
                , Phosphor.arrowBendDoubleUpLeft Bold |> toHtml []
                , Phosphor.arrowBendDoubleUpLeft Fill |> toHtml []
                , Phosphor.arrowBendDoubleUpLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendDoubleUpRight model.weight |> toHtml []
                , Phosphor.arrowBendDoubleUpRight Thin |> toHtml []
                , Phosphor.arrowBendDoubleUpRight Light |> toHtml []
                , Phosphor.arrowBendDoubleUpRight Regular |> toHtml []
                , Phosphor.arrowBendDoubleUpRight Bold |> toHtml []
                , Phosphor.arrowBendDoubleUpRight Fill |> toHtml []
                , Phosphor.arrowBendDoubleUpRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendDownLeft model.weight |> toHtml []
                , Phosphor.arrowBendDownLeft Thin |> toHtml []
                , Phosphor.arrowBendDownLeft Light |> toHtml []
                , Phosphor.arrowBendDownLeft Regular |> toHtml []
                , Phosphor.arrowBendDownLeft Bold |> toHtml []
                , Phosphor.arrowBendDownLeft Fill |> toHtml []
                , Phosphor.arrowBendDownLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendDownRight model.weight |> toHtml []
                , Phosphor.arrowBendDownRight Thin |> toHtml []
                , Phosphor.arrowBendDownRight Light |> toHtml []
                , Phosphor.arrowBendDownRight Regular |> toHtml []
                , Phosphor.arrowBendDownRight Bold |> toHtml []
                , Phosphor.arrowBendDownRight Fill |> toHtml []
                , Phosphor.arrowBendDownRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendLeftDown model.weight |> toHtml []
                , Phosphor.arrowBendLeftDown Thin |> toHtml []
                , Phosphor.arrowBendLeftDown Light |> toHtml []
                , Phosphor.arrowBendLeftDown Regular |> toHtml []
                , Phosphor.arrowBendLeftDown Bold |> toHtml []
                , Phosphor.arrowBendLeftDown Fill |> toHtml []
                , Phosphor.arrowBendLeftDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendLeftUp model.weight |> toHtml []
                , Phosphor.arrowBendLeftUp Thin |> toHtml []
                , Phosphor.arrowBendLeftUp Light |> toHtml []
                , Phosphor.arrowBendLeftUp Regular |> toHtml []
                , Phosphor.arrowBendLeftUp Bold |> toHtml []
                , Phosphor.arrowBendLeftUp Fill |> toHtml []
                , Phosphor.arrowBendLeftUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendRightDown model.weight |> toHtml []
                , Phosphor.arrowBendRightDown Thin |> toHtml []
                , Phosphor.arrowBendRightDown Light |> toHtml []
                , Phosphor.arrowBendRightDown Regular |> toHtml []
                , Phosphor.arrowBendRightDown Bold |> toHtml []
                , Phosphor.arrowBendRightDown Fill |> toHtml []
                , Phosphor.arrowBendRightDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendRightUp model.weight |> toHtml []
                , Phosphor.arrowBendRightUp Thin |> toHtml []
                , Phosphor.arrowBendRightUp Light |> toHtml []
                , Phosphor.arrowBendRightUp Regular |> toHtml []
                , Phosphor.arrowBendRightUp Bold |> toHtml []
                , Phosphor.arrowBendRightUp Fill |> toHtml []
                , Phosphor.arrowBendRightUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendUpLeft model.weight |> toHtml []
                , Phosphor.arrowBendUpLeft Thin |> toHtml []
                , Phosphor.arrowBendUpLeft Light |> toHtml []
                , Phosphor.arrowBendUpLeft Regular |> toHtml []
                , Phosphor.arrowBendUpLeft Bold |> toHtml []
                , Phosphor.arrowBendUpLeft Fill |> toHtml []
                , Phosphor.arrowBendUpLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowBendUpRight model.weight |> toHtml []
                , Phosphor.arrowBendUpRight Thin |> toHtml []
                , Phosphor.arrowBendUpRight Light |> toHtml []
                , Phosphor.arrowBendUpRight Regular |> toHtml []
                , Phosphor.arrowBendUpRight Bold |> toHtml []
                , Phosphor.arrowBendUpRight Fill |> toHtml []
                , Phosphor.arrowBendUpRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCircleDown model.weight |> toHtml []
                , Phosphor.arrowCircleDown Thin |> toHtml []
                , Phosphor.arrowCircleDown Light |> toHtml []
                , Phosphor.arrowCircleDown Regular |> toHtml []
                , Phosphor.arrowCircleDown Bold |> toHtml []
                , Phosphor.arrowCircleDown Fill |> toHtml []
                , Phosphor.arrowCircleDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCircleDownLeft model.weight |> toHtml []
                , Phosphor.arrowCircleDownLeft Thin |> toHtml []
                , Phosphor.arrowCircleDownLeft Light |> toHtml []
                , Phosphor.arrowCircleDownLeft Regular |> toHtml []
                , Phosphor.arrowCircleDownLeft Bold |> toHtml []
                , Phosphor.arrowCircleDownLeft Fill |> toHtml []
                , Phosphor.arrowCircleDownLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCircleDownRight model.weight |> toHtml []
                , Phosphor.arrowCircleDownRight Thin |> toHtml []
                , Phosphor.arrowCircleDownRight Light |> toHtml []
                , Phosphor.arrowCircleDownRight Regular |> toHtml []
                , Phosphor.arrowCircleDownRight Bold |> toHtml []
                , Phosphor.arrowCircleDownRight Fill |> toHtml []
                , Phosphor.arrowCircleDownRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCircleLeft model.weight |> toHtml []
                , Phosphor.arrowCircleLeft Thin |> toHtml []
                , Phosphor.arrowCircleLeft Light |> toHtml []
                , Phosphor.arrowCircleLeft Regular |> toHtml []
                , Phosphor.arrowCircleLeft Bold |> toHtml []
                , Phosphor.arrowCircleLeft Fill |> toHtml []
                , Phosphor.arrowCircleLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCircleRight model.weight |> toHtml []
                , Phosphor.arrowCircleRight Thin |> toHtml []
                , Phosphor.arrowCircleRight Light |> toHtml []
                , Phosphor.arrowCircleRight Regular |> toHtml []
                , Phosphor.arrowCircleRight Bold |> toHtml []
                , Phosphor.arrowCircleRight Fill |> toHtml []
                , Phosphor.arrowCircleRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCircleUp model.weight |> toHtml []
                , Phosphor.arrowCircleUp Thin |> toHtml []
                , Phosphor.arrowCircleUp Light |> toHtml []
                , Phosphor.arrowCircleUp Regular |> toHtml []
                , Phosphor.arrowCircleUp Bold |> toHtml []
                , Phosphor.arrowCircleUp Fill |> toHtml []
                , Phosphor.arrowCircleUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCircleUpLeft model.weight |> toHtml []
                , Phosphor.arrowCircleUpLeft Thin |> toHtml []
                , Phosphor.arrowCircleUpLeft Light |> toHtml []
                , Phosphor.arrowCircleUpLeft Regular |> toHtml []
                , Phosphor.arrowCircleUpLeft Bold |> toHtml []
                , Phosphor.arrowCircleUpLeft Fill |> toHtml []
                , Phosphor.arrowCircleUpLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCircleUpRight model.weight |> toHtml []
                , Phosphor.arrowCircleUpRight Thin |> toHtml []
                , Phosphor.arrowCircleUpRight Light |> toHtml []
                , Phosphor.arrowCircleUpRight Regular |> toHtml []
                , Phosphor.arrowCircleUpRight Bold |> toHtml []
                , Phosphor.arrowCircleUpRight Fill |> toHtml []
                , Phosphor.arrowCircleUpRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowClockwise model.weight |> toHtml []
                , Phosphor.arrowClockwise Thin |> toHtml []
                , Phosphor.arrowClockwise Light |> toHtml []
                , Phosphor.arrowClockwise Regular |> toHtml []
                , Phosphor.arrowClockwise Bold |> toHtml []
                , Phosphor.arrowClockwise Fill |> toHtml []
                , Phosphor.arrowClockwise Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowCounterClockwise model.weight |> toHtml []
                , Phosphor.arrowCounterClockwise Thin |> toHtml []
                , Phosphor.arrowCounterClockwise Light |> toHtml []
                , Phosphor.arrowCounterClockwise Regular |> toHtml []
                , Phosphor.arrowCounterClockwise Bold |> toHtml []
                , Phosphor.arrowCounterClockwise Fill |> toHtml []
                , Phosphor.arrowCounterClockwise Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowDown model.weight |> toHtml []
                , Phosphor.arrowDown Thin |> toHtml []
                , Phosphor.arrowDown Light |> toHtml []
                , Phosphor.arrowDown Regular |> toHtml []
                , Phosphor.arrowDown Bold |> toHtml []
                , Phosphor.arrowDown Fill |> toHtml []
                , Phosphor.arrowDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowDownLeft model.weight |> toHtml []
                , Phosphor.arrowDownLeft Thin |> toHtml []
                , Phosphor.arrowDownLeft Light |> toHtml []
                , Phosphor.arrowDownLeft Regular |> toHtml []
                , Phosphor.arrowDownLeft Bold |> toHtml []
                , Phosphor.arrowDownLeft Fill |> toHtml []
                , Phosphor.arrowDownLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowDownRight model.weight |> toHtml []
                , Phosphor.arrowDownRight Thin |> toHtml []
                , Phosphor.arrowDownRight Light |> toHtml []
                , Phosphor.arrowDownRight Regular |> toHtml []
                , Phosphor.arrowDownRight Bold |> toHtml []
                , Phosphor.arrowDownRight Fill |> toHtml []
                , Phosphor.arrowDownRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowDownLeft model.weight |> toHtml []
                , Phosphor.arrowElbowDownLeft Thin |> toHtml []
                , Phosphor.arrowElbowDownLeft Light |> toHtml []
                , Phosphor.arrowElbowDownLeft Regular |> toHtml []
                , Phosphor.arrowElbowDownLeft Bold |> toHtml []
                , Phosphor.arrowElbowDownLeft Fill |> toHtml []
                , Phosphor.arrowElbowDownLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowDownRight model.weight |> toHtml []
                , Phosphor.arrowElbowDownRight Thin |> toHtml []
                , Phosphor.arrowElbowDownRight Light |> toHtml []
                , Phosphor.arrowElbowDownRight Regular |> toHtml []
                , Phosphor.arrowElbowDownRight Bold |> toHtml []
                , Phosphor.arrowElbowDownRight Fill |> toHtml []
                , Phosphor.arrowElbowDownRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowLeft model.weight |> toHtml []
                , Phosphor.arrowElbowLeft Thin |> toHtml []
                , Phosphor.arrowElbowLeft Light |> toHtml []
                , Phosphor.arrowElbowLeft Regular |> toHtml []
                , Phosphor.arrowElbowLeft Bold |> toHtml []
                , Phosphor.arrowElbowLeft Fill |> toHtml []
                , Phosphor.arrowElbowLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowLeftDown model.weight |> toHtml []
                , Phosphor.arrowElbowLeftDown Thin |> toHtml []
                , Phosphor.arrowElbowLeftDown Light |> toHtml []
                , Phosphor.arrowElbowLeftDown Regular |> toHtml []
                , Phosphor.arrowElbowLeftDown Bold |> toHtml []
                , Phosphor.arrowElbowLeftDown Fill |> toHtml []
                , Phosphor.arrowElbowLeftDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowLeftUp model.weight |> toHtml []
                , Phosphor.arrowElbowLeftUp Thin |> toHtml []
                , Phosphor.arrowElbowLeftUp Light |> toHtml []
                , Phosphor.arrowElbowLeftUp Regular |> toHtml []
                , Phosphor.arrowElbowLeftUp Bold |> toHtml []
                , Phosphor.arrowElbowLeftUp Fill |> toHtml []
                , Phosphor.arrowElbowLeftUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowRight model.weight |> toHtml []
                , Phosphor.arrowElbowRight Thin |> toHtml []
                , Phosphor.arrowElbowRight Light |> toHtml []
                , Phosphor.arrowElbowRight Regular |> toHtml []
                , Phosphor.arrowElbowRight Bold |> toHtml []
                , Phosphor.arrowElbowRight Fill |> toHtml []
                , Phosphor.arrowElbowRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowRightDown model.weight |> toHtml []
                , Phosphor.arrowElbowRightDown Thin |> toHtml []
                , Phosphor.arrowElbowRightDown Light |> toHtml []
                , Phosphor.arrowElbowRightDown Regular |> toHtml []
                , Phosphor.arrowElbowRightDown Bold |> toHtml []
                , Phosphor.arrowElbowRightDown Fill |> toHtml []
                , Phosphor.arrowElbowRightDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowRightUp model.weight |> toHtml []
                , Phosphor.arrowElbowRightUp Thin |> toHtml []
                , Phosphor.arrowElbowRightUp Light |> toHtml []
                , Phosphor.arrowElbowRightUp Regular |> toHtml []
                , Phosphor.arrowElbowRightUp Bold |> toHtml []
                , Phosphor.arrowElbowRightUp Fill |> toHtml []
                , Phosphor.arrowElbowRightUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowUpLeft model.weight |> toHtml []
                , Phosphor.arrowElbowUpLeft Thin |> toHtml []
                , Phosphor.arrowElbowUpLeft Light |> toHtml []
                , Phosphor.arrowElbowUpLeft Regular |> toHtml []
                , Phosphor.arrowElbowUpLeft Bold |> toHtml []
                , Phosphor.arrowElbowUpLeft Fill |> toHtml []
                , Phosphor.arrowElbowUpLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowElbowUpRight model.weight |> toHtml []
                , Phosphor.arrowElbowUpRight Thin |> toHtml []
                , Phosphor.arrowElbowUpRight Light |> toHtml []
                , Phosphor.arrowElbowUpRight Regular |> toHtml []
                , Phosphor.arrowElbowUpRight Bold |> toHtml []
                , Phosphor.arrowElbowUpRight Fill |> toHtml []
                , Phosphor.arrowElbowUpRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatDown model.weight |> toHtml []
                , Phosphor.arrowFatDown Thin |> toHtml []
                , Phosphor.arrowFatDown Light |> toHtml []
                , Phosphor.arrowFatDown Regular |> toHtml []
                , Phosphor.arrowFatDown Bold |> toHtml []
                , Phosphor.arrowFatDown Fill |> toHtml []
                , Phosphor.arrowFatDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLeft model.weight |> toHtml []
                , Phosphor.arrowFatLeft Thin |> toHtml []
                , Phosphor.arrowFatLeft Light |> toHtml []
                , Phosphor.arrowFatLeft Regular |> toHtml []
                , Phosphor.arrowFatLeft Bold |> toHtml []
                , Phosphor.arrowFatLeft Fill |> toHtml []
                , Phosphor.arrowFatLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLineDown model.weight |> toHtml []
                , Phosphor.arrowFatLineDown Thin |> toHtml []
                , Phosphor.arrowFatLineDown Light |> toHtml []
                , Phosphor.arrowFatLineDown Regular |> toHtml []
                , Phosphor.arrowFatLineDown Bold |> toHtml []
                , Phosphor.arrowFatLineDown Fill |> toHtml []
                , Phosphor.arrowFatLineDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLineLeft model.weight |> toHtml []
                , Phosphor.arrowFatLineLeft Thin |> toHtml []
                , Phosphor.arrowFatLineLeft Light |> toHtml []
                , Phosphor.arrowFatLineLeft Regular |> toHtml []
                , Phosphor.arrowFatLineLeft Bold |> toHtml []
                , Phosphor.arrowFatLineLeft Fill |> toHtml []
                , Phosphor.arrowFatLineLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLineRight model.weight |> toHtml []
                , Phosphor.arrowFatLineRight Thin |> toHtml []
                , Phosphor.arrowFatLineRight Light |> toHtml []
                , Phosphor.arrowFatLineRight Regular |> toHtml []
                , Phosphor.arrowFatLineRight Bold |> toHtml []
                , Phosphor.arrowFatLineRight Fill |> toHtml []
                , Phosphor.arrowFatLineRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLineUp model.weight |> toHtml []
                , Phosphor.arrowFatLineUp Thin |> toHtml []
                , Phosphor.arrowFatLineUp Light |> toHtml []
                , Phosphor.arrowFatLineUp Regular |> toHtml []
                , Phosphor.arrowFatLineUp Bold |> toHtml []
                , Phosphor.arrowFatLineUp Fill |> toHtml []
                , Phosphor.arrowFatLineUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLinesDown model.weight |> toHtml []
                , Phosphor.arrowFatLinesDown Thin |> toHtml []
                , Phosphor.arrowFatLinesDown Light |> toHtml []
                , Phosphor.arrowFatLinesDown Regular |> toHtml []
                , Phosphor.arrowFatLinesDown Bold |> toHtml []
                , Phosphor.arrowFatLinesDown Fill |> toHtml []
                , Phosphor.arrowFatLinesDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLinesLeft model.weight |> toHtml []
                , Phosphor.arrowFatLinesLeft Thin |> toHtml []
                , Phosphor.arrowFatLinesLeft Light |> toHtml []
                , Phosphor.arrowFatLinesLeft Regular |> toHtml []
                , Phosphor.arrowFatLinesLeft Bold |> toHtml []
                , Phosphor.arrowFatLinesLeft Fill |> toHtml []
                , Phosphor.arrowFatLinesLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLinesRight model.weight |> toHtml []
                , Phosphor.arrowFatLinesRight Thin |> toHtml []
                , Phosphor.arrowFatLinesRight Light |> toHtml []
                , Phosphor.arrowFatLinesRight Regular |> toHtml []
                , Phosphor.arrowFatLinesRight Bold |> toHtml []
                , Phosphor.arrowFatLinesRight Fill |> toHtml []
                , Phosphor.arrowFatLinesRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatLinesUp model.weight |> toHtml []
                , Phosphor.arrowFatLinesUp Thin |> toHtml []
                , Phosphor.arrowFatLinesUp Light |> toHtml []
                , Phosphor.arrowFatLinesUp Regular |> toHtml []
                , Phosphor.arrowFatLinesUp Bold |> toHtml []
                , Phosphor.arrowFatLinesUp Fill |> toHtml []
                , Phosphor.arrowFatLinesUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatRight model.weight |> toHtml []
                , Phosphor.arrowFatRight Thin |> toHtml []
                , Phosphor.arrowFatRight Light |> toHtml []
                , Phosphor.arrowFatRight Regular |> toHtml []
                , Phosphor.arrowFatRight Bold |> toHtml []
                , Phosphor.arrowFatRight Fill |> toHtml []
                , Phosphor.arrowFatRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowFatUp model.weight |> toHtml []
                , Phosphor.arrowFatUp Thin |> toHtml []
                , Phosphor.arrowFatUp Light |> toHtml []
                , Phosphor.arrowFatUp Regular |> toHtml []
                , Phosphor.arrowFatUp Bold |> toHtml []
                , Phosphor.arrowFatUp Fill |> toHtml []
                , Phosphor.arrowFatUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLeft model.weight |> toHtml []
                , Phosphor.arrowLeft Thin |> toHtml []
                , Phosphor.arrowLeft Light |> toHtml []
                , Phosphor.arrowLeft Regular |> toHtml []
                , Phosphor.arrowLeft Bold |> toHtml []
                , Phosphor.arrowLeft Fill |> toHtml []
                , Phosphor.arrowLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLineDown model.weight |> toHtml []
                , Phosphor.arrowLineDown Thin |> toHtml []
                , Phosphor.arrowLineDown Light |> toHtml []
                , Phosphor.arrowLineDown Regular |> toHtml []
                , Phosphor.arrowLineDown Bold |> toHtml []
                , Phosphor.arrowLineDown Fill |> toHtml []
                , Phosphor.arrowLineDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLineDownLeft model.weight |> toHtml []
                , Phosphor.arrowLineDownLeft Thin |> toHtml []
                , Phosphor.arrowLineDownLeft Light |> toHtml []
                , Phosphor.arrowLineDownLeft Regular |> toHtml []
                , Phosphor.arrowLineDownLeft Bold |> toHtml []
                , Phosphor.arrowLineDownLeft Fill |> toHtml []
                , Phosphor.arrowLineDownLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLineDownRight model.weight |> toHtml []
                , Phosphor.arrowLineDownRight Thin |> toHtml []
                , Phosphor.arrowLineDownRight Light |> toHtml []
                , Phosphor.arrowLineDownRight Regular |> toHtml []
                , Phosphor.arrowLineDownRight Bold |> toHtml []
                , Phosphor.arrowLineDownRight Fill |> toHtml []
                , Phosphor.arrowLineDownRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLineLeft model.weight |> toHtml []
                , Phosphor.arrowLineLeft Thin |> toHtml []
                , Phosphor.arrowLineLeft Light |> toHtml []
                , Phosphor.arrowLineLeft Regular |> toHtml []
                , Phosphor.arrowLineLeft Bold |> toHtml []
                , Phosphor.arrowLineLeft Fill |> toHtml []
                , Phosphor.arrowLineLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLineRight model.weight |> toHtml []
                , Phosphor.arrowLineRight Thin |> toHtml []
                , Phosphor.arrowLineRight Light |> toHtml []
                , Phosphor.arrowLineRight Regular |> toHtml []
                , Phosphor.arrowLineRight Bold |> toHtml []
                , Phosphor.arrowLineRight Fill |> toHtml []
                , Phosphor.arrowLineRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLineUp model.weight |> toHtml []
                , Phosphor.arrowLineUp Thin |> toHtml []
                , Phosphor.arrowLineUp Light |> toHtml []
                , Phosphor.arrowLineUp Regular |> toHtml []
                , Phosphor.arrowLineUp Bold |> toHtml []
                , Phosphor.arrowLineUp Fill |> toHtml []
                , Phosphor.arrowLineUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLineUpLeft model.weight |> toHtml []
                , Phosphor.arrowLineUpLeft Thin |> toHtml []
                , Phosphor.arrowLineUpLeft Light |> toHtml []
                , Phosphor.arrowLineUpLeft Regular |> toHtml []
                , Phosphor.arrowLineUpLeft Bold |> toHtml []
                , Phosphor.arrowLineUpLeft Fill |> toHtml []
                , Phosphor.arrowLineUpLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowLineUpRight model.weight |> toHtml []
                , Phosphor.arrowLineUpRight Thin |> toHtml []
                , Phosphor.arrowLineUpRight Light |> toHtml []
                , Phosphor.arrowLineUpRight Regular |> toHtml []
                , Phosphor.arrowLineUpRight Bold |> toHtml []
                , Phosphor.arrowLineUpRight Fill |> toHtml []
                , Phosphor.arrowLineUpRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowRight model.weight |> toHtml []
                , Phosphor.arrowRight Thin |> toHtml []
                , Phosphor.arrowRight Light |> toHtml []
                , Phosphor.arrowRight Regular |> toHtml []
                , Phosphor.arrowRight Bold |> toHtml []
                , Phosphor.arrowRight Fill |> toHtml []
                , Phosphor.arrowRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareDown model.weight |> toHtml []
                , Phosphor.arrowSquareDown Thin |> toHtml []
                , Phosphor.arrowSquareDown Light |> toHtml []
                , Phosphor.arrowSquareDown Regular |> toHtml []
                , Phosphor.arrowSquareDown Bold |> toHtml []
                , Phosphor.arrowSquareDown Fill |> toHtml []
                , Phosphor.arrowSquareDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareDownLeft model.weight |> toHtml []
                , Phosphor.arrowSquareDownLeft Thin |> toHtml []
                , Phosphor.arrowSquareDownLeft Light |> toHtml []
                , Phosphor.arrowSquareDownLeft Regular |> toHtml []
                , Phosphor.arrowSquareDownLeft Bold |> toHtml []
                , Phosphor.arrowSquareDownLeft Fill |> toHtml []
                , Phosphor.arrowSquareDownLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareDownRight model.weight |> toHtml []
                , Phosphor.arrowSquareDownRight Thin |> toHtml []
                , Phosphor.arrowSquareDownRight Light |> toHtml []
                , Phosphor.arrowSquareDownRight Regular |> toHtml []
                , Phosphor.arrowSquareDownRight Bold |> toHtml []
                , Phosphor.arrowSquareDownRight Fill |> toHtml []
                , Phosphor.arrowSquareDownRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareIn model.weight |> toHtml []
                , Phosphor.arrowSquareIn Thin |> toHtml []
                , Phosphor.arrowSquareIn Light |> toHtml []
                , Phosphor.arrowSquareIn Regular |> toHtml []
                , Phosphor.arrowSquareIn Bold |> toHtml []
                , Phosphor.arrowSquareIn Fill |> toHtml []
                , Phosphor.arrowSquareIn Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareLeft model.weight |> toHtml []
                , Phosphor.arrowSquareLeft Thin |> toHtml []
                , Phosphor.arrowSquareLeft Light |> toHtml []
                , Phosphor.arrowSquareLeft Regular |> toHtml []
                , Phosphor.arrowSquareLeft Bold |> toHtml []
                , Phosphor.arrowSquareLeft Fill |> toHtml []
                , Phosphor.arrowSquareLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareOut model.weight |> toHtml []
                , Phosphor.arrowSquareOut Thin |> toHtml []
                , Phosphor.arrowSquareOut Light |> toHtml []
                , Phosphor.arrowSquareOut Regular |> toHtml []
                , Phosphor.arrowSquareOut Bold |> toHtml []
                , Phosphor.arrowSquareOut Fill |> toHtml []
                , Phosphor.arrowSquareOut Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareRight model.weight |> toHtml []
                , Phosphor.arrowSquareRight Thin |> toHtml []
                , Phosphor.arrowSquareRight Light |> toHtml []
                , Phosphor.arrowSquareRight Regular |> toHtml []
                , Phosphor.arrowSquareRight Bold |> toHtml []
                , Phosphor.arrowSquareRight Fill |> toHtml []
                , Phosphor.arrowSquareRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareUp model.weight |> toHtml []
                , Phosphor.arrowSquareUp Thin |> toHtml []
                , Phosphor.arrowSquareUp Light |> toHtml []
                , Phosphor.arrowSquareUp Regular |> toHtml []
                , Phosphor.arrowSquareUp Bold |> toHtml []
                , Phosphor.arrowSquareUp Fill |> toHtml []
                , Phosphor.arrowSquareUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareUpLeft model.weight |> toHtml []
                , Phosphor.arrowSquareUpLeft Thin |> toHtml []
                , Phosphor.arrowSquareUpLeft Light |> toHtml []
                , Phosphor.arrowSquareUpLeft Regular |> toHtml []
                , Phosphor.arrowSquareUpLeft Bold |> toHtml []
                , Phosphor.arrowSquareUpLeft Fill |> toHtml []
                , Phosphor.arrowSquareUpLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowSquareUpRight model.weight |> toHtml []
                , Phosphor.arrowSquareUpRight Thin |> toHtml []
                , Phosphor.arrowSquareUpRight Light |> toHtml []
                , Phosphor.arrowSquareUpRight Regular |> toHtml []
                , Phosphor.arrowSquareUpRight Bold |> toHtml []
                , Phosphor.arrowSquareUpRight Fill |> toHtml []
                , Phosphor.arrowSquareUpRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowUDownLeft model.weight |> toHtml []
                , Phosphor.arrowUDownLeft Thin |> toHtml []
                , Phosphor.arrowUDownLeft Light |> toHtml []
                , Phosphor.arrowUDownLeft Regular |> toHtml []
                , Phosphor.arrowUDownLeft Bold |> toHtml []
                , Phosphor.arrowUDownLeft Fill |> toHtml []
                , Phosphor.arrowUDownLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowUDownRight model.weight |> toHtml []
                , Phosphor.arrowUDownRight Thin |> toHtml []
                , Phosphor.arrowUDownRight Light |> toHtml []
                , Phosphor.arrowUDownRight Regular |> toHtml []
                , Phosphor.arrowUDownRight Bold |> toHtml []
                , Phosphor.arrowUDownRight Fill |> toHtml []
                , Phosphor.arrowUDownRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowULeftDown model.weight |> toHtml []
                , Phosphor.arrowULeftDown Thin |> toHtml []
                , Phosphor.arrowULeftDown Light |> toHtml []
                , Phosphor.arrowULeftDown Regular |> toHtml []
                , Phosphor.arrowULeftDown Bold |> toHtml []
                , Phosphor.arrowULeftDown Fill |> toHtml []
                , Phosphor.arrowULeftDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowULeftUp model.weight |> toHtml []
                , Phosphor.arrowULeftUp Thin |> toHtml []
                , Phosphor.arrowULeftUp Light |> toHtml []
                , Phosphor.arrowULeftUp Regular |> toHtml []
                , Phosphor.arrowULeftUp Bold |> toHtml []
                , Phosphor.arrowULeftUp Fill |> toHtml []
                , Phosphor.arrowULeftUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowURightDown model.weight |> toHtml []
                , Phosphor.arrowURightDown Thin |> toHtml []
                , Phosphor.arrowURightDown Light |> toHtml []
                , Phosphor.arrowURightDown Regular |> toHtml []
                , Phosphor.arrowURightDown Bold |> toHtml []
                , Phosphor.arrowURightDown Fill |> toHtml []
                , Phosphor.arrowURightDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowURightUp model.weight |> toHtml []
                , Phosphor.arrowURightUp Thin |> toHtml []
                , Phosphor.arrowURightUp Light |> toHtml []
                , Phosphor.arrowURightUp Regular |> toHtml []
                , Phosphor.arrowURightUp Bold |> toHtml []
                , Phosphor.arrowURightUp Fill |> toHtml []
                , Phosphor.arrowURightUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowUUpLeft model.weight |> toHtml []
                , Phosphor.arrowUUpLeft Thin |> toHtml []
                , Phosphor.arrowUUpLeft Light |> toHtml []
                , Phosphor.arrowUUpLeft Regular |> toHtml []
                , Phosphor.arrowUUpLeft Bold |> toHtml []
                , Phosphor.arrowUUpLeft Fill |> toHtml []
                , Phosphor.arrowUUpLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowUUpRight model.weight |> toHtml []
                , Phosphor.arrowUUpRight Thin |> toHtml []
                , Phosphor.arrowUUpRight Light |> toHtml []
                , Phosphor.arrowUUpRight Regular |> toHtml []
                , Phosphor.arrowUUpRight Bold |> toHtml []
                , Phosphor.arrowUUpRight Fill |> toHtml []
                , Phosphor.arrowUUpRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowUp model.weight |> toHtml []
                , Phosphor.arrowUp Thin |> toHtml []
                , Phosphor.arrowUp Light |> toHtml []
                , Phosphor.arrowUp Regular |> toHtml []
                , Phosphor.arrowUp Bold |> toHtml []
                , Phosphor.arrowUp Fill |> toHtml []
                , Phosphor.arrowUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowUpLeft model.weight |> toHtml []
                , Phosphor.arrowUpLeft Thin |> toHtml []
                , Phosphor.arrowUpLeft Light |> toHtml []
                , Phosphor.arrowUpLeft Regular |> toHtml []
                , Phosphor.arrowUpLeft Bold |> toHtml []
                , Phosphor.arrowUpLeft Fill |> toHtml []
                , Phosphor.arrowUpLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowUpRight model.weight |> toHtml []
                , Phosphor.arrowUpRight Thin |> toHtml []
                , Phosphor.arrowUpRight Light |> toHtml []
                , Phosphor.arrowUpRight Regular |> toHtml []
                , Phosphor.arrowUpRight Bold |> toHtml []
                , Phosphor.arrowUpRight Fill |> toHtml []
                , Phosphor.arrowUpRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsClockwise model.weight |> toHtml []
                , Phosphor.arrowsClockwise Thin |> toHtml []
                , Phosphor.arrowsClockwise Light |> toHtml []
                , Phosphor.arrowsClockwise Regular |> toHtml []
                , Phosphor.arrowsClockwise Bold |> toHtml []
                , Phosphor.arrowsClockwise Fill |> toHtml []
                , Phosphor.arrowsClockwise Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsCounterClockwise model.weight |> toHtml []
                , Phosphor.arrowsCounterClockwise Thin |> toHtml []
                , Phosphor.arrowsCounterClockwise Light |> toHtml []
                , Phosphor.arrowsCounterClockwise Regular |> toHtml []
                , Phosphor.arrowsCounterClockwise Bold |> toHtml []
                , Phosphor.arrowsCounterClockwise Fill |> toHtml []
                , Phosphor.arrowsCounterClockwise Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsDownUp model.weight |> toHtml []
                , Phosphor.arrowsDownUp Thin |> toHtml []
                , Phosphor.arrowsDownUp Light |> toHtml []
                , Phosphor.arrowsDownUp Regular |> toHtml []
                , Phosphor.arrowsDownUp Bold |> toHtml []
                , Phosphor.arrowsDownUp Fill |> toHtml []
                , Phosphor.arrowsDownUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsHorizontal model.weight |> toHtml []
                , Phosphor.arrowsHorizontal Thin |> toHtml []
                , Phosphor.arrowsHorizontal Light |> toHtml []
                , Phosphor.arrowsHorizontal Regular |> toHtml []
                , Phosphor.arrowsHorizontal Bold |> toHtml []
                , Phosphor.arrowsHorizontal Fill |> toHtml []
                , Phosphor.arrowsHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsIn model.weight |> toHtml []
                , Phosphor.arrowsIn Thin |> toHtml []
                , Phosphor.arrowsIn Light |> toHtml []
                , Phosphor.arrowsIn Regular |> toHtml []
                , Phosphor.arrowsIn Bold |> toHtml []
                , Phosphor.arrowsIn Fill |> toHtml []
                , Phosphor.arrowsIn Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsInCardinal model.weight |> toHtml []
                , Phosphor.arrowsInCardinal Thin |> toHtml []
                , Phosphor.arrowsInCardinal Light |> toHtml []
                , Phosphor.arrowsInCardinal Regular |> toHtml []
                , Phosphor.arrowsInCardinal Bold |> toHtml []
                , Phosphor.arrowsInCardinal Fill |> toHtml []
                , Phosphor.arrowsInCardinal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsInLineHorizontal model.weight |> toHtml []
                , Phosphor.arrowsInLineHorizontal Thin |> toHtml []
                , Phosphor.arrowsInLineHorizontal Light |> toHtml []
                , Phosphor.arrowsInLineHorizontal Regular |> toHtml []
                , Phosphor.arrowsInLineHorizontal Bold |> toHtml []
                , Phosphor.arrowsInLineHorizontal Fill |> toHtml []
                , Phosphor.arrowsInLineHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsInLineVertical model.weight |> toHtml []
                , Phosphor.arrowsInLineVertical Thin |> toHtml []
                , Phosphor.arrowsInLineVertical Light |> toHtml []
                , Phosphor.arrowsInLineVertical Regular |> toHtml []
                , Phosphor.arrowsInLineVertical Bold |> toHtml []
                , Phosphor.arrowsInLineVertical Fill |> toHtml []
                , Phosphor.arrowsInLineVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsInSimple model.weight |> toHtml []
                , Phosphor.arrowsInSimple Thin |> toHtml []
                , Phosphor.arrowsInSimple Light |> toHtml []
                , Phosphor.arrowsInSimple Regular |> toHtml []
                , Phosphor.arrowsInSimple Bold |> toHtml []
                , Phosphor.arrowsInSimple Fill |> toHtml []
                , Phosphor.arrowsInSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsLeftRight model.weight |> toHtml []
                , Phosphor.arrowsLeftRight Thin |> toHtml []
                , Phosphor.arrowsLeftRight Light |> toHtml []
                , Phosphor.arrowsLeftRight Regular |> toHtml []
                , Phosphor.arrowsLeftRight Bold |> toHtml []
                , Phosphor.arrowsLeftRight Fill |> toHtml []
                , Phosphor.arrowsLeftRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsMerge model.weight |> toHtml []
                , Phosphor.arrowsMerge Thin |> toHtml []
                , Phosphor.arrowsMerge Light |> toHtml []
                , Phosphor.arrowsMerge Regular |> toHtml []
                , Phosphor.arrowsMerge Bold |> toHtml []
                , Phosphor.arrowsMerge Fill |> toHtml []
                , Phosphor.arrowsMerge Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsOut model.weight |> toHtml []
                , Phosphor.arrowsOut Thin |> toHtml []
                , Phosphor.arrowsOut Light |> toHtml []
                , Phosphor.arrowsOut Regular |> toHtml []
                , Phosphor.arrowsOut Bold |> toHtml []
                , Phosphor.arrowsOut Fill |> toHtml []
                , Phosphor.arrowsOut Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsOutCardinal model.weight |> toHtml []
                , Phosphor.arrowsOutCardinal Thin |> toHtml []
                , Phosphor.arrowsOutCardinal Light |> toHtml []
                , Phosphor.arrowsOutCardinal Regular |> toHtml []
                , Phosphor.arrowsOutCardinal Bold |> toHtml []
                , Phosphor.arrowsOutCardinal Fill |> toHtml []
                , Phosphor.arrowsOutCardinal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsOutLineHorizontal model.weight |> toHtml []
                , Phosphor.arrowsOutLineHorizontal Thin |> toHtml []
                , Phosphor.arrowsOutLineHorizontal Light |> toHtml []
                , Phosphor.arrowsOutLineHorizontal Regular |> toHtml []
                , Phosphor.arrowsOutLineHorizontal Bold |> toHtml []
                , Phosphor.arrowsOutLineHorizontal Fill |> toHtml []
                , Phosphor.arrowsOutLineHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsOutLineVertical model.weight |> toHtml []
                , Phosphor.arrowsOutLineVertical Thin |> toHtml []
                , Phosphor.arrowsOutLineVertical Light |> toHtml []
                , Phosphor.arrowsOutLineVertical Regular |> toHtml []
                , Phosphor.arrowsOutLineVertical Bold |> toHtml []
                , Phosphor.arrowsOutLineVertical Fill |> toHtml []
                , Phosphor.arrowsOutLineVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsOutSimple model.weight |> toHtml []
                , Phosphor.arrowsOutSimple Thin |> toHtml []
                , Phosphor.arrowsOutSimple Light |> toHtml []
                , Phosphor.arrowsOutSimple Regular |> toHtml []
                , Phosphor.arrowsOutSimple Bold |> toHtml []
                , Phosphor.arrowsOutSimple Fill |> toHtml []
                , Phosphor.arrowsOutSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsSplit model.weight |> toHtml []
                , Phosphor.arrowsSplit Thin |> toHtml []
                , Phosphor.arrowsSplit Light |> toHtml []
                , Phosphor.arrowsSplit Regular |> toHtml []
                , Phosphor.arrowsSplit Bold |> toHtml []
                , Phosphor.arrowsSplit Fill |> toHtml []
                , Phosphor.arrowsSplit Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.arrowsVertical model.weight |> toHtml []
                , Phosphor.arrowsVertical Thin |> toHtml []
                , Phosphor.arrowsVertical Light |> toHtml []
                , Phosphor.arrowsVertical Regular |> toHtml []
                , Phosphor.arrowsVertical Bold |> toHtml []
                , Phosphor.arrowsVertical Fill |> toHtml []
                , Phosphor.arrowsVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.article model.weight |> toHtml []
                , Phosphor.article Thin |> toHtml []
                , Phosphor.article Light |> toHtml []
                , Phosphor.article Regular |> toHtml []
                , Phosphor.article Bold |> toHtml []
                , Phosphor.article Fill |> toHtml []
                , Phosphor.article Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.articleMedium model.weight |> toHtml []
                , Phosphor.articleMedium Thin |> toHtml []
                , Phosphor.articleMedium Light |> toHtml []
                , Phosphor.articleMedium Regular |> toHtml []
                , Phosphor.articleMedium Bold |> toHtml []
                , Phosphor.articleMedium Fill |> toHtml []
                , Phosphor.articleMedium Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.articleNyTimes model.weight |> toHtml []
                , Phosphor.articleNyTimes Thin |> toHtml []
                , Phosphor.articleNyTimes Light |> toHtml []
                , Phosphor.articleNyTimes Regular |> toHtml []
                , Phosphor.articleNyTimes Bold |> toHtml []
                , Phosphor.articleNyTimes Fill |> toHtml []
                , Phosphor.articleNyTimes Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.asterisk model.weight |> toHtml []
                , Phosphor.asterisk Thin |> toHtml []
                , Phosphor.asterisk Light |> toHtml []
                , Phosphor.asterisk Regular |> toHtml []
                , Phosphor.asterisk Bold |> toHtml []
                , Phosphor.asterisk Fill |> toHtml []
                , Phosphor.asterisk Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.asteriskSimple model.weight |> toHtml []
                , Phosphor.asteriskSimple Thin |> toHtml []
                , Phosphor.asteriskSimple Light |> toHtml []
                , Phosphor.asteriskSimple Regular |> toHtml []
                , Phosphor.asteriskSimple Bold |> toHtml []
                , Phosphor.asteriskSimple Fill |> toHtml []
                , Phosphor.asteriskSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.at model.weight |> toHtml []
                , Phosphor.at Thin |> toHtml []
                , Phosphor.at Light |> toHtml []
                , Phosphor.at Regular |> toHtml []
                , Phosphor.at Bold |> toHtml []
                , Phosphor.at Fill |> toHtml []
                , Phosphor.at Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.atom model.weight |> toHtml []
                , Phosphor.atom Thin |> toHtml []
                , Phosphor.atom Light |> toHtml []
                , Phosphor.atom Regular |> toHtml []
                , Phosphor.atom Bold |> toHtml []
                , Phosphor.atom Fill |> toHtml []
                , Phosphor.atom Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.baby model.weight |> toHtml []
                , Phosphor.baby Thin |> toHtml []
                , Phosphor.baby Light |> toHtml []
                , Phosphor.baby Regular |> toHtml []
                , Phosphor.baby Bold |> toHtml []
                , Phosphor.baby Fill |> toHtml []
                , Phosphor.baby Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.backpack model.weight |> toHtml []
                , Phosphor.backpack Thin |> toHtml []
                , Phosphor.backpack Light |> toHtml []
                , Phosphor.backpack Regular |> toHtml []
                , Phosphor.backpack Bold |> toHtml []
                , Phosphor.backpack Fill |> toHtml []
                , Phosphor.backpack Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.backspace model.weight |> toHtml []
                , Phosphor.backspace Thin |> toHtml []
                , Phosphor.backspace Light |> toHtml []
                , Phosphor.backspace Regular |> toHtml []
                , Phosphor.backspace Bold |> toHtml []
                , Phosphor.backspace Fill |> toHtml []
                , Phosphor.backspace Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bag model.weight |> toHtml []
                , Phosphor.bag Thin |> toHtml []
                , Phosphor.bag Light |> toHtml []
                , Phosphor.bag Regular |> toHtml []
                , Phosphor.bag Bold |> toHtml []
                , Phosphor.bag Fill |> toHtml []
                , Phosphor.bag Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bagSimple model.weight |> toHtml []
                , Phosphor.bagSimple Thin |> toHtml []
                , Phosphor.bagSimple Light |> toHtml []
                , Phosphor.bagSimple Regular |> toHtml []
                , Phosphor.bagSimple Bold |> toHtml []
                , Phosphor.bagSimple Fill |> toHtml []
                , Phosphor.bagSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.balloon model.weight |> toHtml []
                , Phosphor.balloon Thin |> toHtml []
                , Phosphor.balloon Light |> toHtml []
                , Phosphor.balloon Regular |> toHtml []
                , Phosphor.balloon Bold |> toHtml []
                , Phosphor.balloon Fill |> toHtml []
                , Phosphor.balloon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bandaids model.weight |> toHtml []
                , Phosphor.bandaids Thin |> toHtml []
                , Phosphor.bandaids Light |> toHtml []
                , Phosphor.bandaids Regular |> toHtml []
                , Phosphor.bandaids Bold |> toHtml []
                , Phosphor.bandaids Fill |> toHtml []
                , Phosphor.bandaids Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bank model.weight |> toHtml []
                , Phosphor.bank Thin |> toHtml []
                , Phosphor.bank Light |> toHtml []
                , Phosphor.bank Regular |> toHtml []
                , Phosphor.bank Bold |> toHtml []
                , Phosphor.bank Fill |> toHtml []
                , Phosphor.bank Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.barbell model.weight |> toHtml []
                , Phosphor.barbell Thin |> toHtml []
                , Phosphor.barbell Light |> toHtml []
                , Phosphor.barbell Regular |> toHtml []
                , Phosphor.barbell Bold |> toHtml []
                , Phosphor.barbell Fill |> toHtml []
                , Phosphor.barbell Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.barcode model.weight |> toHtml []
                , Phosphor.barcode Thin |> toHtml []
                , Phosphor.barcode Light |> toHtml []
                , Phosphor.barcode Regular |> toHtml []
                , Phosphor.barcode Bold |> toHtml []
                , Phosphor.barcode Fill |> toHtml []
                , Phosphor.barcode Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.barricade model.weight |> toHtml []
                , Phosphor.barricade Thin |> toHtml []
                , Phosphor.barricade Light |> toHtml []
                , Phosphor.barricade Regular |> toHtml []
                , Phosphor.barricade Bold |> toHtml []
                , Phosphor.barricade Fill |> toHtml []
                , Phosphor.barricade Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.baseball model.weight |> toHtml []
                , Phosphor.baseball Thin |> toHtml []
                , Phosphor.baseball Light |> toHtml []
                , Phosphor.baseball Regular |> toHtml []
                , Phosphor.baseball Bold |> toHtml []
                , Phosphor.baseball Fill |> toHtml []
                , Phosphor.baseball Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.baseballCap model.weight |> toHtml []
                , Phosphor.baseballCap Thin |> toHtml []
                , Phosphor.baseballCap Light |> toHtml []
                , Phosphor.baseballCap Regular |> toHtml []
                , Phosphor.baseballCap Bold |> toHtml []
                , Phosphor.baseballCap Fill |> toHtml []
                , Phosphor.baseballCap Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.basket model.weight |> toHtml []
                , Phosphor.basket Thin |> toHtml []
                , Phosphor.basket Light |> toHtml []
                , Phosphor.basket Regular |> toHtml []
                , Phosphor.basket Bold |> toHtml []
                , Phosphor.basket Fill |> toHtml []
                , Phosphor.basket Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.basketball model.weight |> toHtml []
                , Phosphor.basketball Thin |> toHtml []
                , Phosphor.basketball Light |> toHtml []
                , Phosphor.basketball Regular |> toHtml []
                , Phosphor.basketball Bold |> toHtml []
                , Phosphor.basketball Fill |> toHtml []
                , Phosphor.basketball Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bathtub model.weight |> toHtml []
                , Phosphor.bathtub Thin |> toHtml []
                , Phosphor.bathtub Light |> toHtml []
                , Phosphor.bathtub Regular |> toHtml []
                , Phosphor.bathtub Bold |> toHtml []
                , Phosphor.bathtub Fill |> toHtml []
                , Phosphor.bathtub Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryCharging model.weight |> toHtml []
                , Phosphor.batteryCharging Thin |> toHtml []
                , Phosphor.batteryCharging Light |> toHtml []
                , Phosphor.batteryCharging Regular |> toHtml []
                , Phosphor.batteryCharging Bold |> toHtml []
                , Phosphor.batteryCharging Fill |> toHtml []
                , Phosphor.batteryCharging Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryChargingVertical model.weight |> toHtml []
                , Phosphor.batteryChargingVertical Thin |> toHtml []
                , Phosphor.batteryChargingVertical Light |> toHtml []
                , Phosphor.batteryChargingVertical Regular |> toHtml []
                , Phosphor.batteryChargingVertical Bold |> toHtml []
                , Phosphor.batteryChargingVertical Fill |> toHtml []
                , Phosphor.batteryChargingVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryEmpty model.weight |> toHtml []
                , Phosphor.batteryEmpty Thin |> toHtml []
                , Phosphor.batteryEmpty Light |> toHtml []
                , Phosphor.batteryEmpty Regular |> toHtml []
                , Phosphor.batteryEmpty Bold |> toHtml []
                , Phosphor.batteryEmpty Fill |> toHtml []
                , Phosphor.batteryEmpty Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryFull model.weight |> toHtml []
                , Phosphor.batteryFull Thin |> toHtml []
                , Phosphor.batteryFull Light |> toHtml []
                , Phosphor.batteryFull Regular |> toHtml []
                , Phosphor.batteryFull Bold |> toHtml []
                , Phosphor.batteryFull Fill |> toHtml []
                , Phosphor.batteryFull Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryHigh model.weight |> toHtml []
                , Phosphor.batteryHigh Thin |> toHtml []
                , Phosphor.batteryHigh Light |> toHtml []
                , Phosphor.batteryHigh Regular |> toHtml []
                , Phosphor.batteryHigh Bold |> toHtml []
                , Phosphor.batteryHigh Fill |> toHtml []
                , Phosphor.batteryHigh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryLow model.weight |> toHtml []
                , Phosphor.batteryLow Thin |> toHtml []
                , Phosphor.batteryLow Light |> toHtml []
                , Phosphor.batteryLow Regular |> toHtml []
                , Phosphor.batteryLow Bold |> toHtml []
                , Phosphor.batteryLow Fill |> toHtml []
                , Phosphor.batteryLow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryMedium model.weight |> toHtml []
                , Phosphor.batteryMedium Thin |> toHtml []
                , Phosphor.batteryMedium Light |> toHtml []
                , Phosphor.batteryMedium Regular |> toHtml []
                , Phosphor.batteryMedium Bold |> toHtml []
                , Phosphor.batteryMedium Fill |> toHtml []
                , Phosphor.batteryMedium Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryPlus model.weight |> toHtml []
                , Phosphor.batteryPlus Thin |> toHtml []
                , Phosphor.batteryPlus Light |> toHtml []
                , Phosphor.batteryPlus Regular |> toHtml []
                , Phosphor.batteryPlus Bold |> toHtml []
                , Phosphor.batteryPlus Fill |> toHtml []
                , Phosphor.batteryPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryPlusVertical model.weight |> toHtml []
                , Phosphor.batteryPlusVertical Thin |> toHtml []
                , Phosphor.batteryPlusVertical Light |> toHtml []
                , Phosphor.batteryPlusVertical Regular |> toHtml []
                , Phosphor.batteryPlusVertical Bold |> toHtml []
                , Phosphor.batteryPlusVertical Fill |> toHtml []
                , Phosphor.batteryPlusVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryVerticalEmpty model.weight |> toHtml []
                , Phosphor.batteryVerticalEmpty Thin |> toHtml []
                , Phosphor.batteryVerticalEmpty Light |> toHtml []
                , Phosphor.batteryVerticalEmpty Regular |> toHtml []
                , Phosphor.batteryVerticalEmpty Bold |> toHtml []
                , Phosphor.batteryVerticalEmpty Fill |> toHtml []
                , Phosphor.batteryVerticalEmpty Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryVerticalFull model.weight |> toHtml []
                , Phosphor.batteryVerticalFull Thin |> toHtml []
                , Phosphor.batteryVerticalFull Light |> toHtml []
                , Phosphor.batteryVerticalFull Regular |> toHtml []
                , Phosphor.batteryVerticalFull Bold |> toHtml []
                , Phosphor.batteryVerticalFull Fill |> toHtml []
                , Phosphor.batteryVerticalFull Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryVerticalHigh model.weight |> toHtml []
                , Phosphor.batteryVerticalHigh Thin |> toHtml []
                , Phosphor.batteryVerticalHigh Light |> toHtml []
                , Phosphor.batteryVerticalHigh Regular |> toHtml []
                , Phosphor.batteryVerticalHigh Bold |> toHtml []
                , Phosphor.batteryVerticalHigh Fill |> toHtml []
                , Phosphor.batteryVerticalHigh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryVerticalLow model.weight |> toHtml []
                , Phosphor.batteryVerticalLow Thin |> toHtml []
                , Phosphor.batteryVerticalLow Light |> toHtml []
                , Phosphor.batteryVerticalLow Regular |> toHtml []
                , Phosphor.batteryVerticalLow Bold |> toHtml []
                , Phosphor.batteryVerticalLow Fill |> toHtml []
                , Phosphor.batteryVerticalLow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryVerticalMedium model.weight |> toHtml []
                , Phosphor.batteryVerticalMedium Thin |> toHtml []
                , Phosphor.batteryVerticalMedium Light |> toHtml []
                , Phosphor.batteryVerticalMedium Regular |> toHtml []
                , Phosphor.batteryVerticalMedium Bold |> toHtml []
                , Phosphor.batteryVerticalMedium Fill |> toHtml []
                , Phosphor.batteryVerticalMedium Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryWarning model.weight |> toHtml []
                , Phosphor.batteryWarning Thin |> toHtml []
                , Phosphor.batteryWarning Light |> toHtml []
                , Phosphor.batteryWarning Regular |> toHtml []
                , Phosphor.batteryWarning Bold |> toHtml []
                , Phosphor.batteryWarning Fill |> toHtml []
                , Phosphor.batteryWarning Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.batteryWarningVertical model.weight |> toHtml []
                , Phosphor.batteryWarningVertical Thin |> toHtml []
                , Phosphor.batteryWarningVertical Light |> toHtml []
                , Phosphor.batteryWarningVertical Regular |> toHtml []
                , Phosphor.batteryWarningVertical Bold |> toHtml []
                , Phosphor.batteryWarningVertical Fill |> toHtml []
                , Phosphor.batteryWarningVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bed model.weight |> toHtml []
                , Phosphor.bed Thin |> toHtml []
                , Phosphor.bed Light |> toHtml []
                , Phosphor.bed Regular |> toHtml []
                , Phosphor.bed Bold |> toHtml []
                , Phosphor.bed Fill |> toHtml []
                , Phosphor.bed Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.beerBottle model.weight |> toHtml []
                , Phosphor.beerBottle Thin |> toHtml []
                , Phosphor.beerBottle Light |> toHtml []
                , Phosphor.beerBottle Regular |> toHtml []
                , Phosphor.beerBottle Bold |> toHtml []
                , Phosphor.beerBottle Fill |> toHtml []
                , Phosphor.beerBottle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.beerStein model.weight |> toHtml []
                , Phosphor.beerStein Thin |> toHtml []
                , Phosphor.beerStein Light |> toHtml []
                , Phosphor.beerStein Regular |> toHtml []
                , Phosphor.beerStein Bold |> toHtml []
                , Phosphor.beerStein Fill |> toHtml []
                , Phosphor.beerStein Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.behanceLogo model.weight |> toHtml []
                , Phosphor.behanceLogo Thin |> toHtml []
                , Phosphor.behanceLogo Light |> toHtml []
                , Phosphor.behanceLogo Regular |> toHtml []
                , Phosphor.behanceLogo Bold |> toHtml []
                , Phosphor.behanceLogo Fill |> toHtml []
                , Phosphor.behanceLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bell model.weight |> toHtml []
                , Phosphor.bell Thin |> toHtml []
                , Phosphor.bell Light |> toHtml []
                , Phosphor.bell Regular |> toHtml []
                , Phosphor.bell Bold |> toHtml []
                , Phosphor.bell Fill |> toHtml []
                , Phosphor.bell Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bellRinging model.weight |> toHtml []
                , Phosphor.bellRinging Thin |> toHtml []
                , Phosphor.bellRinging Light |> toHtml []
                , Phosphor.bellRinging Regular |> toHtml []
                , Phosphor.bellRinging Bold |> toHtml []
                , Phosphor.bellRinging Fill |> toHtml []
                , Phosphor.bellRinging Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bellSimple model.weight |> toHtml []
                , Phosphor.bellSimple Thin |> toHtml []
                , Phosphor.bellSimple Light |> toHtml []
                , Phosphor.bellSimple Regular |> toHtml []
                , Phosphor.bellSimple Bold |> toHtml []
                , Phosphor.bellSimple Fill |> toHtml []
                , Phosphor.bellSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bellSimpleRinging model.weight |> toHtml []
                , Phosphor.bellSimpleRinging Thin |> toHtml []
                , Phosphor.bellSimpleRinging Light |> toHtml []
                , Phosphor.bellSimpleRinging Regular |> toHtml []
                , Phosphor.bellSimpleRinging Bold |> toHtml []
                , Phosphor.bellSimpleRinging Fill |> toHtml []
                , Phosphor.bellSimpleRinging Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bellSimpleSlash model.weight |> toHtml []
                , Phosphor.bellSimpleSlash Thin |> toHtml []
                , Phosphor.bellSimpleSlash Light |> toHtml []
                , Phosphor.bellSimpleSlash Regular |> toHtml []
                , Phosphor.bellSimpleSlash Bold |> toHtml []
                , Phosphor.bellSimpleSlash Fill |> toHtml []
                , Phosphor.bellSimpleSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bellSimpleZ model.weight |> toHtml []
                , Phosphor.bellSimpleZ Thin |> toHtml []
                , Phosphor.bellSimpleZ Light |> toHtml []
                , Phosphor.bellSimpleZ Regular |> toHtml []
                , Phosphor.bellSimpleZ Bold |> toHtml []
                , Phosphor.bellSimpleZ Fill |> toHtml []
                , Phosphor.bellSimpleZ Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bellSlash model.weight |> toHtml []
                , Phosphor.bellSlash Thin |> toHtml []
                , Phosphor.bellSlash Light |> toHtml []
                , Phosphor.bellSlash Regular |> toHtml []
                , Phosphor.bellSlash Bold |> toHtml []
                , Phosphor.bellSlash Fill |> toHtml []
                , Phosphor.bellSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bellZ model.weight |> toHtml []
                , Phosphor.bellZ Thin |> toHtml []
                , Phosphor.bellZ Light |> toHtml []
                , Phosphor.bellZ Regular |> toHtml []
                , Phosphor.bellZ Bold |> toHtml []
                , Phosphor.bellZ Fill |> toHtml []
                , Phosphor.bellZ Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bezierCurve model.weight |> toHtml []
                , Phosphor.bezierCurve Thin |> toHtml []
                , Phosphor.bezierCurve Light |> toHtml []
                , Phosphor.bezierCurve Regular |> toHtml []
                , Phosphor.bezierCurve Bold |> toHtml []
                , Phosphor.bezierCurve Fill |> toHtml []
                , Phosphor.bezierCurve Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bicycle model.weight |> toHtml []
                , Phosphor.bicycle Thin |> toHtml []
                , Phosphor.bicycle Light |> toHtml []
                , Phosphor.bicycle Regular |> toHtml []
                , Phosphor.bicycle Bold |> toHtml []
                , Phosphor.bicycle Fill |> toHtml []
                , Phosphor.bicycle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.binoculars model.weight |> toHtml []
                , Phosphor.binoculars Thin |> toHtml []
                , Phosphor.binoculars Light |> toHtml []
                , Phosphor.binoculars Regular |> toHtml []
                , Phosphor.binoculars Bold |> toHtml []
                , Phosphor.binoculars Fill |> toHtml []
                , Phosphor.binoculars Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bird model.weight |> toHtml []
                , Phosphor.bird Thin |> toHtml []
                , Phosphor.bird Light |> toHtml []
                , Phosphor.bird Regular |> toHtml []
                , Phosphor.bird Bold |> toHtml []
                , Phosphor.bird Fill |> toHtml []
                , Phosphor.bird Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bluetooth model.weight |> toHtml []
                , Phosphor.bluetooth Thin |> toHtml []
                , Phosphor.bluetooth Light |> toHtml []
                , Phosphor.bluetooth Regular |> toHtml []
                , Phosphor.bluetooth Bold |> toHtml []
                , Phosphor.bluetooth Fill |> toHtml []
                , Phosphor.bluetooth Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bluetoothConnected model.weight |> toHtml []
                , Phosphor.bluetoothConnected Thin |> toHtml []
                , Phosphor.bluetoothConnected Light |> toHtml []
                , Phosphor.bluetoothConnected Regular |> toHtml []
                , Phosphor.bluetoothConnected Bold |> toHtml []
                , Phosphor.bluetoothConnected Fill |> toHtml []
                , Phosphor.bluetoothConnected Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bluetoothSlash model.weight |> toHtml []
                , Phosphor.bluetoothSlash Thin |> toHtml []
                , Phosphor.bluetoothSlash Light |> toHtml []
                , Phosphor.bluetoothSlash Regular |> toHtml []
                , Phosphor.bluetoothSlash Bold |> toHtml []
                , Phosphor.bluetoothSlash Fill |> toHtml []
                , Phosphor.bluetoothSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bluetoothX model.weight |> toHtml []
                , Phosphor.bluetoothX Thin |> toHtml []
                , Phosphor.bluetoothX Light |> toHtml []
                , Phosphor.bluetoothX Regular |> toHtml []
                , Phosphor.bluetoothX Bold |> toHtml []
                , Phosphor.bluetoothX Fill |> toHtml []
                , Phosphor.bluetoothX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.boat model.weight |> toHtml []
                , Phosphor.boat Thin |> toHtml []
                , Phosphor.boat Light |> toHtml []
                , Phosphor.boat Regular |> toHtml []
                , Phosphor.boat Bold |> toHtml []
                , Phosphor.boat Fill |> toHtml []
                , Phosphor.boat Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bone model.weight |> toHtml []
                , Phosphor.bone Thin |> toHtml []
                , Phosphor.bone Light |> toHtml []
                , Phosphor.bone Regular |> toHtml []
                , Phosphor.bone Bold |> toHtml []
                , Phosphor.bone Fill |> toHtml []
                , Phosphor.bone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.book model.weight |> toHtml []
                , Phosphor.book Thin |> toHtml []
                , Phosphor.book Light |> toHtml []
                , Phosphor.book Regular |> toHtml []
                , Phosphor.book Bold |> toHtml []
                , Phosphor.book Fill |> toHtml []
                , Phosphor.book Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bookBookmark model.weight |> toHtml []
                , Phosphor.bookBookmark Thin |> toHtml []
                , Phosphor.bookBookmark Light |> toHtml []
                , Phosphor.bookBookmark Regular |> toHtml []
                , Phosphor.bookBookmark Bold |> toHtml []
                , Phosphor.bookBookmark Fill |> toHtml []
                , Phosphor.bookBookmark Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bookOpen model.weight |> toHtml []
                , Phosphor.bookOpen Thin |> toHtml []
                , Phosphor.bookOpen Light |> toHtml []
                , Phosphor.bookOpen Regular |> toHtml []
                , Phosphor.bookOpen Bold |> toHtml []
                , Phosphor.bookOpen Fill |> toHtml []
                , Phosphor.bookOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bookOpenText model.weight |> toHtml []
                , Phosphor.bookOpenText Thin |> toHtml []
                , Phosphor.bookOpenText Light |> toHtml []
                , Phosphor.bookOpenText Regular |> toHtml []
                , Phosphor.bookOpenText Bold |> toHtml []
                , Phosphor.bookOpenText Fill |> toHtml []
                , Phosphor.bookOpenText Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bookmark model.weight |> toHtml []
                , Phosphor.bookmark Thin |> toHtml []
                , Phosphor.bookmark Light |> toHtml []
                , Phosphor.bookmark Regular |> toHtml []
                , Phosphor.bookmark Bold |> toHtml []
                , Phosphor.bookmark Fill |> toHtml []
                , Phosphor.bookmark Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bookmarkSimple model.weight |> toHtml []
                , Phosphor.bookmarkSimple Thin |> toHtml []
                , Phosphor.bookmarkSimple Light |> toHtml []
                , Phosphor.bookmarkSimple Regular |> toHtml []
                , Phosphor.bookmarkSimple Bold |> toHtml []
                , Phosphor.bookmarkSimple Fill |> toHtml []
                , Phosphor.bookmarkSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bookmarks model.weight |> toHtml []
                , Phosphor.bookmarks Thin |> toHtml []
                , Phosphor.bookmarks Light |> toHtml []
                , Phosphor.bookmarks Regular |> toHtml []
                , Phosphor.bookmarks Bold |> toHtml []
                , Phosphor.bookmarks Fill |> toHtml []
                , Phosphor.bookmarks Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bookmarksSimple model.weight |> toHtml []
                , Phosphor.bookmarksSimple Thin |> toHtml []
                , Phosphor.bookmarksSimple Light |> toHtml []
                , Phosphor.bookmarksSimple Regular |> toHtml []
                , Phosphor.bookmarksSimple Bold |> toHtml []
                , Phosphor.bookmarksSimple Fill |> toHtml []
                , Phosphor.bookmarksSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.books model.weight |> toHtml []
                , Phosphor.books Thin |> toHtml []
                , Phosphor.books Light |> toHtml []
                , Phosphor.books Regular |> toHtml []
                , Phosphor.books Bold |> toHtml []
                , Phosphor.books Fill |> toHtml []
                , Phosphor.books Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.boot model.weight |> toHtml []
                , Phosphor.boot Thin |> toHtml []
                , Phosphor.boot Light |> toHtml []
                , Phosphor.boot Regular |> toHtml []
                , Phosphor.boot Bold |> toHtml []
                , Phosphor.boot Fill |> toHtml []
                , Phosphor.boot Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.boundingBox model.weight |> toHtml []
                , Phosphor.boundingBox Thin |> toHtml []
                , Phosphor.boundingBox Light |> toHtml []
                , Phosphor.boundingBox Regular |> toHtml []
                , Phosphor.boundingBox Bold |> toHtml []
                , Phosphor.boundingBox Fill |> toHtml []
                , Phosphor.boundingBox Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bowlFood model.weight |> toHtml []
                , Phosphor.bowlFood Thin |> toHtml []
                , Phosphor.bowlFood Light |> toHtml []
                , Phosphor.bowlFood Regular |> toHtml []
                , Phosphor.bowlFood Bold |> toHtml []
                , Phosphor.bowlFood Fill |> toHtml []
                , Phosphor.bowlFood Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bracketsAngle model.weight |> toHtml []
                , Phosphor.bracketsAngle Thin |> toHtml []
                , Phosphor.bracketsAngle Light |> toHtml []
                , Phosphor.bracketsAngle Regular |> toHtml []
                , Phosphor.bracketsAngle Bold |> toHtml []
                , Phosphor.bracketsAngle Fill |> toHtml []
                , Phosphor.bracketsAngle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bracketsCurly model.weight |> toHtml []
                , Phosphor.bracketsCurly Thin |> toHtml []
                , Phosphor.bracketsCurly Light |> toHtml []
                , Phosphor.bracketsCurly Regular |> toHtml []
                , Phosphor.bracketsCurly Bold |> toHtml []
                , Phosphor.bracketsCurly Fill |> toHtml []
                , Phosphor.bracketsCurly Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bracketsRound model.weight |> toHtml []
                , Phosphor.bracketsRound Thin |> toHtml []
                , Phosphor.bracketsRound Light |> toHtml []
                , Phosphor.bracketsRound Regular |> toHtml []
                , Phosphor.bracketsRound Bold |> toHtml []
                , Phosphor.bracketsRound Fill |> toHtml []
                , Phosphor.bracketsRound Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bracketsSquare model.weight |> toHtml []
                , Phosphor.bracketsSquare Thin |> toHtml []
                , Phosphor.bracketsSquare Light |> toHtml []
                , Phosphor.bracketsSquare Regular |> toHtml []
                , Phosphor.bracketsSquare Bold |> toHtml []
                , Phosphor.bracketsSquare Fill |> toHtml []
                , Phosphor.bracketsSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.brain model.weight |> toHtml []
                , Phosphor.brain Thin |> toHtml []
                , Phosphor.brain Light |> toHtml []
                , Phosphor.brain Regular |> toHtml []
                , Phosphor.brain Bold |> toHtml []
                , Phosphor.brain Fill |> toHtml []
                , Phosphor.brain Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.brandy model.weight |> toHtml []
                , Phosphor.brandy Thin |> toHtml []
                , Phosphor.brandy Light |> toHtml []
                , Phosphor.brandy Regular |> toHtml []
                , Phosphor.brandy Bold |> toHtml []
                , Phosphor.brandy Fill |> toHtml []
                , Phosphor.brandy Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bridge model.weight |> toHtml []
                , Phosphor.bridge Thin |> toHtml []
                , Phosphor.bridge Light |> toHtml []
                , Phosphor.bridge Regular |> toHtml []
                , Phosphor.bridge Bold |> toHtml []
                , Phosphor.bridge Fill |> toHtml []
                , Phosphor.bridge Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.briefcase model.weight |> toHtml []
                , Phosphor.briefcase Thin |> toHtml []
                , Phosphor.briefcase Light |> toHtml []
                , Phosphor.briefcase Regular |> toHtml []
                , Phosphor.briefcase Bold |> toHtml []
                , Phosphor.briefcase Fill |> toHtml []
                , Phosphor.briefcase Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.briefcaseMetal model.weight |> toHtml []
                , Phosphor.briefcaseMetal Thin |> toHtml []
                , Phosphor.briefcaseMetal Light |> toHtml []
                , Phosphor.briefcaseMetal Regular |> toHtml []
                , Phosphor.briefcaseMetal Bold |> toHtml []
                , Phosphor.briefcaseMetal Fill |> toHtml []
                , Phosphor.briefcaseMetal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.broadcast model.weight |> toHtml []
                , Phosphor.broadcast Thin |> toHtml []
                , Phosphor.broadcast Light |> toHtml []
                , Phosphor.broadcast Regular |> toHtml []
                , Phosphor.broadcast Bold |> toHtml []
                , Phosphor.broadcast Fill |> toHtml []
                , Phosphor.broadcast Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.broom model.weight |> toHtml []
                , Phosphor.broom Thin |> toHtml []
                , Phosphor.broom Light |> toHtml []
                , Phosphor.broom Regular |> toHtml []
                , Phosphor.broom Bold |> toHtml []
                , Phosphor.broom Fill |> toHtml []
                , Phosphor.broom Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.browser model.weight |> toHtml []
                , Phosphor.browser Thin |> toHtml []
                , Phosphor.browser Light |> toHtml []
                , Phosphor.browser Regular |> toHtml []
                , Phosphor.browser Bold |> toHtml []
                , Phosphor.browser Fill |> toHtml []
                , Phosphor.browser Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.browsers model.weight |> toHtml []
                , Phosphor.browsers Thin |> toHtml []
                , Phosphor.browsers Light |> toHtml []
                , Phosphor.browsers Regular |> toHtml []
                , Phosphor.browsers Bold |> toHtml []
                , Phosphor.browsers Fill |> toHtml []
                , Phosphor.browsers Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bugBeetle model.weight |> toHtml []
                , Phosphor.bugBeetle Thin |> toHtml []
                , Phosphor.bugBeetle Light |> toHtml []
                , Phosphor.bugBeetle Regular |> toHtml []
                , Phosphor.bugBeetle Bold |> toHtml []
                , Phosphor.bugBeetle Fill |> toHtml []
                , Phosphor.bugBeetle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bug model.weight |> toHtml []
                , Phosphor.bug Thin |> toHtml []
                , Phosphor.bug Light |> toHtml []
                , Phosphor.bug Regular |> toHtml []
                , Phosphor.bug Bold |> toHtml []
                , Phosphor.bug Fill |> toHtml []
                , Phosphor.bug Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bugDroid model.weight |> toHtml []
                , Phosphor.bugDroid Thin |> toHtml []
                , Phosphor.bugDroid Light |> toHtml []
                , Phosphor.bugDroid Regular |> toHtml []
                , Phosphor.bugDroid Bold |> toHtml []
                , Phosphor.bugDroid Fill |> toHtml []
                , Phosphor.bugDroid Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.buildings model.weight |> toHtml []
                , Phosphor.buildings Thin |> toHtml []
                , Phosphor.buildings Light |> toHtml []
                , Phosphor.buildings Regular |> toHtml []
                , Phosphor.buildings Bold |> toHtml []
                , Phosphor.buildings Fill |> toHtml []
                , Phosphor.buildings Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.bus model.weight |> toHtml []
                , Phosphor.bus Thin |> toHtml []
                , Phosphor.bus Light |> toHtml []
                , Phosphor.bus Regular |> toHtml []
                , Phosphor.bus Bold |> toHtml []
                , Phosphor.bus Fill |> toHtml []
                , Phosphor.bus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.butterfly model.weight |> toHtml []
                , Phosphor.butterfly Thin |> toHtml []
                , Phosphor.butterfly Light |> toHtml []
                , Phosphor.butterfly Regular |> toHtml []
                , Phosphor.butterfly Bold |> toHtml []
                , Phosphor.butterfly Fill |> toHtml []
                , Phosphor.butterfly Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cactus model.weight |> toHtml []
                , Phosphor.cactus Thin |> toHtml []
                , Phosphor.cactus Light |> toHtml []
                , Phosphor.cactus Regular |> toHtml []
                , Phosphor.cactus Bold |> toHtml []
                , Phosphor.cactus Fill |> toHtml []
                , Phosphor.cactus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cake model.weight |> toHtml []
                , Phosphor.cake Thin |> toHtml []
                , Phosphor.cake Light |> toHtml []
                , Phosphor.cake Regular |> toHtml []
                , Phosphor.cake Bold |> toHtml []
                , Phosphor.cake Fill |> toHtml []
                , Phosphor.cake Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.calculator model.weight |> toHtml []
                , Phosphor.calculator Thin |> toHtml []
                , Phosphor.calculator Light |> toHtml []
                , Phosphor.calculator Regular |> toHtml []
                , Phosphor.calculator Bold |> toHtml []
                , Phosphor.calculator Fill |> toHtml []
                , Phosphor.calculator Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.calendarBlank model.weight |> toHtml []
                , Phosphor.calendarBlank Thin |> toHtml []
                , Phosphor.calendarBlank Light |> toHtml []
                , Phosphor.calendarBlank Regular |> toHtml []
                , Phosphor.calendarBlank Bold |> toHtml []
                , Phosphor.calendarBlank Fill |> toHtml []
                , Phosphor.calendarBlank Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.calendar model.weight |> toHtml []
                , Phosphor.calendar Thin |> toHtml []
                , Phosphor.calendar Light |> toHtml []
                , Phosphor.calendar Regular |> toHtml []
                , Phosphor.calendar Bold |> toHtml []
                , Phosphor.calendar Fill |> toHtml []
                , Phosphor.calendar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.calendarCheck model.weight |> toHtml []
                , Phosphor.calendarCheck Thin |> toHtml []
                , Phosphor.calendarCheck Light |> toHtml []
                , Phosphor.calendarCheck Regular |> toHtml []
                , Phosphor.calendarCheck Bold |> toHtml []
                , Phosphor.calendarCheck Fill |> toHtml []
                , Phosphor.calendarCheck Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.calendarPlus model.weight |> toHtml []
                , Phosphor.calendarPlus Thin |> toHtml []
                , Phosphor.calendarPlus Light |> toHtml []
                , Phosphor.calendarPlus Regular |> toHtml []
                , Phosphor.calendarPlus Bold |> toHtml []
                , Phosphor.calendarPlus Fill |> toHtml []
                , Phosphor.calendarPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.calendarX model.weight |> toHtml []
                , Phosphor.calendarX Thin |> toHtml []
                , Phosphor.calendarX Light |> toHtml []
                , Phosphor.calendarX Regular |> toHtml []
                , Phosphor.calendarX Bold |> toHtml []
                , Phosphor.calendarX Fill |> toHtml []
                , Phosphor.calendarX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.callBell model.weight |> toHtml []
                , Phosphor.callBell Thin |> toHtml []
                , Phosphor.callBell Light |> toHtml []
                , Phosphor.callBell Regular |> toHtml []
                , Phosphor.callBell Bold |> toHtml []
                , Phosphor.callBell Fill |> toHtml []
                , Phosphor.callBell Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.camera model.weight |> toHtml []
                , Phosphor.camera Thin |> toHtml []
                , Phosphor.camera Light |> toHtml []
                , Phosphor.camera Regular |> toHtml []
                , Phosphor.camera Bold |> toHtml []
                , Phosphor.camera Fill |> toHtml []
                , Phosphor.camera Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cameraPlus model.weight |> toHtml []
                , Phosphor.cameraPlus Thin |> toHtml []
                , Phosphor.cameraPlus Light |> toHtml []
                , Phosphor.cameraPlus Regular |> toHtml []
                , Phosphor.cameraPlus Bold |> toHtml []
                , Phosphor.cameraPlus Fill |> toHtml []
                , Phosphor.cameraPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cameraRotate model.weight |> toHtml []
                , Phosphor.cameraRotate Thin |> toHtml []
                , Phosphor.cameraRotate Light |> toHtml []
                , Phosphor.cameraRotate Regular |> toHtml []
                , Phosphor.cameraRotate Bold |> toHtml []
                , Phosphor.cameraRotate Fill |> toHtml []
                , Phosphor.cameraRotate Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cameraSlash model.weight |> toHtml []
                , Phosphor.cameraSlash Thin |> toHtml []
                , Phosphor.cameraSlash Light |> toHtml []
                , Phosphor.cameraSlash Regular |> toHtml []
                , Phosphor.cameraSlash Bold |> toHtml []
                , Phosphor.cameraSlash Fill |> toHtml []
                , Phosphor.cameraSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.campfire model.weight |> toHtml []
                , Phosphor.campfire Thin |> toHtml []
                , Phosphor.campfire Light |> toHtml []
                , Phosphor.campfire Regular |> toHtml []
                , Phosphor.campfire Bold |> toHtml []
                , Phosphor.campfire Fill |> toHtml []
                , Phosphor.campfire Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.car model.weight |> toHtml []
                , Phosphor.car Thin |> toHtml []
                , Phosphor.car Light |> toHtml []
                , Phosphor.car Regular |> toHtml []
                , Phosphor.car Bold |> toHtml []
                , Phosphor.car Fill |> toHtml []
                , Phosphor.car Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.carProfile model.weight |> toHtml []
                , Phosphor.carProfile Thin |> toHtml []
                , Phosphor.carProfile Light |> toHtml []
                , Phosphor.carProfile Regular |> toHtml []
                , Phosphor.carProfile Bold |> toHtml []
                , Phosphor.carProfile Fill |> toHtml []
                , Phosphor.carProfile Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.carSimple model.weight |> toHtml []
                , Phosphor.carSimple Thin |> toHtml []
                , Phosphor.carSimple Light |> toHtml []
                , Phosphor.carSimple Regular |> toHtml []
                , Phosphor.carSimple Bold |> toHtml []
                , Phosphor.carSimple Fill |> toHtml []
                , Phosphor.carSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cardholder model.weight |> toHtml []
                , Phosphor.cardholder Thin |> toHtml []
                , Phosphor.cardholder Light |> toHtml []
                , Phosphor.cardholder Regular |> toHtml []
                , Phosphor.cardholder Bold |> toHtml []
                , Phosphor.cardholder Fill |> toHtml []
                , Phosphor.cardholder Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cards model.weight |> toHtml []
                , Phosphor.cards Thin |> toHtml []
                , Phosphor.cards Light |> toHtml []
                , Phosphor.cards Regular |> toHtml []
                , Phosphor.cards Bold |> toHtml []
                , Phosphor.cards Fill |> toHtml []
                , Phosphor.cards Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleDoubleDown model.weight |> toHtml []
                , Phosphor.caretCircleDoubleDown Thin |> toHtml []
                , Phosphor.caretCircleDoubleDown Light |> toHtml []
                , Phosphor.caretCircleDoubleDown Regular |> toHtml []
                , Phosphor.caretCircleDoubleDown Bold |> toHtml []
                , Phosphor.caretCircleDoubleDown Fill |> toHtml []
                , Phosphor.caretCircleDoubleDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleDoubleLeft model.weight |> toHtml []
                , Phosphor.caretCircleDoubleLeft Thin |> toHtml []
                , Phosphor.caretCircleDoubleLeft Light |> toHtml []
                , Phosphor.caretCircleDoubleLeft Regular |> toHtml []
                , Phosphor.caretCircleDoubleLeft Bold |> toHtml []
                , Phosphor.caretCircleDoubleLeft Fill |> toHtml []
                , Phosphor.caretCircleDoubleLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleDoubleRight model.weight |> toHtml []
                , Phosphor.caretCircleDoubleRight Thin |> toHtml []
                , Phosphor.caretCircleDoubleRight Light |> toHtml []
                , Phosphor.caretCircleDoubleRight Regular |> toHtml []
                , Phosphor.caretCircleDoubleRight Bold |> toHtml []
                , Phosphor.caretCircleDoubleRight Fill |> toHtml []
                , Phosphor.caretCircleDoubleRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleDoubleUp model.weight |> toHtml []
                , Phosphor.caretCircleDoubleUp Thin |> toHtml []
                , Phosphor.caretCircleDoubleUp Light |> toHtml []
                , Phosphor.caretCircleDoubleUp Regular |> toHtml []
                , Phosphor.caretCircleDoubleUp Bold |> toHtml []
                , Phosphor.caretCircleDoubleUp Fill |> toHtml []
                , Phosphor.caretCircleDoubleUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleDown model.weight |> toHtml []
                , Phosphor.caretCircleDown Thin |> toHtml []
                , Phosphor.caretCircleDown Light |> toHtml []
                , Phosphor.caretCircleDown Regular |> toHtml []
                , Phosphor.caretCircleDown Bold |> toHtml []
                , Phosphor.caretCircleDown Fill |> toHtml []
                , Phosphor.caretCircleDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleLeft model.weight |> toHtml []
                , Phosphor.caretCircleLeft Thin |> toHtml []
                , Phosphor.caretCircleLeft Light |> toHtml []
                , Phosphor.caretCircleLeft Regular |> toHtml []
                , Phosphor.caretCircleLeft Bold |> toHtml []
                , Phosphor.caretCircleLeft Fill |> toHtml []
                , Phosphor.caretCircleLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleRight model.weight |> toHtml []
                , Phosphor.caretCircleRight Thin |> toHtml []
                , Phosphor.caretCircleRight Light |> toHtml []
                , Phosphor.caretCircleRight Regular |> toHtml []
                , Phosphor.caretCircleRight Bold |> toHtml []
                , Phosphor.caretCircleRight Fill |> toHtml []
                , Phosphor.caretCircleRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleUp model.weight |> toHtml []
                , Phosphor.caretCircleUp Thin |> toHtml []
                , Phosphor.caretCircleUp Light |> toHtml []
                , Phosphor.caretCircleUp Regular |> toHtml []
                , Phosphor.caretCircleUp Bold |> toHtml []
                , Phosphor.caretCircleUp Fill |> toHtml []
                , Phosphor.caretCircleUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretCircleUpDown model.weight |> toHtml []
                , Phosphor.caretCircleUpDown Thin |> toHtml []
                , Phosphor.caretCircleUpDown Light |> toHtml []
                , Phosphor.caretCircleUpDown Regular |> toHtml []
                , Phosphor.caretCircleUpDown Bold |> toHtml []
                , Phosphor.caretCircleUpDown Fill |> toHtml []
                , Phosphor.caretCircleUpDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretDoubleDown model.weight |> toHtml []
                , Phosphor.caretDoubleDown Thin |> toHtml []
                , Phosphor.caretDoubleDown Light |> toHtml []
                , Phosphor.caretDoubleDown Regular |> toHtml []
                , Phosphor.caretDoubleDown Bold |> toHtml []
                , Phosphor.caretDoubleDown Fill |> toHtml []
                , Phosphor.caretDoubleDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretDoubleLeft model.weight |> toHtml []
                , Phosphor.caretDoubleLeft Thin |> toHtml []
                , Phosphor.caretDoubleLeft Light |> toHtml []
                , Phosphor.caretDoubleLeft Regular |> toHtml []
                , Phosphor.caretDoubleLeft Bold |> toHtml []
                , Phosphor.caretDoubleLeft Fill |> toHtml []
                , Phosphor.caretDoubleLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretDoubleRight model.weight |> toHtml []
                , Phosphor.caretDoubleRight Thin |> toHtml []
                , Phosphor.caretDoubleRight Light |> toHtml []
                , Phosphor.caretDoubleRight Regular |> toHtml []
                , Phosphor.caretDoubleRight Bold |> toHtml []
                , Phosphor.caretDoubleRight Fill |> toHtml []
                , Phosphor.caretDoubleRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretDoubleUp model.weight |> toHtml []
                , Phosphor.caretDoubleUp Thin |> toHtml []
                , Phosphor.caretDoubleUp Light |> toHtml []
                , Phosphor.caretDoubleUp Regular |> toHtml []
                , Phosphor.caretDoubleUp Bold |> toHtml []
                , Phosphor.caretDoubleUp Fill |> toHtml []
                , Phosphor.caretDoubleUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretDown model.weight |> toHtml []
                , Phosphor.caretDown Thin |> toHtml []
                , Phosphor.caretDown Light |> toHtml []
                , Phosphor.caretDown Regular |> toHtml []
                , Phosphor.caretDown Bold |> toHtml []
                , Phosphor.caretDown Fill |> toHtml []
                , Phosphor.caretDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretLeft model.weight |> toHtml []
                , Phosphor.caretLeft Thin |> toHtml []
                , Phosphor.caretLeft Light |> toHtml []
                , Phosphor.caretLeft Regular |> toHtml []
                , Phosphor.caretLeft Bold |> toHtml []
                , Phosphor.caretLeft Fill |> toHtml []
                , Phosphor.caretLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretRight model.weight |> toHtml []
                , Phosphor.caretRight Thin |> toHtml []
                , Phosphor.caretRight Light |> toHtml []
                , Phosphor.caretRight Regular |> toHtml []
                , Phosphor.caretRight Bold |> toHtml []
                , Phosphor.caretRight Fill |> toHtml []
                , Phosphor.caretRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretUp model.weight |> toHtml []
                , Phosphor.caretUp Thin |> toHtml []
                , Phosphor.caretUp Light |> toHtml []
                , Phosphor.caretUp Regular |> toHtml []
                , Phosphor.caretUp Bold |> toHtml []
                , Phosphor.caretUp Fill |> toHtml []
                , Phosphor.caretUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.caretUpDown model.weight |> toHtml []
                , Phosphor.caretUpDown Thin |> toHtml []
                , Phosphor.caretUpDown Light |> toHtml []
                , Phosphor.caretUpDown Regular |> toHtml []
                , Phosphor.caretUpDown Bold |> toHtml []
                , Phosphor.caretUpDown Fill |> toHtml []
                , Phosphor.caretUpDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.carrot model.weight |> toHtml []
                , Phosphor.carrot Thin |> toHtml []
                , Phosphor.carrot Light |> toHtml []
                , Phosphor.carrot Regular |> toHtml []
                , Phosphor.carrot Bold |> toHtml []
                , Phosphor.carrot Fill |> toHtml []
                , Phosphor.carrot Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cassetteTape model.weight |> toHtml []
                , Phosphor.cassetteTape Thin |> toHtml []
                , Phosphor.cassetteTape Light |> toHtml []
                , Phosphor.cassetteTape Regular |> toHtml []
                , Phosphor.cassetteTape Bold |> toHtml []
                , Phosphor.cassetteTape Fill |> toHtml []
                , Phosphor.cassetteTape Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.castleTurret model.weight |> toHtml []
                , Phosphor.castleTurret Thin |> toHtml []
                , Phosphor.castleTurret Light |> toHtml []
                , Phosphor.castleTurret Regular |> toHtml []
                , Phosphor.castleTurret Bold |> toHtml []
                , Phosphor.castleTurret Fill |> toHtml []
                , Phosphor.castleTurret Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cat model.weight |> toHtml []
                , Phosphor.cat Thin |> toHtml []
                , Phosphor.cat Light |> toHtml []
                , Phosphor.cat Regular |> toHtml []
                , Phosphor.cat Bold |> toHtml []
                , Phosphor.cat Fill |> toHtml []
                , Phosphor.cat Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cellSignalFull model.weight |> toHtml []
                , Phosphor.cellSignalFull Thin |> toHtml []
                , Phosphor.cellSignalFull Light |> toHtml []
                , Phosphor.cellSignalFull Regular |> toHtml []
                , Phosphor.cellSignalFull Bold |> toHtml []
                , Phosphor.cellSignalFull Fill |> toHtml []
                , Phosphor.cellSignalFull Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cellSignalHigh model.weight |> toHtml []
                , Phosphor.cellSignalHigh Thin |> toHtml []
                , Phosphor.cellSignalHigh Light |> toHtml []
                , Phosphor.cellSignalHigh Regular |> toHtml []
                , Phosphor.cellSignalHigh Bold |> toHtml []
                , Phosphor.cellSignalHigh Fill |> toHtml []
                , Phosphor.cellSignalHigh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cellSignalLow model.weight |> toHtml []
                , Phosphor.cellSignalLow Thin |> toHtml []
                , Phosphor.cellSignalLow Light |> toHtml []
                , Phosphor.cellSignalLow Regular |> toHtml []
                , Phosphor.cellSignalLow Bold |> toHtml []
                , Phosphor.cellSignalLow Fill |> toHtml []
                , Phosphor.cellSignalLow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cellSignalMedium model.weight |> toHtml []
                , Phosphor.cellSignalMedium Thin |> toHtml []
                , Phosphor.cellSignalMedium Light |> toHtml []
                , Phosphor.cellSignalMedium Regular |> toHtml []
                , Phosphor.cellSignalMedium Bold |> toHtml []
                , Phosphor.cellSignalMedium Fill |> toHtml []
                , Phosphor.cellSignalMedium Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cellSignalNone model.weight |> toHtml []
                , Phosphor.cellSignalNone Thin |> toHtml []
                , Phosphor.cellSignalNone Light |> toHtml []
                , Phosphor.cellSignalNone Regular |> toHtml []
                , Phosphor.cellSignalNone Bold |> toHtml []
                , Phosphor.cellSignalNone Fill |> toHtml []
                , Phosphor.cellSignalNone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cellSignalSlash model.weight |> toHtml []
                , Phosphor.cellSignalSlash Thin |> toHtml []
                , Phosphor.cellSignalSlash Light |> toHtml []
                , Phosphor.cellSignalSlash Regular |> toHtml []
                , Phosphor.cellSignalSlash Bold |> toHtml []
                , Phosphor.cellSignalSlash Fill |> toHtml []
                , Phosphor.cellSignalSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cellSignalX model.weight |> toHtml []
                , Phosphor.cellSignalX Thin |> toHtml []
                , Phosphor.cellSignalX Light |> toHtml []
                , Phosphor.cellSignalX Regular |> toHtml []
                , Phosphor.cellSignalX Bold |> toHtml []
                , Phosphor.cellSignalX Fill |> toHtml []
                , Phosphor.cellSignalX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.certificate model.weight |> toHtml []
                , Phosphor.certificate Thin |> toHtml []
                , Phosphor.certificate Light |> toHtml []
                , Phosphor.certificate Regular |> toHtml []
                , Phosphor.certificate Bold |> toHtml []
                , Phosphor.certificate Fill |> toHtml []
                , Phosphor.certificate Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chair model.weight |> toHtml []
                , Phosphor.chair Thin |> toHtml []
                , Phosphor.chair Light |> toHtml []
                , Phosphor.chair Regular |> toHtml []
                , Phosphor.chair Bold |> toHtml []
                , Phosphor.chair Fill |> toHtml []
                , Phosphor.chair Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chalkboard model.weight |> toHtml []
                , Phosphor.chalkboard Thin |> toHtml []
                , Phosphor.chalkboard Light |> toHtml []
                , Phosphor.chalkboard Regular |> toHtml []
                , Phosphor.chalkboard Bold |> toHtml []
                , Phosphor.chalkboard Fill |> toHtml []
                , Phosphor.chalkboard Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chalkboardSimple model.weight |> toHtml []
                , Phosphor.chalkboardSimple Thin |> toHtml []
                , Phosphor.chalkboardSimple Light |> toHtml []
                , Phosphor.chalkboardSimple Regular |> toHtml []
                , Phosphor.chalkboardSimple Bold |> toHtml []
                , Phosphor.chalkboardSimple Fill |> toHtml []
                , Phosphor.chalkboardSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chalkboardTeacher model.weight |> toHtml []
                , Phosphor.chalkboardTeacher Thin |> toHtml []
                , Phosphor.chalkboardTeacher Light |> toHtml []
                , Phosphor.chalkboardTeacher Regular |> toHtml []
                , Phosphor.chalkboardTeacher Bold |> toHtml []
                , Phosphor.chalkboardTeacher Fill |> toHtml []
                , Phosphor.chalkboardTeacher Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.champagne model.weight |> toHtml []
                , Phosphor.champagne Thin |> toHtml []
                , Phosphor.champagne Light |> toHtml []
                , Phosphor.champagne Regular |> toHtml []
                , Phosphor.champagne Bold |> toHtml []
                , Phosphor.champagne Fill |> toHtml []
                , Phosphor.champagne Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chargingStation model.weight |> toHtml []
                , Phosphor.chargingStation Thin |> toHtml []
                , Phosphor.chargingStation Light |> toHtml []
                , Phosphor.chargingStation Regular |> toHtml []
                , Phosphor.chargingStation Bold |> toHtml []
                , Phosphor.chargingStation Fill |> toHtml []
                , Phosphor.chargingStation Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartBar model.weight |> toHtml []
                , Phosphor.chartBar Thin |> toHtml []
                , Phosphor.chartBar Light |> toHtml []
                , Phosphor.chartBar Regular |> toHtml []
                , Phosphor.chartBar Bold |> toHtml []
                , Phosphor.chartBar Fill |> toHtml []
                , Phosphor.chartBar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartBarHorizontal model.weight |> toHtml []
                , Phosphor.chartBarHorizontal Thin |> toHtml []
                , Phosphor.chartBarHorizontal Light |> toHtml []
                , Phosphor.chartBarHorizontal Regular |> toHtml []
                , Phosphor.chartBarHorizontal Bold |> toHtml []
                , Phosphor.chartBarHorizontal Fill |> toHtml []
                , Phosphor.chartBarHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartDonut model.weight |> toHtml []
                , Phosphor.chartDonut Thin |> toHtml []
                , Phosphor.chartDonut Light |> toHtml []
                , Phosphor.chartDonut Regular |> toHtml []
                , Phosphor.chartDonut Bold |> toHtml []
                , Phosphor.chartDonut Fill |> toHtml []
                , Phosphor.chartDonut Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartLine model.weight |> toHtml []
                , Phosphor.chartLine Thin |> toHtml []
                , Phosphor.chartLine Light |> toHtml []
                , Phosphor.chartLine Regular |> toHtml []
                , Phosphor.chartLine Bold |> toHtml []
                , Phosphor.chartLine Fill |> toHtml []
                , Phosphor.chartLine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartLineDown model.weight |> toHtml []
                , Phosphor.chartLineDown Thin |> toHtml []
                , Phosphor.chartLineDown Light |> toHtml []
                , Phosphor.chartLineDown Regular |> toHtml []
                , Phosphor.chartLineDown Bold |> toHtml []
                , Phosphor.chartLineDown Fill |> toHtml []
                , Phosphor.chartLineDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartLineUp model.weight |> toHtml []
                , Phosphor.chartLineUp Thin |> toHtml []
                , Phosphor.chartLineUp Light |> toHtml []
                , Phosphor.chartLineUp Regular |> toHtml []
                , Phosphor.chartLineUp Bold |> toHtml []
                , Phosphor.chartLineUp Fill |> toHtml []
                , Phosphor.chartLineUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartPie model.weight |> toHtml []
                , Phosphor.chartPie Thin |> toHtml []
                , Phosphor.chartPie Light |> toHtml []
                , Phosphor.chartPie Regular |> toHtml []
                , Phosphor.chartPie Bold |> toHtml []
                , Phosphor.chartPie Fill |> toHtml []
                , Phosphor.chartPie Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartPieSlice model.weight |> toHtml []
                , Phosphor.chartPieSlice Thin |> toHtml []
                , Phosphor.chartPieSlice Light |> toHtml []
                , Phosphor.chartPieSlice Regular |> toHtml []
                , Phosphor.chartPieSlice Bold |> toHtml []
                , Phosphor.chartPieSlice Fill |> toHtml []
                , Phosphor.chartPieSlice Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartPolar model.weight |> toHtml []
                , Phosphor.chartPolar Thin |> toHtml []
                , Phosphor.chartPolar Light |> toHtml []
                , Phosphor.chartPolar Regular |> toHtml []
                , Phosphor.chartPolar Bold |> toHtml []
                , Phosphor.chartPolar Fill |> toHtml []
                , Phosphor.chartPolar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chartScatter model.weight |> toHtml []
                , Phosphor.chartScatter Thin |> toHtml []
                , Phosphor.chartScatter Light |> toHtml []
                , Phosphor.chartScatter Regular |> toHtml []
                , Phosphor.chartScatter Bold |> toHtml []
                , Phosphor.chartScatter Fill |> toHtml []
                , Phosphor.chartScatter Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chat model.weight |> toHtml []
                , Phosphor.chat Thin |> toHtml []
                , Phosphor.chat Light |> toHtml []
                , Phosphor.chat Regular |> toHtml []
                , Phosphor.chat Bold |> toHtml []
                , Phosphor.chat Fill |> toHtml []
                , Phosphor.chat Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatCentered model.weight |> toHtml []
                , Phosphor.chatCentered Thin |> toHtml []
                , Phosphor.chatCentered Light |> toHtml []
                , Phosphor.chatCentered Regular |> toHtml []
                , Phosphor.chatCentered Bold |> toHtml []
                , Phosphor.chatCentered Fill |> toHtml []
                , Phosphor.chatCentered Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatCenteredDots model.weight |> toHtml []
                , Phosphor.chatCenteredDots Thin |> toHtml []
                , Phosphor.chatCenteredDots Light |> toHtml []
                , Phosphor.chatCenteredDots Regular |> toHtml []
                , Phosphor.chatCenteredDots Bold |> toHtml []
                , Phosphor.chatCenteredDots Fill |> toHtml []
                , Phosphor.chatCenteredDots Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatCenteredText model.weight |> toHtml []
                , Phosphor.chatCenteredText Thin |> toHtml []
                , Phosphor.chatCenteredText Light |> toHtml []
                , Phosphor.chatCenteredText Regular |> toHtml []
                , Phosphor.chatCenteredText Bold |> toHtml []
                , Phosphor.chatCenteredText Fill |> toHtml []
                , Phosphor.chatCenteredText Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatCircle model.weight |> toHtml []
                , Phosphor.chatCircle Thin |> toHtml []
                , Phosphor.chatCircle Light |> toHtml []
                , Phosphor.chatCircle Regular |> toHtml []
                , Phosphor.chatCircle Bold |> toHtml []
                , Phosphor.chatCircle Fill |> toHtml []
                , Phosphor.chatCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatCircleDots model.weight |> toHtml []
                , Phosphor.chatCircleDots Thin |> toHtml []
                , Phosphor.chatCircleDots Light |> toHtml []
                , Phosphor.chatCircleDots Regular |> toHtml []
                , Phosphor.chatCircleDots Bold |> toHtml []
                , Phosphor.chatCircleDots Fill |> toHtml []
                , Phosphor.chatCircleDots Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatCircleText model.weight |> toHtml []
                , Phosphor.chatCircleText Thin |> toHtml []
                , Phosphor.chatCircleText Light |> toHtml []
                , Phosphor.chatCircleText Regular |> toHtml []
                , Phosphor.chatCircleText Bold |> toHtml []
                , Phosphor.chatCircleText Fill |> toHtml []
                , Phosphor.chatCircleText Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatDots model.weight |> toHtml []
                , Phosphor.chatDots Thin |> toHtml []
                , Phosphor.chatDots Light |> toHtml []
                , Phosphor.chatDots Regular |> toHtml []
                , Phosphor.chatDots Bold |> toHtml []
                , Phosphor.chatDots Fill |> toHtml []
                , Phosphor.chatDots Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatTeardrop model.weight |> toHtml []
                , Phosphor.chatTeardrop Thin |> toHtml []
                , Phosphor.chatTeardrop Light |> toHtml []
                , Phosphor.chatTeardrop Regular |> toHtml []
                , Phosphor.chatTeardrop Bold |> toHtml []
                , Phosphor.chatTeardrop Fill |> toHtml []
                , Phosphor.chatTeardrop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatTeardropDots model.weight |> toHtml []
                , Phosphor.chatTeardropDots Thin |> toHtml []
                , Phosphor.chatTeardropDots Light |> toHtml []
                , Phosphor.chatTeardropDots Regular |> toHtml []
                , Phosphor.chatTeardropDots Bold |> toHtml []
                , Phosphor.chatTeardropDots Fill |> toHtml []
                , Phosphor.chatTeardropDots Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatTeardropText model.weight |> toHtml []
                , Phosphor.chatTeardropText Thin |> toHtml []
                , Phosphor.chatTeardropText Light |> toHtml []
                , Phosphor.chatTeardropText Regular |> toHtml []
                , Phosphor.chatTeardropText Bold |> toHtml []
                , Phosphor.chatTeardropText Fill |> toHtml []
                , Phosphor.chatTeardropText Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatText model.weight |> toHtml []
                , Phosphor.chatText Thin |> toHtml []
                , Phosphor.chatText Light |> toHtml []
                , Phosphor.chatText Regular |> toHtml []
                , Phosphor.chatText Bold |> toHtml []
                , Phosphor.chatText Fill |> toHtml []
                , Phosphor.chatText Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chats model.weight |> toHtml []
                , Phosphor.chats Thin |> toHtml []
                , Phosphor.chats Light |> toHtml []
                , Phosphor.chats Regular |> toHtml []
                , Phosphor.chats Bold |> toHtml []
                , Phosphor.chats Fill |> toHtml []
                , Phosphor.chats Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatsCircle model.weight |> toHtml []
                , Phosphor.chatsCircle Thin |> toHtml []
                , Phosphor.chatsCircle Light |> toHtml []
                , Phosphor.chatsCircle Regular |> toHtml []
                , Phosphor.chatsCircle Bold |> toHtml []
                , Phosphor.chatsCircle Fill |> toHtml []
                , Phosphor.chatsCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.chatsTeardrop model.weight |> toHtml []
                , Phosphor.chatsTeardrop Thin |> toHtml []
                , Phosphor.chatsTeardrop Light |> toHtml []
                , Phosphor.chatsTeardrop Regular |> toHtml []
                , Phosphor.chatsTeardrop Bold |> toHtml []
                , Phosphor.chatsTeardrop Fill |> toHtml []
                , Phosphor.chatsTeardrop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.check model.weight |> toHtml []
                , Phosphor.check Thin |> toHtml []
                , Phosphor.check Light |> toHtml []
                , Phosphor.check Regular |> toHtml []
                , Phosphor.check Bold |> toHtml []
                , Phosphor.check Fill |> toHtml []
                , Phosphor.check Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.checkCircle model.weight |> toHtml []
                , Phosphor.checkCircle Thin |> toHtml []
                , Phosphor.checkCircle Light |> toHtml []
                , Phosphor.checkCircle Regular |> toHtml []
                , Phosphor.checkCircle Bold |> toHtml []
                , Phosphor.checkCircle Fill |> toHtml []
                , Phosphor.checkCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.checkFat model.weight |> toHtml []
                , Phosphor.checkFat Thin |> toHtml []
                , Phosphor.checkFat Light |> toHtml []
                , Phosphor.checkFat Regular |> toHtml []
                , Phosphor.checkFat Bold |> toHtml []
                , Phosphor.checkFat Fill |> toHtml []
                , Phosphor.checkFat Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.checkSquare model.weight |> toHtml []
                , Phosphor.checkSquare Thin |> toHtml []
                , Phosphor.checkSquare Light |> toHtml []
                , Phosphor.checkSquare Regular |> toHtml []
                , Phosphor.checkSquare Bold |> toHtml []
                , Phosphor.checkSquare Fill |> toHtml []
                , Phosphor.checkSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.checkSquareOffset model.weight |> toHtml []
                , Phosphor.checkSquareOffset Thin |> toHtml []
                , Phosphor.checkSquareOffset Light |> toHtml []
                , Phosphor.checkSquareOffset Regular |> toHtml []
                , Phosphor.checkSquareOffset Bold |> toHtml []
                , Phosphor.checkSquareOffset Fill |> toHtml []
                , Phosphor.checkSquareOffset Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.checks model.weight |> toHtml []
                , Phosphor.checks Thin |> toHtml []
                , Phosphor.checks Light |> toHtml []
                , Phosphor.checks Regular |> toHtml []
                , Phosphor.checks Bold |> toHtml []
                , Phosphor.checks Fill |> toHtml []
                , Phosphor.checks Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.church model.weight |> toHtml []
                , Phosphor.church Thin |> toHtml []
                , Phosphor.church Light |> toHtml []
                , Phosphor.church Regular |> toHtml []
                , Phosphor.church Bold |> toHtml []
                , Phosphor.church Fill |> toHtml []
                , Phosphor.church Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circle model.weight |> toHtml []
                , Phosphor.circle Thin |> toHtml []
                , Phosphor.circle Light |> toHtml []
                , Phosphor.circle Regular |> toHtml []
                , Phosphor.circle Bold |> toHtml []
                , Phosphor.circle Fill |> toHtml []
                , Phosphor.circle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circleDashed model.weight |> toHtml []
                , Phosphor.circleDashed Thin |> toHtml []
                , Phosphor.circleDashed Light |> toHtml []
                , Phosphor.circleDashed Regular |> toHtml []
                , Phosphor.circleDashed Bold |> toHtml []
                , Phosphor.circleDashed Fill |> toHtml []
                , Phosphor.circleDashed Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circleHalf model.weight |> toHtml []
                , Phosphor.circleHalf Thin |> toHtml []
                , Phosphor.circleHalf Light |> toHtml []
                , Phosphor.circleHalf Regular |> toHtml []
                , Phosphor.circleHalf Bold |> toHtml []
                , Phosphor.circleHalf Fill |> toHtml []
                , Phosphor.circleHalf Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circleHalfTilt model.weight |> toHtml []
                , Phosphor.circleHalfTilt Thin |> toHtml []
                , Phosphor.circleHalfTilt Light |> toHtml []
                , Phosphor.circleHalfTilt Regular |> toHtml []
                , Phosphor.circleHalfTilt Bold |> toHtml []
                , Phosphor.circleHalfTilt Fill |> toHtml []
                , Phosphor.circleHalfTilt Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circleNotch model.weight |> toHtml []
                , Phosphor.circleNotch Thin |> toHtml []
                , Phosphor.circleNotch Light |> toHtml []
                , Phosphor.circleNotch Regular |> toHtml []
                , Phosphor.circleNotch Bold |> toHtml []
                , Phosphor.circleNotch Fill |> toHtml []
                , Phosphor.circleNotch Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circlesFour model.weight |> toHtml []
                , Phosphor.circlesFour Thin |> toHtml []
                , Phosphor.circlesFour Light |> toHtml []
                , Phosphor.circlesFour Regular |> toHtml []
                , Phosphor.circlesFour Bold |> toHtml []
                , Phosphor.circlesFour Fill |> toHtml []
                , Phosphor.circlesFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circlesThree model.weight |> toHtml []
                , Phosphor.circlesThree Thin |> toHtml []
                , Phosphor.circlesThree Light |> toHtml []
                , Phosphor.circlesThree Regular |> toHtml []
                , Phosphor.circlesThree Bold |> toHtml []
                , Phosphor.circlesThree Fill |> toHtml []
                , Phosphor.circlesThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circlesThreePlus model.weight |> toHtml []
                , Phosphor.circlesThreePlus Thin |> toHtml []
                , Phosphor.circlesThreePlus Light |> toHtml []
                , Phosphor.circlesThreePlus Regular |> toHtml []
                , Phosphor.circlesThreePlus Bold |> toHtml []
                , Phosphor.circlesThreePlus Fill |> toHtml []
                , Phosphor.circlesThreePlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.circuitry model.weight |> toHtml []
                , Phosphor.circuitry Thin |> toHtml []
                , Phosphor.circuitry Light |> toHtml []
                , Phosphor.circuitry Regular |> toHtml []
                , Phosphor.circuitry Bold |> toHtml []
                , Phosphor.circuitry Fill |> toHtml []
                , Phosphor.circuitry Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.clipboard model.weight |> toHtml []
                , Phosphor.clipboard Thin |> toHtml []
                , Phosphor.clipboard Light |> toHtml []
                , Phosphor.clipboard Regular |> toHtml []
                , Phosphor.clipboard Bold |> toHtml []
                , Phosphor.clipboard Fill |> toHtml []
                , Phosphor.clipboard Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.clipboardText model.weight |> toHtml []
                , Phosphor.clipboardText Thin |> toHtml []
                , Phosphor.clipboardText Light |> toHtml []
                , Phosphor.clipboardText Regular |> toHtml []
                , Phosphor.clipboardText Bold |> toHtml []
                , Phosphor.clipboardText Fill |> toHtml []
                , Phosphor.clipboardText Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.clockAfternoon model.weight |> toHtml []
                , Phosphor.clockAfternoon Thin |> toHtml []
                , Phosphor.clockAfternoon Light |> toHtml []
                , Phosphor.clockAfternoon Regular |> toHtml []
                , Phosphor.clockAfternoon Bold |> toHtml []
                , Phosphor.clockAfternoon Fill |> toHtml []
                , Phosphor.clockAfternoon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.clock model.weight |> toHtml []
                , Phosphor.clock Thin |> toHtml []
                , Phosphor.clock Light |> toHtml []
                , Phosphor.clock Regular |> toHtml []
                , Phosphor.clock Bold |> toHtml []
                , Phosphor.clock Fill |> toHtml []
                , Phosphor.clock Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.clockClockwise model.weight |> toHtml []
                , Phosphor.clockClockwise Thin |> toHtml []
                , Phosphor.clockClockwise Light |> toHtml []
                , Phosphor.clockClockwise Regular |> toHtml []
                , Phosphor.clockClockwise Bold |> toHtml []
                , Phosphor.clockClockwise Fill |> toHtml []
                , Phosphor.clockClockwise Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.clockCountdown model.weight |> toHtml []
                , Phosphor.clockCountdown Thin |> toHtml []
                , Phosphor.clockCountdown Light |> toHtml []
                , Phosphor.clockCountdown Regular |> toHtml []
                , Phosphor.clockCountdown Bold |> toHtml []
                , Phosphor.clockCountdown Fill |> toHtml []
                , Phosphor.clockCountdown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.clockCounterClockwise model.weight |> toHtml []
                , Phosphor.clockCounterClockwise Thin |> toHtml []
                , Phosphor.clockCounterClockwise Light |> toHtml []
                , Phosphor.clockCounterClockwise Regular |> toHtml []
                , Phosphor.clockCounterClockwise Bold |> toHtml []
                , Phosphor.clockCounterClockwise Fill |> toHtml []
                , Phosphor.clockCounterClockwise Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.closedCaptioning model.weight |> toHtml []
                , Phosphor.closedCaptioning Thin |> toHtml []
                , Phosphor.closedCaptioning Light |> toHtml []
                , Phosphor.closedCaptioning Regular |> toHtml []
                , Phosphor.closedCaptioning Bold |> toHtml []
                , Phosphor.closedCaptioning Fill |> toHtml []
                , Phosphor.closedCaptioning Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudArrowDown model.weight |> toHtml []
                , Phosphor.cloudArrowDown Thin |> toHtml []
                , Phosphor.cloudArrowDown Light |> toHtml []
                , Phosphor.cloudArrowDown Regular |> toHtml []
                , Phosphor.cloudArrowDown Bold |> toHtml []
                , Phosphor.cloudArrowDown Fill |> toHtml []
                , Phosphor.cloudArrowDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudArrowUp model.weight |> toHtml []
                , Phosphor.cloudArrowUp Thin |> toHtml []
                , Phosphor.cloudArrowUp Light |> toHtml []
                , Phosphor.cloudArrowUp Regular |> toHtml []
                , Phosphor.cloudArrowUp Bold |> toHtml []
                , Phosphor.cloudArrowUp Fill |> toHtml []
                , Phosphor.cloudArrowUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloud model.weight |> toHtml []
                , Phosphor.cloud Thin |> toHtml []
                , Phosphor.cloud Light |> toHtml []
                , Phosphor.cloud Regular |> toHtml []
                , Phosphor.cloud Bold |> toHtml []
                , Phosphor.cloud Fill |> toHtml []
                , Phosphor.cloud Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudCheck model.weight |> toHtml []
                , Phosphor.cloudCheck Thin |> toHtml []
                , Phosphor.cloudCheck Light |> toHtml []
                , Phosphor.cloudCheck Regular |> toHtml []
                , Phosphor.cloudCheck Bold |> toHtml []
                , Phosphor.cloudCheck Fill |> toHtml []
                , Phosphor.cloudCheck Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudFog model.weight |> toHtml []
                , Phosphor.cloudFog Thin |> toHtml []
                , Phosphor.cloudFog Light |> toHtml []
                , Phosphor.cloudFog Regular |> toHtml []
                , Phosphor.cloudFog Bold |> toHtml []
                , Phosphor.cloudFog Fill |> toHtml []
                , Phosphor.cloudFog Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudLightning model.weight |> toHtml []
                , Phosphor.cloudLightning Thin |> toHtml []
                , Phosphor.cloudLightning Light |> toHtml []
                , Phosphor.cloudLightning Regular |> toHtml []
                , Phosphor.cloudLightning Bold |> toHtml []
                , Phosphor.cloudLightning Fill |> toHtml []
                , Phosphor.cloudLightning Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudMoon model.weight |> toHtml []
                , Phosphor.cloudMoon Thin |> toHtml []
                , Phosphor.cloudMoon Light |> toHtml []
                , Phosphor.cloudMoon Regular |> toHtml []
                , Phosphor.cloudMoon Bold |> toHtml []
                , Phosphor.cloudMoon Fill |> toHtml []
                , Phosphor.cloudMoon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudRain model.weight |> toHtml []
                , Phosphor.cloudRain Thin |> toHtml []
                , Phosphor.cloudRain Light |> toHtml []
                , Phosphor.cloudRain Regular |> toHtml []
                , Phosphor.cloudRain Bold |> toHtml []
                , Phosphor.cloudRain Fill |> toHtml []
                , Phosphor.cloudRain Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudSlash model.weight |> toHtml []
                , Phosphor.cloudSlash Thin |> toHtml []
                , Phosphor.cloudSlash Light |> toHtml []
                , Phosphor.cloudSlash Regular |> toHtml []
                , Phosphor.cloudSlash Bold |> toHtml []
                , Phosphor.cloudSlash Fill |> toHtml []
                , Phosphor.cloudSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudSnow model.weight |> toHtml []
                , Phosphor.cloudSnow Thin |> toHtml []
                , Phosphor.cloudSnow Light |> toHtml []
                , Phosphor.cloudSnow Regular |> toHtml []
                , Phosphor.cloudSnow Bold |> toHtml []
                , Phosphor.cloudSnow Fill |> toHtml []
                , Phosphor.cloudSnow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudSun model.weight |> toHtml []
                , Phosphor.cloudSun Thin |> toHtml []
                , Phosphor.cloudSun Light |> toHtml []
                , Phosphor.cloudSun Regular |> toHtml []
                , Phosphor.cloudSun Bold |> toHtml []
                , Phosphor.cloudSun Fill |> toHtml []
                , Phosphor.cloudSun Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudWarning model.weight |> toHtml []
                , Phosphor.cloudWarning Thin |> toHtml []
                , Phosphor.cloudWarning Light |> toHtml []
                , Phosphor.cloudWarning Regular |> toHtml []
                , Phosphor.cloudWarning Bold |> toHtml []
                , Phosphor.cloudWarning Fill |> toHtml []
                , Phosphor.cloudWarning Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cloudX model.weight |> toHtml []
                , Phosphor.cloudX Thin |> toHtml []
                , Phosphor.cloudX Light |> toHtml []
                , Phosphor.cloudX Regular |> toHtml []
                , Phosphor.cloudX Bold |> toHtml []
                , Phosphor.cloudX Fill |> toHtml []
                , Phosphor.cloudX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.club model.weight |> toHtml []
                , Phosphor.club Thin |> toHtml []
                , Phosphor.club Light |> toHtml []
                , Phosphor.club Regular |> toHtml []
                , Phosphor.club Bold |> toHtml []
                , Phosphor.club Fill |> toHtml []
                , Phosphor.club Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.coatHanger model.weight |> toHtml []
                , Phosphor.coatHanger Thin |> toHtml []
                , Phosphor.coatHanger Light |> toHtml []
                , Phosphor.coatHanger Regular |> toHtml []
                , Phosphor.coatHanger Bold |> toHtml []
                , Phosphor.coatHanger Fill |> toHtml []
                , Phosphor.coatHanger Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.codaLogo model.weight |> toHtml []
                , Phosphor.codaLogo Thin |> toHtml []
                , Phosphor.codaLogo Light |> toHtml []
                , Phosphor.codaLogo Regular |> toHtml []
                , Phosphor.codaLogo Bold |> toHtml []
                , Phosphor.codaLogo Fill |> toHtml []
                , Phosphor.codaLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.codeBlock model.weight |> toHtml []
                , Phosphor.codeBlock Thin |> toHtml []
                , Phosphor.codeBlock Light |> toHtml []
                , Phosphor.codeBlock Regular |> toHtml []
                , Phosphor.codeBlock Bold |> toHtml []
                , Phosphor.codeBlock Fill |> toHtml []
                , Phosphor.codeBlock Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.code model.weight |> toHtml []
                , Phosphor.code Thin |> toHtml []
                , Phosphor.code Light |> toHtml []
                , Phosphor.code Regular |> toHtml []
                , Phosphor.code Bold |> toHtml []
                , Phosphor.code Fill |> toHtml []
                , Phosphor.code Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.codeSimple model.weight |> toHtml []
                , Phosphor.codeSimple Thin |> toHtml []
                , Phosphor.codeSimple Light |> toHtml []
                , Phosphor.codeSimple Regular |> toHtml []
                , Phosphor.codeSimple Bold |> toHtml []
                , Phosphor.codeSimple Fill |> toHtml []
                , Phosphor.codeSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.codepenLogo model.weight |> toHtml []
                , Phosphor.codepenLogo Thin |> toHtml []
                , Phosphor.codepenLogo Light |> toHtml []
                , Phosphor.codepenLogo Regular |> toHtml []
                , Phosphor.codepenLogo Bold |> toHtml []
                , Phosphor.codepenLogo Fill |> toHtml []
                , Phosphor.codepenLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.codesandboxLogo model.weight |> toHtml []
                , Phosphor.codesandboxLogo Thin |> toHtml []
                , Phosphor.codesandboxLogo Light |> toHtml []
                , Phosphor.codesandboxLogo Regular |> toHtml []
                , Phosphor.codesandboxLogo Bold |> toHtml []
                , Phosphor.codesandboxLogo Fill |> toHtml []
                , Phosphor.codesandboxLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.coffee model.weight |> toHtml []
                , Phosphor.coffee Thin |> toHtml []
                , Phosphor.coffee Light |> toHtml []
                , Phosphor.coffee Regular |> toHtml []
                , Phosphor.coffee Bold |> toHtml []
                , Phosphor.coffee Fill |> toHtml []
                , Phosphor.coffee Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.coin model.weight |> toHtml []
                , Phosphor.coin Thin |> toHtml []
                , Phosphor.coin Light |> toHtml []
                , Phosphor.coin Regular |> toHtml []
                , Phosphor.coin Bold |> toHtml []
                , Phosphor.coin Fill |> toHtml []
                , Phosphor.coin Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.coinVertical model.weight |> toHtml []
                , Phosphor.coinVertical Thin |> toHtml []
                , Phosphor.coinVertical Light |> toHtml []
                , Phosphor.coinVertical Regular |> toHtml []
                , Phosphor.coinVertical Bold |> toHtml []
                , Phosphor.coinVertical Fill |> toHtml []
                , Phosphor.coinVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.coins model.weight |> toHtml []
                , Phosphor.coins Thin |> toHtml []
                , Phosphor.coins Light |> toHtml []
                , Phosphor.coins Regular |> toHtml []
                , Phosphor.coins Bold |> toHtml []
                , Phosphor.coins Fill |> toHtml []
                , Phosphor.coins Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.columns model.weight |> toHtml []
                , Phosphor.columns Thin |> toHtml []
                , Phosphor.columns Light |> toHtml []
                , Phosphor.columns Regular |> toHtml []
                , Phosphor.columns Bold |> toHtml []
                , Phosphor.columns Fill |> toHtml []
                , Phosphor.columns Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.command model.weight |> toHtml []
                , Phosphor.command Thin |> toHtml []
                , Phosphor.command Light |> toHtml []
                , Phosphor.command Regular |> toHtml []
                , Phosphor.command Bold |> toHtml []
                , Phosphor.command Fill |> toHtml []
                , Phosphor.command Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.compass model.weight |> toHtml []
                , Phosphor.compass Thin |> toHtml []
                , Phosphor.compass Light |> toHtml []
                , Phosphor.compass Regular |> toHtml []
                , Phosphor.compass Bold |> toHtml []
                , Phosphor.compass Fill |> toHtml []
                , Phosphor.compass Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.compassTool model.weight |> toHtml []
                , Phosphor.compassTool Thin |> toHtml []
                , Phosphor.compassTool Light |> toHtml []
                , Phosphor.compassTool Regular |> toHtml []
                , Phosphor.compassTool Bold |> toHtml []
                , Phosphor.compassTool Fill |> toHtml []
                , Phosphor.compassTool Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.computerTower model.weight |> toHtml []
                , Phosphor.computerTower Thin |> toHtml []
                , Phosphor.computerTower Light |> toHtml []
                , Phosphor.computerTower Regular |> toHtml []
                , Phosphor.computerTower Bold |> toHtml []
                , Phosphor.computerTower Fill |> toHtml []
                , Phosphor.computerTower Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.confetti model.weight |> toHtml []
                , Phosphor.confetti Thin |> toHtml []
                , Phosphor.confetti Light |> toHtml []
                , Phosphor.confetti Regular |> toHtml []
                , Phosphor.confetti Bold |> toHtml []
                , Phosphor.confetti Fill |> toHtml []
                , Phosphor.confetti Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.contactlessPayment model.weight |> toHtml []
                , Phosphor.contactlessPayment Thin |> toHtml []
                , Phosphor.contactlessPayment Light |> toHtml []
                , Phosphor.contactlessPayment Regular |> toHtml []
                , Phosphor.contactlessPayment Bold |> toHtml []
                , Phosphor.contactlessPayment Fill |> toHtml []
                , Phosphor.contactlessPayment Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.control model.weight |> toHtml []
                , Phosphor.control Thin |> toHtml []
                , Phosphor.control Light |> toHtml []
                , Phosphor.control Regular |> toHtml []
                , Phosphor.control Bold |> toHtml []
                , Phosphor.control Fill |> toHtml []
                , Phosphor.control Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cookie model.weight |> toHtml []
                , Phosphor.cookie Thin |> toHtml []
                , Phosphor.cookie Light |> toHtml []
                , Phosphor.cookie Regular |> toHtml []
                , Phosphor.cookie Bold |> toHtml []
                , Phosphor.cookie Fill |> toHtml []
                , Phosphor.cookie Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cookingPot model.weight |> toHtml []
                , Phosphor.cookingPot Thin |> toHtml []
                , Phosphor.cookingPot Light |> toHtml []
                , Phosphor.cookingPot Regular |> toHtml []
                , Phosphor.cookingPot Bold |> toHtml []
                , Phosphor.cookingPot Fill |> toHtml []
                , Phosphor.cookingPot Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.copy model.weight |> toHtml []
                , Phosphor.copy Thin |> toHtml []
                , Phosphor.copy Light |> toHtml []
                , Phosphor.copy Regular |> toHtml []
                , Phosphor.copy Bold |> toHtml []
                , Phosphor.copy Fill |> toHtml []
                , Phosphor.copy Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.copySimple model.weight |> toHtml []
                , Phosphor.copySimple Thin |> toHtml []
                , Phosphor.copySimple Light |> toHtml []
                , Phosphor.copySimple Regular |> toHtml []
                , Phosphor.copySimple Bold |> toHtml []
                , Phosphor.copySimple Fill |> toHtml []
                , Phosphor.copySimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.copyleft model.weight |> toHtml []
                , Phosphor.copyleft Thin |> toHtml []
                , Phosphor.copyleft Light |> toHtml []
                , Phosphor.copyleft Regular |> toHtml []
                , Phosphor.copyleft Bold |> toHtml []
                , Phosphor.copyleft Fill |> toHtml []
                , Phosphor.copyleft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.copyright model.weight |> toHtml []
                , Phosphor.copyright Thin |> toHtml []
                , Phosphor.copyright Light |> toHtml []
                , Phosphor.copyright Regular |> toHtml []
                , Phosphor.copyright Bold |> toHtml []
                , Phosphor.copyright Fill |> toHtml []
                , Phosphor.copyright Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cornersIn model.weight |> toHtml []
                , Phosphor.cornersIn Thin |> toHtml []
                , Phosphor.cornersIn Light |> toHtml []
                , Phosphor.cornersIn Regular |> toHtml []
                , Phosphor.cornersIn Bold |> toHtml []
                , Phosphor.cornersIn Fill |> toHtml []
                , Phosphor.cornersIn Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cornersOut model.weight |> toHtml []
                , Phosphor.cornersOut Thin |> toHtml []
                , Phosphor.cornersOut Light |> toHtml []
                , Phosphor.cornersOut Regular |> toHtml []
                , Phosphor.cornersOut Bold |> toHtml []
                , Phosphor.cornersOut Fill |> toHtml []
                , Phosphor.cornersOut Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.couch model.weight |> toHtml []
                , Phosphor.couch Thin |> toHtml []
                , Phosphor.couch Light |> toHtml []
                , Phosphor.couch Regular |> toHtml []
                , Phosphor.couch Bold |> toHtml []
                , Phosphor.couch Fill |> toHtml []
                , Phosphor.couch Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cpu model.weight |> toHtml []
                , Phosphor.cpu Thin |> toHtml []
                , Phosphor.cpu Light |> toHtml []
                , Phosphor.cpu Regular |> toHtml []
                , Phosphor.cpu Bold |> toHtml []
                , Phosphor.cpu Fill |> toHtml []
                , Phosphor.cpu Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.creditCard model.weight |> toHtml []
                , Phosphor.creditCard Thin |> toHtml []
                , Phosphor.creditCard Light |> toHtml []
                , Phosphor.creditCard Regular |> toHtml []
                , Phosphor.creditCard Bold |> toHtml []
                , Phosphor.creditCard Fill |> toHtml []
                , Phosphor.creditCard Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.crop model.weight |> toHtml []
                , Phosphor.crop Thin |> toHtml []
                , Phosphor.crop Light |> toHtml []
                , Phosphor.crop Regular |> toHtml []
                , Phosphor.crop Bold |> toHtml []
                , Phosphor.crop Fill |> toHtml []
                , Phosphor.crop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cross model.weight |> toHtml []
                , Phosphor.cross Thin |> toHtml []
                , Phosphor.cross Light |> toHtml []
                , Phosphor.cross Regular |> toHtml []
                , Phosphor.cross Bold |> toHtml []
                , Phosphor.cross Fill |> toHtml []
                , Phosphor.cross Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.crosshair model.weight |> toHtml []
                , Phosphor.crosshair Thin |> toHtml []
                , Phosphor.crosshair Light |> toHtml []
                , Phosphor.crosshair Regular |> toHtml []
                , Phosphor.crosshair Bold |> toHtml []
                , Phosphor.crosshair Fill |> toHtml []
                , Phosphor.crosshair Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.crosshairSimple model.weight |> toHtml []
                , Phosphor.crosshairSimple Thin |> toHtml []
                , Phosphor.crosshairSimple Light |> toHtml []
                , Phosphor.crosshairSimple Regular |> toHtml []
                , Phosphor.crosshairSimple Bold |> toHtml []
                , Phosphor.crosshairSimple Fill |> toHtml []
                , Phosphor.crosshairSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.crown model.weight |> toHtml []
                , Phosphor.crown Thin |> toHtml []
                , Phosphor.crown Light |> toHtml []
                , Phosphor.crown Regular |> toHtml []
                , Phosphor.crown Bold |> toHtml []
                , Phosphor.crown Fill |> toHtml []
                , Phosphor.crown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.crownSimple model.weight |> toHtml []
                , Phosphor.crownSimple Thin |> toHtml []
                , Phosphor.crownSimple Light |> toHtml []
                , Phosphor.crownSimple Regular |> toHtml []
                , Phosphor.crownSimple Bold |> toHtml []
                , Phosphor.crownSimple Fill |> toHtml []
                , Phosphor.crownSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cube model.weight |> toHtml []
                , Phosphor.cube Thin |> toHtml []
                , Phosphor.cube Light |> toHtml []
                , Phosphor.cube Regular |> toHtml []
                , Phosphor.cube Bold |> toHtml []
                , Phosphor.cube Fill |> toHtml []
                , Phosphor.cube Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cubeFocus model.weight |> toHtml []
                , Phosphor.cubeFocus Thin |> toHtml []
                , Phosphor.cubeFocus Light |> toHtml []
                , Phosphor.cubeFocus Regular |> toHtml []
                , Phosphor.cubeFocus Bold |> toHtml []
                , Phosphor.cubeFocus Fill |> toHtml []
                , Phosphor.cubeFocus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cubeTransparent model.weight |> toHtml []
                , Phosphor.cubeTransparent Thin |> toHtml []
                , Phosphor.cubeTransparent Light |> toHtml []
                , Phosphor.cubeTransparent Regular |> toHtml []
                , Phosphor.cubeTransparent Bold |> toHtml []
                , Phosphor.cubeTransparent Fill |> toHtml []
                , Phosphor.cubeTransparent Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyBtc model.weight |> toHtml []
                , Phosphor.currencyBtc Thin |> toHtml []
                , Phosphor.currencyBtc Light |> toHtml []
                , Phosphor.currencyBtc Regular |> toHtml []
                , Phosphor.currencyBtc Bold |> toHtml []
                , Phosphor.currencyBtc Fill |> toHtml []
                , Phosphor.currencyBtc Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyCircleDollar model.weight |> toHtml []
                , Phosphor.currencyCircleDollar Thin |> toHtml []
                , Phosphor.currencyCircleDollar Light |> toHtml []
                , Phosphor.currencyCircleDollar Regular |> toHtml []
                , Phosphor.currencyCircleDollar Bold |> toHtml []
                , Phosphor.currencyCircleDollar Fill |> toHtml []
                , Phosphor.currencyCircleDollar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyCny model.weight |> toHtml []
                , Phosphor.currencyCny Thin |> toHtml []
                , Phosphor.currencyCny Light |> toHtml []
                , Phosphor.currencyCny Regular |> toHtml []
                , Phosphor.currencyCny Bold |> toHtml []
                , Phosphor.currencyCny Fill |> toHtml []
                , Phosphor.currencyCny Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyDollar model.weight |> toHtml []
                , Phosphor.currencyDollar Thin |> toHtml []
                , Phosphor.currencyDollar Light |> toHtml []
                , Phosphor.currencyDollar Regular |> toHtml []
                , Phosphor.currencyDollar Bold |> toHtml []
                , Phosphor.currencyDollar Fill |> toHtml []
                , Phosphor.currencyDollar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyDollarSimple model.weight |> toHtml []
                , Phosphor.currencyDollarSimple Thin |> toHtml []
                , Phosphor.currencyDollarSimple Light |> toHtml []
                , Phosphor.currencyDollarSimple Regular |> toHtml []
                , Phosphor.currencyDollarSimple Bold |> toHtml []
                , Phosphor.currencyDollarSimple Fill |> toHtml []
                , Phosphor.currencyDollarSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyEth model.weight |> toHtml []
                , Phosphor.currencyEth Thin |> toHtml []
                , Phosphor.currencyEth Light |> toHtml []
                , Phosphor.currencyEth Regular |> toHtml []
                , Phosphor.currencyEth Bold |> toHtml []
                , Phosphor.currencyEth Fill |> toHtml []
                , Phosphor.currencyEth Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyEur model.weight |> toHtml []
                , Phosphor.currencyEur Thin |> toHtml []
                , Phosphor.currencyEur Light |> toHtml []
                , Phosphor.currencyEur Regular |> toHtml []
                , Phosphor.currencyEur Bold |> toHtml []
                , Phosphor.currencyEur Fill |> toHtml []
                , Phosphor.currencyEur Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyGbp model.weight |> toHtml []
                , Phosphor.currencyGbp Thin |> toHtml []
                , Phosphor.currencyGbp Light |> toHtml []
                , Phosphor.currencyGbp Regular |> toHtml []
                , Phosphor.currencyGbp Bold |> toHtml []
                , Phosphor.currencyGbp Fill |> toHtml []
                , Phosphor.currencyGbp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyInr model.weight |> toHtml []
                , Phosphor.currencyInr Thin |> toHtml []
                , Phosphor.currencyInr Light |> toHtml []
                , Phosphor.currencyInr Regular |> toHtml []
                , Phosphor.currencyInr Bold |> toHtml []
                , Phosphor.currencyInr Fill |> toHtml []
                , Phosphor.currencyInr Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyJpy model.weight |> toHtml []
                , Phosphor.currencyJpy Thin |> toHtml []
                , Phosphor.currencyJpy Light |> toHtml []
                , Phosphor.currencyJpy Regular |> toHtml []
                , Phosphor.currencyJpy Bold |> toHtml []
                , Phosphor.currencyJpy Fill |> toHtml []
                , Phosphor.currencyJpy Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyKrw model.weight |> toHtml []
                , Phosphor.currencyKrw Thin |> toHtml []
                , Phosphor.currencyKrw Light |> toHtml []
                , Phosphor.currencyKrw Regular |> toHtml []
                , Phosphor.currencyKrw Bold |> toHtml []
                , Phosphor.currencyKrw Fill |> toHtml []
                , Phosphor.currencyKrw Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyKzt model.weight |> toHtml []
                , Phosphor.currencyKzt Thin |> toHtml []
                , Phosphor.currencyKzt Light |> toHtml []
                , Phosphor.currencyKzt Regular |> toHtml []
                , Phosphor.currencyKzt Bold |> toHtml []
                , Phosphor.currencyKzt Fill |> toHtml []
                , Phosphor.currencyKzt Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyNgn model.weight |> toHtml []
                , Phosphor.currencyNgn Thin |> toHtml []
                , Phosphor.currencyNgn Light |> toHtml []
                , Phosphor.currencyNgn Regular |> toHtml []
                , Phosphor.currencyNgn Bold |> toHtml []
                , Phosphor.currencyNgn Fill |> toHtml []
                , Phosphor.currencyNgn Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.currencyRub model.weight |> toHtml []
                , Phosphor.currencyRub Thin |> toHtml []
                , Phosphor.currencyRub Light |> toHtml []
                , Phosphor.currencyRub Regular |> toHtml []
                , Phosphor.currencyRub Bold |> toHtml []
                , Phosphor.currencyRub Fill |> toHtml []
                , Phosphor.currencyRub Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cursor model.weight |> toHtml []
                , Phosphor.cursor Thin |> toHtml []
                , Phosphor.cursor Light |> toHtml []
                , Phosphor.cursor Regular |> toHtml []
                , Phosphor.cursor Bold |> toHtml []
                , Phosphor.cursor Fill |> toHtml []
                , Phosphor.cursor Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cursorClick model.weight |> toHtml []
                , Phosphor.cursorClick Thin |> toHtml []
                , Phosphor.cursorClick Light |> toHtml []
                , Phosphor.cursorClick Regular |> toHtml []
                , Phosphor.cursorClick Bold |> toHtml []
                , Phosphor.cursorClick Fill |> toHtml []
                , Phosphor.cursorClick Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cursorText model.weight |> toHtml []
                , Phosphor.cursorText Thin |> toHtml []
                , Phosphor.cursorText Light |> toHtml []
                , Phosphor.cursorText Regular |> toHtml []
                , Phosphor.cursorText Bold |> toHtml []
                , Phosphor.cursorText Fill |> toHtml []
                , Phosphor.cursorText Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.cylinder model.weight |> toHtml []
                , Phosphor.cylinder Thin |> toHtml []
                , Phosphor.cylinder Light |> toHtml []
                , Phosphor.cylinder Regular |> toHtml []
                , Phosphor.cylinder Bold |> toHtml []
                , Phosphor.cylinder Fill |> toHtml []
                , Phosphor.cylinder Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.database model.weight |> toHtml []
                , Phosphor.database Thin |> toHtml []
                , Phosphor.database Light |> toHtml []
                , Phosphor.database Regular |> toHtml []
                , Phosphor.database Bold |> toHtml []
                , Phosphor.database Fill |> toHtml []
                , Phosphor.database Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.desktop model.weight |> toHtml []
                , Phosphor.desktop Thin |> toHtml []
                , Phosphor.desktop Light |> toHtml []
                , Phosphor.desktop Regular |> toHtml []
                , Phosphor.desktop Bold |> toHtml []
                , Phosphor.desktop Fill |> toHtml []
                , Phosphor.desktop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.desktopTower model.weight |> toHtml []
                , Phosphor.desktopTower Thin |> toHtml []
                , Phosphor.desktopTower Light |> toHtml []
                , Phosphor.desktopTower Regular |> toHtml []
                , Phosphor.desktopTower Bold |> toHtml []
                , Phosphor.desktopTower Fill |> toHtml []
                , Phosphor.desktopTower Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.detective model.weight |> toHtml []
                , Phosphor.detective Thin |> toHtml []
                , Phosphor.detective Light |> toHtml []
                , Phosphor.detective Regular |> toHtml []
                , Phosphor.detective Bold |> toHtml []
                , Phosphor.detective Fill |> toHtml []
                , Phosphor.detective Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.devToLogo model.weight |> toHtml []
                , Phosphor.devToLogo Thin |> toHtml []
                , Phosphor.devToLogo Light |> toHtml []
                , Phosphor.devToLogo Regular |> toHtml []
                , Phosphor.devToLogo Bold |> toHtml []
                , Phosphor.devToLogo Fill |> toHtml []
                , Phosphor.devToLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.deviceMobile model.weight |> toHtml []
                , Phosphor.deviceMobile Thin |> toHtml []
                , Phosphor.deviceMobile Light |> toHtml []
                , Phosphor.deviceMobile Regular |> toHtml []
                , Phosphor.deviceMobile Bold |> toHtml []
                , Phosphor.deviceMobile Fill |> toHtml []
                , Phosphor.deviceMobile Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.deviceMobileCamera model.weight |> toHtml []
                , Phosphor.deviceMobileCamera Thin |> toHtml []
                , Phosphor.deviceMobileCamera Light |> toHtml []
                , Phosphor.deviceMobileCamera Regular |> toHtml []
                , Phosphor.deviceMobileCamera Bold |> toHtml []
                , Phosphor.deviceMobileCamera Fill |> toHtml []
                , Phosphor.deviceMobileCamera Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.deviceMobileSpeaker model.weight |> toHtml []
                , Phosphor.deviceMobileSpeaker Thin |> toHtml []
                , Phosphor.deviceMobileSpeaker Light |> toHtml []
                , Phosphor.deviceMobileSpeaker Regular |> toHtml []
                , Phosphor.deviceMobileSpeaker Bold |> toHtml []
                , Phosphor.deviceMobileSpeaker Fill |> toHtml []
                , Phosphor.deviceMobileSpeaker Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.deviceTablet model.weight |> toHtml []
                , Phosphor.deviceTablet Thin |> toHtml []
                , Phosphor.deviceTablet Light |> toHtml []
                , Phosphor.deviceTablet Regular |> toHtml []
                , Phosphor.deviceTablet Bold |> toHtml []
                , Phosphor.deviceTablet Fill |> toHtml []
                , Phosphor.deviceTablet Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.deviceTabletCamera model.weight |> toHtml []
                , Phosphor.deviceTabletCamera Thin |> toHtml []
                , Phosphor.deviceTabletCamera Light |> toHtml []
                , Phosphor.deviceTabletCamera Regular |> toHtml []
                , Phosphor.deviceTabletCamera Bold |> toHtml []
                , Phosphor.deviceTabletCamera Fill |> toHtml []
                , Phosphor.deviceTabletCamera Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.deviceTabletSpeaker model.weight |> toHtml []
                , Phosphor.deviceTabletSpeaker Thin |> toHtml []
                , Phosphor.deviceTabletSpeaker Light |> toHtml []
                , Phosphor.deviceTabletSpeaker Regular |> toHtml []
                , Phosphor.deviceTabletSpeaker Bold |> toHtml []
                , Phosphor.deviceTabletSpeaker Fill |> toHtml []
                , Phosphor.deviceTabletSpeaker Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.devices model.weight |> toHtml []
                , Phosphor.devices Thin |> toHtml []
                , Phosphor.devices Light |> toHtml []
                , Phosphor.devices Regular |> toHtml []
                , Phosphor.devices Bold |> toHtml []
                , Phosphor.devices Fill |> toHtml []
                , Phosphor.devices Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.diamond model.weight |> toHtml []
                , Phosphor.diamond Thin |> toHtml []
                , Phosphor.diamond Light |> toHtml []
                , Phosphor.diamond Regular |> toHtml []
                , Phosphor.diamond Bold |> toHtml []
                , Phosphor.diamond Fill |> toHtml []
                , Phosphor.diamond Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.diamondsFour model.weight |> toHtml []
                , Phosphor.diamondsFour Thin |> toHtml []
                , Phosphor.diamondsFour Light |> toHtml []
                , Phosphor.diamondsFour Regular |> toHtml []
                , Phosphor.diamondsFour Bold |> toHtml []
                , Phosphor.diamondsFour Fill |> toHtml []
                , Phosphor.diamondsFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.diceFive model.weight |> toHtml []
                , Phosphor.diceFive Thin |> toHtml []
                , Phosphor.diceFive Light |> toHtml []
                , Phosphor.diceFive Regular |> toHtml []
                , Phosphor.diceFive Bold |> toHtml []
                , Phosphor.diceFive Fill |> toHtml []
                , Phosphor.diceFive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.diceFour model.weight |> toHtml []
                , Phosphor.diceFour Thin |> toHtml []
                , Phosphor.diceFour Light |> toHtml []
                , Phosphor.diceFour Regular |> toHtml []
                , Phosphor.diceFour Bold |> toHtml []
                , Phosphor.diceFour Fill |> toHtml []
                , Phosphor.diceFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.diceOne model.weight |> toHtml []
                , Phosphor.diceOne Thin |> toHtml []
                , Phosphor.diceOne Light |> toHtml []
                , Phosphor.diceOne Regular |> toHtml []
                , Phosphor.diceOne Bold |> toHtml []
                , Phosphor.diceOne Fill |> toHtml []
                , Phosphor.diceOne Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.diceSix model.weight |> toHtml []
                , Phosphor.diceSix Thin |> toHtml []
                , Phosphor.diceSix Light |> toHtml []
                , Phosphor.diceSix Regular |> toHtml []
                , Phosphor.diceSix Bold |> toHtml []
                , Phosphor.diceSix Fill |> toHtml []
                , Phosphor.diceSix Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.diceThree model.weight |> toHtml []
                , Phosphor.diceThree Thin |> toHtml []
                , Phosphor.diceThree Light |> toHtml []
                , Phosphor.diceThree Regular |> toHtml []
                , Phosphor.diceThree Bold |> toHtml []
                , Phosphor.diceThree Fill |> toHtml []
                , Phosphor.diceThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.diceTwo model.weight |> toHtml []
                , Phosphor.diceTwo Thin |> toHtml []
                , Phosphor.diceTwo Light |> toHtml []
                , Phosphor.diceTwo Regular |> toHtml []
                , Phosphor.diceTwo Bold |> toHtml []
                , Phosphor.diceTwo Fill |> toHtml []
                , Phosphor.diceTwo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.disc model.weight |> toHtml []
                , Phosphor.disc Thin |> toHtml []
                , Phosphor.disc Light |> toHtml []
                , Phosphor.disc Regular |> toHtml []
                , Phosphor.disc Bold |> toHtml []
                , Phosphor.disc Fill |> toHtml []
                , Phosphor.disc Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.discordLogo model.weight |> toHtml []
                , Phosphor.discordLogo Thin |> toHtml []
                , Phosphor.discordLogo Light |> toHtml []
                , Phosphor.discordLogo Regular |> toHtml []
                , Phosphor.discordLogo Bold |> toHtml []
                , Phosphor.discordLogo Fill |> toHtml []
                , Phosphor.discordLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.divide model.weight |> toHtml []
                , Phosphor.divide Thin |> toHtml []
                , Phosphor.divide Light |> toHtml []
                , Phosphor.divide Regular |> toHtml []
                , Phosphor.divide Bold |> toHtml []
                , Phosphor.divide Fill |> toHtml []
                , Phosphor.divide Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dna model.weight |> toHtml []
                , Phosphor.dna Thin |> toHtml []
                , Phosphor.dna Light |> toHtml []
                , Phosphor.dna Regular |> toHtml []
                , Phosphor.dna Bold |> toHtml []
                , Phosphor.dna Fill |> toHtml []
                , Phosphor.dna Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dog model.weight |> toHtml []
                , Phosphor.dog Thin |> toHtml []
                , Phosphor.dog Light |> toHtml []
                , Phosphor.dog Regular |> toHtml []
                , Phosphor.dog Bold |> toHtml []
                , Phosphor.dog Fill |> toHtml []
                , Phosphor.dog Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.door model.weight |> toHtml []
                , Phosphor.door Thin |> toHtml []
                , Phosphor.door Light |> toHtml []
                , Phosphor.door Regular |> toHtml []
                , Phosphor.door Bold |> toHtml []
                , Phosphor.door Fill |> toHtml []
                , Phosphor.door Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.doorOpen model.weight |> toHtml []
                , Phosphor.doorOpen Thin |> toHtml []
                , Phosphor.doorOpen Light |> toHtml []
                , Phosphor.doorOpen Regular |> toHtml []
                , Phosphor.doorOpen Bold |> toHtml []
                , Phosphor.doorOpen Fill |> toHtml []
                , Phosphor.doorOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dot model.weight |> toHtml []
                , Phosphor.dot Thin |> toHtml []
                , Phosphor.dot Light |> toHtml []
                , Phosphor.dot Regular |> toHtml []
                , Phosphor.dot Bold |> toHtml []
                , Phosphor.dot Fill |> toHtml []
                , Phosphor.dot Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotOutline model.weight |> toHtml []
                , Phosphor.dotOutline Thin |> toHtml []
                , Phosphor.dotOutline Light |> toHtml []
                , Phosphor.dotOutline Regular |> toHtml []
                , Phosphor.dotOutline Bold |> toHtml []
                , Phosphor.dotOutline Fill |> toHtml []
                , Phosphor.dotOutline Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsNine model.weight |> toHtml []
                , Phosphor.dotsNine Thin |> toHtml []
                , Phosphor.dotsNine Light |> toHtml []
                , Phosphor.dotsNine Regular |> toHtml []
                , Phosphor.dotsNine Bold |> toHtml []
                , Phosphor.dotsNine Fill |> toHtml []
                , Phosphor.dotsNine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsSix model.weight |> toHtml []
                , Phosphor.dotsSix Thin |> toHtml []
                , Phosphor.dotsSix Light |> toHtml []
                , Phosphor.dotsSix Regular |> toHtml []
                , Phosphor.dotsSix Bold |> toHtml []
                , Phosphor.dotsSix Fill |> toHtml []
                , Phosphor.dotsSix Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsSixVertical model.weight |> toHtml []
                , Phosphor.dotsSixVertical Thin |> toHtml []
                , Phosphor.dotsSixVertical Light |> toHtml []
                , Phosphor.dotsSixVertical Regular |> toHtml []
                , Phosphor.dotsSixVertical Bold |> toHtml []
                , Phosphor.dotsSixVertical Fill |> toHtml []
                , Phosphor.dotsSixVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsThree model.weight |> toHtml []
                , Phosphor.dotsThree Thin |> toHtml []
                , Phosphor.dotsThree Light |> toHtml []
                , Phosphor.dotsThree Regular |> toHtml []
                , Phosphor.dotsThree Bold |> toHtml []
                , Phosphor.dotsThree Fill |> toHtml []
                , Phosphor.dotsThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsThreeCircle model.weight |> toHtml []
                , Phosphor.dotsThreeCircle Thin |> toHtml []
                , Phosphor.dotsThreeCircle Light |> toHtml []
                , Phosphor.dotsThreeCircle Regular |> toHtml []
                , Phosphor.dotsThreeCircle Bold |> toHtml []
                , Phosphor.dotsThreeCircle Fill |> toHtml []
                , Phosphor.dotsThreeCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsThreeCircleVertical model.weight |> toHtml []
                , Phosphor.dotsThreeCircleVertical Thin |> toHtml []
                , Phosphor.dotsThreeCircleVertical Light |> toHtml []
                , Phosphor.dotsThreeCircleVertical Regular |> toHtml []
                , Phosphor.dotsThreeCircleVertical Bold |> toHtml []
                , Phosphor.dotsThreeCircleVertical Fill |> toHtml []
                , Phosphor.dotsThreeCircleVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsThreeOutline model.weight |> toHtml []
                , Phosphor.dotsThreeOutline Thin |> toHtml []
                , Phosphor.dotsThreeOutline Light |> toHtml []
                , Phosphor.dotsThreeOutline Regular |> toHtml []
                , Phosphor.dotsThreeOutline Bold |> toHtml []
                , Phosphor.dotsThreeOutline Fill |> toHtml []
                , Phosphor.dotsThreeOutline Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsThreeOutlineVertical model.weight |> toHtml []
                , Phosphor.dotsThreeOutlineVertical Thin |> toHtml []
                , Phosphor.dotsThreeOutlineVertical Light |> toHtml []
                , Phosphor.dotsThreeOutlineVertical Regular |> toHtml []
                , Phosphor.dotsThreeOutlineVertical Bold |> toHtml []
                , Phosphor.dotsThreeOutlineVertical Fill |> toHtml []
                , Phosphor.dotsThreeOutlineVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dotsThreeVertical model.weight |> toHtml []
                , Phosphor.dotsThreeVertical Thin |> toHtml []
                , Phosphor.dotsThreeVertical Light |> toHtml []
                , Phosphor.dotsThreeVertical Regular |> toHtml []
                , Phosphor.dotsThreeVertical Bold |> toHtml []
                , Phosphor.dotsThreeVertical Fill |> toHtml []
                , Phosphor.dotsThreeVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.download model.weight |> toHtml []
                , Phosphor.download Thin |> toHtml []
                , Phosphor.download Light |> toHtml []
                , Phosphor.download Regular |> toHtml []
                , Phosphor.download Bold |> toHtml []
                , Phosphor.download Fill |> toHtml []
                , Phosphor.download Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.downloadSimple model.weight |> toHtml []
                , Phosphor.downloadSimple Thin |> toHtml []
                , Phosphor.downloadSimple Light |> toHtml []
                , Phosphor.downloadSimple Regular |> toHtml []
                , Phosphor.downloadSimple Bold |> toHtml []
                , Phosphor.downloadSimple Fill |> toHtml []
                , Phosphor.downloadSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dress model.weight |> toHtml []
                , Phosphor.dress Thin |> toHtml []
                , Phosphor.dress Light |> toHtml []
                , Phosphor.dress Regular |> toHtml []
                , Phosphor.dress Bold |> toHtml []
                , Phosphor.dress Fill |> toHtml []
                , Phosphor.dress Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dribbbleLogo model.weight |> toHtml []
                , Phosphor.dribbbleLogo Thin |> toHtml []
                , Phosphor.dribbbleLogo Light |> toHtml []
                , Phosphor.dribbbleLogo Regular |> toHtml []
                , Phosphor.dribbbleLogo Bold |> toHtml []
                , Phosphor.dribbbleLogo Fill |> toHtml []
                , Phosphor.dribbbleLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.drop model.weight |> toHtml []
                , Phosphor.drop Thin |> toHtml []
                , Phosphor.drop Light |> toHtml []
                , Phosphor.drop Regular |> toHtml []
                , Phosphor.drop Bold |> toHtml []
                , Phosphor.drop Fill |> toHtml []
                , Phosphor.drop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dropHalf model.weight |> toHtml []
                , Phosphor.dropHalf Thin |> toHtml []
                , Phosphor.dropHalf Light |> toHtml []
                , Phosphor.dropHalf Regular |> toHtml []
                , Phosphor.dropHalf Bold |> toHtml []
                , Phosphor.dropHalf Fill |> toHtml []
                , Phosphor.dropHalf Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dropHalfBottom model.weight |> toHtml []
                , Phosphor.dropHalfBottom Thin |> toHtml []
                , Phosphor.dropHalfBottom Light |> toHtml []
                , Phosphor.dropHalfBottom Regular |> toHtml []
                , Phosphor.dropHalfBottom Bold |> toHtml []
                , Phosphor.dropHalfBottom Fill |> toHtml []
                , Phosphor.dropHalfBottom Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.dropboxLogo model.weight |> toHtml []
                , Phosphor.dropboxLogo Thin |> toHtml []
                , Phosphor.dropboxLogo Light |> toHtml []
                , Phosphor.dropboxLogo Regular |> toHtml []
                , Phosphor.dropboxLogo Bold |> toHtml []
                , Phosphor.dropboxLogo Fill |> toHtml []
                , Phosphor.dropboxLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.ear model.weight |> toHtml []
                , Phosphor.ear Thin |> toHtml []
                , Phosphor.ear Light |> toHtml []
                , Phosphor.ear Regular |> toHtml []
                , Phosphor.ear Bold |> toHtml []
                , Phosphor.ear Fill |> toHtml []
                , Phosphor.ear Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.earSlash model.weight |> toHtml []
                , Phosphor.earSlash Thin |> toHtml []
                , Phosphor.earSlash Light |> toHtml []
                , Phosphor.earSlash Regular |> toHtml []
                , Phosphor.earSlash Bold |> toHtml []
                , Phosphor.earSlash Fill |> toHtml []
                , Phosphor.earSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.egg model.weight |> toHtml []
                , Phosphor.egg Thin |> toHtml []
                , Phosphor.egg Light |> toHtml []
                , Phosphor.egg Regular |> toHtml []
                , Phosphor.egg Bold |> toHtml []
                , Phosphor.egg Fill |> toHtml []
                , Phosphor.egg Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eggCrack model.weight |> toHtml []
                , Phosphor.eggCrack Thin |> toHtml []
                , Phosphor.eggCrack Light |> toHtml []
                , Phosphor.eggCrack Regular |> toHtml []
                , Phosphor.eggCrack Bold |> toHtml []
                , Phosphor.eggCrack Fill |> toHtml []
                , Phosphor.eggCrack Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eject model.weight |> toHtml []
                , Phosphor.eject Thin |> toHtml []
                , Phosphor.eject Light |> toHtml []
                , Phosphor.eject Regular |> toHtml []
                , Phosphor.eject Bold |> toHtml []
                , Phosphor.eject Fill |> toHtml []
                , Phosphor.eject Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.ejectSimple model.weight |> toHtml []
                , Phosphor.ejectSimple Thin |> toHtml []
                , Phosphor.ejectSimple Light |> toHtml []
                , Phosphor.ejectSimple Regular |> toHtml []
                , Phosphor.ejectSimple Bold |> toHtml []
                , Phosphor.ejectSimple Fill |> toHtml []
                , Phosphor.ejectSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.elevator model.weight |> toHtml []
                , Phosphor.elevator Thin |> toHtml []
                , Phosphor.elevator Light |> toHtml []
                , Phosphor.elevator Regular |> toHtml []
                , Phosphor.elevator Bold |> toHtml []
                , Phosphor.elevator Fill |> toHtml []
                , Phosphor.elevator Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.engine model.weight |> toHtml []
                , Phosphor.engine Thin |> toHtml []
                , Phosphor.engine Light |> toHtml []
                , Phosphor.engine Regular |> toHtml []
                , Phosphor.engine Bold |> toHtml []
                , Phosphor.engine Fill |> toHtml []
                , Phosphor.engine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.envelope model.weight |> toHtml []
                , Phosphor.envelope Thin |> toHtml []
                , Phosphor.envelope Light |> toHtml []
                , Phosphor.envelope Regular |> toHtml []
                , Phosphor.envelope Bold |> toHtml []
                , Phosphor.envelope Fill |> toHtml []
                , Phosphor.envelope Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.envelopeOpen model.weight |> toHtml []
                , Phosphor.envelopeOpen Thin |> toHtml []
                , Phosphor.envelopeOpen Light |> toHtml []
                , Phosphor.envelopeOpen Regular |> toHtml []
                , Phosphor.envelopeOpen Bold |> toHtml []
                , Phosphor.envelopeOpen Fill |> toHtml []
                , Phosphor.envelopeOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.envelopeSimple model.weight |> toHtml []
                , Phosphor.envelopeSimple Thin |> toHtml []
                , Phosphor.envelopeSimple Light |> toHtml []
                , Phosphor.envelopeSimple Regular |> toHtml []
                , Phosphor.envelopeSimple Bold |> toHtml []
                , Phosphor.envelopeSimple Fill |> toHtml []
                , Phosphor.envelopeSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.envelopeSimpleOpen model.weight |> toHtml []
                , Phosphor.envelopeSimpleOpen Thin |> toHtml []
                , Phosphor.envelopeSimpleOpen Light |> toHtml []
                , Phosphor.envelopeSimpleOpen Regular |> toHtml []
                , Phosphor.envelopeSimpleOpen Bold |> toHtml []
                , Phosphor.envelopeSimpleOpen Fill |> toHtml []
                , Phosphor.envelopeSimpleOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.equalizer model.weight |> toHtml []
                , Phosphor.equalizer Thin |> toHtml []
                , Phosphor.equalizer Light |> toHtml []
                , Phosphor.equalizer Regular |> toHtml []
                , Phosphor.equalizer Bold |> toHtml []
                , Phosphor.equalizer Fill |> toHtml []
                , Phosphor.equalizer Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.equals model.weight |> toHtml []
                , Phosphor.equals Thin |> toHtml []
                , Phosphor.equals Light |> toHtml []
                , Phosphor.equals Regular |> toHtml []
                , Phosphor.equals Bold |> toHtml []
                , Phosphor.equals Fill |> toHtml []
                , Phosphor.equals Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eraser model.weight |> toHtml []
                , Phosphor.eraser Thin |> toHtml []
                , Phosphor.eraser Light |> toHtml []
                , Phosphor.eraser Regular |> toHtml []
                , Phosphor.eraser Bold |> toHtml []
                , Phosphor.eraser Fill |> toHtml []
                , Phosphor.eraser Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.escalatorDown model.weight |> toHtml []
                , Phosphor.escalatorDown Thin |> toHtml []
                , Phosphor.escalatorDown Light |> toHtml []
                , Phosphor.escalatorDown Regular |> toHtml []
                , Phosphor.escalatorDown Bold |> toHtml []
                , Phosphor.escalatorDown Fill |> toHtml []
                , Phosphor.escalatorDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.escalatorUp model.weight |> toHtml []
                , Phosphor.escalatorUp Thin |> toHtml []
                , Phosphor.escalatorUp Light |> toHtml []
                , Phosphor.escalatorUp Regular |> toHtml []
                , Phosphor.escalatorUp Bold |> toHtml []
                , Phosphor.escalatorUp Fill |> toHtml []
                , Phosphor.escalatorUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.exam model.weight |> toHtml []
                , Phosphor.exam Thin |> toHtml []
                , Phosphor.exam Light |> toHtml []
                , Phosphor.exam Regular |> toHtml []
                , Phosphor.exam Bold |> toHtml []
                , Phosphor.exam Fill |> toHtml []
                , Phosphor.exam Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.exclude model.weight |> toHtml []
                , Phosphor.exclude Thin |> toHtml []
                , Phosphor.exclude Light |> toHtml []
                , Phosphor.exclude Regular |> toHtml []
                , Phosphor.exclude Bold |> toHtml []
                , Phosphor.exclude Fill |> toHtml []
                , Phosphor.exclude Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.excludeSquare model.weight |> toHtml []
                , Phosphor.excludeSquare Thin |> toHtml []
                , Phosphor.excludeSquare Light |> toHtml []
                , Phosphor.excludeSquare Regular |> toHtml []
                , Phosphor.excludeSquare Bold |> toHtml []
                , Phosphor.excludeSquare Fill |> toHtml []
                , Phosphor.excludeSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.export model.weight |> toHtml []
                , Phosphor.export Thin |> toHtml []
                , Phosphor.export Light |> toHtml []
                , Phosphor.export Regular |> toHtml []
                , Phosphor.export Bold |> toHtml []
                , Phosphor.export Fill |> toHtml []
                , Phosphor.export Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eye model.weight |> toHtml []
                , Phosphor.eye Thin |> toHtml []
                , Phosphor.eye Light |> toHtml []
                , Phosphor.eye Regular |> toHtml []
                , Phosphor.eye Bold |> toHtml []
                , Phosphor.eye Fill |> toHtml []
                , Phosphor.eye Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eyeClosed model.weight |> toHtml []
                , Phosphor.eyeClosed Thin |> toHtml []
                , Phosphor.eyeClosed Light |> toHtml []
                , Phosphor.eyeClosed Regular |> toHtml []
                , Phosphor.eyeClosed Bold |> toHtml []
                , Phosphor.eyeClosed Fill |> toHtml []
                , Phosphor.eyeClosed Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eyeSlash model.weight |> toHtml []
                , Phosphor.eyeSlash Thin |> toHtml []
                , Phosphor.eyeSlash Light |> toHtml []
                , Phosphor.eyeSlash Regular |> toHtml []
                , Phosphor.eyeSlash Bold |> toHtml []
                , Phosphor.eyeSlash Fill |> toHtml []
                , Phosphor.eyeSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eyedropper model.weight |> toHtml []
                , Phosphor.eyedropper Thin |> toHtml []
                , Phosphor.eyedropper Light |> toHtml []
                , Phosphor.eyedropper Regular |> toHtml []
                , Phosphor.eyedropper Bold |> toHtml []
                , Phosphor.eyedropper Fill |> toHtml []
                , Phosphor.eyedropper Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eyedropperSample model.weight |> toHtml []
                , Phosphor.eyedropperSample Thin |> toHtml []
                , Phosphor.eyedropperSample Light |> toHtml []
                , Phosphor.eyedropperSample Regular |> toHtml []
                , Phosphor.eyedropperSample Bold |> toHtml []
                , Phosphor.eyedropperSample Fill |> toHtml []
                , Phosphor.eyedropperSample Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.eyeglasses model.weight |> toHtml []
                , Phosphor.eyeglasses Thin |> toHtml []
                , Phosphor.eyeglasses Light |> toHtml []
                , Phosphor.eyeglasses Regular |> toHtml []
                , Phosphor.eyeglasses Bold |> toHtml []
                , Phosphor.eyeglasses Fill |> toHtml []
                , Phosphor.eyeglasses Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.faceMask model.weight |> toHtml []
                , Phosphor.faceMask Thin |> toHtml []
                , Phosphor.faceMask Light |> toHtml []
                , Phosphor.faceMask Regular |> toHtml []
                , Phosphor.faceMask Bold |> toHtml []
                , Phosphor.faceMask Fill |> toHtml []
                , Phosphor.faceMask Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.facebookLogo model.weight |> toHtml []
                , Phosphor.facebookLogo Thin |> toHtml []
                , Phosphor.facebookLogo Light |> toHtml []
                , Phosphor.facebookLogo Regular |> toHtml []
                , Phosphor.facebookLogo Bold |> toHtml []
                , Phosphor.facebookLogo Fill |> toHtml []
                , Phosphor.facebookLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.factory model.weight |> toHtml []
                , Phosphor.factory Thin |> toHtml []
                , Phosphor.factory Light |> toHtml []
                , Phosphor.factory Regular |> toHtml []
                , Phosphor.factory Bold |> toHtml []
                , Phosphor.factory Fill |> toHtml []
                , Phosphor.factory Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.faders model.weight |> toHtml []
                , Phosphor.faders Thin |> toHtml []
                , Phosphor.faders Light |> toHtml []
                , Phosphor.faders Regular |> toHtml []
                , Phosphor.faders Bold |> toHtml []
                , Phosphor.faders Fill |> toHtml []
                , Phosphor.faders Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fadersHorizontal model.weight |> toHtml []
                , Phosphor.fadersHorizontal Thin |> toHtml []
                , Phosphor.fadersHorizontal Light |> toHtml []
                , Phosphor.fadersHorizontal Regular |> toHtml []
                , Phosphor.fadersHorizontal Bold |> toHtml []
                , Phosphor.fadersHorizontal Fill |> toHtml []
                , Phosphor.fadersHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fan model.weight |> toHtml []
                , Phosphor.fan Thin |> toHtml []
                , Phosphor.fan Light |> toHtml []
                , Phosphor.fan Regular |> toHtml []
                , Phosphor.fan Bold |> toHtml []
                , Phosphor.fan Fill |> toHtml []
                , Phosphor.fan Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fastForward model.weight |> toHtml []
                , Phosphor.fastForward Thin |> toHtml []
                , Phosphor.fastForward Light |> toHtml []
                , Phosphor.fastForward Regular |> toHtml []
                , Phosphor.fastForward Bold |> toHtml []
                , Phosphor.fastForward Fill |> toHtml []
                , Phosphor.fastForward Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fastForwardCircle model.weight |> toHtml []
                , Phosphor.fastForwardCircle Thin |> toHtml []
                , Phosphor.fastForwardCircle Light |> toHtml []
                , Phosphor.fastForwardCircle Regular |> toHtml []
                , Phosphor.fastForwardCircle Bold |> toHtml []
                , Phosphor.fastForwardCircle Fill |> toHtml []
                , Phosphor.fastForwardCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.feather model.weight |> toHtml []
                , Phosphor.feather Thin |> toHtml []
                , Phosphor.feather Light |> toHtml []
                , Phosphor.feather Regular |> toHtml []
                , Phosphor.feather Bold |> toHtml []
                , Phosphor.feather Fill |> toHtml []
                , Phosphor.feather Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.figmaLogo model.weight |> toHtml []
                , Phosphor.figmaLogo Thin |> toHtml []
                , Phosphor.figmaLogo Light |> toHtml []
                , Phosphor.figmaLogo Regular |> toHtml []
                , Phosphor.figmaLogo Bold |> toHtml []
                , Phosphor.figmaLogo Fill |> toHtml []
                , Phosphor.figmaLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileArchive model.weight |> toHtml []
                , Phosphor.fileArchive Thin |> toHtml []
                , Phosphor.fileArchive Light |> toHtml []
                , Phosphor.fileArchive Regular |> toHtml []
                , Phosphor.fileArchive Bold |> toHtml []
                , Phosphor.fileArchive Fill |> toHtml []
                , Phosphor.fileArchive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileArrowDown model.weight |> toHtml []
                , Phosphor.fileArrowDown Thin |> toHtml []
                , Phosphor.fileArrowDown Light |> toHtml []
                , Phosphor.fileArrowDown Regular |> toHtml []
                , Phosphor.fileArrowDown Bold |> toHtml []
                , Phosphor.fileArrowDown Fill |> toHtml []
                , Phosphor.fileArrowDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileArrowUp model.weight |> toHtml []
                , Phosphor.fileArrowUp Thin |> toHtml []
                , Phosphor.fileArrowUp Light |> toHtml []
                , Phosphor.fileArrowUp Regular |> toHtml []
                , Phosphor.fileArrowUp Bold |> toHtml []
                , Phosphor.fileArrowUp Fill |> toHtml []
                , Phosphor.fileArrowUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileAudio model.weight |> toHtml []
                , Phosphor.fileAudio Thin |> toHtml []
                , Phosphor.fileAudio Light |> toHtml []
                , Phosphor.fileAudio Regular |> toHtml []
                , Phosphor.fileAudio Bold |> toHtml []
                , Phosphor.fileAudio Fill |> toHtml []
                , Phosphor.fileAudio Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.file model.weight |> toHtml []
                , Phosphor.file Thin |> toHtml []
                , Phosphor.file Light |> toHtml []
                , Phosphor.file Regular |> toHtml []
                , Phosphor.file Bold |> toHtml []
                , Phosphor.file Fill |> toHtml []
                , Phosphor.file Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileCloud model.weight |> toHtml []
                , Phosphor.fileCloud Thin |> toHtml []
                , Phosphor.fileCloud Light |> toHtml []
                , Phosphor.fileCloud Regular |> toHtml []
                , Phosphor.fileCloud Bold |> toHtml []
                , Phosphor.fileCloud Fill |> toHtml []
                , Phosphor.fileCloud Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileCode model.weight |> toHtml []
                , Phosphor.fileCode Thin |> toHtml []
                , Phosphor.fileCode Light |> toHtml []
                , Phosphor.fileCode Regular |> toHtml []
                , Phosphor.fileCode Bold |> toHtml []
                , Phosphor.fileCode Fill |> toHtml []
                , Phosphor.fileCode Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileCss model.weight |> toHtml []
                , Phosphor.fileCss Thin |> toHtml []
                , Phosphor.fileCss Light |> toHtml []
                , Phosphor.fileCss Regular |> toHtml []
                , Phosphor.fileCss Bold |> toHtml []
                , Phosphor.fileCss Fill |> toHtml []
                , Phosphor.fileCss Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileCsv model.weight |> toHtml []
                , Phosphor.fileCsv Thin |> toHtml []
                , Phosphor.fileCsv Light |> toHtml []
                , Phosphor.fileCsv Regular |> toHtml []
                , Phosphor.fileCsv Bold |> toHtml []
                , Phosphor.fileCsv Fill |> toHtml []
                , Phosphor.fileCsv Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileDashed model.weight |> toHtml []
                , Phosphor.fileDashed Thin |> toHtml []
                , Phosphor.fileDashed Light |> toHtml []
                , Phosphor.fileDashed Regular |> toHtml []
                , Phosphor.fileDashed Bold |> toHtml []
                , Phosphor.fileDashed Fill |> toHtml []
                , Phosphor.fileDashed Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileDoc model.weight |> toHtml []
                , Phosphor.fileDoc Thin |> toHtml []
                , Phosphor.fileDoc Light |> toHtml []
                , Phosphor.fileDoc Regular |> toHtml []
                , Phosphor.fileDoc Bold |> toHtml []
                , Phosphor.fileDoc Fill |> toHtml []
                , Phosphor.fileDoc Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileHtml model.weight |> toHtml []
                , Phosphor.fileHtml Thin |> toHtml []
                , Phosphor.fileHtml Light |> toHtml []
                , Phosphor.fileHtml Regular |> toHtml []
                , Phosphor.fileHtml Bold |> toHtml []
                , Phosphor.fileHtml Fill |> toHtml []
                , Phosphor.fileHtml Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileImage model.weight |> toHtml []
                , Phosphor.fileImage Thin |> toHtml []
                , Phosphor.fileImage Light |> toHtml []
                , Phosphor.fileImage Regular |> toHtml []
                , Phosphor.fileImage Bold |> toHtml []
                , Phosphor.fileImage Fill |> toHtml []
                , Phosphor.fileImage Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileJpg model.weight |> toHtml []
                , Phosphor.fileJpg Thin |> toHtml []
                , Phosphor.fileJpg Light |> toHtml []
                , Phosphor.fileJpg Regular |> toHtml []
                , Phosphor.fileJpg Bold |> toHtml []
                , Phosphor.fileJpg Fill |> toHtml []
                , Phosphor.fileJpg Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileJs model.weight |> toHtml []
                , Phosphor.fileJs Thin |> toHtml []
                , Phosphor.fileJs Light |> toHtml []
                , Phosphor.fileJs Regular |> toHtml []
                , Phosphor.fileJs Bold |> toHtml []
                , Phosphor.fileJs Fill |> toHtml []
                , Phosphor.fileJs Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileJsx model.weight |> toHtml []
                , Phosphor.fileJsx Thin |> toHtml []
                , Phosphor.fileJsx Light |> toHtml []
                , Phosphor.fileJsx Regular |> toHtml []
                , Phosphor.fileJsx Bold |> toHtml []
                , Phosphor.fileJsx Fill |> toHtml []
                , Phosphor.fileJsx Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileLock model.weight |> toHtml []
                , Phosphor.fileLock Thin |> toHtml []
                , Phosphor.fileLock Light |> toHtml []
                , Phosphor.fileLock Regular |> toHtml []
                , Phosphor.fileLock Bold |> toHtml []
                , Phosphor.fileLock Fill |> toHtml []
                , Phosphor.fileLock Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileMagnifyingGlass model.weight |> toHtml []
                , Phosphor.fileMagnifyingGlass Thin |> toHtml []
                , Phosphor.fileMagnifyingGlass Light |> toHtml []
                , Phosphor.fileMagnifyingGlass Regular |> toHtml []
                , Phosphor.fileMagnifyingGlass Bold |> toHtml []
                , Phosphor.fileMagnifyingGlass Fill |> toHtml []
                , Phosphor.fileMagnifyingGlass Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileMinus model.weight |> toHtml []
                , Phosphor.fileMinus Thin |> toHtml []
                , Phosphor.fileMinus Light |> toHtml []
                , Phosphor.fileMinus Regular |> toHtml []
                , Phosphor.fileMinus Bold |> toHtml []
                , Phosphor.fileMinus Fill |> toHtml []
                , Phosphor.fileMinus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.filePdf model.weight |> toHtml []
                , Phosphor.filePdf Thin |> toHtml []
                , Phosphor.filePdf Light |> toHtml []
                , Phosphor.filePdf Regular |> toHtml []
                , Phosphor.filePdf Bold |> toHtml []
                , Phosphor.filePdf Fill |> toHtml []
                , Phosphor.filePdf Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.filePlus model.weight |> toHtml []
                , Phosphor.filePlus Thin |> toHtml []
                , Phosphor.filePlus Light |> toHtml []
                , Phosphor.filePlus Regular |> toHtml []
                , Phosphor.filePlus Bold |> toHtml []
                , Phosphor.filePlus Fill |> toHtml []
                , Phosphor.filePlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.filePng model.weight |> toHtml []
                , Phosphor.filePng Thin |> toHtml []
                , Phosphor.filePng Light |> toHtml []
                , Phosphor.filePng Regular |> toHtml []
                , Phosphor.filePng Bold |> toHtml []
                , Phosphor.filePng Fill |> toHtml []
                , Phosphor.filePng Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.filePpt model.weight |> toHtml []
                , Phosphor.filePpt Thin |> toHtml []
                , Phosphor.filePpt Light |> toHtml []
                , Phosphor.filePpt Regular |> toHtml []
                , Phosphor.filePpt Bold |> toHtml []
                , Phosphor.filePpt Fill |> toHtml []
                , Phosphor.filePpt Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileRs model.weight |> toHtml []
                , Phosphor.fileRs Thin |> toHtml []
                , Phosphor.fileRs Light |> toHtml []
                , Phosphor.fileRs Regular |> toHtml []
                , Phosphor.fileRs Bold |> toHtml []
                , Phosphor.fileRs Fill |> toHtml []
                , Phosphor.fileRs Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileSql model.weight |> toHtml []
                , Phosphor.fileSql Thin |> toHtml []
                , Phosphor.fileSql Light |> toHtml []
                , Phosphor.fileSql Regular |> toHtml []
                , Phosphor.fileSql Bold |> toHtml []
                , Phosphor.fileSql Fill |> toHtml []
                , Phosphor.fileSql Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileSvg model.weight |> toHtml []
                , Phosphor.fileSvg Thin |> toHtml []
                , Phosphor.fileSvg Light |> toHtml []
                , Phosphor.fileSvg Regular |> toHtml []
                , Phosphor.fileSvg Bold |> toHtml []
                , Phosphor.fileSvg Fill |> toHtml []
                , Phosphor.fileSvg Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileText model.weight |> toHtml []
                , Phosphor.fileText Thin |> toHtml []
                , Phosphor.fileText Light |> toHtml []
                , Phosphor.fileText Regular |> toHtml []
                , Phosphor.fileText Bold |> toHtml []
                , Phosphor.fileText Fill |> toHtml []
                , Phosphor.fileText Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileTs model.weight |> toHtml []
                , Phosphor.fileTs Thin |> toHtml []
                , Phosphor.fileTs Light |> toHtml []
                , Phosphor.fileTs Regular |> toHtml []
                , Phosphor.fileTs Bold |> toHtml []
                , Phosphor.fileTs Fill |> toHtml []
                , Phosphor.fileTs Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileTsx model.weight |> toHtml []
                , Phosphor.fileTsx Thin |> toHtml []
                , Phosphor.fileTsx Light |> toHtml []
                , Phosphor.fileTsx Regular |> toHtml []
                , Phosphor.fileTsx Bold |> toHtml []
                , Phosphor.fileTsx Fill |> toHtml []
                , Phosphor.fileTsx Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileVideo model.weight |> toHtml []
                , Phosphor.fileVideo Thin |> toHtml []
                , Phosphor.fileVideo Light |> toHtml []
                , Phosphor.fileVideo Regular |> toHtml []
                , Phosphor.fileVideo Bold |> toHtml []
                , Phosphor.fileVideo Fill |> toHtml []
                , Phosphor.fileVideo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileVue model.weight |> toHtml []
                , Phosphor.fileVue Thin |> toHtml []
                , Phosphor.fileVue Light |> toHtml []
                , Phosphor.fileVue Regular |> toHtml []
                , Phosphor.fileVue Bold |> toHtml []
                , Phosphor.fileVue Fill |> toHtml []
                , Phosphor.fileVue Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileX model.weight |> toHtml []
                , Phosphor.fileX Thin |> toHtml []
                , Phosphor.fileX Light |> toHtml []
                , Phosphor.fileX Regular |> toHtml []
                , Phosphor.fileX Bold |> toHtml []
                , Phosphor.fileX Fill |> toHtml []
                , Phosphor.fileX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileXls model.weight |> toHtml []
                , Phosphor.fileXls Thin |> toHtml []
                , Phosphor.fileXls Light |> toHtml []
                , Phosphor.fileXls Regular |> toHtml []
                , Phosphor.fileXls Bold |> toHtml []
                , Phosphor.fileXls Fill |> toHtml []
                , Phosphor.fileXls Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fileZip model.weight |> toHtml []
                , Phosphor.fileZip Thin |> toHtml []
                , Phosphor.fileZip Light |> toHtml []
                , Phosphor.fileZip Regular |> toHtml []
                , Phosphor.fileZip Bold |> toHtml []
                , Phosphor.fileZip Fill |> toHtml []
                , Phosphor.fileZip Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.files model.weight |> toHtml []
                , Phosphor.files Thin |> toHtml []
                , Phosphor.files Light |> toHtml []
                , Phosphor.files Regular |> toHtml []
                , Phosphor.files Bold |> toHtml []
                , Phosphor.files Fill |> toHtml []
                , Phosphor.files Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.filmReel model.weight |> toHtml []
                , Phosphor.filmReel Thin |> toHtml []
                , Phosphor.filmReel Light |> toHtml []
                , Phosphor.filmReel Regular |> toHtml []
                , Phosphor.filmReel Bold |> toHtml []
                , Phosphor.filmReel Fill |> toHtml []
                , Phosphor.filmReel Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.filmScript model.weight |> toHtml []
                , Phosphor.filmScript Thin |> toHtml []
                , Phosphor.filmScript Light |> toHtml []
                , Phosphor.filmScript Regular |> toHtml []
                , Phosphor.filmScript Bold |> toHtml []
                , Phosphor.filmScript Fill |> toHtml []
                , Phosphor.filmScript Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.filmSlate model.weight |> toHtml []
                , Phosphor.filmSlate Thin |> toHtml []
                , Phosphor.filmSlate Light |> toHtml []
                , Phosphor.filmSlate Regular |> toHtml []
                , Phosphor.filmSlate Bold |> toHtml []
                , Phosphor.filmSlate Fill |> toHtml []
                , Phosphor.filmSlate Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.filmStrip model.weight |> toHtml []
                , Phosphor.filmStrip Thin |> toHtml []
                , Phosphor.filmStrip Light |> toHtml []
                , Phosphor.filmStrip Regular |> toHtml []
                , Phosphor.filmStrip Bold |> toHtml []
                , Phosphor.filmStrip Fill |> toHtml []
                , Phosphor.filmStrip Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fingerprint model.weight |> toHtml []
                , Phosphor.fingerprint Thin |> toHtml []
                , Phosphor.fingerprint Light |> toHtml []
                , Phosphor.fingerprint Regular |> toHtml []
                , Phosphor.fingerprint Bold |> toHtml []
                , Phosphor.fingerprint Fill |> toHtml []
                , Phosphor.fingerprint Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fingerprintSimple model.weight |> toHtml []
                , Phosphor.fingerprintSimple Thin |> toHtml []
                , Phosphor.fingerprintSimple Light |> toHtml []
                , Phosphor.fingerprintSimple Regular |> toHtml []
                , Phosphor.fingerprintSimple Bold |> toHtml []
                , Phosphor.fingerprintSimple Fill |> toHtml []
                , Phosphor.fingerprintSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.finnTheHuman model.weight |> toHtml []
                , Phosphor.finnTheHuman Thin |> toHtml []
                , Phosphor.finnTheHuman Light |> toHtml []
                , Phosphor.finnTheHuman Regular |> toHtml []
                , Phosphor.finnTheHuman Bold |> toHtml []
                , Phosphor.finnTheHuman Fill |> toHtml []
                , Phosphor.finnTheHuman Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fire model.weight |> toHtml []
                , Phosphor.fire Thin |> toHtml []
                , Phosphor.fire Light |> toHtml []
                , Phosphor.fire Regular |> toHtml []
                , Phosphor.fire Bold |> toHtml []
                , Phosphor.fire Fill |> toHtml []
                , Phosphor.fire Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fireExtinguisher model.weight |> toHtml []
                , Phosphor.fireExtinguisher Thin |> toHtml []
                , Phosphor.fireExtinguisher Light |> toHtml []
                , Phosphor.fireExtinguisher Regular |> toHtml []
                , Phosphor.fireExtinguisher Bold |> toHtml []
                , Phosphor.fireExtinguisher Fill |> toHtml []
                , Phosphor.fireExtinguisher Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fireSimple model.weight |> toHtml []
                , Phosphor.fireSimple Thin |> toHtml []
                , Phosphor.fireSimple Light |> toHtml []
                , Phosphor.fireSimple Regular |> toHtml []
                , Phosphor.fireSimple Bold |> toHtml []
                , Phosphor.fireSimple Fill |> toHtml []
                , Phosphor.fireSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.firstAid model.weight |> toHtml []
                , Phosphor.firstAid Thin |> toHtml []
                , Phosphor.firstAid Light |> toHtml []
                , Phosphor.firstAid Regular |> toHtml []
                , Phosphor.firstAid Bold |> toHtml []
                , Phosphor.firstAid Fill |> toHtml []
                , Phosphor.firstAid Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.firstAidKit model.weight |> toHtml []
                , Phosphor.firstAidKit Thin |> toHtml []
                , Phosphor.firstAidKit Light |> toHtml []
                , Phosphor.firstAidKit Regular |> toHtml []
                , Phosphor.firstAidKit Bold |> toHtml []
                , Phosphor.firstAidKit Fill |> toHtml []
                , Phosphor.firstAidKit Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fish model.weight |> toHtml []
                , Phosphor.fish Thin |> toHtml []
                , Phosphor.fish Light |> toHtml []
                , Phosphor.fish Regular |> toHtml []
                , Phosphor.fish Bold |> toHtml []
                , Phosphor.fish Fill |> toHtml []
                , Phosphor.fish Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.fishSimple model.weight |> toHtml []
                , Phosphor.fishSimple Thin |> toHtml []
                , Phosphor.fishSimple Light |> toHtml []
                , Phosphor.fishSimple Regular |> toHtml []
                , Phosphor.fishSimple Bold |> toHtml []
                , Phosphor.fishSimple Fill |> toHtml []
                , Phosphor.fishSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flagBanner model.weight |> toHtml []
                , Phosphor.flagBanner Thin |> toHtml []
                , Phosphor.flagBanner Light |> toHtml []
                , Phosphor.flagBanner Regular |> toHtml []
                , Phosphor.flagBanner Bold |> toHtml []
                , Phosphor.flagBanner Fill |> toHtml []
                , Phosphor.flagBanner Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flag model.weight |> toHtml []
                , Phosphor.flag Thin |> toHtml []
                , Phosphor.flag Light |> toHtml []
                , Phosphor.flag Regular |> toHtml []
                , Phosphor.flag Bold |> toHtml []
                , Phosphor.flag Fill |> toHtml []
                , Phosphor.flag Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flagCheckered model.weight |> toHtml []
                , Phosphor.flagCheckered Thin |> toHtml []
                , Phosphor.flagCheckered Light |> toHtml []
                , Phosphor.flagCheckered Regular |> toHtml []
                , Phosphor.flagCheckered Bold |> toHtml []
                , Phosphor.flagCheckered Fill |> toHtml []
                , Phosphor.flagCheckered Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flagPennant model.weight |> toHtml []
                , Phosphor.flagPennant Thin |> toHtml []
                , Phosphor.flagPennant Light |> toHtml []
                , Phosphor.flagPennant Regular |> toHtml []
                , Phosphor.flagPennant Bold |> toHtml []
                , Phosphor.flagPennant Fill |> toHtml []
                , Phosphor.flagPennant Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flame model.weight |> toHtml []
                , Phosphor.flame Thin |> toHtml []
                , Phosphor.flame Light |> toHtml []
                , Phosphor.flame Regular |> toHtml []
                , Phosphor.flame Bold |> toHtml []
                , Phosphor.flame Fill |> toHtml []
                , Phosphor.flame Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flashlight model.weight |> toHtml []
                , Phosphor.flashlight Thin |> toHtml []
                , Phosphor.flashlight Light |> toHtml []
                , Phosphor.flashlight Regular |> toHtml []
                , Phosphor.flashlight Bold |> toHtml []
                , Phosphor.flashlight Fill |> toHtml []
                , Phosphor.flashlight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flask model.weight |> toHtml []
                , Phosphor.flask Thin |> toHtml []
                , Phosphor.flask Light |> toHtml []
                , Phosphor.flask Regular |> toHtml []
                , Phosphor.flask Bold |> toHtml []
                , Phosphor.flask Fill |> toHtml []
                , Phosphor.flask Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.floppyDiskBack model.weight |> toHtml []
                , Phosphor.floppyDiskBack Thin |> toHtml []
                , Phosphor.floppyDiskBack Light |> toHtml []
                , Phosphor.floppyDiskBack Regular |> toHtml []
                , Phosphor.floppyDiskBack Bold |> toHtml []
                , Phosphor.floppyDiskBack Fill |> toHtml []
                , Phosphor.floppyDiskBack Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.floppyDisk model.weight |> toHtml []
                , Phosphor.floppyDisk Thin |> toHtml []
                , Phosphor.floppyDisk Light |> toHtml []
                , Phosphor.floppyDisk Regular |> toHtml []
                , Phosphor.floppyDisk Bold |> toHtml []
                , Phosphor.floppyDisk Fill |> toHtml []
                , Phosphor.floppyDisk Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flowArrow model.weight |> toHtml []
                , Phosphor.flowArrow Thin |> toHtml []
                , Phosphor.flowArrow Light |> toHtml []
                , Phosphor.flowArrow Regular |> toHtml []
                , Phosphor.flowArrow Bold |> toHtml []
                , Phosphor.flowArrow Fill |> toHtml []
                , Phosphor.flowArrow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flower model.weight |> toHtml []
                , Phosphor.flower Thin |> toHtml []
                , Phosphor.flower Light |> toHtml []
                , Phosphor.flower Regular |> toHtml []
                , Phosphor.flower Bold |> toHtml []
                , Phosphor.flower Fill |> toHtml []
                , Phosphor.flower Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flowerLotus model.weight |> toHtml []
                , Phosphor.flowerLotus Thin |> toHtml []
                , Phosphor.flowerLotus Light |> toHtml []
                , Phosphor.flowerLotus Regular |> toHtml []
                , Phosphor.flowerLotus Bold |> toHtml []
                , Phosphor.flowerLotus Fill |> toHtml []
                , Phosphor.flowerLotus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flowerTulip model.weight |> toHtml []
                , Phosphor.flowerTulip Thin |> toHtml []
                , Phosphor.flowerTulip Light |> toHtml []
                , Phosphor.flowerTulip Regular |> toHtml []
                , Phosphor.flowerTulip Bold |> toHtml []
                , Phosphor.flowerTulip Fill |> toHtml []
                , Phosphor.flowerTulip Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.flyingSaucer model.weight |> toHtml []
                , Phosphor.flyingSaucer Thin |> toHtml []
                , Phosphor.flyingSaucer Light |> toHtml []
                , Phosphor.flyingSaucer Regular |> toHtml []
                , Phosphor.flyingSaucer Bold |> toHtml []
                , Phosphor.flyingSaucer Fill |> toHtml []
                , Phosphor.flyingSaucer Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folder model.weight |> toHtml []
                , Phosphor.folder Thin |> toHtml []
                , Phosphor.folder Light |> toHtml []
                , Phosphor.folder Regular |> toHtml []
                , Phosphor.folder Bold |> toHtml []
                , Phosphor.folder Fill |> toHtml []
                , Phosphor.folder Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderDashed model.weight |> toHtml []
                , Phosphor.folderDashed Thin |> toHtml []
                , Phosphor.folderDashed Light |> toHtml []
                , Phosphor.folderDashed Regular |> toHtml []
                , Phosphor.folderDashed Bold |> toHtml []
                , Phosphor.folderDashed Fill |> toHtml []
                , Phosphor.folderDashed Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderLock model.weight |> toHtml []
                , Phosphor.folderLock Thin |> toHtml []
                , Phosphor.folderLock Light |> toHtml []
                , Phosphor.folderLock Regular |> toHtml []
                , Phosphor.folderLock Bold |> toHtml []
                , Phosphor.folderLock Fill |> toHtml []
                , Phosphor.folderLock Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderMinus model.weight |> toHtml []
                , Phosphor.folderMinus Thin |> toHtml []
                , Phosphor.folderMinus Light |> toHtml []
                , Phosphor.folderMinus Regular |> toHtml []
                , Phosphor.folderMinus Bold |> toHtml []
                , Phosphor.folderMinus Fill |> toHtml []
                , Phosphor.folderMinus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderNotch model.weight |> toHtml []
                , Phosphor.folderNotch Thin |> toHtml []
                , Phosphor.folderNotch Light |> toHtml []
                , Phosphor.folderNotch Regular |> toHtml []
                , Phosphor.folderNotch Bold |> toHtml []
                , Phosphor.folderNotch Fill |> toHtml []
                , Phosphor.folderNotch Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderNotchMinus model.weight |> toHtml []
                , Phosphor.folderNotchMinus Thin |> toHtml []
                , Phosphor.folderNotchMinus Light |> toHtml []
                , Phosphor.folderNotchMinus Regular |> toHtml []
                , Phosphor.folderNotchMinus Bold |> toHtml []
                , Phosphor.folderNotchMinus Fill |> toHtml []
                , Phosphor.folderNotchMinus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderNotchOpen model.weight |> toHtml []
                , Phosphor.folderNotchOpen Thin |> toHtml []
                , Phosphor.folderNotchOpen Light |> toHtml []
                , Phosphor.folderNotchOpen Regular |> toHtml []
                , Phosphor.folderNotchOpen Bold |> toHtml []
                , Phosphor.folderNotchOpen Fill |> toHtml []
                , Phosphor.folderNotchOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderNotchPlus model.weight |> toHtml []
                , Phosphor.folderNotchPlus Thin |> toHtml []
                , Phosphor.folderNotchPlus Light |> toHtml []
                , Phosphor.folderNotchPlus Regular |> toHtml []
                , Phosphor.folderNotchPlus Bold |> toHtml []
                , Phosphor.folderNotchPlus Fill |> toHtml []
                , Phosphor.folderNotchPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderOpen model.weight |> toHtml []
                , Phosphor.folderOpen Thin |> toHtml []
                , Phosphor.folderOpen Light |> toHtml []
                , Phosphor.folderOpen Regular |> toHtml []
                , Phosphor.folderOpen Bold |> toHtml []
                , Phosphor.folderOpen Fill |> toHtml []
                , Phosphor.folderOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderPlus model.weight |> toHtml []
                , Phosphor.folderPlus Thin |> toHtml []
                , Phosphor.folderPlus Light |> toHtml []
                , Phosphor.folderPlus Regular |> toHtml []
                , Phosphor.folderPlus Bold |> toHtml []
                , Phosphor.folderPlus Fill |> toHtml []
                , Phosphor.folderPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderSimple model.weight |> toHtml []
                , Phosphor.folderSimple Thin |> toHtml []
                , Phosphor.folderSimple Light |> toHtml []
                , Phosphor.folderSimple Regular |> toHtml []
                , Phosphor.folderSimple Bold |> toHtml []
                , Phosphor.folderSimple Fill |> toHtml []
                , Phosphor.folderSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderSimpleDashed model.weight |> toHtml []
                , Phosphor.folderSimpleDashed Thin |> toHtml []
                , Phosphor.folderSimpleDashed Light |> toHtml []
                , Phosphor.folderSimpleDashed Regular |> toHtml []
                , Phosphor.folderSimpleDashed Bold |> toHtml []
                , Phosphor.folderSimpleDashed Fill |> toHtml []
                , Phosphor.folderSimpleDashed Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderSimpleLock model.weight |> toHtml []
                , Phosphor.folderSimpleLock Thin |> toHtml []
                , Phosphor.folderSimpleLock Light |> toHtml []
                , Phosphor.folderSimpleLock Regular |> toHtml []
                , Phosphor.folderSimpleLock Bold |> toHtml []
                , Phosphor.folderSimpleLock Fill |> toHtml []
                , Phosphor.folderSimpleLock Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderSimpleMinus model.weight |> toHtml []
                , Phosphor.folderSimpleMinus Thin |> toHtml []
                , Phosphor.folderSimpleMinus Light |> toHtml []
                , Phosphor.folderSimpleMinus Regular |> toHtml []
                , Phosphor.folderSimpleMinus Bold |> toHtml []
                , Phosphor.folderSimpleMinus Fill |> toHtml []
                , Phosphor.folderSimpleMinus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderSimplePlus model.weight |> toHtml []
                , Phosphor.folderSimplePlus Thin |> toHtml []
                , Phosphor.folderSimplePlus Light |> toHtml []
                , Phosphor.folderSimplePlus Regular |> toHtml []
                , Phosphor.folderSimplePlus Bold |> toHtml []
                , Phosphor.folderSimplePlus Fill |> toHtml []
                , Phosphor.folderSimplePlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderSimpleStar model.weight |> toHtml []
                , Phosphor.folderSimpleStar Thin |> toHtml []
                , Phosphor.folderSimpleStar Light |> toHtml []
                , Phosphor.folderSimpleStar Regular |> toHtml []
                , Phosphor.folderSimpleStar Bold |> toHtml []
                , Phosphor.folderSimpleStar Fill |> toHtml []
                , Phosphor.folderSimpleStar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderSimpleUser model.weight |> toHtml []
                , Phosphor.folderSimpleUser Thin |> toHtml []
                , Phosphor.folderSimpleUser Light |> toHtml []
                , Phosphor.folderSimpleUser Regular |> toHtml []
                , Phosphor.folderSimpleUser Bold |> toHtml []
                , Phosphor.folderSimpleUser Fill |> toHtml []
                , Phosphor.folderSimpleUser Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderStar model.weight |> toHtml []
                , Phosphor.folderStar Thin |> toHtml []
                , Phosphor.folderStar Light |> toHtml []
                , Phosphor.folderStar Regular |> toHtml []
                , Phosphor.folderStar Bold |> toHtml []
                , Phosphor.folderStar Fill |> toHtml []
                , Phosphor.folderStar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folderUser model.weight |> toHtml []
                , Phosphor.folderUser Thin |> toHtml []
                , Phosphor.folderUser Light |> toHtml []
                , Phosphor.folderUser Regular |> toHtml []
                , Phosphor.folderUser Bold |> toHtml []
                , Phosphor.folderUser Fill |> toHtml []
                , Phosphor.folderUser Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.folders model.weight |> toHtml []
                , Phosphor.folders Thin |> toHtml []
                , Phosphor.folders Light |> toHtml []
                , Phosphor.folders Regular |> toHtml []
                , Phosphor.folders Bold |> toHtml []
                , Phosphor.folders Fill |> toHtml []
                , Phosphor.folders Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.football model.weight |> toHtml []
                , Phosphor.football Thin |> toHtml []
                , Phosphor.football Light |> toHtml []
                , Phosphor.football Regular |> toHtml []
                , Phosphor.football Bold |> toHtml []
                , Phosphor.football Fill |> toHtml []
                , Phosphor.football Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.footprints model.weight |> toHtml []
                , Phosphor.footprints Thin |> toHtml []
                , Phosphor.footprints Light |> toHtml []
                , Phosphor.footprints Regular |> toHtml []
                , Phosphor.footprints Bold |> toHtml []
                , Phosphor.footprints Fill |> toHtml []
                , Phosphor.footprints Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.forkKnife model.weight |> toHtml []
                , Phosphor.forkKnife Thin |> toHtml []
                , Phosphor.forkKnife Light |> toHtml []
                , Phosphor.forkKnife Regular |> toHtml []
                , Phosphor.forkKnife Bold |> toHtml []
                , Phosphor.forkKnife Fill |> toHtml []
                , Phosphor.forkKnife Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.frameCorners model.weight |> toHtml []
                , Phosphor.frameCorners Thin |> toHtml []
                , Phosphor.frameCorners Light |> toHtml []
                , Phosphor.frameCorners Regular |> toHtml []
                , Phosphor.frameCorners Bold |> toHtml []
                , Phosphor.frameCorners Fill |> toHtml []
                , Phosphor.frameCorners Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.framerLogo model.weight |> toHtml []
                , Phosphor.framerLogo Thin |> toHtml []
                , Phosphor.framerLogo Light |> toHtml []
                , Phosphor.framerLogo Regular |> toHtml []
                , Phosphor.framerLogo Bold |> toHtml []
                , Phosphor.framerLogo Fill |> toHtml []
                , Phosphor.framerLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.function model.weight |> toHtml []
                , Phosphor.function Thin |> toHtml []
                , Phosphor.function Light |> toHtml []
                , Phosphor.function Regular |> toHtml []
                , Phosphor.function Bold |> toHtml []
                , Phosphor.function Fill |> toHtml []
                , Phosphor.function Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.funnel model.weight |> toHtml []
                , Phosphor.funnel Thin |> toHtml []
                , Phosphor.funnel Light |> toHtml []
                , Phosphor.funnel Regular |> toHtml []
                , Phosphor.funnel Bold |> toHtml []
                , Phosphor.funnel Fill |> toHtml []
                , Phosphor.funnel Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.funnelSimple model.weight |> toHtml []
                , Phosphor.funnelSimple Thin |> toHtml []
                , Phosphor.funnelSimple Light |> toHtml []
                , Phosphor.funnelSimple Regular |> toHtml []
                , Phosphor.funnelSimple Bold |> toHtml []
                , Phosphor.funnelSimple Fill |> toHtml []
                , Phosphor.funnelSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gameController model.weight |> toHtml []
                , Phosphor.gameController Thin |> toHtml []
                , Phosphor.gameController Light |> toHtml []
                , Phosphor.gameController Regular |> toHtml []
                , Phosphor.gameController Bold |> toHtml []
                , Phosphor.gameController Fill |> toHtml []
                , Phosphor.gameController Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.garage model.weight |> toHtml []
                , Phosphor.garage Thin |> toHtml []
                , Phosphor.garage Light |> toHtml []
                , Phosphor.garage Regular |> toHtml []
                , Phosphor.garage Bold |> toHtml []
                , Phosphor.garage Fill |> toHtml []
                , Phosphor.garage Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gasCan model.weight |> toHtml []
                , Phosphor.gasCan Thin |> toHtml []
                , Phosphor.gasCan Light |> toHtml []
                , Phosphor.gasCan Regular |> toHtml []
                , Phosphor.gasCan Bold |> toHtml []
                , Phosphor.gasCan Fill |> toHtml []
                , Phosphor.gasCan Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gasPump model.weight |> toHtml []
                , Phosphor.gasPump Thin |> toHtml []
                , Phosphor.gasPump Light |> toHtml []
                , Phosphor.gasPump Regular |> toHtml []
                , Phosphor.gasPump Bold |> toHtml []
                , Phosphor.gasPump Fill |> toHtml []
                , Phosphor.gasPump Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gauge model.weight |> toHtml []
                , Phosphor.gauge Thin |> toHtml []
                , Phosphor.gauge Light |> toHtml []
                , Phosphor.gauge Regular |> toHtml []
                , Phosphor.gauge Bold |> toHtml []
                , Phosphor.gauge Fill |> toHtml []
                , Phosphor.gauge Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gavel model.weight |> toHtml []
                , Phosphor.gavel Thin |> toHtml []
                , Phosphor.gavel Light |> toHtml []
                , Phosphor.gavel Regular |> toHtml []
                , Phosphor.gavel Bold |> toHtml []
                , Phosphor.gavel Fill |> toHtml []
                , Phosphor.gavel Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gear model.weight |> toHtml []
                , Phosphor.gear Thin |> toHtml []
                , Phosphor.gear Light |> toHtml []
                , Phosphor.gear Regular |> toHtml []
                , Phosphor.gear Bold |> toHtml []
                , Phosphor.gear Fill |> toHtml []
                , Phosphor.gear Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gearFine model.weight |> toHtml []
                , Phosphor.gearFine Thin |> toHtml []
                , Phosphor.gearFine Light |> toHtml []
                , Phosphor.gearFine Regular |> toHtml []
                , Phosphor.gearFine Bold |> toHtml []
                , Phosphor.gearFine Fill |> toHtml []
                , Phosphor.gearFine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gearSix model.weight |> toHtml []
                , Phosphor.gearSix Thin |> toHtml []
                , Phosphor.gearSix Light |> toHtml []
                , Phosphor.gearSix Regular |> toHtml []
                , Phosphor.gearSix Bold |> toHtml []
                , Phosphor.gearSix Fill |> toHtml []
                , Phosphor.gearSix Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.genderFemale model.weight |> toHtml []
                , Phosphor.genderFemale Thin |> toHtml []
                , Phosphor.genderFemale Light |> toHtml []
                , Phosphor.genderFemale Regular |> toHtml []
                , Phosphor.genderFemale Bold |> toHtml []
                , Phosphor.genderFemale Fill |> toHtml []
                , Phosphor.genderFemale Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.genderIntersex model.weight |> toHtml []
                , Phosphor.genderIntersex Thin |> toHtml []
                , Phosphor.genderIntersex Light |> toHtml []
                , Phosphor.genderIntersex Regular |> toHtml []
                , Phosphor.genderIntersex Bold |> toHtml []
                , Phosphor.genderIntersex Fill |> toHtml []
                , Phosphor.genderIntersex Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.genderMale model.weight |> toHtml []
                , Phosphor.genderMale Thin |> toHtml []
                , Phosphor.genderMale Light |> toHtml []
                , Phosphor.genderMale Regular |> toHtml []
                , Phosphor.genderMale Bold |> toHtml []
                , Phosphor.genderMale Fill |> toHtml []
                , Phosphor.genderMale Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.genderNeuter model.weight |> toHtml []
                , Phosphor.genderNeuter Thin |> toHtml []
                , Phosphor.genderNeuter Light |> toHtml []
                , Phosphor.genderNeuter Regular |> toHtml []
                , Phosphor.genderNeuter Bold |> toHtml []
                , Phosphor.genderNeuter Fill |> toHtml []
                , Phosphor.genderNeuter Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.genderNonbinary model.weight |> toHtml []
                , Phosphor.genderNonbinary Thin |> toHtml []
                , Phosphor.genderNonbinary Light |> toHtml []
                , Phosphor.genderNonbinary Regular |> toHtml []
                , Phosphor.genderNonbinary Bold |> toHtml []
                , Phosphor.genderNonbinary Fill |> toHtml []
                , Phosphor.genderNonbinary Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.genderTransgender model.weight |> toHtml []
                , Phosphor.genderTransgender Thin |> toHtml []
                , Phosphor.genderTransgender Light |> toHtml []
                , Phosphor.genderTransgender Regular |> toHtml []
                , Phosphor.genderTransgender Bold |> toHtml []
                , Phosphor.genderTransgender Fill |> toHtml []
                , Phosphor.genderTransgender Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.ghost model.weight |> toHtml []
                , Phosphor.ghost Thin |> toHtml []
                , Phosphor.ghost Light |> toHtml []
                , Phosphor.ghost Regular |> toHtml []
                , Phosphor.ghost Bold |> toHtml []
                , Phosphor.ghost Fill |> toHtml []
                , Phosphor.ghost Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gif model.weight |> toHtml []
                , Phosphor.gif Thin |> toHtml []
                , Phosphor.gif Light |> toHtml []
                , Phosphor.gif Regular |> toHtml []
                , Phosphor.gif Bold |> toHtml []
                , Phosphor.gif Fill |> toHtml []
                , Phosphor.gif Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gift model.weight |> toHtml []
                , Phosphor.gift Thin |> toHtml []
                , Phosphor.gift Light |> toHtml []
                , Phosphor.gift Regular |> toHtml []
                , Phosphor.gift Bold |> toHtml []
                , Phosphor.gift Fill |> toHtml []
                , Phosphor.gift Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gitBranch model.weight |> toHtml []
                , Phosphor.gitBranch Thin |> toHtml []
                , Phosphor.gitBranch Light |> toHtml []
                , Phosphor.gitBranch Regular |> toHtml []
                , Phosphor.gitBranch Bold |> toHtml []
                , Phosphor.gitBranch Fill |> toHtml []
                , Phosphor.gitBranch Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gitCommit model.weight |> toHtml []
                , Phosphor.gitCommit Thin |> toHtml []
                , Phosphor.gitCommit Light |> toHtml []
                , Phosphor.gitCommit Regular |> toHtml []
                , Phosphor.gitCommit Bold |> toHtml []
                , Phosphor.gitCommit Fill |> toHtml []
                , Phosphor.gitCommit Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gitDiff model.weight |> toHtml []
                , Phosphor.gitDiff Thin |> toHtml []
                , Phosphor.gitDiff Light |> toHtml []
                , Phosphor.gitDiff Regular |> toHtml []
                , Phosphor.gitDiff Bold |> toHtml []
                , Phosphor.gitDiff Fill |> toHtml []
                , Phosphor.gitDiff Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gitFork model.weight |> toHtml []
                , Phosphor.gitFork Thin |> toHtml []
                , Phosphor.gitFork Light |> toHtml []
                , Phosphor.gitFork Regular |> toHtml []
                , Phosphor.gitFork Bold |> toHtml []
                , Phosphor.gitFork Fill |> toHtml []
                , Phosphor.gitFork Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gitMerge model.weight |> toHtml []
                , Phosphor.gitMerge Thin |> toHtml []
                , Phosphor.gitMerge Light |> toHtml []
                , Phosphor.gitMerge Regular |> toHtml []
                , Phosphor.gitMerge Bold |> toHtml []
                , Phosphor.gitMerge Fill |> toHtml []
                , Phosphor.gitMerge Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gitPullRequest model.weight |> toHtml []
                , Phosphor.gitPullRequest Thin |> toHtml []
                , Phosphor.gitPullRequest Light |> toHtml []
                , Phosphor.gitPullRequest Regular |> toHtml []
                , Phosphor.gitPullRequest Bold |> toHtml []
                , Phosphor.gitPullRequest Fill |> toHtml []
                , Phosphor.gitPullRequest Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.githubLogo model.weight |> toHtml []
                , Phosphor.githubLogo Thin |> toHtml []
                , Phosphor.githubLogo Light |> toHtml []
                , Phosphor.githubLogo Regular |> toHtml []
                , Phosphor.githubLogo Bold |> toHtml []
                , Phosphor.githubLogo Fill |> toHtml []
                , Phosphor.githubLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gitlabLogo model.weight |> toHtml []
                , Phosphor.gitlabLogo Thin |> toHtml []
                , Phosphor.gitlabLogo Light |> toHtml []
                , Phosphor.gitlabLogo Regular |> toHtml []
                , Phosphor.gitlabLogo Bold |> toHtml []
                , Phosphor.gitlabLogo Fill |> toHtml []
                , Phosphor.gitlabLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gitlabLogoSimple model.weight |> toHtml []
                , Phosphor.gitlabLogoSimple Thin |> toHtml []
                , Phosphor.gitlabLogoSimple Light |> toHtml []
                , Phosphor.gitlabLogoSimple Regular |> toHtml []
                , Phosphor.gitlabLogoSimple Bold |> toHtml []
                , Phosphor.gitlabLogoSimple Fill |> toHtml []
                , Phosphor.gitlabLogoSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.globe model.weight |> toHtml []
                , Phosphor.globe Thin |> toHtml []
                , Phosphor.globe Light |> toHtml []
                , Phosphor.globe Regular |> toHtml []
                , Phosphor.globe Bold |> toHtml []
                , Phosphor.globe Fill |> toHtml []
                , Phosphor.globe Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.globeHemisphereEast model.weight |> toHtml []
                , Phosphor.globeHemisphereEast Thin |> toHtml []
                , Phosphor.globeHemisphereEast Light |> toHtml []
                , Phosphor.globeHemisphereEast Regular |> toHtml []
                , Phosphor.globeHemisphereEast Bold |> toHtml []
                , Phosphor.globeHemisphereEast Fill |> toHtml []
                , Phosphor.globeHemisphereEast Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.globeHemisphereWest model.weight |> toHtml []
                , Phosphor.globeHemisphereWest Thin |> toHtml []
                , Phosphor.globeHemisphereWest Light |> toHtml []
                , Phosphor.globeHemisphereWest Regular |> toHtml []
                , Phosphor.globeHemisphereWest Bold |> toHtml []
                , Phosphor.globeHemisphereWest Fill |> toHtml []
                , Phosphor.globeHemisphereWest Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.globeSimple model.weight |> toHtml []
                , Phosphor.globeSimple Thin |> toHtml []
                , Phosphor.globeSimple Light |> toHtml []
                , Phosphor.globeSimple Regular |> toHtml []
                , Phosphor.globeSimple Bold |> toHtml []
                , Phosphor.globeSimple Fill |> toHtml []
                , Phosphor.globeSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.globeStand model.weight |> toHtml []
                , Phosphor.globeStand Thin |> toHtml []
                , Phosphor.globeStand Light |> toHtml []
                , Phosphor.globeStand Regular |> toHtml []
                , Phosphor.globeStand Bold |> toHtml []
                , Phosphor.globeStand Fill |> toHtml []
                , Phosphor.globeStand Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.goggles model.weight |> toHtml []
                , Phosphor.goggles Thin |> toHtml []
                , Phosphor.goggles Light |> toHtml []
                , Phosphor.goggles Regular |> toHtml []
                , Phosphor.goggles Bold |> toHtml []
                , Phosphor.goggles Fill |> toHtml []
                , Phosphor.goggles Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.goodreadsLogo model.weight |> toHtml []
                , Phosphor.goodreadsLogo Thin |> toHtml []
                , Phosphor.goodreadsLogo Light |> toHtml []
                , Phosphor.goodreadsLogo Regular |> toHtml []
                , Phosphor.goodreadsLogo Bold |> toHtml []
                , Phosphor.goodreadsLogo Fill |> toHtml []
                , Phosphor.goodreadsLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.googleCardboardLogo model.weight |> toHtml []
                , Phosphor.googleCardboardLogo Thin |> toHtml []
                , Phosphor.googleCardboardLogo Light |> toHtml []
                , Phosphor.googleCardboardLogo Regular |> toHtml []
                , Phosphor.googleCardboardLogo Bold |> toHtml []
                , Phosphor.googleCardboardLogo Fill |> toHtml []
                , Phosphor.googleCardboardLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.googleChromeLogo model.weight |> toHtml []
                , Phosphor.googleChromeLogo Thin |> toHtml []
                , Phosphor.googleChromeLogo Light |> toHtml []
                , Phosphor.googleChromeLogo Regular |> toHtml []
                , Phosphor.googleChromeLogo Bold |> toHtml []
                , Phosphor.googleChromeLogo Fill |> toHtml []
                , Phosphor.googleChromeLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.googleDriveLogo model.weight |> toHtml []
                , Phosphor.googleDriveLogo Thin |> toHtml []
                , Phosphor.googleDriveLogo Light |> toHtml []
                , Phosphor.googleDriveLogo Regular |> toHtml []
                , Phosphor.googleDriveLogo Bold |> toHtml []
                , Phosphor.googleDriveLogo Fill |> toHtml []
                , Phosphor.googleDriveLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.googleLogo model.weight |> toHtml []
                , Phosphor.googleLogo Thin |> toHtml []
                , Phosphor.googleLogo Light |> toHtml []
                , Phosphor.googleLogo Regular |> toHtml []
                , Phosphor.googleLogo Bold |> toHtml []
                , Phosphor.googleLogo Fill |> toHtml []
                , Phosphor.googleLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.googlePhotosLogo model.weight |> toHtml []
                , Phosphor.googlePhotosLogo Thin |> toHtml []
                , Phosphor.googlePhotosLogo Light |> toHtml []
                , Phosphor.googlePhotosLogo Regular |> toHtml []
                , Phosphor.googlePhotosLogo Bold |> toHtml []
                , Phosphor.googlePhotosLogo Fill |> toHtml []
                , Phosphor.googlePhotosLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.googlePlayLogo model.weight |> toHtml []
                , Phosphor.googlePlayLogo Thin |> toHtml []
                , Phosphor.googlePlayLogo Light |> toHtml []
                , Phosphor.googlePlayLogo Regular |> toHtml []
                , Phosphor.googlePlayLogo Bold |> toHtml []
                , Phosphor.googlePlayLogo Fill |> toHtml []
                , Phosphor.googlePlayLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.googlePodcastsLogo model.weight |> toHtml []
                , Phosphor.googlePodcastsLogo Thin |> toHtml []
                , Phosphor.googlePodcastsLogo Light |> toHtml []
                , Phosphor.googlePodcastsLogo Regular |> toHtml []
                , Phosphor.googlePodcastsLogo Bold |> toHtml []
                , Phosphor.googlePodcastsLogo Fill |> toHtml []
                , Phosphor.googlePodcastsLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gradient model.weight |> toHtml []
                , Phosphor.gradient Thin |> toHtml []
                , Phosphor.gradient Light |> toHtml []
                , Phosphor.gradient Regular |> toHtml []
                , Phosphor.gradient Bold |> toHtml []
                , Phosphor.gradient Fill |> toHtml []
                , Phosphor.gradient Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.graduationCap model.weight |> toHtml []
                , Phosphor.graduationCap Thin |> toHtml []
                , Phosphor.graduationCap Light |> toHtml []
                , Phosphor.graduationCap Regular |> toHtml []
                , Phosphor.graduationCap Bold |> toHtml []
                , Phosphor.graduationCap Fill |> toHtml []
                , Phosphor.graduationCap Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.grains model.weight |> toHtml []
                , Phosphor.grains Thin |> toHtml []
                , Phosphor.grains Light |> toHtml []
                , Phosphor.grains Regular |> toHtml []
                , Phosphor.grains Bold |> toHtml []
                , Phosphor.grains Fill |> toHtml []
                , Phosphor.grains Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.grainsSlash model.weight |> toHtml []
                , Phosphor.grainsSlash Thin |> toHtml []
                , Phosphor.grainsSlash Light |> toHtml []
                , Phosphor.grainsSlash Regular |> toHtml []
                , Phosphor.grainsSlash Bold |> toHtml []
                , Phosphor.grainsSlash Fill |> toHtml []
                , Phosphor.grainsSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.graph model.weight |> toHtml []
                , Phosphor.graph Thin |> toHtml []
                , Phosphor.graph Light |> toHtml []
                , Phosphor.graph Regular |> toHtml []
                , Phosphor.graph Bold |> toHtml []
                , Phosphor.graph Fill |> toHtml []
                , Phosphor.graph Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gridFour model.weight |> toHtml []
                , Phosphor.gridFour Thin |> toHtml []
                , Phosphor.gridFour Light |> toHtml []
                , Phosphor.gridFour Regular |> toHtml []
                , Phosphor.gridFour Bold |> toHtml []
                , Phosphor.gridFour Fill |> toHtml []
                , Phosphor.gridFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.gridNine model.weight |> toHtml []
                , Phosphor.gridNine Thin |> toHtml []
                , Phosphor.gridNine Light |> toHtml []
                , Phosphor.gridNine Regular |> toHtml []
                , Phosphor.gridNine Bold |> toHtml []
                , Phosphor.gridNine Fill |> toHtml []
                , Phosphor.gridNine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.guitar model.weight |> toHtml []
                , Phosphor.guitar Thin |> toHtml []
                , Phosphor.guitar Light |> toHtml []
                , Phosphor.guitar Regular |> toHtml []
                , Phosphor.guitar Bold |> toHtml []
                , Phosphor.guitar Fill |> toHtml []
                , Phosphor.guitar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hamburger model.weight |> toHtml []
                , Phosphor.hamburger Thin |> toHtml []
                , Phosphor.hamburger Light |> toHtml []
                , Phosphor.hamburger Regular |> toHtml []
                , Phosphor.hamburger Bold |> toHtml []
                , Phosphor.hamburger Fill |> toHtml []
                , Phosphor.hamburger Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hammer model.weight |> toHtml []
                , Phosphor.hammer Thin |> toHtml []
                , Phosphor.hammer Light |> toHtml []
                , Phosphor.hammer Regular |> toHtml []
                , Phosphor.hammer Bold |> toHtml []
                , Phosphor.hammer Fill |> toHtml []
                , Phosphor.hammer Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hand model.weight |> toHtml []
                , Phosphor.hand Thin |> toHtml []
                , Phosphor.hand Light |> toHtml []
                , Phosphor.hand Regular |> toHtml []
                , Phosphor.hand Bold |> toHtml []
                , Phosphor.hand Fill |> toHtml []
                , Phosphor.hand Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handCoins model.weight |> toHtml []
                , Phosphor.handCoins Thin |> toHtml []
                , Phosphor.handCoins Light |> toHtml []
                , Phosphor.handCoins Regular |> toHtml []
                , Phosphor.handCoins Bold |> toHtml []
                , Phosphor.handCoins Fill |> toHtml []
                , Phosphor.handCoins Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handEye model.weight |> toHtml []
                , Phosphor.handEye Thin |> toHtml []
                , Phosphor.handEye Light |> toHtml []
                , Phosphor.handEye Regular |> toHtml []
                , Phosphor.handEye Bold |> toHtml []
                , Phosphor.handEye Fill |> toHtml []
                , Phosphor.handEye Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handFist model.weight |> toHtml []
                , Phosphor.handFist Thin |> toHtml []
                , Phosphor.handFist Light |> toHtml []
                , Phosphor.handFist Regular |> toHtml []
                , Phosphor.handFist Bold |> toHtml []
                , Phosphor.handFist Fill |> toHtml []
                , Phosphor.handFist Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handGrabbing model.weight |> toHtml []
                , Phosphor.handGrabbing Thin |> toHtml []
                , Phosphor.handGrabbing Light |> toHtml []
                , Phosphor.handGrabbing Regular |> toHtml []
                , Phosphor.handGrabbing Bold |> toHtml []
                , Phosphor.handGrabbing Fill |> toHtml []
                , Phosphor.handGrabbing Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handHeart model.weight |> toHtml []
                , Phosphor.handHeart Thin |> toHtml []
                , Phosphor.handHeart Light |> toHtml []
                , Phosphor.handHeart Regular |> toHtml []
                , Phosphor.handHeart Bold |> toHtml []
                , Phosphor.handHeart Fill |> toHtml []
                , Phosphor.handHeart Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handPalm model.weight |> toHtml []
                , Phosphor.handPalm Thin |> toHtml []
                , Phosphor.handPalm Light |> toHtml []
                , Phosphor.handPalm Regular |> toHtml []
                , Phosphor.handPalm Bold |> toHtml []
                , Phosphor.handPalm Fill |> toHtml []
                , Phosphor.handPalm Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handPointing model.weight |> toHtml []
                , Phosphor.handPointing Thin |> toHtml []
                , Phosphor.handPointing Light |> toHtml []
                , Phosphor.handPointing Regular |> toHtml []
                , Phosphor.handPointing Bold |> toHtml []
                , Phosphor.handPointing Fill |> toHtml []
                , Phosphor.handPointing Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handSoap model.weight |> toHtml []
                , Phosphor.handSoap Thin |> toHtml []
                , Phosphor.handSoap Light |> toHtml []
                , Phosphor.handSoap Regular |> toHtml []
                , Phosphor.handSoap Bold |> toHtml []
                , Phosphor.handSoap Fill |> toHtml []
                , Phosphor.handSoap Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handSwipeLeft model.weight |> toHtml []
                , Phosphor.handSwipeLeft Thin |> toHtml []
                , Phosphor.handSwipeLeft Light |> toHtml []
                , Phosphor.handSwipeLeft Regular |> toHtml []
                , Phosphor.handSwipeLeft Bold |> toHtml []
                , Phosphor.handSwipeLeft Fill |> toHtml []
                , Phosphor.handSwipeLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handSwipeRight model.weight |> toHtml []
                , Phosphor.handSwipeRight Thin |> toHtml []
                , Phosphor.handSwipeRight Light |> toHtml []
                , Phosphor.handSwipeRight Regular |> toHtml []
                , Phosphor.handSwipeRight Bold |> toHtml []
                , Phosphor.handSwipeRight Fill |> toHtml []
                , Phosphor.handSwipeRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handTap model.weight |> toHtml []
                , Phosphor.handTap Thin |> toHtml []
                , Phosphor.handTap Light |> toHtml []
                , Phosphor.handTap Regular |> toHtml []
                , Phosphor.handTap Bold |> toHtml []
                , Phosphor.handTap Fill |> toHtml []
                , Phosphor.handTap Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handWaving model.weight |> toHtml []
                , Phosphor.handWaving Thin |> toHtml []
                , Phosphor.handWaving Light |> toHtml []
                , Phosphor.handWaving Regular |> toHtml []
                , Phosphor.handWaving Bold |> toHtml []
                , Phosphor.handWaving Fill |> toHtml []
                , Phosphor.handWaving Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handbag model.weight |> toHtml []
                , Phosphor.handbag Thin |> toHtml []
                , Phosphor.handbag Light |> toHtml []
                , Phosphor.handbag Regular |> toHtml []
                , Phosphor.handbag Bold |> toHtml []
                , Phosphor.handbag Fill |> toHtml []
                , Phosphor.handbag Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handbagSimple model.weight |> toHtml []
                , Phosphor.handbagSimple Thin |> toHtml []
                , Phosphor.handbagSimple Light |> toHtml []
                , Phosphor.handbagSimple Regular |> toHtml []
                , Phosphor.handbagSimple Bold |> toHtml []
                , Phosphor.handbagSimple Fill |> toHtml []
                , Phosphor.handbagSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handsClapping model.weight |> toHtml []
                , Phosphor.handsClapping Thin |> toHtml []
                , Phosphor.handsClapping Light |> toHtml []
                , Phosphor.handsClapping Regular |> toHtml []
                , Phosphor.handsClapping Bold |> toHtml []
                , Phosphor.handsClapping Fill |> toHtml []
                , Phosphor.handsClapping Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handsPraying model.weight |> toHtml []
                , Phosphor.handsPraying Thin |> toHtml []
                , Phosphor.handsPraying Light |> toHtml []
                , Phosphor.handsPraying Regular |> toHtml []
                , Phosphor.handsPraying Bold |> toHtml []
                , Phosphor.handsPraying Fill |> toHtml []
                , Phosphor.handsPraying Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.handshake model.weight |> toHtml []
                , Phosphor.handshake Thin |> toHtml []
                , Phosphor.handshake Light |> toHtml []
                , Phosphor.handshake Regular |> toHtml []
                , Phosphor.handshake Bold |> toHtml []
                , Phosphor.handshake Fill |> toHtml []
                , Phosphor.handshake Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hardDrive model.weight |> toHtml []
                , Phosphor.hardDrive Thin |> toHtml []
                , Phosphor.hardDrive Light |> toHtml []
                , Phosphor.hardDrive Regular |> toHtml []
                , Phosphor.hardDrive Bold |> toHtml []
                , Phosphor.hardDrive Fill |> toHtml []
                , Phosphor.hardDrive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hardDrives model.weight |> toHtml []
                , Phosphor.hardDrives Thin |> toHtml []
                , Phosphor.hardDrives Light |> toHtml []
                , Phosphor.hardDrives Regular |> toHtml []
                , Phosphor.hardDrives Bold |> toHtml []
                , Phosphor.hardDrives Fill |> toHtml []
                , Phosphor.hardDrives Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hash model.weight |> toHtml []
                , Phosphor.hash Thin |> toHtml []
                , Phosphor.hash Light |> toHtml []
                , Phosphor.hash Regular |> toHtml []
                , Phosphor.hash Bold |> toHtml []
                , Phosphor.hash Fill |> toHtml []
                , Phosphor.hash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hashStraight model.weight |> toHtml []
                , Phosphor.hashStraight Thin |> toHtml []
                , Phosphor.hashStraight Light |> toHtml []
                , Phosphor.hashStraight Regular |> toHtml []
                , Phosphor.hashStraight Bold |> toHtml []
                , Phosphor.hashStraight Fill |> toHtml []
                , Phosphor.hashStraight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.headlights model.weight |> toHtml []
                , Phosphor.headlights Thin |> toHtml []
                , Phosphor.headlights Light |> toHtml []
                , Phosphor.headlights Regular |> toHtml []
                , Phosphor.headlights Bold |> toHtml []
                , Phosphor.headlights Fill |> toHtml []
                , Phosphor.headlights Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.headphones model.weight |> toHtml []
                , Phosphor.headphones Thin |> toHtml []
                , Phosphor.headphones Light |> toHtml []
                , Phosphor.headphones Regular |> toHtml []
                , Phosphor.headphones Bold |> toHtml []
                , Phosphor.headphones Fill |> toHtml []
                , Phosphor.headphones Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.headset model.weight |> toHtml []
                , Phosphor.headset Thin |> toHtml []
                , Phosphor.headset Light |> toHtml []
                , Phosphor.headset Regular |> toHtml []
                , Phosphor.headset Bold |> toHtml []
                , Phosphor.headset Fill |> toHtml []
                , Phosphor.headset Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.heart model.weight |> toHtml []
                , Phosphor.heart Thin |> toHtml []
                , Phosphor.heart Light |> toHtml []
                , Phosphor.heart Regular |> toHtml []
                , Phosphor.heart Bold |> toHtml []
                , Phosphor.heart Fill |> toHtml []
                , Phosphor.heart Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.heartBreak model.weight |> toHtml []
                , Phosphor.heartBreak Thin |> toHtml []
                , Phosphor.heartBreak Light |> toHtml []
                , Phosphor.heartBreak Regular |> toHtml []
                , Phosphor.heartBreak Bold |> toHtml []
                , Phosphor.heartBreak Fill |> toHtml []
                , Phosphor.heartBreak Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.heartHalf model.weight |> toHtml []
                , Phosphor.heartHalf Thin |> toHtml []
                , Phosphor.heartHalf Light |> toHtml []
                , Phosphor.heartHalf Regular |> toHtml []
                , Phosphor.heartHalf Bold |> toHtml []
                , Phosphor.heartHalf Fill |> toHtml []
                , Phosphor.heartHalf Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.heartStraight model.weight |> toHtml []
                , Phosphor.heartStraight Thin |> toHtml []
                , Phosphor.heartStraight Light |> toHtml []
                , Phosphor.heartStraight Regular |> toHtml []
                , Phosphor.heartStraight Bold |> toHtml []
                , Phosphor.heartStraight Fill |> toHtml []
                , Phosphor.heartStraight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.heartStraightBreak model.weight |> toHtml []
                , Phosphor.heartStraightBreak Thin |> toHtml []
                , Phosphor.heartStraightBreak Light |> toHtml []
                , Phosphor.heartStraightBreak Regular |> toHtml []
                , Phosphor.heartStraightBreak Bold |> toHtml []
                , Phosphor.heartStraightBreak Fill |> toHtml []
                , Phosphor.heartStraightBreak Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.heartbeat model.weight |> toHtml []
                , Phosphor.heartbeat Thin |> toHtml []
                , Phosphor.heartbeat Light |> toHtml []
                , Phosphor.heartbeat Regular |> toHtml []
                , Phosphor.heartbeat Bold |> toHtml []
                , Phosphor.heartbeat Fill |> toHtml []
                , Phosphor.heartbeat Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hexagon model.weight |> toHtml []
                , Phosphor.hexagon Thin |> toHtml []
                , Phosphor.hexagon Light |> toHtml []
                , Phosphor.hexagon Regular |> toHtml []
                , Phosphor.hexagon Bold |> toHtml []
                , Phosphor.hexagon Fill |> toHtml []
                , Phosphor.hexagon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.highHeel model.weight |> toHtml []
                , Phosphor.highHeel Thin |> toHtml []
                , Phosphor.highHeel Light |> toHtml []
                , Phosphor.highHeel Regular |> toHtml []
                , Phosphor.highHeel Bold |> toHtml []
                , Phosphor.highHeel Fill |> toHtml []
                , Phosphor.highHeel Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.highlighterCircle model.weight |> toHtml []
                , Phosphor.highlighterCircle Thin |> toHtml []
                , Phosphor.highlighterCircle Light |> toHtml []
                , Phosphor.highlighterCircle Regular |> toHtml []
                , Phosphor.highlighterCircle Bold |> toHtml []
                , Phosphor.highlighterCircle Fill |> toHtml []
                , Phosphor.highlighterCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hoodie model.weight |> toHtml []
                , Phosphor.hoodie Thin |> toHtml []
                , Phosphor.hoodie Light |> toHtml []
                , Phosphor.hoodie Regular |> toHtml []
                , Phosphor.hoodie Bold |> toHtml []
                , Phosphor.hoodie Fill |> toHtml []
                , Phosphor.hoodie Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.horse model.weight |> toHtml []
                , Phosphor.horse Thin |> toHtml []
                , Phosphor.horse Light |> toHtml []
                , Phosphor.horse Regular |> toHtml []
                , Phosphor.horse Bold |> toHtml []
                , Phosphor.horse Fill |> toHtml []
                , Phosphor.horse Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hourglass model.weight |> toHtml []
                , Phosphor.hourglass Thin |> toHtml []
                , Phosphor.hourglass Light |> toHtml []
                , Phosphor.hourglass Regular |> toHtml []
                , Phosphor.hourglass Bold |> toHtml []
                , Phosphor.hourglass Fill |> toHtml []
                , Phosphor.hourglass Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hourglassHigh model.weight |> toHtml []
                , Phosphor.hourglassHigh Thin |> toHtml []
                , Phosphor.hourglassHigh Light |> toHtml []
                , Phosphor.hourglassHigh Regular |> toHtml []
                , Phosphor.hourglassHigh Bold |> toHtml []
                , Phosphor.hourglassHigh Fill |> toHtml []
                , Phosphor.hourglassHigh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hourglassLow model.weight |> toHtml []
                , Phosphor.hourglassLow Thin |> toHtml []
                , Phosphor.hourglassLow Light |> toHtml []
                , Phosphor.hourglassLow Regular |> toHtml []
                , Phosphor.hourglassLow Bold |> toHtml []
                , Phosphor.hourglassLow Fill |> toHtml []
                , Phosphor.hourglassLow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hourglassMedium model.weight |> toHtml []
                , Phosphor.hourglassMedium Thin |> toHtml []
                , Phosphor.hourglassMedium Light |> toHtml []
                , Phosphor.hourglassMedium Regular |> toHtml []
                , Phosphor.hourglassMedium Bold |> toHtml []
                , Phosphor.hourglassMedium Fill |> toHtml []
                , Phosphor.hourglassMedium Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hourglassSimple model.weight |> toHtml []
                , Phosphor.hourglassSimple Thin |> toHtml []
                , Phosphor.hourglassSimple Light |> toHtml []
                , Phosphor.hourglassSimple Regular |> toHtml []
                , Phosphor.hourglassSimple Bold |> toHtml []
                , Phosphor.hourglassSimple Fill |> toHtml []
                , Phosphor.hourglassSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hourglassSimpleHigh model.weight |> toHtml []
                , Phosphor.hourglassSimpleHigh Thin |> toHtml []
                , Phosphor.hourglassSimpleHigh Light |> toHtml []
                , Phosphor.hourglassSimpleHigh Regular |> toHtml []
                , Phosphor.hourglassSimpleHigh Bold |> toHtml []
                , Phosphor.hourglassSimpleHigh Fill |> toHtml []
                , Phosphor.hourglassSimpleHigh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hourglassSimpleLow model.weight |> toHtml []
                , Phosphor.hourglassSimpleLow Thin |> toHtml []
                , Phosphor.hourglassSimpleLow Light |> toHtml []
                , Phosphor.hourglassSimpleLow Regular |> toHtml []
                , Phosphor.hourglassSimpleLow Bold |> toHtml []
                , Phosphor.hourglassSimpleLow Fill |> toHtml []
                , Phosphor.hourglassSimpleLow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.hourglassSimpleMedium model.weight |> toHtml []
                , Phosphor.hourglassSimpleMedium Thin |> toHtml []
                , Phosphor.hourglassSimpleMedium Light |> toHtml []
                , Phosphor.hourglassSimpleMedium Regular |> toHtml []
                , Phosphor.hourglassSimpleMedium Bold |> toHtml []
                , Phosphor.hourglassSimpleMedium Fill |> toHtml []
                , Phosphor.hourglassSimpleMedium Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.house model.weight |> toHtml []
                , Phosphor.house Thin |> toHtml []
                , Phosphor.house Light |> toHtml []
                , Phosphor.house Regular |> toHtml []
                , Phosphor.house Bold |> toHtml []
                , Phosphor.house Fill |> toHtml []
                , Phosphor.house Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.houseLine model.weight |> toHtml []
                , Phosphor.houseLine Thin |> toHtml []
                , Phosphor.houseLine Light |> toHtml []
                , Phosphor.houseLine Regular |> toHtml []
                , Phosphor.houseLine Bold |> toHtml []
                , Phosphor.houseLine Fill |> toHtml []
                , Phosphor.houseLine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.houseSimple model.weight |> toHtml []
                , Phosphor.houseSimple Thin |> toHtml []
                , Phosphor.houseSimple Light |> toHtml []
                , Phosphor.houseSimple Regular |> toHtml []
                , Phosphor.houseSimple Bold |> toHtml []
                , Phosphor.houseSimple Fill |> toHtml []
                , Phosphor.houseSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.iceCream model.weight |> toHtml []
                , Phosphor.iceCream Thin |> toHtml []
                , Phosphor.iceCream Light |> toHtml []
                , Phosphor.iceCream Regular |> toHtml []
                , Phosphor.iceCream Bold |> toHtml []
                , Phosphor.iceCream Fill |> toHtml []
                , Phosphor.iceCream Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.identificationBadge model.weight |> toHtml []
                , Phosphor.identificationBadge Thin |> toHtml []
                , Phosphor.identificationBadge Light |> toHtml []
                , Phosphor.identificationBadge Regular |> toHtml []
                , Phosphor.identificationBadge Bold |> toHtml []
                , Phosphor.identificationBadge Fill |> toHtml []
                , Phosphor.identificationBadge Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.identificationCard model.weight |> toHtml []
                , Phosphor.identificationCard Thin |> toHtml []
                , Phosphor.identificationCard Light |> toHtml []
                , Phosphor.identificationCard Regular |> toHtml []
                , Phosphor.identificationCard Bold |> toHtml []
                , Phosphor.identificationCard Fill |> toHtml []
                , Phosphor.identificationCard Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.image model.weight |> toHtml []
                , Phosphor.image Thin |> toHtml []
                , Phosphor.image Light |> toHtml []
                , Phosphor.image Regular |> toHtml []
                , Phosphor.image Bold |> toHtml []
                , Phosphor.image Fill |> toHtml []
                , Phosphor.image Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.imageSquare model.weight |> toHtml []
                , Phosphor.imageSquare Thin |> toHtml []
                , Phosphor.imageSquare Light |> toHtml []
                , Phosphor.imageSquare Regular |> toHtml []
                , Phosphor.imageSquare Bold |> toHtml []
                , Phosphor.imageSquare Fill |> toHtml []
                , Phosphor.imageSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.images model.weight |> toHtml []
                , Phosphor.images Thin |> toHtml []
                , Phosphor.images Light |> toHtml []
                , Phosphor.images Regular |> toHtml []
                , Phosphor.images Bold |> toHtml []
                , Phosphor.images Fill |> toHtml []
                , Phosphor.images Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.imagesSquare model.weight |> toHtml []
                , Phosphor.imagesSquare Thin |> toHtml []
                , Phosphor.imagesSquare Light |> toHtml []
                , Phosphor.imagesSquare Regular |> toHtml []
                , Phosphor.imagesSquare Bold |> toHtml []
                , Phosphor.imagesSquare Fill |> toHtml []
                , Phosphor.imagesSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.infinity model.weight |> toHtml []
                , Phosphor.infinity Thin |> toHtml []
                , Phosphor.infinity Light |> toHtml []
                , Phosphor.infinity Regular |> toHtml []
                , Phosphor.infinity Bold |> toHtml []
                , Phosphor.infinity Fill |> toHtml []
                , Phosphor.infinity Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.info model.weight |> toHtml []
                , Phosphor.info Thin |> toHtml []
                , Phosphor.info Light |> toHtml []
                , Phosphor.info Regular |> toHtml []
                , Phosphor.info Bold |> toHtml []
                , Phosphor.info Fill |> toHtml []
                , Phosphor.info Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.instagramLogo model.weight |> toHtml []
                , Phosphor.instagramLogo Thin |> toHtml []
                , Phosphor.instagramLogo Light |> toHtml []
                , Phosphor.instagramLogo Regular |> toHtml []
                , Phosphor.instagramLogo Bold |> toHtml []
                , Phosphor.instagramLogo Fill |> toHtml []
                , Phosphor.instagramLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.intersect model.weight |> toHtml []
                , Phosphor.intersect Thin |> toHtml []
                , Phosphor.intersect Light |> toHtml []
                , Phosphor.intersect Regular |> toHtml []
                , Phosphor.intersect Bold |> toHtml []
                , Phosphor.intersect Fill |> toHtml []
                , Phosphor.intersect Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.intersectSquare model.weight |> toHtml []
                , Phosphor.intersectSquare Thin |> toHtml []
                , Phosphor.intersectSquare Light |> toHtml []
                , Phosphor.intersectSquare Regular |> toHtml []
                , Phosphor.intersectSquare Bold |> toHtml []
                , Phosphor.intersectSquare Fill |> toHtml []
                , Phosphor.intersectSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.intersectThree model.weight |> toHtml []
                , Phosphor.intersectThree Thin |> toHtml []
                , Phosphor.intersectThree Light |> toHtml []
                , Phosphor.intersectThree Regular |> toHtml []
                , Phosphor.intersectThree Bold |> toHtml []
                , Phosphor.intersectThree Fill |> toHtml []
                , Phosphor.intersectThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.jeep model.weight |> toHtml []
                , Phosphor.jeep Thin |> toHtml []
                , Phosphor.jeep Light |> toHtml []
                , Phosphor.jeep Regular |> toHtml []
                , Phosphor.jeep Bold |> toHtml []
                , Phosphor.jeep Fill |> toHtml []
                , Phosphor.jeep Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.kanban model.weight |> toHtml []
                , Phosphor.kanban Thin |> toHtml []
                , Phosphor.kanban Light |> toHtml []
                , Phosphor.kanban Regular |> toHtml []
                , Phosphor.kanban Bold |> toHtml []
                , Phosphor.kanban Fill |> toHtml []
                , Phosphor.kanban Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.key model.weight |> toHtml []
                , Phosphor.key Thin |> toHtml []
                , Phosphor.key Light |> toHtml []
                , Phosphor.key Regular |> toHtml []
                , Phosphor.key Bold |> toHtml []
                , Phosphor.key Fill |> toHtml []
                , Phosphor.key Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.keyReturn model.weight |> toHtml []
                , Phosphor.keyReturn Thin |> toHtml []
                , Phosphor.keyReturn Light |> toHtml []
                , Phosphor.keyReturn Regular |> toHtml []
                , Phosphor.keyReturn Bold |> toHtml []
                , Phosphor.keyReturn Fill |> toHtml []
                , Phosphor.keyReturn Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.keyboard model.weight |> toHtml []
                , Phosphor.keyboard Thin |> toHtml []
                , Phosphor.keyboard Light |> toHtml []
                , Phosphor.keyboard Regular |> toHtml []
                , Phosphor.keyboard Bold |> toHtml []
                , Phosphor.keyboard Fill |> toHtml []
                , Phosphor.keyboard Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.keyhole model.weight |> toHtml []
                , Phosphor.keyhole Thin |> toHtml []
                , Phosphor.keyhole Light |> toHtml []
                , Phosphor.keyhole Regular |> toHtml []
                , Phosphor.keyhole Bold |> toHtml []
                , Phosphor.keyhole Fill |> toHtml []
                , Phosphor.keyhole Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.knife model.weight |> toHtml []
                , Phosphor.knife Thin |> toHtml []
                , Phosphor.knife Light |> toHtml []
                , Phosphor.knife Regular |> toHtml []
                , Phosphor.knife Bold |> toHtml []
                , Phosphor.knife Fill |> toHtml []
                , Phosphor.knife Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.ladder model.weight |> toHtml []
                , Phosphor.ladder Thin |> toHtml []
                , Phosphor.ladder Light |> toHtml []
                , Phosphor.ladder Regular |> toHtml []
                , Phosphor.ladder Bold |> toHtml []
                , Phosphor.ladder Fill |> toHtml []
                , Phosphor.ladder Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.ladderSimple model.weight |> toHtml []
                , Phosphor.ladderSimple Thin |> toHtml []
                , Phosphor.ladderSimple Light |> toHtml []
                , Phosphor.ladderSimple Regular |> toHtml []
                , Phosphor.ladderSimple Bold |> toHtml []
                , Phosphor.ladderSimple Fill |> toHtml []
                , Phosphor.ladderSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lamp model.weight |> toHtml []
                , Phosphor.lamp Thin |> toHtml []
                , Phosphor.lamp Light |> toHtml []
                , Phosphor.lamp Regular |> toHtml []
                , Phosphor.lamp Bold |> toHtml []
                , Phosphor.lamp Fill |> toHtml []
                , Phosphor.lamp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.laptop model.weight |> toHtml []
                , Phosphor.laptop Thin |> toHtml []
                , Phosphor.laptop Light |> toHtml []
                , Phosphor.laptop Regular |> toHtml []
                , Phosphor.laptop Bold |> toHtml []
                , Phosphor.laptop Fill |> toHtml []
                , Phosphor.laptop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.layout model.weight |> toHtml []
                , Phosphor.layout Thin |> toHtml []
                , Phosphor.layout Light |> toHtml []
                , Phosphor.layout Regular |> toHtml []
                , Phosphor.layout Bold |> toHtml []
                , Phosphor.layout Fill |> toHtml []
                , Phosphor.layout Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.leaf model.weight |> toHtml []
                , Phosphor.leaf Thin |> toHtml []
                , Phosphor.leaf Light |> toHtml []
                , Phosphor.leaf Regular |> toHtml []
                , Phosphor.leaf Bold |> toHtml []
                , Phosphor.leaf Fill |> toHtml []
                , Phosphor.leaf Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lifebuoy model.weight |> toHtml []
                , Phosphor.lifebuoy Thin |> toHtml []
                , Phosphor.lifebuoy Light |> toHtml []
                , Phosphor.lifebuoy Regular |> toHtml []
                , Phosphor.lifebuoy Bold |> toHtml []
                , Phosphor.lifebuoy Fill |> toHtml []
                , Phosphor.lifebuoy Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lightbulb model.weight |> toHtml []
                , Phosphor.lightbulb Thin |> toHtml []
                , Phosphor.lightbulb Light |> toHtml []
                , Phosphor.lightbulb Regular |> toHtml []
                , Phosphor.lightbulb Bold |> toHtml []
                , Phosphor.lightbulb Fill |> toHtml []
                , Phosphor.lightbulb Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lightbulbFilament model.weight |> toHtml []
                , Phosphor.lightbulbFilament Thin |> toHtml []
                , Phosphor.lightbulbFilament Light |> toHtml []
                , Phosphor.lightbulbFilament Regular |> toHtml []
                , Phosphor.lightbulbFilament Bold |> toHtml []
                , Phosphor.lightbulbFilament Fill |> toHtml []
                , Phosphor.lightbulbFilament Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lighthouse model.weight |> toHtml []
                , Phosphor.lighthouse Thin |> toHtml []
                , Phosphor.lighthouse Light |> toHtml []
                , Phosphor.lighthouse Regular |> toHtml []
                , Phosphor.lighthouse Bold |> toHtml []
                , Phosphor.lighthouse Fill |> toHtml []
                , Phosphor.lighthouse Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lightningA model.weight |> toHtml []
                , Phosphor.lightningA Thin |> toHtml []
                , Phosphor.lightningA Light |> toHtml []
                , Phosphor.lightningA Regular |> toHtml []
                , Phosphor.lightningA Bold |> toHtml []
                , Phosphor.lightningA Fill |> toHtml []
                , Phosphor.lightningA Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lightning model.weight |> toHtml []
                , Phosphor.lightning Thin |> toHtml []
                , Phosphor.lightning Light |> toHtml []
                , Phosphor.lightning Regular |> toHtml []
                , Phosphor.lightning Bold |> toHtml []
                , Phosphor.lightning Fill |> toHtml []
                , Phosphor.lightning Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lightningSlash model.weight |> toHtml []
                , Phosphor.lightningSlash Thin |> toHtml []
                , Phosphor.lightningSlash Light |> toHtml []
                , Phosphor.lightningSlash Regular |> toHtml []
                , Phosphor.lightningSlash Bold |> toHtml []
                , Phosphor.lightningSlash Fill |> toHtml []
                , Phosphor.lightningSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lineSegment model.weight |> toHtml []
                , Phosphor.lineSegment Thin |> toHtml []
                , Phosphor.lineSegment Light |> toHtml []
                , Phosphor.lineSegment Regular |> toHtml []
                , Phosphor.lineSegment Bold |> toHtml []
                , Phosphor.lineSegment Fill |> toHtml []
                , Phosphor.lineSegment Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lineSegments model.weight |> toHtml []
                , Phosphor.lineSegments Thin |> toHtml []
                , Phosphor.lineSegments Light |> toHtml []
                , Phosphor.lineSegments Regular |> toHtml []
                , Phosphor.lineSegments Bold |> toHtml []
                , Phosphor.lineSegments Fill |> toHtml []
                , Phosphor.lineSegments Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.link model.weight |> toHtml []
                , Phosphor.link Thin |> toHtml []
                , Phosphor.link Light |> toHtml []
                , Phosphor.link Regular |> toHtml []
                , Phosphor.link Bold |> toHtml []
                , Phosphor.link Fill |> toHtml []
                , Phosphor.link Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.linkBreak model.weight |> toHtml []
                , Phosphor.linkBreak Thin |> toHtml []
                , Phosphor.linkBreak Light |> toHtml []
                , Phosphor.linkBreak Regular |> toHtml []
                , Phosphor.linkBreak Bold |> toHtml []
                , Phosphor.linkBreak Fill |> toHtml []
                , Phosphor.linkBreak Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.linkSimple model.weight |> toHtml []
                , Phosphor.linkSimple Thin |> toHtml []
                , Phosphor.linkSimple Light |> toHtml []
                , Phosphor.linkSimple Regular |> toHtml []
                , Phosphor.linkSimple Bold |> toHtml []
                , Phosphor.linkSimple Fill |> toHtml []
                , Phosphor.linkSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.linkSimpleBreak model.weight |> toHtml []
                , Phosphor.linkSimpleBreak Thin |> toHtml []
                , Phosphor.linkSimpleBreak Light |> toHtml []
                , Phosphor.linkSimpleBreak Regular |> toHtml []
                , Phosphor.linkSimpleBreak Bold |> toHtml []
                , Phosphor.linkSimpleBreak Fill |> toHtml []
                , Phosphor.linkSimpleBreak Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.linkSimpleHorizontal model.weight |> toHtml []
                , Phosphor.linkSimpleHorizontal Thin |> toHtml []
                , Phosphor.linkSimpleHorizontal Light |> toHtml []
                , Phosphor.linkSimpleHorizontal Regular |> toHtml []
                , Phosphor.linkSimpleHorizontal Bold |> toHtml []
                , Phosphor.linkSimpleHorizontal Fill |> toHtml []
                , Phosphor.linkSimpleHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.linkSimpleHorizontalBreak model.weight |> toHtml []
                , Phosphor.linkSimpleHorizontalBreak Thin |> toHtml []
                , Phosphor.linkSimpleHorizontalBreak Light |> toHtml []
                , Phosphor.linkSimpleHorizontalBreak Regular |> toHtml []
                , Phosphor.linkSimpleHorizontalBreak Bold |> toHtml []
                , Phosphor.linkSimpleHorizontalBreak Fill |> toHtml []
                , Phosphor.linkSimpleHorizontalBreak Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.linkedinLogo model.weight |> toHtml []
                , Phosphor.linkedinLogo Thin |> toHtml []
                , Phosphor.linkedinLogo Light |> toHtml []
                , Phosphor.linkedinLogo Regular |> toHtml []
                , Phosphor.linkedinLogo Bold |> toHtml []
                , Phosphor.linkedinLogo Fill |> toHtml []
                , Phosphor.linkedinLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.linuxLogo model.weight |> toHtml []
                , Phosphor.linuxLogo Thin |> toHtml []
                , Phosphor.linuxLogo Light |> toHtml []
                , Phosphor.linuxLogo Regular |> toHtml []
                , Phosphor.linuxLogo Bold |> toHtml []
                , Phosphor.linuxLogo Fill |> toHtml []
                , Phosphor.linuxLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.list model.weight |> toHtml []
                , Phosphor.list Thin |> toHtml []
                , Phosphor.list Light |> toHtml []
                , Phosphor.list Regular |> toHtml []
                , Phosphor.list Bold |> toHtml []
                , Phosphor.list Fill |> toHtml []
                , Phosphor.list Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.listBullets model.weight |> toHtml []
                , Phosphor.listBullets Thin |> toHtml []
                , Phosphor.listBullets Light |> toHtml []
                , Phosphor.listBullets Regular |> toHtml []
                , Phosphor.listBullets Bold |> toHtml []
                , Phosphor.listBullets Fill |> toHtml []
                , Phosphor.listBullets Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.listChecks model.weight |> toHtml []
                , Phosphor.listChecks Thin |> toHtml []
                , Phosphor.listChecks Light |> toHtml []
                , Phosphor.listChecks Regular |> toHtml []
                , Phosphor.listChecks Bold |> toHtml []
                , Phosphor.listChecks Fill |> toHtml []
                , Phosphor.listChecks Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.listDashes model.weight |> toHtml []
                , Phosphor.listDashes Thin |> toHtml []
                , Phosphor.listDashes Light |> toHtml []
                , Phosphor.listDashes Regular |> toHtml []
                , Phosphor.listDashes Bold |> toHtml []
                , Phosphor.listDashes Fill |> toHtml []
                , Phosphor.listDashes Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.listMagnifyingGlass model.weight |> toHtml []
                , Phosphor.listMagnifyingGlass Thin |> toHtml []
                , Phosphor.listMagnifyingGlass Light |> toHtml []
                , Phosphor.listMagnifyingGlass Regular |> toHtml []
                , Phosphor.listMagnifyingGlass Bold |> toHtml []
                , Phosphor.listMagnifyingGlass Fill |> toHtml []
                , Phosphor.listMagnifyingGlass Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.listNumbers model.weight |> toHtml []
                , Phosphor.listNumbers Thin |> toHtml []
                , Phosphor.listNumbers Light |> toHtml []
                , Phosphor.listNumbers Regular |> toHtml []
                , Phosphor.listNumbers Bold |> toHtml []
                , Phosphor.listNumbers Fill |> toHtml []
                , Phosphor.listNumbers Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.listPlus model.weight |> toHtml []
                , Phosphor.listPlus Thin |> toHtml []
                , Phosphor.listPlus Light |> toHtml []
                , Phosphor.listPlus Regular |> toHtml []
                , Phosphor.listPlus Bold |> toHtml []
                , Phosphor.listPlus Fill |> toHtml []
                , Phosphor.listPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lock model.weight |> toHtml []
                , Phosphor.lock Thin |> toHtml []
                , Phosphor.lock Light |> toHtml []
                , Phosphor.lock Regular |> toHtml []
                , Phosphor.lock Bold |> toHtml []
                , Phosphor.lock Fill |> toHtml []
                , Phosphor.lock Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lockKey model.weight |> toHtml []
                , Phosphor.lockKey Thin |> toHtml []
                , Phosphor.lockKey Light |> toHtml []
                , Phosphor.lockKey Regular |> toHtml []
                , Phosphor.lockKey Bold |> toHtml []
                , Phosphor.lockKey Fill |> toHtml []
                , Phosphor.lockKey Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lockKeyOpen model.weight |> toHtml []
                , Phosphor.lockKeyOpen Thin |> toHtml []
                , Phosphor.lockKeyOpen Light |> toHtml []
                , Phosphor.lockKeyOpen Regular |> toHtml []
                , Phosphor.lockKeyOpen Bold |> toHtml []
                , Phosphor.lockKeyOpen Fill |> toHtml []
                , Phosphor.lockKeyOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lockLaminated model.weight |> toHtml []
                , Phosphor.lockLaminated Thin |> toHtml []
                , Phosphor.lockLaminated Light |> toHtml []
                , Phosphor.lockLaminated Regular |> toHtml []
                , Phosphor.lockLaminated Bold |> toHtml []
                , Phosphor.lockLaminated Fill |> toHtml []
                , Phosphor.lockLaminated Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lockLaminatedOpen model.weight |> toHtml []
                , Phosphor.lockLaminatedOpen Thin |> toHtml []
                , Phosphor.lockLaminatedOpen Light |> toHtml []
                , Phosphor.lockLaminatedOpen Regular |> toHtml []
                , Phosphor.lockLaminatedOpen Bold |> toHtml []
                , Phosphor.lockLaminatedOpen Fill |> toHtml []
                , Phosphor.lockLaminatedOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lockOpen model.weight |> toHtml []
                , Phosphor.lockOpen Thin |> toHtml []
                , Phosphor.lockOpen Light |> toHtml []
                , Phosphor.lockOpen Regular |> toHtml []
                , Phosphor.lockOpen Bold |> toHtml []
                , Phosphor.lockOpen Fill |> toHtml []
                , Phosphor.lockOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lockSimple model.weight |> toHtml []
                , Phosphor.lockSimple Thin |> toHtml []
                , Phosphor.lockSimple Light |> toHtml []
                , Phosphor.lockSimple Regular |> toHtml []
                , Phosphor.lockSimple Bold |> toHtml []
                , Phosphor.lockSimple Fill |> toHtml []
                , Phosphor.lockSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lockSimpleOpen model.weight |> toHtml []
                , Phosphor.lockSimpleOpen Thin |> toHtml []
                , Phosphor.lockSimpleOpen Light |> toHtml []
                , Phosphor.lockSimpleOpen Regular |> toHtml []
                , Phosphor.lockSimpleOpen Bold |> toHtml []
                , Phosphor.lockSimpleOpen Fill |> toHtml []
                , Phosphor.lockSimpleOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.lockers model.weight |> toHtml []
                , Phosphor.lockers Thin |> toHtml []
                , Phosphor.lockers Light |> toHtml []
                , Phosphor.lockers Regular |> toHtml []
                , Phosphor.lockers Bold |> toHtml []
                , Phosphor.lockers Fill |> toHtml []
                , Phosphor.lockers Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.magicWand model.weight |> toHtml []
                , Phosphor.magicWand Thin |> toHtml []
                , Phosphor.magicWand Light |> toHtml []
                , Phosphor.magicWand Regular |> toHtml []
                , Phosphor.magicWand Bold |> toHtml []
                , Phosphor.magicWand Fill |> toHtml []
                , Phosphor.magicWand Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.magnet model.weight |> toHtml []
                , Phosphor.magnet Thin |> toHtml []
                , Phosphor.magnet Light |> toHtml []
                , Phosphor.magnet Regular |> toHtml []
                , Phosphor.magnet Bold |> toHtml []
                , Phosphor.magnet Fill |> toHtml []
                , Phosphor.magnet Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.magnetStraight model.weight |> toHtml []
                , Phosphor.magnetStraight Thin |> toHtml []
                , Phosphor.magnetStraight Light |> toHtml []
                , Phosphor.magnetStraight Regular |> toHtml []
                , Phosphor.magnetStraight Bold |> toHtml []
                , Phosphor.magnetStraight Fill |> toHtml []
                , Phosphor.magnetStraight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.magnifyingGlass model.weight |> toHtml []
                , Phosphor.magnifyingGlass Thin |> toHtml []
                , Phosphor.magnifyingGlass Light |> toHtml []
                , Phosphor.magnifyingGlass Regular |> toHtml []
                , Phosphor.magnifyingGlass Bold |> toHtml []
                , Phosphor.magnifyingGlass Fill |> toHtml []
                , Phosphor.magnifyingGlass Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.magnifyingGlassMinus model.weight |> toHtml []
                , Phosphor.magnifyingGlassMinus Thin |> toHtml []
                , Phosphor.magnifyingGlassMinus Light |> toHtml []
                , Phosphor.magnifyingGlassMinus Regular |> toHtml []
                , Phosphor.magnifyingGlassMinus Bold |> toHtml []
                , Phosphor.magnifyingGlassMinus Fill |> toHtml []
                , Phosphor.magnifyingGlassMinus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.magnifyingGlassPlus model.weight |> toHtml []
                , Phosphor.magnifyingGlassPlus Thin |> toHtml []
                , Phosphor.magnifyingGlassPlus Light |> toHtml []
                , Phosphor.magnifyingGlassPlus Regular |> toHtml []
                , Phosphor.magnifyingGlassPlus Bold |> toHtml []
                , Phosphor.magnifyingGlassPlus Fill |> toHtml []
                , Phosphor.magnifyingGlassPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mapPin model.weight |> toHtml []
                , Phosphor.mapPin Thin |> toHtml []
                , Phosphor.mapPin Light |> toHtml []
                , Phosphor.mapPin Regular |> toHtml []
                , Phosphor.mapPin Bold |> toHtml []
                , Phosphor.mapPin Fill |> toHtml []
                , Phosphor.mapPin Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mapPinLine model.weight |> toHtml []
                , Phosphor.mapPinLine Thin |> toHtml []
                , Phosphor.mapPinLine Light |> toHtml []
                , Phosphor.mapPinLine Regular |> toHtml []
                , Phosphor.mapPinLine Bold |> toHtml []
                , Phosphor.mapPinLine Fill |> toHtml []
                , Phosphor.mapPinLine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mapTrifold model.weight |> toHtml []
                , Phosphor.mapTrifold Thin |> toHtml []
                , Phosphor.mapTrifold Light |> toHtml []
                , Phosphor.mapTrifold Regular |> toHtml []
                , Phosphor.mapTrifold Bold |> toHtml []
                , Phosphor.mapTrifold Fill |> toHtml []
                , Phosphor.mapTrifold Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.markerCircle model.weight |> toHtml []
                , Phosphor.markerCircle Thin |> toHtml []
                , Phosphor.markerCircle Light |> toHtml []
                , Phosphor.markerCircle Regular |> toHtml []
                , Phosphor.markerCircle Bold |> toHtml []
                , Phosphor.markerCircle Fill |> toHtml []
                , Phosphor.markerCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.martini model.weight |> toHtml []
                , Phosphor.martini Thin |> toHtml []
                , Phosphor.martini Light |> toHtml []
                , Phosphor.martini Regular |> toHtml []
                , Phosphor.martini Bold |> toHtml []
                , Phosphor.martini Fill |> toHtml []
                , Phosphor.martini Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.maskHappy model.weight |> toHtml []
                , Phosphor.maskHappy Thin |> toHtml []
                , Phosphor.maskHappy Light |> toHtml []
                , Phosphor.maskHappy Regular |> toHtml []
                , Phosphor.maskHappy Bold |> toHtml []
                , Phosphor.maskHappy Fill |> toHtml []
                , Phosphor.maskHappy Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.maskSad model.weight |> toHtml []
                , Phosphor.maskSad Thin |> toHtml []
                , Phosphor.maskSad Light |> toHtml []
                , Phosphor.maskSad Regular |> toHtml []
                , Phosphor.maskSad Bold |> toHtml []
                , Phosphor.maskSad Fill |> toHtml []
                , Phosphor.maskSad Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mathOperations model.weight |> toHtml []
                , Phosphor.mathOperations Thin |> toHtml []
                , Phosphor.mathOperations Light |> toHtml []
                , Phosphor.mathOperations Regular |> toHtml []
                , Phosphor.mathOperations Bold |> toHtml []
                , Phosphor.mathOperations Fill |> toHtml []
                , Phosphor.mathOperations Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.medal model.weight |> toHtml []
                , Phosphor.medal Thin |> toHtml []
                , Phosphor.medal Light |> toHtml []
                , Phosphor.medal Regular |> toHtml []
                , Phosphor.medal Bold |> toHtml []
                , Phosphor.medal Fill |> toHtml []
                , Phosphor.medal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.medalMilitary model.weight |> toHtml []
                , Phosphor.medalMilitary Thin |> toHtml []
                , Phosphor.medalMilitary Light |> toHtml []
                , Phosphor.medalMilitary Regular |> toHtml []
                , Phosphor.medalMilitary Bold |> toHtml []
                , Phosphor.medalMilitary Fill |> toHtml []
                , Phosphor.medalMilitary Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mediumLogo model.weight |> toHtml []
                , Phosphor.mediumLogo Thin |> toHtml []
                , Phosphor.mediumLogo Light |> toHtml []
                , Phosphor.mediumLogo Regular |> toHtml []
                , Phosphor.mediumLogo Bold |> toHtml []
                , Phosphor.mediumLogo Fill |> toHtml []
                , Phosphor.mediumLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.megaphone model.weight |> toHtml []
                , Phosphor.megaphone Thin |> toHtml []
                , Phosphor.megaphone Light |> toHtml []
                , Phosphor.megaphone Regular |> toHtml []
                , Phosphor.megaphone Bold |> toHtml []
                , Phosphor.megaphone Fill |> toHtml []
                , Phosphor.megaphone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.megaphoneSimple model.weight |> toHtml []
                , Phosphor.megaphoneSimple Thin |> toHtml []
                , Phosphor.megaphoneSimple Light |> toHtml []
                , Phosphor.megaphoneSimple Regular |> toHtml []
                , Phosphor.megaphoneSimple Bold |> toHtml []
                , Phosphor.megaphoneSimple Fill |> toHtml []
                , Phosphor.megaphoneSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.messengerLogo model.weight |> toHtml []
                , Phosphor.messengerLogo Thin |> toHtml []
                , Phosphor.messengerLogo Light |> toHtml []
                , Phosphor.messengerLogo Regular |> toHtml []
                , Phosphor.messengerLogo Bold |> toHtml []
                , Phosphor.messengerLogo Fill |> toHtml []
                , Phosphor.messengerLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.metaLogo model.weight |> toHtml []
                , Phosphor.metaLogo Thin |> toHtml []
                , Phosphor.metaLogo Light |> toHtml []
                , Phosphor.metaLogo Regular |> toHtml []
                , Phosphor.metaLogo Bold |> toHtml []
                , Phosphor.metaLogo Fill |> toHtml []
                , Phosphor.metaLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.metronome model.weight |> toHtml []
                , Phosphor.metronome Thin |> toHtml []
                , Phosphor.metronome Light |> toHtml []
                , Phosphor.metronome Regular |> toHtml []
                , Phosphor.metronome Bold |> toHtml []
                , Phosphor.metronome Fill |> toHtml []
                , Phosphor.metronome Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.microphone model.weight |> toHtml []
                , Phosphor.microphone Thin |> toHtml []
                , Phosphor.microphone Light |> toHtml []
                , Phosphor.microphone Regular |> toHtml []
                , Phosphor.microphone Bold |> toHtml []
                , Phosphor.microphone Fill |> toHtml []
                , Phosphor.microphone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.microphoneSlash model.weight |> toHtml []
                , Phosphor.microphoneSlash Thin |> toHtml []
                , Phosphor.microphoneSlash Light |> toHtml []
                , Phosphor.microphoneSlash Regular |> toHtml []
                , Phosphor.microphoneSlash Bold |> toHtml []
                , Phosphor.microphoneSlash Fill |> toHtml []
                , Phosphor.microphoneSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.microphoneStage model.weight |> toHtml []
                , Phosphor.microphoneStage Thin |> toHtml []
                , Phosphor.microphoneStage Light |> toHtml []
                , Phosphor.microphoneStage Regular |> toHtml []
                , Phosphor.microphoneStage Bold |> toHtml []
                , Phosphor.microphoneStage Fill |> toHtml []
                , Phosphor.microphoneStage Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.microsoftExcelLogo model.weight |> toHtml []
                , Phosphor.microsoftExcelLogo Thin |> toHtml []
                , Phosphor.microsoftExcelLogo Light |> toHtml []
                , Phosphor.microsoftExcelLogo Regular |> toHtml []
                , Phosphor.microsoftExcelLogo Bold |> toHtml []
                , Phosphor.microsoftExcelLogo Fill |> toHtml []
                , Phosphor.microsoftExcelLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.microsoftOutlookLogo model.weight |> toHtml []
                , Phosphor.microsoftOutlookLogo Thin |> toHtml []
                , Phosphor.microsoftOutlookLogo Light |> toHtml []
                , Phosphor.microsoftOutlookLogo Regular |> toHtml []
                , Phosphor.microsoftOutlookLogo Bold |> toHtml []
                , Phosphor.microsoftOutlookLogo Fill |> toHtml []
                , Phosphor.microsoftOutlookLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.microsoftPowerpointLogo model.weight |> toHtml []
                , Phosphor.microsoftPowerpointLogo Thin |> toHtml []
                , Phosphor.microsoftPowerpointLogo Light |> toHtml []
                , Phosphor.microsoftPowerpointLogo Regular |> toHtml []
                , Phosphor.microsoftPowerpointLogo Bold |> toHtml []
                , Phosphor.microsoftPowerpointLogo Fill |> toHtml []
                , Phosphor.microsoftPowerpointLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.microsoftTeamsLogo model.weight |> toHtml []
                , Phosphor.microsoftTeamsLogo Thin |> toHtml []
                , Phosphor.microsoftTeamsLogo Light |> toHtml []
                , Phosphor.microsoftTeamsLogo Regular |> toHtml []
                , Phosphor.microsoftTeamsLogo Bold |> toHtml []
                , Phosphor.microsoftTeamsLogo Fill |> toHtml []
                , Phosphor.microsoftTeamsLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.microsoftWordLogo model.weight |> toHtml []
                , Phosphor.microsoftWordLogo Thin |> toHtml []
                , Phosphor.microsoftWordLogo Light |> toHtml []
                , Phosphor.microsoftWordLogo Regular |> toHtml []
                , Phosphor.microsoftWordLogo Bold |> toHtml []
                , Phosphor.microsoftWordLogo Fill |> toHtml []
                , Phosphor.microsoftWordLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.minus model.weight |> toHtml []
                , Phosphor.minus Thin |> toHtml []
                , Phosphor.minus Light |> toHtml []
                , Phosphor.minus Regular |> toHtml []
                , Phosphor.minus Bold |> toHtml []
                , Phosphor.minus Fill |> toHtml []
                , Phosphor.minus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.minusCircle model.weight |> toHtml []
                , Phosphor.minusCircle Thin |> toHtml []
                , Phosphor.minusCircle Light |> toHtml []
                , Phosphor.minusCircle Regular |> toHtml []
                , Phosphor.minusCircle Bold |> toHtml []
                , Phosphor.minusCircle Fill |> toHtml []
                , Phosphor.minusCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.minusSquare model.weight |> toHtml []
                , Phosphor.minusSquare Thin |> toHtml []
                , Phosphor.minusSquare Light |> toHtml []
                , Phosphor.minusSquare Regular |> toHtml []
                , Phosphor.minusSquare Bold |> toHtml []
                , Phosphor.minusSquare Fill |> toHtml []
                , Phosphor.minusSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.money model.weight |> toHtml []
                , Phosphor.money Thin |> toHtml []
                , Phosphor.money Light |> toHtml []
                , Phosphor.money Regular |> toHtml []
                , Phosphor.money Bold |> toHtml []
                , Phosphor.money Fill |> toHtml []
                , Phosphor.money Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.monitor model.weight |> toHtml []
                , Phosphor.monitor Thin |> toHtml []
                , Phosphor.monitor Light |> toHtml []
                , Phosphor.monitor Regular |> toHtml []
                , Phosphor.monitor Bold |> toHtml []
                , Phosphor.monitor Fill |> toHtml []
                , Phosphor.monitor Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.monitorPlay model.weight |> toHtml []
                , Phosphor.monitorPlay Thin |> toHtml []
                , Phosphor.monitorPlay Light |> toHtml []
                , Phosphor.monitorPlay Regular |> toHtml []
                , Phosphor.monitorPlay Bold |> toHtml []
                , Phosphor.monitorPlay Fill |> toHtml []
                , Phosphor.monitorPlay Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.moon model.weight |> toHtml []
                , Phosphor.moon Thin |> toHtml []
                , Phosphor.moon Light |> toHtml []
                , Phosphor.moon Regular |> toHtml []
                , Phosphor.moon Bold |> toHtml []
                , Phosphor.moon Fill |> toHtml []
                , Phosphor.moon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.moonStars model.weight |> toHtml []
                , Phosphor.moonStars Thin |> toHtml []
                , Phosphor.moonStars Light |> toHtml []
                , Phosphor.moonStars Regular |> toHtml []
                , Phosphor.moonStars Bold |> toHtml []
                , Phosphor.moonStars Fill |> toHtml []
                , Phosphor.moonStars Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.moped model.weight |> toHtml []
                , Phosphor.moped Thin |> toHtml []
                , Phosphor.moped Light |> toHtml []
                , Phosphor.moped Regular |> toHtml []
                , Phosphor.moped Bold |> toHtml []
                , Phosphor.moped Fill |> toHtml []
                , Phosphor.moped Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mopedFront model.weight |> toHtml []
                , Phosphor.mopedFront Thin |> toHtml []
                , Phosphor.mopedFront Light |> toHtml []
                , Phosphor.mopedFront Regular |> toHtml []
                , Phosphor.mopedFront Bold |> toHtml []
                , Phosphor.mopedFront Fill |> toHtml []
                , Phosphor.mopedFront Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mosque model.weight |> toHtml []
                , Phosphor.mosque Thin |> toHtml []
                , Phosphor.mosque Light |> toHtml []
                , Phosphor.mosque Regular |> toHtml []
                , Phosphor.mosque Bold |> toHtml []
                , Phosphor.mosque Fill |> toHtml []
                , Phosphor.mosque Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.motorcycle model.weight |> toHtml []
                , Phosphor.motorcycle Thin |> toHtml []
                , Phosphor.motorcycle Light |> toHtml []
                , Phosphor.motorcycle Regular |> toHtml []
                , Phosphor.motorcycle Bold |> toHtml []
                , Phosphor.motorcycle Fill |> toHtml []
                , Phosphor.motorcycle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mountains model.weight |> toHtml []
                , Phosphor.mountains Thin |> toHtml []
                , Phosphor.mountains Light |> toHtml []
                , Phosphor.mountains Regular |> toHtml []
                , Phosphor.mountains Bold |> toHtml []
                , Phosphor.mountains Fill |> toHtml []
                , Phosphor.mountains Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mouse model.weight |> toHtml []
                , Phosphor.mouse Thin |> toHtml []
                , Phosphor.mouse Light |> toHtml []
                , Phosphor.mouse Regular |> toHtml []
                , Phosphor.mouse Bold |> toHtml []
                , Phosphor.mouse Fill |> toHtml []
                , Phosphor.mouse Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.mouseSimple model.weight |> toHtml []
                , Phosphor.mouseSimple Thin |> toHtml []
                , Phosphor.mouseSimple Light |> toHtml []
                , Phosphor.mouseSimple Regular |> toHtml []
                , Phosphor.mouseSimple Bold |> toHtml []
                , Phosphor.mouseSimple Fill |> toHtml []
                , Phosphor.mouseSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.musicNote model.weight |> toHtml []
                , Phosphor.musicNote Thin |> toHtml []
                , Phosphor.musicNote Light |> toHtml []
                , Phosphor.musicNote Regular |> toHtml []
                , Phosphor.musicNote Bold |> toHtml []
                , Phosphor.musicNote Fill |> toHtml []
                , Phosphor.musicNote Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.musicNoteSimple model.weight |> toHtml []
                , Phosphor.musicNoteSimple Thin |> toHtml []
                , Phosphor.musicNoteSimple Light |> toHtml []
                , Phosphor.musicNoteSimple Regular |> toHtml []
                , Phosphor.musicNoteSimple Bold |> toHtml []
                , Phosphor.musicNoteSimple Fill |> toHtml []
                , Phosphor.musicNoteSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.musicNotes model.weight |> toHtml []
                , Phosphor.musicNotes Thin |> toHtml []
                , Phosphor.musicNotes Light |> toHtml []
                , Phosphor.musicNotes Regular |> toHtml []
                , Phosphor.musicNotes Bold |> toHtml []
                , Phosphor.musicNotes Fill |> toHtml []
                , Phosphor.musicNotes Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.musicNotesPlus model.weight |> toHtml []
                , Phosphor.musicNotesPlus Thin |> toHtml []
                , Phosphor.musicNotesPlus Light |> toHtml []
                , Phosphor.musicNotesPlus Regular |> toHtml []
                , Phosphor.musicNotesPlus Bold |> toHtml []
                , Phosphor.musicNotesPlus Fill |> toHtml []
                , Phosphor.musicNotesPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.musicNotesSimple model.weight |> toHtml []
                , Phosphor.musicNotesSimple Thin |> toHtml []
                , Phosphor.musicNotesSimple Light |> toHtml []
                , Phosphor.musicNotesSimple Regular |> toHtml []
                , Phosphor.musicNotesSimple Bold |> toHtml []
                , Phosphor.musicNotesSimple Fill |> toHtml []
                , Phosphor.musicNotesSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.navigationArrow model.weight |> toHtml []
                , Phosphor.navigationArrow Thin |> toHtml []
                , Phosphor.navigationArrow Light |> toHtml []
                , Phosphor.navigationArrow Regular |> toHtml []
                , Phosphor.navigationArrow Bold |> toHtml []
                , Phosphor.navigationArrow Fill |> toHtml []
                , Phosphor.navigationArrow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.needle model.weight |> toHtml []
                , Phosphor.needle Thin |> toHtml []
                , Phosphor.needle Light |> toHtml []
                , Phosphor.needle Regular |> toHtml []
                , Phosphor.needle Bold |> toHtml []
                , Phosphor.needle Fill |> toHtml []
                , Phosphor.needle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.newspaper model.weight |> toHtml []
                , Phosphor.newspaper Thin |> toHtml []
                , Phosphor.newspaper Light |> toHtml []
                , Phosphor.newspaper Regular |> toHtml []
                , Phosphor.newspaper Bold |> toHtml []
                , Phosphor.newspaper Fill |> toHtml []
                , Phosphor.newspaper Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.newspaperClipping model.weight |> toHtml []
                , Phosphor.newspaperClipping Thin |> toHtml []
                , Phosphor.newspaperClipping Light |> toHtml []
                , Phosphor.newspaperClipping Regular |> toHtml []
                , Phosphor.newspaperClipping Bold |> toHtml []
                , Phosphor.newspaperClipping Fill |> toHtml []
                , Phosphor.newspaperClipping Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.notches model.weight |> toHtml []
                , Phosphor.notches Thin |> toHtml []
                , Phosphor.notches Light |> toHtml []
                , Phosphor.notches Regular |> toHtml []
                , Phosphor.notches Bold |> toHtml []
                , Phosphor.notches Fill |> toHtml []
                , Phosphor.notches Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.noteBlank model.weight |> toHtml []
                , Phosphor.noteBlank Thin |> toHtml []
                , Phosphor.noteBlank Light |> toHtml []
                , Phosphor.noteBlank Regular |> toHtml []
                , Phosphor.noteBlank Bold |> toHtml []
                , Phosphor.noteBlank Fill |> toHtml []
                , Phosphor.noteBlank Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.note model.weight |> toHtml []
                , Phosphor.note Thin |> toHtml []
                , Phosphor.note Light |> toHtml []
                , Phosphor.note Regular |> toHtml []
                , Phosphor.note Bold |> toHtml []
                , Phosphor.note Fill |> toHtml []
                , Phosphor.note Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.notePencil model.weight |> toHtml []
                , Phosphor.notePencil Thin |> toHtml []
                , Phosphor.notePencil Light |> toHtml []
                , Phosphor.notePencil Regular |> toHtml []
                , Phosphor.notePencil Bold |> toHtml []
                , Phosphor.notePencil Fill |> toHtml []
                , Phosphor.notePencil Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.notebook model.weight |> toHtml []
                , Phosphor.notebook Thin |> toHtml []
                , Phosphor.notebook Light |> toHtml []
                , Phosphor.notebook Regular |> toHtml []
                , Phosphor.notebook Bold |> toHtml []
                , Phosphor.notebook Fill |> toHtml []
                , Phosphor.notebook Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.notepad model.weight |> toHtml []
                , Phosphor.notepad Thin |> toHtml []
                , Phosphor.notepad Light |> toHtml []
                , Phosphor.notepad Regular |> toHtml []
                , Phosphor.notepad Bold |> toHtml []
                , Phosphor.notepad Fill |> toHtml []
                , Phosphor.notepad Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.notification model.weight |> toHtml []
                , Phosphor.notification Thin |> toHtml []
                , Phosphor.notification Light |> toHtml []
                , Phosphor.notification Regular |> toHtml []
                , Phosphor.notification Bold |> toHtml []
                , Phosphor.notification Fill |> toHtml []
                , Phosphor.notification Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.notionLogo model.weight |> toHtml []
                , Phosphor.notionLogo Thin |> toHtml []
                , Phosphor.notionLogo Light |> toHtml []
                , Phosphor.notionLogo Regular |> toHtml []
                , Phosphor.notionLogo Bold |> toHtml []
                , Phosphor.notionLogo Fill |> toHtml []
                , Phosphor.notionLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleEight model.weight |> toHtml []
                , Phosphor.numberCircleEight Thin |> toHtml []
                , Phosphor.numberCircleEight Light |> toHtml []
                , Phosphor.numberCircleEight Regular |> toHtml []
                , Phosphor.numberCircleEight Bold |> toHtml []
                , Phosphor.numberCircleEight Fill |> toHtml []
                , Phosphor.numberCircleEight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleFive model.weight |> toHtml []
                , Phosphor.numberCircleFive Thin |> toHtml []
                , Phosphor.numberCircleFive Light |> toHtml []
                , Phosphor.numberCircleFive Regular |> toHtml []
                , Phosphor.numberCircleFive Bold |> toHtml []
                , Phosphor.numberCircleFive Fill |> toHtml []
                , Phosphor.numberCircleFive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleFour model.weight |> toHtml []
                , Phosphor.numberCircleFour Thin |> toHtml []
                , Phosphor.numberCircleFour Light |> toHtml []
                , Phosphor.numberCircleFour Regular |> toHtml []
                , Phosphor.numberCircleFour Bold |> toHtml []
                , Phosphor.numberCircleFour Fill |> toHtml []
                , Phosphor.numberCircleFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleNine model.weight |> toHtml []
                , Phosphor.numberCircleNine Thin |> toHtml []
                , Phosphor.numberCircleNine Light |> toHtml []
                , Phosphor.numberCircleNine Regular |> toHtml []
                , Phosphor.numberCircleNine Bold |> toHtml []
                , Phosphor.numberCircleNine Fill |> toHtml []
                , Phosphor.numberCircleNine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleOne model.weight |> toHtml []
                , Phosphor.numberCircleOne Thin |> toHtml []
                , Phosphor.numberCircleOne Light |> toHtml []
                , Phosphor.numberCircleOne Regular |> toHtml []
                , Phosphor.numberCircleOne Bold |> toHtml []
                , Phosphor.numberCircleOne Fill |> toHtml []
                , Phosphor.numberCircleOne Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleSeven model.weight |> toHtml []
                , Phosphor.numberCircleSeven Thin |> toHtml []
                , Phosphor.numberCircleSeven Light |> toHtml []
                , Phosphor.numberCircleSeven Regular |> toHtml []
                , Phosphor.numberCircleSeven Bold |> toHtml []
                , Phosphor.numberCircleSeven Fill |> toHtml []
                , Phosphor.numberCircleSeven Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleSix model.weight |> toHtml []
                , Phosphor.numberCircleSix Thin |> toHtml []
                , Phosphor.numberCircleSix Light |> toHtml []
                , Phosphor.numberCircleSix Regular |> toHtml []
                , Phosphor.numberCircleSix Bold |> toHtml []
                , Phosphor.numberCircleSix Fill |> toHtml []
                , Phosphor.numberCircleSix Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleThree model.weight |> toHtml []
                , Phosphor.numberCircleThree Thin |> toHtml []
                , Phosphor.numberCircleThree Light |> toHtml []
                , Phosphor.numberCircleThree Regular |> toHtml []
                , Phosphor.numberCircleThree Bold |> toHtml []
                , Phosphor.numberCircleThree Fill |> toHtml []
                , Phosphor.numberCircleThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleTwo model.weight |> toHtml []
                , Phosphor.numberCircleTwo Thin |> toHtml []
                , Phosphor.numberCircleTwo Light |> toHtml []
                , Phosphor.numberCircleTwo Regular |> toHtml []
                , Phosphor.numberCircleTwo Bold |> toHtml []
                , Phosphor.numberCircleTwo Fill |> toHtml []
                , Phosphor.numberCircleTwo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberCircleZero model.weight |> toHtml []
                , Phosphor.numberCircleZero Thin |> toHtml []
                , Phosphor.numberCircleZero Light |> toHtml []
                , Phosphor.numberCircleZero Regular |> toHtml []
                , Phosphor.numberCircleZero Bold |> toHtml []
                , Phosphor.numberCircleZero Fill |> toHtml []
                , Phosphor.numberCircleZero Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberEight model.weight |> toHtml []
                , Phosphor.numberEight Thin |> toHtml []
                , Phosphor.numberEight Light |> toHtml []
                , Phosphor.numberEight Regular |> toHtml []
                , Phosphor.numberEight Bold |> toHtml []
                , Phosphor.numberEight Fill |> toHtml []
                , Phosphor.numberEight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberFive model.weight |> toHtml []
                , Phosphor.numberFive Thin |> toHtml []
                , Phosphor.numberFive Light |> toHtml []
                , Phosphor.numberFive Regular |> toHtml []
                , Phosphor.numberFive Bold |> toHtml []
                , Phosphor.numberFive Fill |> toHtml []
                , Phosphor.numberFive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberFour model.weight |> toHtml []
                , Phosphor.numberFour Thin |> toHtml []
                , Phosphor.numberFour Light |> toHtml []
                , Phosphor.numberFour Regular |> toHtml []
                , Phosphor.numberFour Bold |> toHtml []
                , Phosphor.numberFour Fill |> toHtml []
                , Phosphor.numberFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberNine model.weight |> toHtml []
                , Phosphor.numberNine Thin |> toHtml []
                , Phosphor.numberNine Light |> toHtml []
                , Phosphor.numberNine Regular |> toHtml []
                , Phosphor.numberNine Bold |> toHtml []
                , Phosphor.numberNine Fill |> toHtml []
                , Phosphor.numberNine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberOne model.weight |> toHtml []
                , Phosphor.numberOne Thin |> toHtml []
                , Phosphor.numberOne Light |> toHtml []
                , Phosphor.numberOne Regular |> toHtml []
                , Phosphor.numberOne Bold |> toHtml []
                , Phosphor.numberOne Fill |> toHtml []
                , Phosphor.numberOne Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSeven model.weight |> toHtml []
                , Phosphor.numberSeven Thin |> toHtml []
                , Phosphor.numberSeven Light |> toHtml []
                , Phosphor.numberSeven Regular |> toHtml []
                , Phosphor.numberSeven Bold |> toHtml []
                , Phosphor.numberSeven Fill |> toHtml []
                , Phosphor.numberSeven Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSix model.weight |> toHtml []
                , Phosphor.numberSix Thin |> toHtml []
                , Phosphor.numberSix Light |> toHtml []
                , Phosphor.numberSix Regular |> toHtml []
                , Phosphor.numberSix Bold |> toHtml []
                , Phosphor.numberSix Fill |> toHtml []
                , Phosphor.numberSix Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareEight model.weight |> toHtml []
                , Phosphor.numberSquareEight Thin |> toHtml []
                , Phosphor.numberSquareEight Light |> toHtml []
                , Phosphor.numberSquareEight Regular |> toHtml []
                , Phosphor.numberSquareEight Bold |> toHtml []
                , Phosphor.numberSquareEight Fill |> toHtml []
                , Phosphor.numberSquareEight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareFive model.weight |> toHtml []
                , Phosphor.numberSquareFive Thin |> toHtml []
                , Phosphor.numberSquareFive Light |> toHtml []
                , Phosphor.numberSquareFive Regular |> toHtml []
                , Phosphor.numberSquareFive Bold |> toHtml []
                , Phosphor.numberSquareFive Fill |> toHtml []
                , Phosphor.numberSquareFive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareFour model.weight |> toHtml []
                , Phosphor.numberSquareFour Thin |> toHtml []
                , Phosphor.numberSquareFour Light |> toHtml []
                , Phosphor.numberSquareFour Regular |> toHtml []
                , Phosphor.numberSquareFour Bold |> toHtml []
                , Phosphor.numberSquareFour Fill |> toHtml []
                , Phosphor.numberSquareFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareNine model.weight |> toHtml []
                , Phosphor.numberSquareNine Thin |> toHtml []
                , Phosphor.numberSquareNine Light |> toHtml []
                , Phosphor.numberSquareNine Regular |> toHtml []
                , Phosphor.numberSquareNine Bold |> toHtml []
                , Phosphor.numberSquareNine Fill |> toHtml []
                , Phosphor.numberSquareNine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareOne model.weight |> toHtml []
                , Phosphor.numberSquareOne Thin |> toHtml []
                , Phosphor.numberSquareOne Light |> toHtml []
                , Phosphor.numberSquareOne Regular |> toHtml []
                , Phosphor.numberSquareOne Bold |> toHtml []
                , Phosphor.numberSquareOne Fill |> toHtml []
                , Phosphor.numberSquareOne Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareSeven model.weight |> toHtml []
                , Phosphor.numberSquareSeven Thin |> toHtml []
                , Phosphor.numberSquareSeven Light |> toHtml []
                , Phosphor.numberSquareSeven Regular |> toHtml []
                , Phosphor.numberSquareSeven Bold |> toHtml []
                , Phosphor.numberSquareSeven Fill |> toHtml []
                , Phosphor.numberSquareSeven Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareSix model.weight |> toHtml []
                , Phosphor.numberSquareSix Thin |> toHtml []
                , Phosphor.numberSquareSix Light |> toHtml []
                , Phosphor.numberSquareSix Regular |> toHtml []
                , Phosphor.numberSquareSix Bold |> toHtml []
                , Phosphor.numberSquareSix Fill |> toHtml []
                , Phosphor.numberSquareSix Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareThree model.weight |> toHtml []
                , Phosphor.numberSquareThree Thin |> toHtml []
                , Phosphor.numberSquareThree Light |> toHtml []
                , Phosphor.numberSquareThree Regular |> toHtml []
                , Phosphor.numberSquareThree Bold |> toHtml []
                , Phosphor.numberSquareThree Fill |> toHtml []
                , Phosphor.numberSquareThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareTwo model.weight |> toHtml []
                , Phosphor.numberSquareTwo Thin |> toHtml []
                , Phosphor.numberSquareTwo Light |> toHtml []
                , Phosphor.numberSquareTwo Regular |> toHtml []
                , Phosphor.numberSquareTwo Bold |> toHtml []
                , Phosphor.numberSquareTwo Fill |> toHtml []
                , Phosphor.numberSquareTwo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberSquareZero model.weight |> toHtml []
                , Phosphor.numberSquareZero Thin |> toHtml []
                , Phosphor.numberSquareZero Light |> toHtml []
                , Phosphor.numberSquareZero Regular |> toHtml []
                , Phosphor.numberSquareZero Bold |> toHtml []
                , Phosphor.numberSquareZero Fill |> toHtml []
                , Phosphor.numberSquareZero Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberThree model.weight |> toHtml []
                , Phosphor.numberThree Thin |> toHtml []
                , Phosphor.numberThree Light |> toHtml []
                , Phosphor.numberThree Regular |> toHtml []
                , Phosphor.numberThree Bold |> toHtml []
                , Phosphor.numberThree Fill |> toHtml []
                , Phosphor.numberThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberTwo model.weight |> toHtml []
                , Phosphor.numberTwo Thin |> toHtml []
                , Phosphor.numberTwo Light |> toHtml []
                , Phosphor.numberTwo Regular |> toHtml []
                , Phosphor.numberTwo Bold |> toHtml []
                , Phosphor.numberTwo Fill |> toHtml []
                , Phosphor.numberTwo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.numberZero model.weight |> toHtml []
                , Phosphor.numberZero Thin |> toHtml []
                , Phosphor.numberZero Light |> toHtml []
                , Phosphor.numberZero Regular |> toHtml []
                , Phosphor.numberZero Bold |> toHtml []
                , Phosphor.numberZero Fill |> toHtml []
                , Phosphor.numberZero Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.nut model.weight |> toHtml []
                , Phosphor.nut Thin |> toHtml []
                , Phosphor.nut Light |> toHtml []
                , Phosphor.nut Regular |> toHtml []
                , Phosphor.nut Bold |> toHtml []
                , Phosphor.nut Fill |> toHtml []
                , Phosphor.nut Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.nyTimesLogo model.weight |> toHtml []
                , Phosphor.nyTimesLogo Thin |> toHtml []
                , Phosphor.nyTimesLogo Light |> toHtml []
                , Phosphor.nyTimesLogo Regular |> toHtml []
                , Phosphor.nyTimesLogo Bold |> toHtml []
                , Phosphor.nyTimesLogo Fill |> toHtml []
                , Phosphor.nyTimesLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.octagon model.weight |> toHtml []
                , Phosphor.octagon Thin |> toHtml []
                , Phosphor.octagon Light |> toHtml []
                , Phosphor.octagon Regular |> toHtml []
                , Phosphor.octagon Bold |> toHtml []
                , Phosphor.octagon Fill |> toHtml []
                , Phosphor.octagon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.officeChair model.weight |> toHtml []
                , Phosphor.officeChair Thin |> toHtml []
                , Phosphor.officeChair Light |> toHtml []
                , Phosphor.officeChair Regular |> toHtml []
                , Phosphor.officeChair Bold |> toHtml []
                , Phosphor.officeChair Fill |> toHtml []
                , Phosphor.officeChair Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.option model.weight |> toHtml []
                , Phosphor.option Thin |> toHtml []
                , Phosphor.option Light |> toHtml []
                , Phosphor.option Regular |> toHtml []
                , Phosphor.option Bold |> toHtml []
                , Phosphor.option Fill |> toHtml []
                , Phosphor.option Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.orangeSlice model.weight |> toHtml []
                , Phosphor.orangeSlice Thin |> toHtml []
                , Phosphor.orangeSlice Light |> toHtml []
                , Phosphor.orangeSlice Regular |> toHtml []
                , Phosphor.orangeSlice Bold |> toHtml []
                , Phosphor.orangeSlice Fill |> toHtml []
                , Phosphor.orangeSlice Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.package model.weight |> toHtml []
                , Phosphor.package Thin |> toHtml []
                , Phosphor.package Light |> toHtml []
                , Phosphor.package Regular |> toHtml []
                , Phosphor.package Bold |> toHtml []
                , Phosphor.package Fill |> toHtml []
                , Phosphor.package Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paintBrush model.weight |> toHtml []
                , Phosphor.paintBrush Thin |> toHtml []
                , Phosphor.paintBrush Light |> toHtml []
                , Phosphor.paintBrush Regular |> toHtml []
                , Phosphor.paintBrush Bold |> toHtml []
                , Phosphor.paintBrush Fill |> toHtml []
                , Phosphor.paintBrush Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paintBrushBroad model.weight |> toHtml []
                , Phosphor.paintBrushBroad Thin |> toHtml []
                , Phosphor.paintBrushBroad Light |> toHtml []
                , Phosphor.paintBrushBroad Regular |> toHtml []
                , Phosphor.paintBrushBroad Bold |> toHtml []
                , Phosphor.paintBrushBroad Fill |> toHtml []
                , Phosphor.paintBrushBroad Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paintBrushHousehold model.weight |> toHtml []
                , Phosphor.paintBrushHousehold Thin |> toHtml []
                , Phosphor.paintBrushHousehold Light |> toHtml []
                , Phosphor.paintBrushHousehold Regular |> toHtml []
                , Phosphor.paintBrushHousehold Bold |> toHtml []
                , Phosphor.paintBrushHousehold Fill |> toHtml []
                , Phosphor.paintBrushHousehold Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paintBucket model.weight |> toHtml []
                , Phosphor.paintBucket Thin |> toHtml []
                , Phosphor.paintBucket Light |> toHtml []
                , Phosphor.paintBucket Regular |> toHtml []
                , Phosphor.paintBucket Bold |> toHtml []
                , Phosphor.paintBucket Fill |> toHtml []
                , Phosphor.paintBucket Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paintRoller model.weight |> toHtml []
                , Phosphor.paintRoller Thin |> toHtml []
                , Phosphor.paintRoller Light |> toHtml []
                , Phosphor.paintRoller Regular |> toHtml []
                , Phosphor.paintRoller Bold |> toHtml []
                , Phosphor.paintRoller Fill |> toHtml []
                , Phosphor.paintRoller Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.palette model.weight |> toHtml []
                , Phosphor.palette Thin |> toHtml []
                , Phosphor.palette Light |> toHtml []
                , Phosphor.palette Regular |> toHtml []
                , Phosphor.palette Bold |> toHtml []
                , Phosphor.palette Fill |> toHtml []
                , Phosphor.palette Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pants model.weight |> toHtml []
                , Phosphor.pants Thin |> toHtml []
                , Phosphor.pants Light |> toHtml []
                , Phosphor.pants Regular |> toHtml []
                , Phosphor.pants Bold |> toHtml []
                , Phosphor.pants Fill |> toHtml []
                , Phosphor.pants Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paperPlane model.weight |> toHtml []
                , Phosphor.paperPlane Thin |> toHtml []
                , Phosphor.paperPlane Light |> toHtml []
                , Phosphor.paperPlane Regular |> toHtml []
                , Phosphor.paperPlane Bold |> toHtml []
                , Phosphor.paperPlane Fill |> toHtml []
                , Phosphor.paperPlane Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paperPlaneRight model.weight |> toHtml []
                , Phosphor.paperPlaneRight Thin |> toHtml []
                , Phosphor.paperPlaneRight Light |> toHtml []
                , Phosphor.paperPlaneRight Regular |> toHtml []
                , Phosphor.paperPlaneRight Bold |> toHtml []
                , Phosphor.paperPlaneRight Fill |> toHtml []
                , Phosphor.paperPlaneRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paperPlaneTilt model.weight |> toHtml []
                , Phosphor.paperPlaneTilt Thin |> toHtml []
                , Phosphor.paperPlaneTilt Light |> toHtml []
                , Phosphor.paperPlaneTilt Regular |> toHtml []
                , Phosphor.paperPlaneTilt Bold |> toHtml []
                , Phosphor.paperPlaneTilt Fill |> toHtml []
                , Phosphor.paperPlaneTilt Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paperclip model.weight |> toHtml []
                , Phosphor.paperclip Thin |> toHtml []
                , Phosphor.paperclip Light |> toHtml []
                , Phosphor.paperclip Regular |> toHtml []
                , Phosphor.paperclip Bold |> toHtml []
                , Phosphor.paperclip Fill |> toHtml []
                , Phosphor.paperclip Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paperclipHorizontal model.weight |> toHtml []
                , Phosphor.paperclipHorizontal Thin |> toHtml []
                , Phosphor.paperclipHorizontal Light |> toHtml []
                , Phosphor.paperclipHorizontal Regular |> toHtml []
                , Phosphor.paperclipHorizontal Bold |> toHtml []
                , Phosphor.paperclipHorizontal Fill |> toHtml []
                , Phosphor.paperclipHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.parachute model.weight |> toHtml []
                , Phosphor.parachute Thin |> toHtml []
                , Phosphor.parachute Light |> toHtml []
                , Phosphor.parachute Regular |> toHtml []
                , Phosphor.parachute Bold |> toHtml []
                , Phosphor.parachute Fill |> toHtml []
                , Phosphor.parachute Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paragraph model.weight |> toHtml []
                , Phosphor.paragraph Thin |> toHtml []
                , Phosphor.paragraph Light |> toHtml []
                , Phosphor.paragraph Regular |> toHtml []
                , Phosphor.paragraph Bold |> toHtml []
                , Phosphor.paragraph Fill |> toHtml []
                , Phosphor.paragraph Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.parallelogram model.weight |> toHtml []
                , Phosphor.parallelogram Thin |> toHtml []
                , Phosphor.parallelogram Light |> toHtml []
                , Phosphor.parallelogram Regular |> toHtml []
                , Phosphor.parallelogram Bold |> toHtml []
                , Phosphor.parallelogram Fill |> toHtml []
                , Phosphor.parallelogram Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.park model.weight |> toHtml []
                , Phosphor.park Thin |> toHtml []
                , Phosphor.park Light |> toHtml []
                , Phosphor.park Regular |> toHtml []
                , Phosphor.park Bold |> toHtml []
                , Phosphor.park Fill |> toHtml []
                , Phosphor.park Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.password model.weight |> toHtml []
                , Phosphor.password Thin |> toHtml []
                , Phosphor.password Light |> toHtml []
                , Phosphor.password Regular |> toHtml []
                , Phosphor.password Bold |> toHtml []
                , Phosphor.password Fill |> toHtml []
                , Phosphor.password Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.path model.weight |> toHtml []
                , Phosphor.path Thin |> toHtml []
                , Phosphor.path Light |> toHtml []
                , Phosphor.path Regular |> toHtml []
                , Phosphor.path Bold |> toHtml []
                , Phosphor.path Fill |> toHtml []
                , Phosphor.path Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.patreonLogo model.weight |> toHtml []
                , Phosphor.patreonLogo Thin |> toHtml []
                , Phosphor.patreonLogo Light |> toHtml []
                , Phosphor.patreonLogo Regular |> toHtml []
                , Phosphor.patreonLogo Bold |> toHtml []
                , Phosphor.patreonLogo Fill |> toHtml []
                , Phosphor.patreonLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pause model.weight |> toHtml []
                , Phosphor.pause Thin |> toHtml []
                , Phosphor.pause Light |> toHtml []
                , Phosphor.pause Regular |> toHtml []
                , Phosphor.pause Bold |> toHtml []
                , Phosphor.pause Fill |> toHtml []
                , Phosphor.pause Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pauseCircle model.weight |> toHtml []
                , Phosphor.pauseCircle Thin |> toHtml []
                , Phosphor.pauseCircle Light |> toHtml []
                , Phosphor.pauseCircle Regular |> toHtml []
                , Phosphor.pauseCircle Bold |> toHtml []
                , Phosphor.pauseCircle Fill |> toHtml []
                , Phosphor.pauseCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pawPrint model.weight |> toHtml []
                , Phosphor.pawPrint Thin |> toHtml []
                , Phosphor.pawPrint Light |> toHtml []
                , Phosphor.pawPrint Regular |> toHtml []
                , Phosphor.pawPrint Bold |> toHtml []
                , Phosphor.pawPrint Fill |> toHtml []
                , Phosphor.pawPrint Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.paypalLogo model.weight |> toHtml []
                , Phosphor.paypalLogo Thin |> toHtml []
                , Phosphor.paypalLogo Light |> toHtml []
                , Phosphor.paypalLogo Regular |> toHtml []
                , Phosphor.paypalLogo Bold |> toHtml []
                , Phosphor.paypalLogo Fill |> toHtml []
                , Phosphor.paypalLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.peace model.weight |> toHtml []
                , Phosphor.peace Thin |> toHtml []
                , Phosphor.peace Light |> toHtml []
                , Phosphor.peace Regular |> toHtml []
                , Phosphor.peace Bold |> toHtml []
                , Phosphor.peace Fill |> toHtml []
                , Phosphor.peace Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pen model.weight |> toHtml []
                , Phosphor.pen Thin |> toHtml []
                , Phosphor.pen Light |> toHtml []
                , Phosphor.pen Regular |> toHtml []
                , Phosphor.pen Bold |> toHtml []
                , Phosphor.pen Fill |> toHtml []
                , Phosphor.pen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.penNib model.weight |> toHtml []
                , Phosphor.penNib Thin |> toHtml []
                , Phosphor.penNib Light |> toHtml []
                , Phosphor.penNib Regular |> toHtml []
                , Phosphor.penNib Bold |> toHtml []
                , Phosphor.penNib Fill |> toHtml []
                , Phosphor.penNib Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.penNibStraight model.weight |> toHtml []
                , Phosphor.penNibStraight Thin |> toHtml []
                , Phosphor.penNibStraight Light |> toHtml []
                , Phosphor.penNibStraight Regular |> toHtml []
                , Phosphor.penNibStraight Bold |> toHtml []
                , Phosphor.penNibStraight Fill |> toHtml []
                , Phosphor.penNibStraight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pencil model.weight |> toHtml []
                , Phosphor.pencil Thin |> toHtml []
                , Phosphor.pencil Light |> toHtml []
                , Phosphor.pencil Regular |> toHtml []
                , Phosphor.pencil Bold |> toHtml []
                , Phosphor.pencil Fill |> toHtml []
                , Phosphor.pencil Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pencilCircle model.weight |> toHtml []
                , Phosphor.pencilCircle Thin |> toHtml []
                , Phosphor.pencilCircle Light |> toHtml []
                , Phosphor.pencilCircle Regular |> toHtml []
                , Phosphor.pencilCircle Bold |> toHtml []
                , Phosphor.pencilCircle Fill |> toHtml []
                , Phosphor.pencilCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pencilLine model.weight |> toHtml []
                , Phosphor.pencilLine Thin |> toHtml []
                , Phosphor.pencilLine Light |> toHtml []
                , Phosphor.pencilLine Regular |> toHtml []
                , Phosphor.pencilLine Bold |> toHtml []
                , Phosphor.pencilLine Fill |> toHtml []
                , Phosphor.pencilLine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pencilSimple model.weight |> toHtml []
                , Phosphor.pencilSimple Thin |> toHtml []
                , Phosphor.pencilSimple Light |> toHtml []
                , Phosphor.pencilSimple Regular |> toHtml []
                , Phosphor.pencilSimple Bold |> toHtml []
                , Phosphor.pencilSimple Fill |> toHtml []
                , Phosphor.pencilSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pencilSimpleLine model.weight |> toHtml []
                , Phosphor.pencilSimpleLine Thin |> toHtml []
                , Phosphor.pencilSimpleLine Light |> toHtml []
                , Phosphor.pencilSimpleLine Regular |> toHtml []
                , Phosphor.pencilSimpleLine Bold |> toHtml []
                , Phosphor.pencilSimpleLine Fill |> toHtml []
                , Phosphor.pencilSimpleLine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pencilSimpleSlash model.weight |> toHtml []
                , Phosphor.pencilSimpleSlash Thin |> toHtml []
                , Phosphor.pencilSimpleSlash Light |> toHtml []
                , Phosphor.pencilSimpleSlash Regular |> toHtml []
                , Phosphor.pencilSimpleSlash Bold |> toHtml []
                , Phosphor.pencilSimpleSlash Fill |> toHtml []
                , Phosphor.pencilSimpleSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pencilSlash model.weight |> toHtml []
                , Phosphor.pencilSlash Thin |> toHtml []
                , Phosphor.pencilSlash Light |> toHtml []
                , Phosphor.pencilSlash Regular |> toHtml []
                , Phosphor.pencilSlash Bold |> toHtml []
                , Phosphor.pencilSlash Fill |> toHtml []
                , Phosphor.pencilSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pentagram model.weight |> toHtml []
                , Phosphor.pentagram Thin |> toHtml []
                , Phosphor.pentagram Light |> toHtml []
                , Phosphor.pentagram Regular |> toHtml []
                , Phosphor.pentagram Bold |> toHtml []
                , Phosphor.pentagram Fill |> toHtml []
                , Phosphor.pentagram Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pepper model.weight |> toHtml []
                , Phosphor.pepper Thin |> toHtml []
                , Phosphor.pepper Light |> toHtml []
                , Phosphor.pepper Regular |> toHtml []
                , Phosphor.pepper Bold |> toHtml []
                , Phosphor.pepper Fill |> toHtml []
                , Phosphor.pepper Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.percent model.weight |> toHtml []
                , Phosphor.percent Thin |> toHtml []
                , Phosphor.percent Light |> toHtml []
                , Phosphor.percent Regular |> toHtml []
                , Phosphor.percent Bold |> toHtml []
                , Phosphor.percent Fill |> toHtml []
                , Phosphor.percent Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.personArmsSpread model.weight |> toHtml []
                , Phosphor.personArmsSpread Thin |> toHtml []
                , Phosphor.personArmsSpread Light |> toHtml []
                , Phosphor.personArmsSpread Regular |> toHtml []
                , Phosphor.personArmsSpread Bold |> toHtml []
                , Phosphor.personArmsSpread Fill |> toHtml []
                , Phosphor.personArmsSpread Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.person model.weight |> toHtml []
                , Phosphor.person Thin |> toHtml []
                , Phosphor.person Light |> toHtml []
                , Phosphor.person Regular |> toHtml []
                , Phosphor.person Bold |> toHtml []
                , Phosphor.person Fill |> toHtml []
                , Phosphor.person Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.personSimpleBike model.weight |> toHtml []
                , Phosphor.personSimpleBike Thin |> toHtml []
                , Phosphor.personSimpleBike Light |> toHtml []
                , Phosphor.personSimpleBike Regular |> toHtml []
                , Phosphor.personSimpleBike Bold |> toHtml []
                , Phosphor.personSimpleBike Fill |> toHtml []
                , Phosphor.personSimpleBike Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.personSimple model.weight |> toHtml []
                , Phosphor.personSimple Thin |> toHtml []
                , Phosphor.personSimple Light |> toHtml []
                , Phosphor.personSimple Regular |> toHtml []
                , Phosphor.personSimple Bold |> toHtml []
                , Phosphor.personSimple Fill |> toHtml []
                , Phosphor.personSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.personSimpleRun model.weight |> toHtml []
                , Phosphor.personSimpleRun Thin |> toHtml []
                , Phosphor.personSimpleRun Light |> toHtml []
                , Phosphor.personSimpleRun Regular |> toHtml []
                , Phosphor.personSimpleRun Bold |> toHtml []
                , Phosphor.personSimpleRun Fill |> toHtml []
                , Phosphor.personSimpleRun Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.personSimpleThrow model.weight |> toHtml []
                , Phosphor.personSimpleThrow Thin |> toHtml []
                , Phosphor.personSimpleThrow Light |> toHtml []
                , Phosphor.personSimpleThrow Regular |> toHtml []
                , Phosphor.personSimpleThrow Bold |> toHtml []
                , Phosphor.personSimpleThrow Fill |> toHtml []
                , Phosphor.personSimpleThrow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.personSimpleWalk model.weight |> toHtml []
                , Phosphor.personSimpleWalk Thin |> toHtml []
                , Phosphor.personSimpleWalk Light |> toHtml []
                , Phosphor.personSimpleWalk Regular |> toHtml []
                , Phosphor.personSimpleWalk Bold |> toHtml []
                , Phosphor.personSimpleWalk Fill |> toHtml []
                , Phosphor.personSimpleWalk Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.perspective model.weight |> toHtml []
                , Phosphor.perspective Thin |> toHtml []
                , Phosphor.perspective Light |> toHtml []
                , Phosphor.perspective Regular |> toHtml []
                , Phosphor.perspective Bold |> toHtml []
                , Phosphor.perspective Fill |> toHtml []
                , Phosphor.perspective Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phone model.weight |> toHtml []
                , Phosphor.phone Thin |> toHtml []
                , Phosphor.phone Light |> toHtml []
                , Phosphor.phone Regular |> toHtml []
                , Phosphor.phone Bold |> toHtml []
                , Phosphor.phone Fill |> toHtml []
                , Phosphor.phone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phoneCall model.weight |> toHtml []
                , Phosphor.phoneCall Thin |> toHtml []
                , Phosphor.phoneCall Light |> toHtml []
                , Phosphor.phoneCall Regular |> toHtml []
                , Phosphor.phoneCall Bold |> toHtml []
                , Phosphor.phoneCall Fill |> toHtml []
                , Phosphor.phoneCall Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phoneDisconnect model.weight |> toHtml []
                , Phosphor.phoneDisconnect Thin |> toHtml []
                , Phosphor.phoneDisconnect Light |> toHtml []
                , Phosphor.phoneDisconnect Regular |> toHtml []
                , Phosphor.phoneDisconnect Bold |> toHtml []
                , Phosphor.phoneDisconnect Fill |> toHtml []
                , Phosphor.phoneDisconnect Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phoneIncoming model.weight |> toHtml []
                , Phosphor.phoneIncoming Thin |> toHtml []
                , Phosphor.phoneIncoming Light |> toHtml []
                , Phosphor.phoneIncoming Regular |> toHtml []
                , Phosphor.phoneIncoming Bold |> toHtml []
                , Phosphor.phoneIncoming Fill |> toHtml []
                , Phosphor.phoneIncoming Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phoneOutgoing model.weight |> toHtml []
                , Phosphor.phoneOutgoing Thin |> toHtml []
                , Phosphor.phoneOutgoing Light |> toHtml []
                , Phosphor.phoneOutgoing Regular |> toHtml []
                , Phosphor.phoneOutgoing Bold |> toHtml []
                , Phosphor.phoneOutgoing Fill |> toHtml []
                , Phosphor.phoneOutgoing Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phonePlus model.weight |> toHtml []
                , Phosphor.phonePlus Thin |> toHtml []
                , Phosphor.phonePlus Light |> toHtml []
                , Phosphor.phonePlus Regular |> toHtml []
                , Phosphor.phonePlus Bold |> toHtml []
                , Phosphor.phonePlus Fill |> toHtml []
                , Phosphor.phonePlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phoneSlash model.weight |> toHtml []
                , Phosphor.phoneSlash Thin |> toHtml []
                , Phosphor.phoneSlash Light |> toHtml []
                , Phosphor.phoneSlash Regular |> toHtml []
                , Phosphor.phoneSlash Bold |> toHtml []
                , Phosphor.phoneSlash Fill |> toHtml []
                , Phosphor.phoneSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phoneX model.weight |> toHtml []
                , Phosphor.phoneX Thin |> toHtml []
                , Phosphor.phoneX Light |> toHtml []
                , Phosphor.phoneX Regular |> toHtml []
                , Phosphor.phoneX Bold |> toHtml []
                , Phosphor.phoneX Fill |> toHtml []
                , Phosphor.phoneX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.phosphorLogo model.weight |> toHtml []
                , Phosphor.phosphorLogo Thin |> toHtml []
                , Phosphor.phosphorLogo Light |> toHtml []
                , Phosphor.phosphorLogo Regular |> toHtml []
                , Phosphor.phosphorLogo Bold |> toHtml []
                , Phosphor.phosphorLogo Fill |> toHtml []
                , Phosphor.phosphorLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pi model.weight |> toHtml []
                , Phosphor.pi Thin |> toHtml []
                , Phosphor.pi Light |> toHtml []
                , Phosphor.pi Regular |> toHtml []
                , Phosphor.pi Bold |> toHtml []
                , Phosphor.pi Fill |> toHtml []
                , Phosphor.pi Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pianoKeys model.weight |> toHtml []
                , Phosphor.pianoKeys Thin |> toHtml []
                , Phosphor.pianoKeys Light |> toHtml []
                , Phosphor.pianoKeys Regular |> toHtml []
                , Phosphor.pianoKeys Bold |> toHtml []
                , Phosphor.pianoKeys Fill |> toHtml []
                , Phosphor.pianoKeys Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pictureInPicture model.weight |> toHtml []
                , Phosphor.pictureInPicture Thin |> toHtml []
                , Phosphor.pictureInPicture Light |> toHtml []
                , Phosphor.pictureInPicture Regular |> toHtml []
                , Phosphor.pictureInPicture Bold |> toHtml []
                , Phosphor.pictureInPicture Fill |> toHtml []
                , Phosphor.pictureInPicture Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.piggyBank model.weight |> toHtml []
                , Phosphor.piggyBank Thin |> toHtml []
                , Phosphor.piggyBank Light |> toHtml []
                , Phosphor.piggyBank Regular |> toHtml []
                , Phosphor.piggyBank Bold |> toHtml []
                , Phosphor.piggyBank Fill |> toHtml []
                , Phosphor.piggyBank Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pill model.weight |> toHtml []
                , Phosphor.pill Thin |> toHtml []
                , Phosphor.pill Light |> toHtml []
                , Phosphor.pill Regular |> toHtml []
                , Phosphor.pill Bold |> toHtml []
                , Phosphor.pill Fill |> toHtml []
                , Phosphor.pill Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pinterestLogo model.weight |> toHtml []
                , Phosphor.pinterestLogo Thin |> toHtml []
                , Phosphor.pinterestLogo Light |> toHtml []
                , Phosphor.pinterestLogo Regular |> toHtml []
                , Phosphor.pinterestLogo Bold |> toHtml []
                , Phosphor.pinterestLogo Fill |> toHtml []
                , Phosphor.pinterestLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pinwheel model.weight |> toHtml []
                , Phosphor.pinwheel Thin |> toHtml []
                , Phosphor.pinwheel Light |> toHtml []
                , Phosphor.pinwheel Regular |> toHtml []
                , Phosphor.pinwheel Bold |> toHtml []
                , Phosphor.pinwheel Fill |> toHtml []
                , Phosphor.pinwheel Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pizza model.weight |> toHtml []
                , Phosphor.pizza Thin |> toHtml []
                , Phosphor.pizza Light |> toHtml []
                , Phosphor.pizza Regular |> toHtml []
                , Phosphor.pizza Bold |> toHtml []
                , Phosphor.pizza Fill |> toHtml []
                , Phosphor.pizza Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.placeholder model.weight |> toHtml []
                , Phosphor.placeholder Thin |> toHtml []
                , Phosphor.placeholder Light |> toHtml []
                , Phosphor.placeholder Regular |> toHtml []
                , Phosphor.placeholder Bold |> toHtml []
                , Phosphor.placeholder Fill |> toHtml []
                , Phosphor.placeholder Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.planet model.weight |> toHtml []
                , Phosphor.planet Thin |> toHtml []
                , Phosphor.planet Light |> toHtml []
                , Phosphor.planet Regular |> toHtml []
                , Phosphor.planet Bold |> toHtml []
                , Phosphor.planet Fill |> toHtml []
                , Phosphor.planet Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plant model.weight |> toHtml []
                , Phosphor.plant Thin |> toHtml []
                , Phosphor.plant Light |> toHtml []
                , Phosphor.plant Regular |> toHtml []
                , Phosphor.plant Bold |> toHtml []
                , Phosphor.plant Fill |> toHtml []
                , Phosphor.plant Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.play model.weight |> toHtml []
                , Phosphor.play Thin |> toHtml []
                , Phosphor.play Light |> toHtml []
                , Phosphor.play Regular |> toHtml []
                , Phosphor.play Bold |> toHtml []
                , Phosphor.play Fill |> toHtml []
                , Phosphor.play Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.playCircle model.weight |> toHtml []
                , Phosphor.playCircle Thin |> toHtml []
                , Phosphor.playCircle Light |> toHtml []
                , Phosphor.playCircle Regular |> toHtml []
                , Phosphor.playCircle Bold |> toHtml []
                , Phosphor.playCircle Fill |> toHtml []
                , Phosphor.playCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.playPause model.weight |> toHtml []
                , Phosphor.playPause Thin |> toHtml []
                , Phosphor.playPause Light |> toHtml []
                , Phosphor.playPause Regular |> toHtml []
                , Phosphor.playPause Bold |> toHtml []
                , Phosphor.playPause Fill |> toHtml []
                , Phosphor.playPause Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.playlist model.weight |> toHtml []
                , Phosphor.playlist Thin |> toHtml []
                , Phosphor.playlist Light |> toHtml []
                , Phosphor.playlist Regular |> toHtml []
                , Phosphor.playlist Bold |> toHtml []
                , Phosphor.playlist Fill |> toHtml []
                , Phosphor.playlist Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plug model.weight |> toHtml []
                , Phosphor.plug Thin |> toHtml []
                , Phosphor.plug Light |> toHtml []
                , Phosphor.plug Regular |> toHtml []
                , Phosphor.plug Bold |> toHtml []
                , Phosphor.plug Fill |> toHtml []
                , Phosphor.plug Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plugCharging model.weight |> toHtml []
                , Phosphor.plugCharging Thin |> toHtml []
                , Phosphor.plugCharging Light |> toHtml []
                , Phosphor.plugCharging Regular |> toHtml []
                , Phosphor.plugCharging Bold |> toHtml []
                , Phosphor.plugCharging Fill |> toHtml []
                , Phosphor.plugCharging Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plugs model.weight |> toHtml []
                , Phosphor.plugs Thin |> toHtml []
                , Phosphor.plugs Light |> toHtml []
                , Phosphor.plugs Regular |> toHtml []
                , Phosphor.plugs Bold |> toHtml []
                , Phosphor.plugs Fill |> toHtml []
                , Phosphor.plugs Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plugsConnected model.weight |> toHtml []
                , Phosphor.plugsConnected Thin |> toHtml []
                , Phosphor.plugsConnected Light |> toHtml []
                , Phosphor.plugsConnected Regular |> toHtml []
                , Phosphor.plugsConnected Bold |> toHtml []
                , Phosphor.plugsConnected Fill |> toHtml []
                , Phosphor.plugsConnected Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plus model.weight |> toHtml []
                , Phosphor.plus Thin |> toHtml []
                , Phosphor.plus Light |> toHtml []
                , Phosphor.plus Regular |> toHtml []
                , Phosphor.plus Bold |> toHtml []
                , Phosphor.plus Fill |> toHtml []
                , Phosphor.plus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plusCircle model.weight |> toHtml []
                , Phosphor.plusCircle Thin |> toHtml []
                , Phosphor.plusCircle Light |> toHtml []
                , Phosphor.plusCircle Regular |> toHtml []
                , Phosphor.plusCircle Bold |> toHtml []
                , Phosphor.plusCircle Fill |> toHtml []
                , Phosphor.plusCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plusMinus model.weight |> toHtml []
                , Phosphor.plusMinus Thin |> toHtml []
                , Phosphor.plusMinus Light |> toHtml []
                , Phosphor.plusMinus Regular |> toHtml []
                , Phosphor.plusMinus Bold |> toHtml []
                , Phosphor.plusMinus Fill |> toHtml []
                , Phosphor.plusMinus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.plusSquare model.weight |> toHtml []
                , Phosphor.plusSquare Thin |> toHtml []
                , Phosphor.plusSquare Light |> toHtml []
                , Phosphor.plusSquare Regular |> toHtml []
                , Phosphor.plusSquare Bold |> toHtml []
                , Phosphor.plusSquare Fill |> toHtml []
                , Phosphor.plusSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pokerChip model.weight |> toHtml []
                , Phosphor.pokerChip Thin |> toHtml []
                , Phosphor.pokerChip Light |> toHtml []
                , Phosphor.pokerChip Regular |> toHtml []
                , Phosphor.pokerChip Bold |> toHtml []
                , Phosphor.pokerChip Fill |> toHtml []
                , Phosphor.pokerChip Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.policeCar model.weight |> toHtml []
                , Phosphor.policeCar Thin |> toHtml []
                , Phosphor.policeCar Light |> toHtml []
                , Phosphor.policeCar Regular |> toHtml []
                , Phosphor.policeCar Bold |> toHtml []
                , Phosphor.policeCar Fill |> toHtml []
                , Phosphor.policeCar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.polygon model.weight |> toHtml []
                , Phosphor.polygon Thin |> toHtml []
                , Phosphor.polygon Light |> toHtml []
                , Phosphor.polygon Regular |> toHtml []
                , Phosphor.polygon Bold |> toHtml []
                , Phosphor.polygon Fill |> toHtml []
                , Phosphor.polygon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.popcorn model.weight |> toHtml []
                , Phosphor.popcorn Thin |> toHtml []
                , Phosphor.popcorn Light |> toHtml []
                , Phosphor.popcorn Regular |> toHtml []
                , Phosphor.popcorn Bold |> toHtml []
                , Phosphor.popcorn Fill |> toHtml []
                , Phosphor.popcorn Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pottedPlant model.weight |> toHtml []
                , Phosphor.pottedPlant Thin |> toHtml []
                , Phosphor.pottedPlant Light |> toHtml []
                , Phosphor.pottedPlant Regular |> toHtml []
                , Phosphor.pottedPlant Bold |> toHtml []
                , Phosphor.pottedPlant Fill |> toHtml []
                , Phosphor.pottedPlant Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.power model.weight |> toHtml []
                , Phosphor.power Thin |> toHtml []
                , Phosphor.power Light |> toHtml []
                , Phosphor.power Regular |> toHtml []
                , Phosphor.power Bold |> toHtml []
                , Phosphor.power Fill |> toHtml []
                , Phosphor.power Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.prescription model.weight |> toHtml []
                , Phosphor.prescription Thin |> toHtml []
                , Phosphor.prescription Light |> toHtml []
                , Phosphor.prescription Regular |> toHtml []
                , Phosphor.prescription Bold |> toHtml []
                , Phosphor.prescription Fill |> toHtml []
                , Phosphor.prescription Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.presentation model.weight |> toHtml []
                , Phosphor.presentation Thin |> toHtml []
                , Phosphor.presentation Light |> toHtml []
                , Phosphor.presentation Regular |> toHtml []
                , Phosphor.presentation Bold |> toHtml []
                , Phosphor.presentation Fill |> toHtml []
                , Phosphor.presentation Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.presentationChart model.weight |> toHtml []
                , Phosphor.presentationChart Thin |> toHtml []
                , Phosphor.presentationChart Light |> toHtml []
                , Phosphor.presentationChart Regular |> toHtml []
                , Phosphor.presentationChart Bold |> toHtml []
                , Phosphor.presentationChart Fill |> toHtml []
                , Phosphor.presentationChart Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.printer model.weight |> toHtml []
                , Phosphor.printer Thin |> toHtml []
                , Phosphor.printer Light |> toHtml []
                , Phosphor.printer Regular |> toHtml []
                , Phosphor.printer Bold |> toHtml []
                , Phosphor.printer Fill |> toHtml []
                , Phosphor.printer Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.prohibit model.weight |> toHtml []
                , Phosphor.prohibit Thin |> toHtml []
                , Phosphor.prohibit Light |> toHtml []
                , Phosphor.prohibit Regular |> toHtml []
                , Phosphor.prohibit Bold |> toHtml []
                , Phosphor.prohibit Fill |> toHtml []
                , Phosphor.prohibit Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.prohibitInset model.weight |> toHtml []
                , Phosphor.prohibitInset Thin |> toHtml []
                , Phosphor.prohibitInset Light |> toHtml []
                , Phosphor.prohibitInset Regular |> toHtml []
                , Phosphor.prohibitInset Bold |> toHtml []
                , Phosphor.prohibitInset Fill |> toHtml []
                , Phosphor.prohibitInset Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.projectorScreen model.weight |> toHtml []
                , Phosphor.projectorScreen Thin |> toHtml []
                , Phosphor.projectorScreen Light |> toHtml []
                , Phosphor.projectorScreen Regular |> toHtml []
                , Phosphor.projectorScreen Bold |> toHtml []
                , Phosphor.projectorScreen Fill |> toHtml []
                , Phosphor.projectorScreen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.projectorScreenChart model.weight |> toHtml []
                , Phosphor.projectorScreenChart Thin |> toHtml []
                , Phosphor.projectorScreenChart Light |> toHtml []
                , Phosphor.projectorScreenChart Regular |> toHtml []
                , Phosphor.projectorScreenChart Bold |> toHtml []
                , Phosphor.projectorScreenChart Fill |> toHtml []
                , Phosphor.projectorScreenChart Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pulse model.weight |> toHtml []
                , Phosphor.pulse Thin |> toHtml []
                , Phosphor.pulse Light |> toHtml []
                , Phosphor.pulse Regular |> toHtml []
                , Phosphor.pulse Bold |> toHtml []
                , Phosphor.pulse Fill |> toHtml []
                , Phosphor.pulse Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pushPin model.weight |> toHtml []
                , Phosphor.pushPin Thin |> toHtml []
                , Phosphor.pushPin Light |> toHtml []
                , Phosphor.pushPin Regular |> toHtml []
                , Phosphor.pushPin Bold |> toHtml []
                , Phosphor.pushPin Fill |> toHtml []
                , Phosphor.pushPin Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pushPinSimple model.weight |> toHtml []
                , Phosphor.pushPinSimple Thin |> toHtml []
                , Phosphor.pushPinSimple Light |> toHtml []
                , Phosphor.pushPinSimple Regular |> toHtml []
                , Phosphor.pushPinSimple Bold |> toHtml []
                , Phosphor.pushPinSimple Fill |> toHtml []
                , Phosphor.pushPinSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pushPinSimpleSlash model.weight |> toHtml []
                , Phosphor.pushPinSimpleSlash Thin |> toHtml []
                , Phosphor.pushPinSimpleSlash Light |> toHtml []
                , Phosphor.pushPinSimpleSlash Regular |> toHtml []
                , Phosphor.pushPinSimpleSlash Bold |> toHtml []
                , Phosphor.pushPinSimpleSlash Fill |> toHtml []
                , Phosphor.pushPinSimpleSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.pushPinSlash model.weight |> toHtml []
                , Phosphor.pushPinSlash Thin |> toHtml []
                , Phosphor.pushPinSlash Light |> toHtml []
                , Phosphor.pushPinSlash Regular |> toHtml []
                , Phosphor.pushPinSlash Bold |> toHtml []
                , Phosphor.pushPinSlash Fill |> toHtml []
                , Phosphor.pushPinSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.puzzlePiece model.weight |> toHtml []
                , Phosphor.puzzlePiece Thin |> toHtml []
                , Phosphor.puzzlePiece Light |> toHtml []
                , Phosphor.puzzlePiece Regular |> toHtml []
                , Phosphor.puzzlePiece Bold |> toHtml []
                , Phosphor.puzzlePiece Fill |> toHtml []
                , Phosphor.puzzlePiece Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.qrCode model.weight |> toHtml []
                , Phosphor.qrCode Thin |> toHtml []
                , Phosphor.qrCode Light |> toHtml []
                , Phosphor.qrCode Regular |> toHtml []
                , Phosphor.qrCode Bold |> toHtml []
                , Phosphor.qrCode Fill |> toHtml []
                , Phosphor.qrCode Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.question model.weight |> toHtml []
                , Phosphor.question Thin |> toHtml []
                , Phosphor.question Light |> toHtml []
                , Phosphor.question Regular |> toHtml []
                , Phosphor.question Bold |> toHtml []
                , Phosphor.question Fill |> toHtml []
                , Phosphor.question Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.queue model.weight |> toHtml []
                , Phosphor.queue Thin |> toHtml []
                , Phosphor.queue Light |> toHtml []
                , Phosphor.queue Regular |> toHtml []
                , Phosphor.queue Bold |> toHtml []
                , Phosphor.queue Fill |> toHtml []
                , Phosphor.queue Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.quotes model.weight |> toHtml []
                , Phosphor.quotes Thin |> toHtml []
                , Phosphor.quotes Light |> toHtml []
                , Phosphor.quotes Regular |> toHtml []
                , Phosphor.quotes Bold |> toHtml []
                , Phosphor.quotes Fill |> toHtml []
                , Phosphor.quotes Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.radical model.weight |> toHtml []
                , Phosphor.radical Thin |> toHtml []
                , Phosphor.radical Light |> toHtml []
                , Phosphor.radical Regular |> toHtml []
                , Phosphor.radical Bold |> toHtml []
                , Phosphor.radical Fill |> toHtml []
                , Phosphor.radical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.radio model.weight |> toHtml []
                , Phosphor.radio Thin |> toHtml []
                , Phosphor.radio Light |> toHtml []
                , Phosphor.radio Regular |> toHtml []
                , Phosphor.radio Bold |> toHtml []
                , Phosphor.radio Fill |> toHtml []
                , Phosphor.radio Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.radioButton model.weight |> toHtml []
                , Phosphor.radioButton Thin |> toHtml []
                , Phosphor.radioButton Light |> toHtml []
                , Phosphor.radioButton Regular |> toHtml []
                , Phosphor.radioButton Bold |> toHtml []
                , Phosphor.radioButton Fill |> toHtml []
                , Phosphor.radioButton Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.radioactive model.weight |> toHtml []
                , Phosphor.radioactive Thin |> toHtml []
                , Phosphor.radioactive Light |> toHtml []
                , Phosphor.radioactive Regular |> toHtml []
                , Phosphor.radioactive Bold |> toHtml []
                , Phosphor.radioactive Fill |> toHtml []
                , Phosphor.radioactive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rainbow model.weight |> toHtml []
                , Phosphor.rainbow Thin |> toHtml []
                , Phosphor.rainbow Light |> toHtml []
                , Phosphor.rainbow Regular |> toHtml []
                , Phosphor.rainbow Bold |> toHtml []
                , Phosphor.rainbow Fill |> toHtml []
                , Phosphor.rainbow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rainbowCloud model.weight |> toHtml []
                , Phosphor.rainbowCloud Thin |> toHtml []
                , Phosphor.rainbowCloud Light |> toHtml []
                , Phosphor.rainbowCloud Regular |> toHtml []
                , Phosphor.rainbowCloud Bold |> toHtml []
                , Phosphor.rainbowCloud Fill |> toHtml []
                , Phosphor.rainbowCloud Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.readCvLogo model.weight |> toHtml []
                , Phosphor.readCvLogo Thin |> toHtml []
                , Phosphor.readCvLogo Light |> toHtml []
                , Phosphor.readCvLogo Regular |> toHtml []
                , Phosphor.readCvLogo Bold |> toHtml []
                , Phosphor.readCvLogo Fill |> toHtml []
                , Phosphor.readCvLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.receipt model.weight |> toHtml []
                , Phosphor.receipt Thin |> toHtml []
                , Phosphor.receipt Light |> toHtml []
                , Phosphor.receipt Regular |> toHtml []
                , Phosphor.receipt Bold |> toHtml []
                , Phosphor.receipt Fill |> toHtml []
                , Phosphor.receipt Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.receiptX model.weight |> toHtml []
                , Phosphor.receiptX Thin |> toHtml []
                , Phosphor.receiptX Light |> toHtml []
                , Phosphor.receiptX Regular |> toHtml []
                , Phosphor.receiptX Bold |> toHtml []
                , Phosphor.receiptX Fill |> toHtml []
                , Phosphor.receiptX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.record model.weight |> toHtml []
                , Phosphor.record Thin |> toHtml []
                , Phosphor.record Light |> toHtml []
                , Phosphor.record Regular |> toHtml []
                , Phosphor.record Bold |> toHtml []
                , Phosphor.record Fill |> toHtml []
                , Phosphor.record Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rectangle model.weight |> toHtml []
                , Phosphor.rectangle Thin |> toHtml []
                , Phosphor.rectangle Light |> toHtml []
                , Phosphor.rectangle Regular |> toHtml []
                , Phosphor.rectangle Bold |> toHtml []
                , Phosphor.rectangle Fill |> toHtml []
                , Phosphor.rectangle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.recycle model.weight |> toHtml []
                , Phosphor.recycle Thin |> toHtml []
                , Phosphor.recycle Light |> toHtml []
                , Phosphor.recycle Regular |> toHtml []
                , Phosphor.recycle Bold |> toHtml []
                , Phosphor.recycle Fill |> toHtml []
                , Phosphor.recycle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.redditLogo model.weight |> toHtml []
                , Phosphor.redditLogo Thin |> toHtml []
                , Phosphor.redditLogo Light |> toHtml []
                , Phosphor.redditLogo Regular |> toHtml []
                , Phosphor.redditLogo Bold |> toHtml []
                , Phosphor.redditLogo Fill |> toHtml []
                , Phosphor.redditLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.repeat model.weight |> toHtml []
                , Phosphor.repeat Thin |> toHtml []
                , Phosphor.repeat Light |> toHtml []
                , Phosphor.repeat Regular |> toHtml []
                , Phosphor.repeat Bold |> toHtml []
                , Phosphor.repeat Fill |> toHtml []
                , Phosphor.repeat Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.repeatOnce model.weight |> toHtml []
                , Phosphor.repeatOnce Thin |> toHtml []
                , Phosphor.repeatOnce Light |> toHtml []
                , Phosphor.repeatOnce Regular |> toHtml []
                , Phosphor.repeatOnce Bold |> toHtml []
                , Phosphor.repeatOnce Fill |> toHtml []
                , Phosphor.repeatOnce Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rewind model.weight |> toHtml []
                , Phosphor.rewind Thin |> toHtml []
                , Phosphor.rewind Light |> toHtml []
                , Phosphor.rewind Regular |> toHtml []
                , Phosphor.rewind Bold |> toHtml []
                , Phosphor.rewind Fill |> toHtml []
                , Phosphor.rewind Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rewindCircle model.weight |> toHtml []
                , Phosphor.rewindCircle Thin |> toHtml []
                , Phosphor.rewindCircle Light |> toHtml []
                , Phosphor.rewindCircle Regular |> toHtml []
                , Phosphor.rewindCircle Bold |> toHtml []
                , Phosphor.rewindCircle Fill |> toHtml []
                , Phosphor.rewindCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.roadHorizon model.weight |> toHtml []
                , Phosphor.roadHorizon Thin |> toHtml []
                , Phosphor.roadHorizon Light |> toHtml []
                , Phosphor.roadHorizon Regular |> toHtml []
                , Phosphor.roadHorizon Bold |> toHtml []
                , Phosphor.roadHorizon Fill |> toHtml []
                , Phosphor.roadHorizon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.robot model.weight |> toHtml []
                , Phosphor.robot Thin |> toHtml []
                , Phosphor.robot Light |> toHtml []
                , Phosphor.robot Regular |> toHtml []
                , Phosphor.robot Bold |> toHtml []
                , Phosphor.robot Fill |> toHtml []
                , Phosphor.robot Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rocket model.weight |> toHtml []
                , Phosphor.rocket Thin |> toHtml []
                , Phosphor.rocket Light |> toHtml []
                , Phosphor.rocket Regular |> toHtml []
                , Phosphor.rocket Bold |> toHtml []
                , Phosphor.rocket Fill |> toHtml []
                , Phosphor.rocket Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rocketLaunch model.weight |> toHtml []
                , Phosphor.rocketLaunch Thin |> toHtml []
                , Phosphor.rocketLaunch Light |> toHtml []
                , Phosphor.rocketLaunch Regular |> toHtml []
                , Phosphor.rocketLaunch Bold |> toHtml []
                , Phosphor.rocketLaunch Fill |> toHtml []
                , Phosphor.rocketLaunch Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rows model.weight |> toHtml []
                , Phosphor.rows Thin |> toHtml []
                , Phosphor.rows Light |> toHtml []
                , Phosphor.rows Regular |> toHtml []
                , Phosphor.rows Bold |> toHtml []
                , Phosphor.rows Fill |> toHtml []
                , Phosphor.rows Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rss model.weight |> toHtml []
                , Phosphor.rss Thin |> toHtml []
                , Phosphor.rss Light |> toHtml []
                , Phosphor.rss Regular |> toHtml []
                , Phosphor.rss Bold |> toHtml []
                , Phosphor.rss Fill |> toHtml []
                , Phosphor.rss Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rssSimple model.weight |> toHtml []
                , Phosphor.rssSimple Thin |> toHtml []
                , Phosphor.rssSimple Light |> toHtml []
                , Phosphor.rssSimple Regular |> toHtml []
                , Phosphor.rssSimple Bold |> toHtml []
                , Phosphor.rssSimple Fill |> toHtml []
                , Phosphor.rssSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.rug model.weight |> toHtml []
                , Phosphor.rug Thin |> toHtml []
                , Phosphor.rug Light |> toHtml []
                , Phosphor.rug Regular |> toHtml []
                , Phosphor.rug Bold |> toHtml []
                , Phosphor.rug Fill |> toHtml []
                , Phosphor.rug Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.ruler model.weight |> toHtml []
                , Phosphor.ruler Thin |> toHtml []
                , Phosphor.ruler Light |> toHtml []
                , Phosphor.ruler Regular |> toHtml []
                , Phosphor.ruler Bold |> toHtml []
                , Phosphor.ruler Fill |> toHtml []
                , Phosphor.ruler Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.scales model.weight |> toHtml []
                , Phosphor.scales Thin |> toHtml []
                , Phosphor.scales Light |> toHtml []
                , Phosphor.scales Regular |> toHtml []
                , Phosphor.scales Bold |> toHtml []
                , Phosphor.scales Fill |> toHtml []
                , Phosphor.scales Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.scan model.weight |> toHtml []
                , Phosphor.scan Thin |> toHtml []
                , Phosphor.scan Light |> toHtml []
                , Phosphor.scan Regular |> toHtml []
                , Phosphor.scan Bold |> toHtml []
                , Phosphor.scan Fill |> toHtml []
                , Phosphor.scan Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.scissors model.weight |> toHtml []
                , Phosphor.scissors Thin |> toHtml []
                , Phosphor.scissors Light |> toHtml []
                , Phosphor.scissors Regular |> toHtml []
                , Phosphor.scissors Bold |> toHtml []
                , Phosphor.scissors Fill |> toHtml []
                , Phosphor.scissors Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.scooter model.weight |> toHtml []
                , Phosphor.scooter Thin |> toHtml []
                , Phosphor.scooter Light |> toHtml []
                , Phosphor.scooter Regular |> toHtml []
                , Phosphor.scooter Bold |> toHtml []
                , Phosphor.scooter Fill |> toHtml []
                , Phosphor.scooter Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.screencast model.weight |> toHtml []
                , Phosphor.screencast Thin |> toHtml []
                , Phosphor.screencast Light |> toHtml []
                , Phosphor.screencast Regular |> toHtml []
                , Phosphor.screencast Bold |> toHtml []
                , Phosphor.screencast Fill |> toHtml []
                , Phosphor.screencast Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.scribbleLoop model.weight |> toHtml []
                , Phosphor.scribbleLoop Thin |> toHtml []
                , Phosphor.scribbleLoop Light |> toHtml []
                , Phosphor.scribbleLoop Regular |> toHtml []
                , Phosphor.scribbleLoop Bold |> toHtml []
                , Phosphor.scribbleLoop Fill |> toHtml []
                , Phosphor.scribbleLoop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.scroll model.weight |> toHtml []
                , Phosphor.scroll Thin |> toHtml []
                , Phosphor.scroll Light |> toHtml []
                , Phosphor.scroll Regular |> toHtml []
                , Phosphor.scroll Bold |> toHtml []
                , Phosphor.scroll Fill |> toHtml []
                , Phosphor.scroll Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.seal model.weight |> toHtml []
                , Phosphor.seal Thin |> toHtml []
                , Phosphor.seal Light |> toHtml []
                , Phosphor.seal Regular |> toHtml []
                , Phosphor.seal Bold |> toHtml []
                , Phosphor.seal Fill |> toHtml []
                , Phosphor.seal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sealCheck model.weight |> toHtml []
                , Phosphor.sealCheck Thin |> toHtml []
                , Phosphor.sealCheck Light |> toHtml []
                , Phosphor.sealCheck Regular |> toHtml []
                , Phosphor.sealCheck Bold |> toHtml []
                , Phosphor.sealCheck Fill |> toHtml []
                , Phosphor.sealCheck Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sealQuestion model.weight |> toHtml []
                , Phosphor.sealQuestion Thin |> toHtml []
                , Phosphor.sealQuestion Light |> toHtml []
                , Phosphor.sealQuestion Regular |> toHtml []
                , Phosphor.sealQuestion Bold |> toHtml []
                , Phosphor.sealQuestion Fill |> toHtml []
                , Phosphor.sealQuestion Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sealWarning model.weight |> toHtml []
                , Phosphor.sealWarning Thin |> toHtml []
                , Phosphor.sealWarning Light |> toHtml []
                , Phosphor.sealWarning Regular |> toHtml []
                , Phosphor.sealWarning Bold |> toHtml []
                , Phosphor.sealWarning Fill |> toHtml []
                , Phosphor.sealWarning Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.selectionAll model.weight |> toHtml []
                , Phosphor.selectionAll Thin |> toHtml []
                , Phosphor.selectionAll Light |> toHtml []
                , Phosphor.selectionAll Regular |> toHtml []
                , Phosphor.selectionAll Bold |> toHtml []
                , Phosphor.selectionAll Fill |> toHtml []
                , Phosphor.selectionAll Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.selectionBackground model.weight |> toHtml []
                , Phosphor.selectionBackground Thin |> toHtml []
                , Phosphor.selectionBackground Light |> toHtml []
                , Phosphor.selectionBackground Regular |> toHtml []
                , Phosphor.selectionBackground Bold |> toHtml []
                , Phosphor.selectionBackground Fill |> toHtml []
                , Phosphor.selectionBackground Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.selection model.weight |> toHtml []
                , Phosphor.selection Thin |> toHtml []
                , Phosphor.selection Light |> toHtml []
                , Phosphor.selection Regular |> toHtml []
                , Phosphor.selection Bold |> toHtml []
                , Phosphor.selection Fill |> toHtml []
                , Phosphor.selection Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.selectionForeground model.weight |> toHtml []
                , Phosphor.selectionForeground Thin |> toHtml []
                , Phosphor.selectionForeground Light |> toHtml []
                , Phosphor.selectionForeground Regular |> toHtml []
                , Phosphor.selectionForeground Bold |> toHtml []
                , Phosphor.selectionForeground Fill |> toHtml []
                , Phosphor.selectionForeground Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.selectionInverse model.weight |> toHtml []
                , Phosphor.selectionInverse Thin |> toHtml []
                , Phosphor.selectionInverse Light |> toHtml []
                , Phosphor.selectionInverse Regular |> toHtml []
                , Phosphor.selectionInverse Bold |> toHtml []
                , Phosphor.selectionInverse Fill |> toHtml []
                , Phosphor.selectionInverse Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.selectionPlus model.weight |> toHtml []
                , Phosphor.selectionPlus Thin |> toHtml []
                , Phosphor.selectionPlus Light |> toHtml []
                , Phosphor.selectionPlus Regular |> toHtml []
                , Phosphor.selectionPlus Bold |> toHtml []
                , Phosphor.selectionPlus Fill |> toHtml []
                , Phosphor.selectionPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.selectionSlash model.weight |> toHtml []
                , Phosphor.selectionSlash Thin |> toHtml []
                , Phosphor.selectionSlash Light |> toHtml []
                , Phosphor.selectionSlash Regular |> toHtml []
                , Phosphor.selectionSlash Bold |> toHtml []
                , Phosphor.selectionSlash Fill |> toHtml []
                , Phosphor.selectionSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shapes model.weight |> toHtml []
                , Phosphor.shapes Thin |> toHtml []
                , Phosphor.shapes Light |> toHtml []
                , Phosphor.shapes Regular |> toHtml []
                , Phosphor.shapes Bold |> toHtml []
                , Phosphor.shapes Fill |> toHtml []
                , Phosphor.shapes Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.share model.weight |> toHtml []
                , Phosphor.share Thin |> toHtml []
                , Phosphor.share Light |> toHtml []
                , Phosphor.share Regular |> toHtml []
                , Phosphor.share Bold |> toHtml []
                , Phosphor.share Fill |> toHtml []
                , Phosphor.share Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shareFat model.weight |> toHtml []
                , Phosphor.shareFat Thin |> toHtml []
                , Phosphor.shareFat Light |> toHtml []
                , Phosphor.shareFat Regular |> toHtml []
                , Phosphor.shareFat Bold |> toHtml []
                , Phosphor.shareFat Fill |> toHtml []
                , Phosphor.shareFat Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shareNetwork model.weight |> toHtml []
                , Phosphor.shareNetwork Thin |> toHtml []
                , Phosphor.shareNetwork Light |> toHtml []
                , Phosphor.shareNetwork Regular |> toHtml []
                , Phosphor.shareNetwork Bold |> toHtml []
                , Phosphor.shareNetwork Fill |> toHtml []
                , Phosphor.shareNetwork Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shield model.weight |> toHtml []
                , Phosphor.shield Thin |> toHtml []
                , Phosphor.shield Light |> toHtml []
                , Phosphor.shield Regular |> toHtml []
                , Phosphor.shield Bold |> toHtml []
                , Phosphor.shield Fill |> toHtml []
                , Phosphor.shield Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shieldCheck model.weight |> toHtml []
                , Phosphor.shieldCheck Thin |> toHtml []
                , Phosphor.shieldCheck Light |> toHtml []
                , Phosphor.shieldCheck Regular |> toHtml []
                , Phosphor.shieldCheck Bold |> toHtml []
                , Phosphor.shieldCheck Fill |> toHtml []
                , Phosphor.shieldCheck Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shieldCheckered model.weight |> toHtml []
                , Phosphor.shieldCheckered Thin |> toHtml []
                , Phosphor.shieldCheckered Light |> toHtml []
                , Phosphor.shieldCheckered Regular |> toHtml []
                , Phosphor.shieldCheckered Bold |> toHtml []
                , Phosphor.shieldCheckered Fill |> toHtml []
                , Phosphor.shieldCheckered Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shieldChevron model.weight |> toHtml []
                , Phosphor.shieldChevron Thin |> toHtml []
                , Phosphor.shieldChevron Light |> toHtml []
                , Phosphor.shieldChevron Regular |> toHtml []
                , Phosphor.shieldChevron Bold |> toHtml []
                , Phosphor.shieldChevron Fill |> toHtml []
                , Phosphor.shieldChevron Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shieldPlus model.weight |> toHtml []
                , Phosphor.shieldPlus Thin |> toHtml []
                , Phosphor.shieldPlus Light |> toHtml []
                , Phosphor.shieldPlus Regular |> toHtml []
                , Phosphor.shieldPlus Bold |> toHtml []
                , Phosphor.shieldPlus Fill |> toHtml []
                , Phosphor.shieldPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shieldSlash model.weight |> toHtml []
                , Phosphor.shieldSlash Thin |> toHtml []
                , Phosphor.shieldSlash Light |> toHtml []
                , Phosphor.shieldSlash Regular |> toHtml []
                , Phosphor.shieldSlash Bold |> toHtml []
                , Phosphor.shieldSlash Fill |> toHtml []
                , Phosphor.shieldSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shieldStar model.weight |> toHtml []
                , Phosphor.shieldStar Thin |> toHtml []
                , Phosphor.shieldStar Light |> toHtml []
                , Phosphor.shieldStar Regular |> toHtml []
                , Phosphor.shieldStar Bold |> toHtml []
                , Phosphor.shieldStar Fill |> toHtml []
                , Phosphor.shieldStar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shieldWarning model.weight |> toHtml []
                , Phosphor.shieldWarning Thin |> toHtml []
                , Phosphor.shieldWarning Light |> toHtml []
                , Phosphor.shieldWarning Regular |> toHtml []
                , Phosphor.shieldWarning Bold |> toHtml []
                , Phosphor.shieldWarning Fill |> toHtml []
                , Phosphor.shieldWarning Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shirtFolded model.weight |> toHtml []
                , Phosphor.shirtFolded Thin |> toHtml []
                , Phosphor.shirtFolded Light |> toHtml []
                , Phosphor.shirtFolded Regular |> toHtml []
                , Phosphor.shirtFolded Bold |> toHtml []
                , Phosphor.shirtFolded Fill |> toHtml []
                , Phosphor.shirtFolded Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shootingStar model.weight |> toHtml []
                , Phosphor.shootingStar Thin |> toHtml []
                , Phosphor.shootingStar Light |> toHtml []
                , Phosphor.shootingStar Regular |> toHtml []
                , Phosphor.shootingStar Bold |> toHtml []
                , Phosphor.shootingStar Fill |> toHtml []
                , Phosphor.shootingStar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shoppingBag model.weight |> toHtml []
                , Phosphor.shoppingBag Thin |> toHtml []
                , Phosphor.shoppingBag Light |> toHtml []
                , Phosphor.shoppingBag Regular |> toHtml []
                , Phosphor.shoppingBag Bold |> toHtml []
                , Phosphor.shoppingBag Fill |> toHtml []
                , Phosphor.shoppingBag Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shoppingBagOpen model.weight |> toHtml []
                , Phosphor.shoppingBagOpen Thin |> toHtml []
                , Phosphor.shoppingBagOpen Light |> toHtml []
                , Phosphor.shoppingBagOpen Regular |> toHtml []
                , Phosphor.shoppingBagOpen Bold |> toHtml []
                , Phosphor.shoppingBagOpen Fill |> toHtml []
                , Phosphor.shoppingBagOpen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shoppingCart model.weight |> toHtml []
                , Phosphor.shoppingCart Thin |> toHtml []
                , Phosphor.shoppingCart Light |> toHtml []
                , Phosphor.shoppingCart Regular |> toHtml []
                , Phosphor.shoppingCart Bold |> toHtml []
                , Phosphor.shoppingCart Fill |> toHtml []
                , Phosphor.shoppingCart Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shoppingCartSimple model.weight |> toHtml []
                , Phosphor.shoppingCartSimple Thin |> toHtml []
                , Phosphor.shoppingCartSimple Light |> toHtml []
                , Phosphor.shoppingCartSimple Regular |> toHtml []
                , Phosphor.shoppingCartSimple Bold |> toHtml []
                , Phosphor.shoppingCartSimple Fill |> toHtml []
                , Phosphor.shoppingCartSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shower model.weight |> toHtml []
                , Phosphor.shower Thin |> toHtml []
                , Phosphor.shower Light |> toHtml []
                , Phosphor.shower Regular |> toHtml []
                , Phosphor.shower Bold |> toHtml []
                , Phosphor.shower Fill |> toHtml []
                , Phosphor.shower Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shrimp model.weight |> toHtml []
                , Phosphor.shrimp Thin |> toHtml []
                , Phosphor.shrimp Light |> toHtml []
                , Phosphor.shrimp Regular |> toHtml []
                , Phosphor.shrimp Bold |> toHtml []
                , Phosphor.shrimp Fill |> toHtml []
                , Phosphor.shrimp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shuffleAngular model.weight |> toHtml []
                , Phosphor.shuffleAngular Thin |> toHtml []
                , Phosphor.shuffleAngular Light |> toHtml []
                , Phosphor.shuffleAngular Regular |> toHtml []
                , Phosphor.shuffleAngular Bold |> toHtml []
                , Phosphor.shuffleAngular Fill |> toHtml []
                , Phosphor.shuffleAngular Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shuffle model.weight |> toHtml []
                , Phosphor.shuffle Thin |> toHtml []
                , Phosphor.shuffle Light |> toHtml []
                , Phosphor.shuffle Regular |> toHtml []
                , Phosphor.shuffle Bold |> toHtml []
                , Phosphor.shuffle Fill |> toHtml []
                , Phosphor.shuffle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.shuffleSimple model.weight |> toHtml []
                , Phosphor.shuffleSimple Thin |> toHtml []
                , Phosphor.shuffleSimple Light |> toHtml []
                , Phosphor.shuffleSimple Regular |> toHtml []
                , Phosphor.shuffleSimple Bold |> toHtml []
                , Phosphor.shuffleSimple Fill |> toHtml []
                , Phosphor.shuffleSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sidebar model.weight |> toHtml []
                , Phosphor.sidebar Thin |> toHtml []
                , Phosphor.sidebar Light |> toHtml []
                , Phosphor.sidebar Regular |> toHtml []
                , Phosphor.sidebar Bold |> toHtml []
                , Phosphor.sidebar Fill |> toHtml []
                , Phosphor.sidebar Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sidebarSimple model.weight |> toHtml []
                , Phosphor.sidebarSimple Thin |> toHtml []
                , Phosphor.sidebarSimple Light |> toHtml []
                , Phosphor.sidebarSimple Regular |> toHtml []
                , Phosphor.sidebarSimple Bold |> toHtml []
                , Phosphor.sidebarSimple Fill |> toHtml []
                , Phosphor.sidebarSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sigma model.weight |> toHtml []
                , Phosphor.sigma Thin |> toHtml []
                , Phosphor.sigma Light |> toHtml []
                , Phosphor.sigma Regular |> toHtml []
                , Phosphor.sigma Bold |> toHtml []
                , Phosphor.sigma Fill |> toHtml []
                , Phosphor.sigma Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.signIn model.weight |> toHtml []
                , Phosphor.signIn Thin |> toHtml []
                , Phosphor.signIn Light |> toHtml []
                , Phosphor.signIn Regular |> toHtml []
                , Phosphor.signIn Bold |> toHtml []
                , Phosphor.signIn Fill |> toHtml []
                , Phosphor.signIn Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.signOut model.weight |> toHtml []
                , Phosphor.signOut Thin |> toHtml []
                , Phosphor.signOut Light |> toHtml []
                , Phosphor.signOut Regular |> toHtml []
                , Phosphor.signOut Bold |> toHtml []
                , Phosphor.signOut Fill |> toHtml []
                , Phosphor.signOut Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.signature model.weight |> toHtml []
                , Phosphor.signature Thin |> toHtml []
                , Phosphor.signature Light |> toHtml []
                , Phosphor.signature Regular |> toHtml []
                , Phosphor.signature Bold |> toHtml []
                , Phosphor.signature Fill |> toHtml []
                , Phosphor.signature Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.signpost model.weight |> toHtml []
                , Phosphor.signpost Thin |> toHtml []
                , Phosphor.signpost Light |> toHtml []
                , Phosphor.signpost Regular |> toHtml []
                , Phosphor.signpost Bold |> toHtml []
                , Phosphor.signpost Fill |> toHtml []
                , Phosphor.signpost Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.simCard model.weight |> toHtml []
                , Phosphor.simCard Thin |> toHtml []
                , Phosphor.simCard Light |> toHtml []
                , Phosphor.simCard Regular |> toHtml []
                , Phosphor.simCard Bold |> toHtml []
                , Phosphor.simCard Fill |> toHtml []
                , Phosphor.simCard Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.siren model.weight |> toHtml []
                , Phosphor.siren Thin |> toHtml []
                , Phosphor.siren Light |> toHtml []
                , Phosphor.siren Regular |> toHtml []
                , Phosphor.siren Bold |> toHtml []
                , Phosphor.siren Fill |> toHtml []
                , Phosphor.siren Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sketchLogo model.weight |> toHtml []
                , Phosphor.sketchLogo Thin |> toHtml []
                , Phosphor.sketchLogo Light |> toHtml []
                , Phosphor.sketchLogo Regular |> toHtml []
                , Phosphor.sketchLogo Bold |> toHtml []
                , Phosphor.sketchLogo Fill |> toHtml []
                , Phosphor.sketchLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.skipBack model.weight |> toHtml []
                , Phosphor.skipBack Thin |> toHtml []
                , Phosphor.skipBack Light |> toHtml []
                , Phosphor.skipBack Regular |> toHtml []
                , Phosphor.skipBack Bold |> toHtml []
                , Phosphor.skipBack Fill |> toHtml []
                , Phosphor.skipBack Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.skipBackCircle model.weight |> toHtml []
                , Phosphor.skipBackCircle Thin |> toHtml []
                , Phosphor.skipBackCircle Light |> toHtml []
                , Phosphor.skipBackCircle Regular |> toHtml []
                , Phosphor.skipBackCircle Bold |> toHtml []
                , Phosphor.skipBackCircle Fill |> toHtml []
                , Phosphor.skipBackCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.skipForward model.weight |> toHtml []
                , Phosphor.skipForward Thin |> toHtml []
                , Phosphor.skipForward Light |> toHtml []
                , Phosphor.skipForward Regular |> toHtml []
                , Phosphor.skipForward Bold |> toHtml []
                , Phosphor.skipForward Fill |> toHtml []
                , Phosphor.skipForward Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.skipForwardCircle model.weight |> toHtml []
                , Phosphor.skipForwardCircle Thin |> toHtml []
                , Phosphor.skipForwardCircle Light |> toHtml []
                , Phosphor.skipForwardCircle Regular |> toHtml []
                , Phosphor.skipForwardCircle Bold |> toHtml []
                , Phosphor.skipForwardCircle Fill |> toHtml []
                , Phosphor.skipForwardCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.skull model.weight |> toHtml []
                , Phosphor.skull Thin |> toHtml []
                , Phosphor.skull Light |> toHtml []
                , Phosphor.skull Regular |> toHtml []
                , Phosphor.skull Bold |> toHtml []
                , Phosphor.skull Fill |> toHtml []
                , Phosphor.skull Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.slackLogo model.weight |> toHtml []
                , Phosphor.slackLogo Thin |> toHtml []
                , Phosphor.slackLogo Light |> toHtml []
                , Phosphor.slackLogo Regular |> toHtml []
                , Phosphor.slackLogo Bold |> toHtml []
                , Phosphor.slackLogo Fill |> toHtml []
                , Phosphor.slackLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sliders model.weight |> toHtml []
                , Phosphor.sliders Thin |> toHtml []
                , Phosphor.sliders Light |> toHtml []
                , Phosphor.sliders Regular |> toHtml []
                , Phosphor.sliders Bold |> toHtml []
                , Phosphor.sliders Fill |> toHtml []
                , Phosphor.sliders Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.slidersHorizontal model.weight |> toHtml []
                , Phosphor.slidersHorizontal Thin |> toHtml []
                , Phosphor.slidersHorizontal Light |> toHtml []
                , Phosphor.slidersHorizontal Regular |> toHtml []
                , Phosphor.slidersHorizontal Bold |> toHtml []
                , Phosphor.slidersHorizontal Fill |> toHtml []
                , Phosphor.slidersHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.slideshow model.weight |> toHtml []
                , Phosphor.slideshow Thin |> toHtml []
                , Phosphor.slideshow Light |> toHtml []
                , Phosphor.slideshow Regular |> toHtml []
                , Phosphor.slideshow Bold |> toHtml []
                , Phosphor.slideshow Fill |> toHtml []
                , Phosphor.slideshow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smileyAngry model.weight |> toHtml []
                , Phosphor.smileyAngry Thin |> toHtml []
                , Phosphor.smileyAngry Light |> toHtml []
                , Phosphor.smileyAngry Regular |> toHtml []
                , Phosphor.smileyAngry Bold |> toHtml []
                , Phosphor.smileyAngry Fill |> toHtml []
                , Phosphor.smileyAngry Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smileyBlank model.weight |> toHtml []
                , Phosphor.smileyBlank Thin |> toHtml []
                , Phosphor.smileyBlank Light |> toHtml []
                , Phosphor.smileyBlank Regular |> toHtml []
                , Phosphor.smileyBlank Bold |> toHtml []
                , Phosphor.smileyBlank Fill |> toHtml []
                , Phosphor.smileyBlank Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smiley model.weight |> toHtml []
                , Phosphor.smiley Thin |> toHtml []
                , Phosphor.smiley Light |> toHtml []
                , Phosphor.smiley Regular |> toHtml []
                , Phosphor.smiley Bold |> toHtml []
                , Phosphor.smiley Fill |> toHtml []
                , Phosphor.smiley Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smileyMeh model.weight |> toHtml []
                , Phosphor.smileyMeh Thin |> toHtml []
                , Phosphor.smileyMeh Light |> toHtml []
                , Phosphor.smileyMeh Regular |> toHtml []
                , Phosphor.smileyMeh Bold |> toHtml []
                , Phosphor.smileyMeh Fill |> toHtml []
                , Phosphor.smileyMeh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smileyNervous model.weight |> toHtml []
                , Phosphor.smileyNervous Thin |> toHtml []
                , Phosphor.smileyNervous Light |> toHtml []
                , Phosphor.smileyNervous Regular |> toHtml []
                , Phosphor.smileyNervous Bold |> toHtml []
                , Phosphor.smileyNervous Fill |> toHtml []
                , Phosphor.smileyNervous Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smileySad model.weight |> toHtml []
                , Phosphor.smileySad Thin |> toHtml []
                , Phosphor.smileySad Light |> toHtml []
                , Phosphor.smileySad Regular |> toHtml []
                , Phosphor.smileySad Bold |> toHtml []
                , Phosphor.smileySad Fill |> toHtml []
                , Phosphor.smileySad Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smileySticker model.weight |> toHtml []
                , Phosphor.smileySticker Thin |> toHtml []
                , Phosphor.smileySticker Light |> toHtml []
                , Phosphor.smileySticker Regular |> toHtml []
                , Phosphor.smileySticker Bold |> toHtml []
                , Phosphor.smileySticker Fill |> toHtml []
                , Phosphor.smileySticker Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smileyWink model.weight |> toHtml []
                , Phosphor.smileyWink Thin |> toHtml []
                , Phosphor.smileyWink Light |> toHtml []
                , Phosphor.smileyWink Regular |> toHtml []
                , Phosphor.smileyWink Bold |> toHtml []
                , Phosphor.smileyWink Fill |> toHtml []
                , Phosphor.smileyWink Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.smileyXEyes model.weight |> toHtml []
                , Phosphor.smileyXEyes Thin |> toHtml []
                , Phosphor.smileyXEyes Light |> toHtml []
                , Phosphor.smileyXEyes Regular |> toHtml []
                , Phosphor.smileyXEyes Bold |> toHtml []
                , Phosphor.smileyXEyes Fill |> toHtml []
                , Phosphor.smileyXEyes Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.snapchatLogo model.weight |> toHtml []
                , Phosphor.snapchatLogo Thin |> toHtml []
                , Phosphor.snapchatLogo Light |> toHtml []
                , Phosphor.snapchatLogo Regular |> toHtml []
                , Phosphor.snapchatLogo Bold |> toHtml []
                , Phosphor.snapchatLogo Fill |> toHtml []
                , Phosphor.snapchatLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sneaker model.weight |> toHtml []
                , Phosphor.sneaker Thin |> toHtml []
                , Phosphor.sneaker Light |> toHtml []
                , Phosphor.sneaker Regular |> toHtml []
                , Phosphor.sneaker Bold |> toHtml []
                , Phosphor.sneaker Fill |> toHtml []
                , Phosphor.sneaker Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sneakerMove model.weight |> toHtml []
                , Phosphor.sneakerMove Thin |> toHtml []
                , Phosphor.sneakerMove Light |> toHtml []
                , Phosphor.sneakerMove Regular |> toHtml []
                , Phosphor.sneakerMove Bold |> toHtml []
                , Phosphor.sneakerMove Fill |> toHtml []
                , Phosphor.sneakerMove Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.snowflake model.weight |> toHtml []
                , Phosphor.snowflake Thin |> toHtml []
                , Phosphor.snowflake Light |> toHtml []
                , Phosphor.snowflake Regular |> toHtml []
                , Phosphor.snowflake Bold |> toHtml []
                , Phosphor.snowflake Fill |> toHtml []
                , Phosphor.snowflake Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.soccerBall model.weight |> toHtml []
                , Phosphor.soccerBall Thin |> toHtml []
                , Phosphor.soccerBall Light |> toHtml []
                , Phosphor.soccerBall Regular |> toHtml []
                , Phosphor.soccerBall Bold |> toHtml []
                , Phosphor.soccerBall Fill |> toHtml []
                , Phosphor.soccerBall Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sortAscending model.weight |> toHtml []
                , Phosphor.sortAscending Thin |> toHtml []
                , Phosphor.sortAscending Light |> toHtml []
                , Phosphor.sortAscending Regular |> toHtml []
                , Phosphor.sortAscending Bold |> toHtml []
                , Phosphor.sortAscending Fill |> toHtml []
                , Phosphor.sortAscending Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sortDescending model.weight |> toHtml []
                , Phosphor.sortDescending Thin |> toHtml []
                , Phosphor.sortDescending Light |> toHtml []
                , Phosphor.sortDescending Regular |> toHtml []
                , Phosphor.sortDescending Bold |> toHtml []
                , Phosphor.sortDescending Fill |> toHtml []
                , Phosphor.sortDescending Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.soundcloudLogo model.weight |> toHtml []
                , Phosphor.soundcloudLogo Thin |> toHtml []
                , Phosphor.soundcloudLogo Light |> toHtml []
                , Phosphor.soundcloudLogo Regular |> toHtml []
                , Phosphor.soundcloudLogo Bold |> toHtml []
                , Phosphor.soundcloudLogo Fill |> toHtml []
                , Phosphor.soundcloudLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.spade model.weight |> toHtml []
                , Phosphor.spade Thin |> toHtml []
                , Phosphor.spade Light |> toHtml []
                , Phosphor.spade Regular |> toHtml []
                , Phosphor.spade Bold |> toHtml []
                , Phosphor.spade Fill |> toHtml []
                , Phosphor.spade Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sparkle model.weight |> toHtml []
                , Phosphor.sparkle Thin |> toHtml []
                , Phosphor.sparkle Light |> toHtml []
                , Phosphor.sparkle Regular |> toHtml []
                , Phosphor.sparkle Bold |> toHtml []
                , Phosphor.sparkle Fill |> toHtml []
                , Phosphor.sparkle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerHifi model.weight |> toHtml []
                , Phosphor.speakerHifi Thin |> toHtml []
                , Phosphor.speakerHifi Light |> toHtml []
                , Phosphor.speakerHifi Regular |> toHtml []
                , Phosphor.speakerHifi Bold |> toHtml []
                , Phosphor.speakerHifi Fill |> toHtml []
                , Phosphor.speakerHifi Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerHigh model.weight |> toHtml []
                , Phosphor.speakerHigh Thin |> toHtml []
                , Phosphor.speakerHigh Light |> toHtml []
                , Phosphor.speakerHigh Regular |> toHtml []
                , Phosphor.speakerHigh Bold |> toHtml []
                , Phosphor.speakerHigh Fill |> toHtml []
                , Phosphor.speakerHigh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerLow model.weight |> toHtml []
                , Phosphor.speakerLow Thin |> toHtml []
                , Phosphor.speakerLow Light |> toHtml []
                , Phosphor.speakerLow Regular |> toHtml []
                , Phosphor.speakerLow Bold |> toHtml []
                , Phosphor.speakerLow Fill |> toHtml []
                , Phosphor.speakerLow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerNone model.weight |> toHtml []
                , Phosphor.speakerNone Thin |> toHtml []
                , Phosphor.speakerNone Light |> toHtml []
                , Phosphor.speakerNone Regular |> toHtml []
                , Phosphor.speakerNone Bold |> toHtml []
                , Phosphor.speakerNone Fill |> toHtml []
                , Phosphor.speakerNone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerSimpleHigh model.weight |> toHtml []
                , Phosphor.speakerSimpleHigh Thin |> toHtml []
                , Phosphor.speakerSimpleHigh Light |> toHtml []
                , Phosphor.speakerSimpleHigh Regular |> toHtml []
                , Phosphor.speakerSimpleHigh Bold |> toHtml []
                , Phosphor.speakerSimpleHigh Fill |> toHtml []
                , Phosphor.speakerSimpleHigh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerSimpleLow model.weight |> toHtml []
                , Phosphor.speakerSimpleLow Thin |> toHtml []
                , Phosphor.speakerSimpleLow Light |> toHtml []
                , Phosphor.speakerSimpleLow Regular |> toHtml []
                , Phosphor.speakerSimpleLow Bold |> toHtml []
                , Phosphor.speakerSimpleLow Fill |> toHtml []
                , Phosphor.speakerSimpleLow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerSimpleNone model.weight |> toHtml []
                , Phosphor.speakerSimpleNone Thin |> toHtml []
                , Phosphor.speakerSimpleNone Light |> toHtml []
                , Phosphor.speakerSimpleNone Regular |> toHtml []
                , Phosphor.speakerSimpleNone Bold |> toHtml []
                , Phosphor.speakerSimpleNone Fill |> toHtml []
                , Phosphor.speakerSimpleNone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerSimpleSlash model.weight |> toHtml []
                , Phosphor.speakerSimpleSlash Thin |> toHtml []
                , Phosphor.speakerSimpleSlash Light |> toHtml []
                , Phosphor.speakerSimpleSlash Regular |> toHtml []
                , Phosphor.speakerSimpleSlash Bold |> toHtml []
                , Phosphor.speakerSimpleSlash Fill |> toHtml []
                , Phosphor.speakerSimpleSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerSimpleX model.weight |> toHtml []
                , Phosphor.speakerSimpleX Thin |> toHtml []
                , Phosphor.speakerSimpleX Light |> toHtml []
                , Phosphor.speakerSimpleX Regular |> toHtml []
                , Phosphor.speakerSimpleX Bold |> toHtml []
                , Phosphor.speakerSimpleX Fill |> toHtml []
                , Phosphor.speakerSimpleX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerSlash model.weight |> toHtml []
                , Phosphor.speakerSlash Thin |> toHtml []
                , Phosphor.speakerSlash Light |> toHtml []
                , Phosphor.speakerSlash Regular |> toHtml []
                , Phosphor.speakerSlash Bold |> toHtml []
                , Phosphor.speakerSlash Fill |> toHtml []
                , Phosphor.speakerSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.speakerX model.weight |> toHtml []
                , Phosphor.speakerX Thin |> toHtml []
                , Phosphor.speakerX Light |> toHtml []
                , Phosphor.speakerX Regular |> toHtml []
                , Phosphor.speakerX Bold |> toHtml []
                , Phosphor.speakerX Fill |> toHtml []
                , Phosphor.speakerX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.spinner model.weight |> toHtml []
                , Phosphor.spinner Thin |> toHtml []
                , Phosphor.spinner Light |> toHtml []
                , Phosphor.spinner Regular |> toHtml []
                , Phosphor.spinner Bold |> toHtml []
                , Phosphor.spinner Fill |> toHtml []
                , Phosphor.spinner Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.spinnerGap model.weight |> toHtml []
                , Phosphor.spinnerGap Thin |> toHtml []
                , Phosphor.spinnerGap Light |> toHtml []
                , Phosphor.spinnerGap Regular |> toHtml []
                , Phosphor.spinnerGap Bold |> toHtml []
                , Phosphor.spinnerGap Fill |> toHtml []
                , Phosphor.spinnerGap Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.spiral model.weight |> toHtml []
                , Phosphor.spiral Thin |> toHtml []
                , Phosphor.spiral Light |> toHtml []
                , Phosphor.spiral Regular |> toHtml []
                , Phosphor.spiral Bold |> toHtml []
                , Phosphor.spiral Fill |> toHtml []
                , Phosphor.spiral Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.splitHorizontal model.weight |> toHtml []
                , Phosphor.splitHorizontal Thin |> toHtml []
                , Phosphor.splitHorizontal Light |> toHtml []
                , Phosphor.splitHorizontal Regular |> toHtml []
                , Phosphor.splitHorizontal Bold |> toHtml []
                , Phosphor.splitHorizontal Fill |> toHtml []
                , Phosphor.splitHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.splitVertical model.weight |> toHtml []
                , Phosphor.splitVertical Thin |> toHtml []
                , Phosphor.splitVertical Light |> toHtml []
                , Phosphor.splitVertical Regular |> toHtml []
                , Phosphor.splitVertical Bold |> toHtml []
                , Phosphor.splitVertical Fill |> toHtml []
                , Phosphor.splitVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.spotifyLogo model.weight |> toHtml []
                , Phosphor.spotifyLogo Thin |> toHtml []
                , Phosphor.spotifyLogo Light |> toHtml []
                , Phosphor.spotifyLogo Regular |> toHtml []
                , Phosphor.spotifyLogo Bold |> toHtml []
                , Phosphor.spotifyLogo Fill |> toHtml []
                , Phosphor.spotifyLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.square model.weight |> toHtml []
                , Phosphor.square Thin |> toHtml []
                , Phosphor.square Light |> toHtml []
                , Phosphor.square Regular |> toHtml []
                , Phosphor.square Bold |> toHtml []
                , Phosphor.square Fill |> toHtml []
                , Phosphor.square Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.squareHalf model.weight |> toHtml []
                , Phosphor.squareHalf Thin |> toHtml []
                , Phosphor.squareHalf Light |> toHtml []
                , Phosphor.squareHalf Regular |> toHtml []
                , Phosphor.squareHalf Bold |> toHtml []
                , Phosphor.squareHalf Fill |> toHtml []
                , Phosphor.squareHalf Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.squareHalfBottom model.weight |> toHtml []
                , Phosphor.squareHalfBottom Thin |> toHtml []
                , Phosphor.squareHalfBottom Light |> toHtml []
                , Phosphor.squareHalfBottom Regular |> toHtml []
                , Phosphor.squareHalfBottom Bold |> toHtml []
                , Phosphor.squareHalfBottom Fill |> toHtml []
                , Phosphor.squareHalfBottom Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.squareLogo model.weight |> toHtml []
                , Phosphor.squareLogo Thin |> toHtml []
                , Phosphor.squareLogo Light |> toHtml []
                , Phosphor.squareLogo Regular |> toHtml []
                , Phosphor.squareLogo Bold |> toHtml []
                , Phosphor.squareLogo Fill |> toHtml []
                , Phosphor.squareLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.squareSplitHorizontal model.weight |> toHtml []
                , Phosphor.squareSplitHorizontal Thin |> toHtml []
                , Phosphor.squareSplitHorizontal Light |> toHtml []
                , Phosphor.squareSplitHorizontal Regular |> toHtml []
                , Phosphor.squareSplitHorizontal Bold |> toHtml []
                , Phosphor.squareSplitHorizontal Fill |> toHtml []
                , Phosphor.squareSplitHorizontal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.squareSplitVertical model.weight |> toHtml []
                , Phosphor.squareSplitVertical Thin |> toHtml []
                , Phosphor.squareSplitVertical Light |> toHtml []
                , Phosphor.squareSplitVertical Regular |> toHtml []
                , Phosphor.squareSplitVertical Bold |> toHtml []
                , Phosphor.squareSplitVertical Fill |> toHtml []
                , Phosphor.squareSplitVertical Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.squaresFour model.weight |> toHtml []
                , Phosphor.squaresFour Thin |> toHtml []
                , Phosphor.squaresFour Light |> toHtml []
                , Phosphor.squaresFour Regular |> toHtml []
                , Phosphor.squaresFour Bold |> toHtml []
                , Phosphor.squaresFour Fill |> toHtml []
                , Phosphor.squaresFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stack model.weight |> toHtml []
                , Phosphor.stack Thin |> toHtml []
                , Phosphor.stack Light |> toHtml []
                , Phosphor.stack Regular |> toHtml []
                , Phosphor.stack Bold |> toHtml []
                , Phosphor.stack Fill |> toHtml []
                , Phosphor.stack Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stackOverflowLogo model.weight |> toHtml []
                , Phosphor.stackOverflowLogo Thin |> toHtml []
                , Phosphor.stackOverflowLogo Light |> toHtml []
                , Phosphor.stackOverflowLogo Regular |> toHtml []
                , Phosphor.stackOverflowLogo Bold |> toHtml []
                , Phosphor.stackOverflowLogo Fill |> toHtml []
                , Phosphor.stackOverflowLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stackSimple model.weight |> toHtml []
                , Phosphor.stackSimple Thin |> toHtml []
                , Phosphor.stackSimple Light |> toHtml []
                , Phosphor.stackSimple Regular |> toHtml []
                , Phosphor.stackSimple Bold |> toHtml []
                , Phosphor.stackSimple Fill |> toHtml []
                , Phosphor.stackSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stairs model.weight |> toHtml []
                , Phosphor.stairs Thin |> toHtml []
                , Phosphor.stairs Light |> toHtml []
                , Phosphor.stairs Regular |> toHtml []
                , Phosphor.stairs Bold |> toHtml []
                , Phosphor.stairs Fill |> toHtml []
                , Phosphor.stairs Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stamp model.weight |> toHtml []
                , Phosphor.stamp Thin |> toHtml []
                , Phosphor.stamp Light |> toHtml []
                , Phosphor.stamp Regular |> toHtml []
                , Phosphor.stamp Bold |> toHtml []
                , Phosphor.stamp Fill |> toHtml []
                , Phosphor.stamp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.starAndCrescent model.weight |> toHtml []
                , Phosphor.starAndCrescent Thin |> toHtml []
                , Phosphor.starAndCrescent Light |> toHtml []
                , Phosphor.starAndCrescent Regular |> toHtml []
                , Phosphor.starAndCrescent Bold |> toHtml []
                , Phosphor.starAndCrescent Fill |> toHtml []
                , Phosphor.starAndCrescent Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.star model.weight |> toHtml []
                , Phosphor.star Thin |> toHtml []
                , Phosphor.star Light |> toHtml []
                , Phosphor.star Regular |> toHtml []
                , Phosphor.star Bold |> toHtml []
                , Phosphor.star Fill |> toHtml []
                , Phosphor.star Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.starFour model.weight |> toHtml []
                , Phosphor.starFour Thin |> toHtml []
                , Phosphor.starFour Light |> toHtml []
                , Phosphor.starFour Regular |> toHtml []
                , Phosphor.starFour Bold |> toHtml []
                , Phosphor.starFour Fill |> toHtml []
                , Phosphor.starFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.starHalf model.weight |> toHtml []
                , Phosphor.starHalf Thin |> toHtml []
                , Phosphor.starHalf Light |> toHtml []
                , Phosphor.starHalf Regular |> toHtml []
                , Phosphor.starHalf Bold |> toHtml []
                , Phosphor.starHalf Fill |> toHtml []
                , Phosphor.starHalf Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.starOfDavid model.weight |> toHtml []
                , Phosphor.starOfDavid Thin |> toHtml []
                , Phosphor.starOfDavid Light |> toHtml []
                , Phosphor.starOfDavid Regular |> toHtml []
                , Phosphor.starOfDavid Bold |> toHtml []
                , Phosphor.starOfDavid Fill |> toHtml []
                , Phosphor.starOfDavid Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.steeringWheel model.weight |> toHtml []
                , Phosphor.steeringWheel Thin |> toHtml []
                , Phosphor.steeringWheel Light |> toHtml []
                , Phosphor.steeringWheel Regular |> toHtml []
                , Phosphor.steeringWheel Bold |> toHtml []
                , Phosphor.steeringWheel Fill |> toHtml []
                , Phosphor.steeringWheel Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.steps model.weight |> toHtml []
                , Phosphor.steps Thin |> toHtml []
                , Phosphor.steps Light |> toHtml []
                , Phosphor.steps Regular |> toHtml []
                , Phosphor.steps Bold |> toHtml []
                , Phosphor.steps Fill |> toHtml []
                , Phosphor.steps Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stethoscope model.weight |> toHtml []
                , Phosphor.stethoscope Thin |> toHtml []
                , Phosphor.stethoscope Light |> toHtml []
                , Phosphor.stethoscope Regular |> toHtml []
                , Phosphor.stethoscope Bold |> toHtml []
                , Phosphor.stethoscope Fill |> toHtml []
                , Phosphor.stethoscope Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sticker model.weight |> toHtml []
                , Phosphor.sticker Thin |> toHtml []
                , Phosphor.sticker Light |> toHtml []
                , Phosphor.sticker Regular |> toHtml []
                , Phosphor.sticker Bold |> toHtml []
                , Phosphor.sticker Fill |> toHtml []
                , Phosphor.sticker Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stool model.weight |> toHtml []
                , Phosphor.stool Thin |> toHtml []
                , Phosphor.stool Light |> toHtml []
                , Phosphor.stool Regular |> toHtml []
                , Phosphor.stool Bold |> toHtml []
                , Phosphor.stool Fill |> toHtml []
                , Phosphor.stool Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stop model.weight |> toHtml []
                , Phosphor.stop Thin |> toHtml []
                , Phosphor.stop Light |> toHtml []
                , Phosphor.stop Regular |> toHtml []
                , Phosphor.stop Bold |> toHtml []
                , Phosphor.stop Fill |> toHtml []
                , Phosphor.stop Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stopCircle model.weight |> toHtml []
                , Phosphor.stopCircle Thin |> toHtml []
                , Phosphor.stopCircle Light |> toHtml []
                , Phosphor.stopCircle Regular |> toHtml []
                , Phosphor.stopCircle Bold |> toHtml []
                , Phosphor.stopCircle Fill |> toHtml []
                , Phosphor.stopCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.storefront model.weight |> toHtml []
                , Phosphor.storefront Thin |> toHtml []
                , Phosphor.storefront Light |> toHtml []
                , Phosphor.storefront Regular |> toHtml []
                , Phosphor.storefront Bold |> toHtml []
                , Phosphor.storefront Fill |> toHtml []
                , Phosphor.storefront Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.strategy model.weight |> toHtml []
                , Phosphor.strategy Thin |> toHtml []
                , Phosphor.strategy Light |> toHtml []
                , Phosphor.strategy Regular |> toHtml []
                , Phosphor.strategy Bold |> toHtml []
                , Phosphor.strategy Fill |> toHtml []
                , Phosphor.strategy Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.stripeLogo model.weight |> toHtml []
                , Phosphor.stripeLogo Thin |> toHtml []
                , Phosphor.stripeLogo Light |> toHtml []
                , Phosphor.stripeLogo Regular |> toHtml []
                , Phosphor.stripeLogo Bold |> toHtml []
                , Phosphor.stripeLogo Fill |> toHtml []
                , Phosphor.stripeLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.student model.weight |> toHtml []
                , Phosphor.student Thin |> toHtml []
                , Phosphor.student Light |> toHtml []
                , Phosphor.student Regular |> toHtml []
                , Phosphor.student Bold |> toHtml []
                , Phosphor.student Fill |> toHtml []
                , Phosphor.student Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.subtitles model.weight |> toHtml []
                , Phosphor.subtitles Thin |> toHtml []
                , Phosphor.subtitles Light |> toHtml []
                , Phosphor.subtitles Regular |> toHtml []
                , Phosphor.subtitles Bold |> toHtml []
                , Phosphor.subtitles Fill |> toHtml []
                , Phosphor.subtitles Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.subtract model.weight |> toHtml []
                , Phosphor.subtract Thin |> toHtml []
                , Phosphor.subtract Light |> toHtml []
                , Phosphor.subtract Regular |> toHtml []
                , Phosphor.subtract Bold |> toHtml []
                , Phosphor.subtract Fill |> toHtml []
                , Phosphor.subtract Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.subtractSquare model.weight |> toHtml []
                , Phosphor.subtractSquare Thin |> toHtml []
                , Phosphor.subtractSquare Light |> toHtml []
                , Phosphor.subtractSquare Regular |> toHtml []
                , Phosphor.subtractSquare Bold |> toHtml []
                , Phosphor.subtractSquare Fill |> toHtml []
                , Phosphor.subtractSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.suitcase model.weight |> toHtml []
                , Phosphor.suitcase Thin |> toHtml []
                , Phosphor.suitcase Light |> toHtml []
                , Phosphor.suitcase Regular |> toHtml []
                , Phosphor.suitcase Bold |> toHtml []
                , Phosphor.suitcase Fill |> toHtml []
                , Phosphor.suitcase Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.suitcaseRolling model.weight |> toHtml []
                , Phosphor.suitcaseRolling Thin |> toHtml []
                , Phosphor.suitcaseRolling Light |> toHtml []
                , Phosphor.suitcaseRolling Regular |> toHtml []
                , Phosphor.suitcaseRolling Bold |> toHtml []
                , Phosphor.suitcaseRolling Fill |> toHtml []
                , Phosphor.suitcaseRolling Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.suitcaseSimple model.weight |> toHtml []
                , Phosphor.suitcaseSimple Thin |> toHtml []
                , Phosphor.suitcaseSimple Light |> toHtml []
                , Phosphor.suitcaseSimple Regular |> toHtml []
                , Phosphor.suitcaseSimple Bold |> toHtml []
                , Phosphor.suitcaseSimple Fill |> toHtml []
                , Phosphor.suitcaseSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sun model.weight |> toHtml []
                , Phosphor.sun Thin |> toHtml []
                , Phosphor.sun Light |> toHtml []
                , Phosphor.sun Regular |> toHtml []
                , Phosphor.sun Bold |> toHtml []
                , Phosphor.sun Fill |> toHtml []
                , Phosphor.sun Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sunDim model.weight |> toHtml []
                , Phosphor.sunDim Thin |> toHtml []
                , Phosphor.sunDim Light |> toHtml []
                , Phosphor.sunDim Regular |> toHtml []
                , Phosphor.sunDim Bold |> toHtml []
                , Phosphor.sunDim Fill |> toHtml []
                , Phosphor.sunDim Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sunHorizon model.weight |> toHtml []
                , Phosphor.sunHorizon Thin |> toHtml []
                , Phosphor.sunHorizon Light |> toHtml []
                , Phosphor.sunHorizon Regular |> toHtml []
                , Phosphor.sunHorizon Bold |> toHtml []
                , Phosphor.sunHorizon Fill |> toHtml []
                , Phosphor.sunHorizon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sunglasses model.weight |> toHtml []
                , Phosphor.sunglasses Thin |> toHtml []
                , Phosphor.sunglasses Light |> toHtml []
                , Phosphor.sunglasses Regular |> toHtml []
                , Phosphor.sunglasses Bold |> toHtml []
                , Phosphor.sunglasses Fill |> toHtml []
                , Phosphor.sunglasses Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.swap model.weight |> toHtml []
                , Phosphor.swap Thin |> toHtml []
                , Phosphor.swap Light |> toHtml []
                , Phosphor.swap Regular |> toHtml []
                , Phosphor.swap Bold |> toHtml []
                , Phosphor.swap Fill |> toHtml []
                , Phosphor.swap Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.swatches model.weight |> toHtml []
                , Phosphor.swatches Thin |> toHtml []
                , Phosphor.swatches Light |> toHtml []
                , Phosphor.swatches Regular |> toHtml []
                , Phosphor.swatches Bold |> toHtml []
                , Phosphor.swatches Fill |> toHtml []
                , Phosphor.swatches Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.swimmingPool model.weight |> toHtml []
                , Phosphor.swimmingPool Thin |> toHtml []
                , Phosphor.swimmingPool Light |> toHtml []
                , Phosphor.swimmingPool Regular |> toHtml []
                , Phosphor.swimmingPool Bold |> toHtml []
                , Phosphor.swimmingPool Fill |> toHtml []
                , Phosphor.swimmingPool Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.sword model.weight |> toHtml []
                , Phosphor.sword Thin |> toHtml []
                , Phosphor.sword Light |> toHtml []
                , Phosphor.sword Regular |> toHtml []
                , Phosphor.sword Bold |> toHtml []
                , Phosphor.sword Fill |> toHtml []
                , Phosphor.sword Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.synagogue model.weight |> toHtml []
                , Phosphor.synagogue Thin |> toHtml []
                , Phosphor.synagogue Light |> toHtml []
                , Phosphor.synagogue Regular |> toHtml []
                , Phosphor.synagogue Bold |> toHtml []
                , Phosphor.synagogue Fill |> toHtml []
                , Phosphor.synagogue Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.syringe model.weight |> toHtml []
                , Phosphor.syringe Thin |> toHtml []
                , Phosphor.syringe Light |> toHtml []
                , Phosphor.syringe Regular |> toHtml []
                , Phosphor.syringe Bold |> toHtml []
                , Phosphor.syringe Fill |> toHtml []
                , Phosphor.syringe Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tShirt model.weight |> toHtml []
                , Phosphor.tShirt Thin |> toHtml []
                , Phosphor.tShirt Light |> toHtml []
                , Phosphor.tShirt Regular |> toHtml []
                , Phosphor.tShirt Bold |> toHtml []
                , Phosphor.tShirt Fill |> toHtml []
                , Phosphor.tShirt Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.table model.weight |> toHtml []
                , Phosphor.table Thin |> toHtml []
                , Phosphor.table Light |> toHtml []
                , Phosphor.table Regular |> toHtml []
                , Phosphor.table Bold |> toHtml []
                , Phosphor.table Fill |> toHtml []
                , Phosphor.table Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tabs model.weight |> toHtml []
                , Phosphor.tabs Thin |> toHtml []
                , Phosphor.tabs Light |> toHtml []
                , Phosphor.tabs Regular |> toHtml []
                , Phosphor.tabs Bold |> toHtml []
                , Phosphor.tabs Fill |> toHtml []
                , Phosphor.tabs Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tag model.weight |> toHtml []
                , Phosphor.tag Thin |> toHtml []
                , Phosphor.tag Light |> toHtml []
                , Phosphor.tag Regular |> toHtml []
                , Phosphor.tag Bold |> toHtml []
                , Phosphor.tag Fill |> toHtml []
                , Phosphor.tag Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tagChevron model.weight |> toHtml []
                , Phosphor.tagChevron Thin |> toHtml []
                , Phosphor.tagChevron Light |> toHtml []
                , Phosphor.tagChevron Regular |> toHtml []
                , Phosphor.tagChevron Bold |> toHtml []
                , Phosphor.tagChevron Fill |> toHtml []
                , Phosphor.tagChevron Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tagSimple model.weight |> toHtml []
                , Phosphor.tagSimple Thin |> toHtml []
                , Phosphor.tagSimple Light |> toHtml []
                , Phosphor.tagSimple Regular |> toHtml []
                , Phosphor.tagSimple Bold |> toHtml []
                , Phosphor.tagSimple Fill |> toHtml []
                , Phosphor.tagSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.target model.weight |> toHtml []
                , Phosphor.target Thin |> toHtml []
                , Phosphor.target Light |> toHtml []
                , Phosphor.target Regular |> toHtml []
                , Phosphor.target Bold |> toHtml []
                , Phosphor.target Fill |> toHtml []
                , Phosphor.target Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.taxi model.weight |> toHtml []
                , Phosphor.taxi Thin |> toHtml []
                , Phosphor.taxi Light |> toHtml []
                , Phosphor.taxi Regular |> toHtml []
                , Phosphor.taxi Bold |> toHtml []
                , Phosphor.taxi Fill |> toHtml []
                , Phosphor.taxi Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.telegramLogo model.weight |> toHtml []
                , Phosphor.telegramLogo Thin |> toHtml []
                , Phosphor.telegramLogo Light |> toHtml []
                , Phosphor.telegramLogo Regular |> toHtml []
                , Phosphor.telegramLogo Bold |> toHtml []
                , Phosphor.telegramLogo Fill |> toHtml []
                , Phosphor.telegramLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.television model.weight |> toHtml []
                , Phosphor.television Thin |> toHtml []
                , Phosphor.television Light |> toHtml []
                , Phosphor.television Regular |> toHtml []
                , Phosphor.television Bold |> toHtml []
                , Phosphor.television Fill |> toHtml []
                , Phosphor.television Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.televisionSimple model.weight |> toHtml []
                , Phosphor.televisionSimple Thin |> toHtml []
                , Phosphor.televisionSimple Light |> toHtml []
                , Phosphor.televisionSimple Regular |> toHtml []
                , Phosphor.televisionSimple Bold |> toHtml []
                , Phosphor.televisionSimple Fill |> toHtml []
                , Phosphor.televisionSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tennisBall model.weight |> toHtml []
                , Phosphor.tennisBall Thin |> toHtml []
                , Phosphor.tennisBall Light |> toHtml []
                , Phosphor.tennisBall Regular |> toHtml []
                , Phosphor.tennisBall Bold |> toHtml []
                , Phosphor.tennisBall Fill |> toHtml []
                , Phosphor.tennisBall Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tent model.weight |> toHtml []
                , Phosphor.tent Thin |> toHtml []
                , Phosphor.tent Light |> toHtml []
                , Phosphor.tent Regular |> toHtml []
                , Phosphor.tent Bold |> toHtml []
                , Phosphor.tent Fill |> toHtml []
                , Phosphor.tent Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.terminal model.weight |> toHtml []
                , Phosphor.terminal Thin |> toHtml []
                , Phosphor.terminal Light |> toHtml []
                , Phosphor.terminal Regular |> toHtml []
                , Phosphor.terminal Bold |> toHtml []
                , Phosphor.terminal Fill |> toHtml []
                , Phosphor.terminal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.terminalWindow model.weight |> toHtml []
                , Phosphor.terminalWindow Thin |> toHtml []
                , Phosphor.terminalWindow Light |> toHtml []
                , Phosphor.terminalWindow Regular |> toHtml []
                , Phosphor.terminalWindow Bold |> toHtml []
                , Phosphor.terminalWindow Fill |> toHtml []
                , Phosphor.terminalWindow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.testTube model.weight |> toHtml []
                , Phosphor.testTube Thin |> toHtml []
                , Phosphor.testTube Light |> toHtml []
                , Phosphor.testTube Regular |> toHtml []
                , Phosphor.testTube Bold |> toHtml []
                , Phosphor.testTube Fill |> toHtml []
                , Phosphor.testTube Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textAUnderline model.weight |> toHtml []
                , Phosphor.textAUnderline Thin |> toHtml []
                , Phosphor.textAUnderline Light |> toHtml []
                , Phosphor.textAUnderline Regular |> toHtml []
                , Phosphor.textAUnderline Bold |> toHtml []
                , Phosphor.textAUnderline Fill |> toHtml []
                , Phosphor.textAUnderline Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textAa model.weight |> toHtml []
                , Phosphor.textAa Thin |> toHtml []
                , Phosphor.textAa Light |> toHtml []
                , Phosphor.textAa Regular |> toHtml []
                , Phosphor.textAa Bold |> toHtml []
                , Phosphor.textAa Fill |> toHtml []
                , Phosphor.textAa Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textAlignCenter model.weight |> toHtml []
                , Phosphor.textAlignCenter Thin |> toHtml []
                , Phosphor.textAlignCenter Light |> toHtml []
                , Phosphor.textAlignCenter Regular |> toHtml []
                , Phosphor.textAlignCenter Bold |> toHtml []
                , Phosphor.textAlignCenter Fill |> toHtml []
                , Phosphor.textAlignCenter Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textAlignJustify model.weight |> toHtml []
                , Phosphor.textAlignJustify Thin |> toHtml []
                , Phosphor.textAlignJustify Light |> toHtml []
                , Phosphor.textAlignJustify Regular |> toHtml []
                , Phosphor.textAlignJustify Bold |> toHtml []
                , Phosphor.textAlignJustify Fill |> toHtml []
                , Phosphor.textAlignJustify Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textAlignLeft model.weight |> toHtml []
                , Phosphor.textAlignLeft Thin |> toHtml []
                , Phosphor.textAlignLeft Light |> toHtml []
                , Phosphor.textAlignLeft Regular |> toHtml []
                , Phosphor.textAlignLeft Bold |> toHtml []
                , Phosphor.textAlignLeft Fill |> toHtml []
                , Phosphor.textAlignLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textAlignRight model.weight |> toHtml []
                , Phosphor.textAlignRight Thin |> toHtml []
                , Phosphor.textAlignRight Light |> toHtml []
                , Phosphor.textAlignRight Regular |> toHtml []
                , Phosphor.textAlignRight Bold |> toHtml []
                , Phosphor.textAlignRight Fill |> toHtml []
                , Phosphor.textAlignRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textB model.weight |> toHtml []
                , Phosphor.textB Thin |> toHtml []
                , Phosphor.textB Light |> toHtml []
                , Phosphor.textB Regular |> toHtml []
                , Phosphor.textB Bold |> toHtml []
                , Phosphor.textB Fill |> toHtml []
                , Phosphor.textB Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textColumns model.weight |> toHtml []
                , Phosphor.textColumns Thin |> toHtml []
                , Phosphor.textColumns Light |> toHtml []
                , Phosphor.textColumns Regular |> toHtml []
                , Phosphor.textColumns Bold |> toHtml []
                , Phosphor.textColumns Fill |> toHtml []
                , Phosphor.textColumns Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textH model.weight |> toHtml []
                , Phosphor.textH Thin |> toHtml []
                , Phosphor.textH Light |> toHtml []
                , Phosphor.textH Regular |> toHtml []
                , Phosphor.textH Bold |> toHtml []
                , Phosphor.textH Fill |> toHtml []
                , Phosphor.textH Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textHFive model.weight |> toHtml []
                , Phosphor.textHFive Thin |> toHtml []
                , Phosphor.textHFive Light |> toHtml []
                , Phosphor.textHFive Regular |> toHtml []
                , Phosphor.textHFive Bold |> toHtml []
                , Phosphor.textHFive Fill |> toHtml []
                , Phosphor.textHFive Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textHFour model.weight |> toHtml []
                , Phosphor.textHFour Thin |> toHtml []
                , Phosphor.textHFour Light |> toHtml []
                , Phosphor.textHFour Regular |> toHtml []
                , Phosphor.textHFour Bold |> toHtml []
                , Phosphor.textHFour Fill |> toHtml []
                , Phosphor.textHFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textHOne model.weight |> toHtml []
                , Phosphor.textHOne Thin |> toHtml []
                , Phosphor.textHOne Light |> toHtml []
                , Phosphor.textHOne Regular |> toHtml []
                , Phosphor.textHOne Bold |> toHtml []
                , Phosphor.textHOne Fill |> toHtml []
                , Phosphor.textHOne Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textHSix model.weight |> toHtml []
                , Phosphor.textHSix Thin |> toHtml []
                , Phosphor.textHSix Light |> toHtml []
                , Phosphor.textHSix Regular |> toHtml []
                , Phosphor.textHSix Bold |> toHtml []
                , Phosphor.textHSix Fill |> toHtml []
                , Phosphor.textHSix Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textHThree model.weight |> toHtml []
                , Phosphor.textHThree Thin |> toHtml []
                , Phosphor.textHThree Light |> toHtml []
                , Phosphor.textHThree Regular |> toHtml []
                , Phosphor.textHThree Bold |> toHtml []
                , Phosphor.textHThree Fill |> toHtml []
                , Phosphor.textHThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textHTwo model.weight |> toHtml []
                , Phosphor.textHTwo Thin |> toHtml []
                , Phosphor.textHTwo Light |> toHtml []
                , Phosphor.textHTwo Regular |> toHtml []
                , Phosphor.textHTwo Bold |> toHtml []
                , Phosphor.textHTwo Fill |> toHtml []
                , Phosphor.textHTwo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textIndent model.weight |> toHtml []
                , Phosphor.textIndent Thin |> toHtml []
                , Phosphor.textIndent Light |> toHtml []
                , Phosphor.textIndent Regular |> toHtml []
                , Phosphor.textIndent Bold |> toHtml []
                , Phosphor.textIndent Fill |> toHtml []
                , Phosphor.textIndent Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textItalic model.weight |> toHtml []
                , Phosphor.textItalic Thin |> toHtml []
                , Phosphor.textItalic Light |> toHtml []
                , Phosphor.textItalic Regular |> toHtml []
                , Phosphor.textItalic Bold |> toHtml []
                , Phosphor.textItalic Fill |> toHtml []
                , Phosphor.textItalic Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textOutdent model.weight |> toHtml []
                , Phosphor.textOutdent Thin |> toHtml []
                , Phosphor.textOutdent Light |> toHtml []
                , Phosphor.textOutdent Regular |> toHtml []
                , Phosphor.textOutdent Bold |> toHtml []
                , Phosphor.textOutdent Fill |> toHtml []
                , Phosphor.textOutdent Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textStrikethrough model.weight |> toHtml []
                , Phosphor.textStrikethrough Thin |> toHtml []
                , Phosphor.textStrikethrough Light |> toHtml []
                , Phosphor.textStrikethrough Regular |> toHtml []
                , Phosphor.textStrikethrough Bold |> toHtml []
                , Phosphor.textStrikethrough Fill |> toHtml []
                , Phosphor.textStrikethrough Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textT model.weight |> toHtml []
                , Phosphor.textT Thin |> toHtml []
                , Phosphor.textT Light |> toHtml []
                , Phosphor.textT Regular |> toHtml []
                , Phosphor.textT Bold |> toHtml []
                , Phosphor.textT Fill |> toHtml []
                , Phosphor.textT Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textUnderline model.weight |> toHtml []
                , Phosphor.textUnderline Thin |> toHtml []
                , Phosphor.textUnderline Light |> toHtml []
                , Phosphor.textUnderline Regular |> toHtml []
                , Phosphor.textUnderline Bold |> toHtml []
                , Phosphor.textUnderline Fill |> toHtml []
                , Phosphor.textUnderline Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.textbox model.weight |> toHtml []
                , Phosphor.textbox Thin |> toHtml []
                , Phosphor.textbox Light |> toHtml []
                , Phosphor.textbox Regular |> toHtml []
                , Phosphor.textbox Bold |> toHtml []
                , Phosphor.textbox Fill |> toHtml []
                , Phosphor.textbox Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.thermometer model.weight |> toHtml []
                , Phosphor.thermometer Thin |> toHtml []
                , Phosphor.thermometer Light |> toHtml []
                , Phosphor.thermometer Regular |> toHtml []
                , Phosphor.thermometer Bold |> toHtml []
                , Phosphor.thermometer Fill |> toHtml []
                , Phosphor.thermometer Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.thermometerCold model.weight |> toHtml []
                , Phosphor.thermometerCold Thin |> toHtml []
                , Phosphor.thermometerCold Light |> toHtml []
                , Phosphor.thermometerCold Regular |> toHtml []
                , Phosphor.thermometerCold Bold |> toHtml []
                , Phosphor.thermometerCold Fill |> toHtml []
                , Phosphor.thermometerCold Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.thermometerHot model.weight |> toHtml []
                , Phosphor.thermometerHot Thin |> toHtml []
                , Phosphor.thermometerHot Light |> toHtml []
                , Phosphor.thermometerHot Regular |> toHtml []
                , Phosphor.thermometerHot Bold |> toHtml []
                , Phosphor.thermometerHot Fill |> toHtml []
                , Phosphor.thermometerHot Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.thermometerSimple model.weight |> toHtml []
                , Phosphor.thermometerSimple Thin |> toHtml []
                , Phosphor.thermometerSimple Light |> toHtml []
                , Phosphor.thermometerSimple Regular |> toHtml []
                , Phosphor.thermometerSimple Bold |> toHtml []
                , Phosphor.thermometerSimple Fill |> toHtml []
                , Phosphor.thermometerSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.thumbsDown model.weight |> toHtml []
                , Phosphor.thumbsDown Thin |> toHtml []
                , Phosphor.thumbsDown Light |> toHtml []
                , Phosphor.thumbsDown Regular |> toHtml []
                , Phosphor.thumbsDown Bold |> toHtml []
                , Phosphor.thumbsDown Fill |> toHtml []
                , Phosphor.thumbsDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.thumbsUp model.weight |> toHtml []
                , Phosphor.thumbsUp Thin |> toHtml []
                , Phosphor.thumbsUp Light |> toHtml []
                , Phosphor.thumbsUp Regular |> toHtml []
                , Phosphor.thumbsUp Bold |> toHtml []
                , Phosphor.thumbsUp Fill |> toHtml []
                , Phosphor.thumbsUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.ticket model.weight |> toHtml []
                , Phosphor.ticket Thin |> toHtml []
                , Phosphor.ticket Light |> toHtml []
                , Phosphor.ticket Regular |> toHtml []
                , Phosphor.ticket Bold |> toHtml []
                , Phosphor.ticket Fill |> toHtml []
                , Phosphor.ticket Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tidalLogo model.weight |> toHtml []
                , Phosphor.tidalLogo Thin |> toHtml []
                , Phosphor.tidalLogo Light |> toHtml []
                , Phosphor.tidalLogo Regular |> toHtml []
                , Phosphor.tidalLogo Bold |> toHtml []
                , Phosphor.tidalLogo Fill |> toHtml []
                , Phosphor.tidalLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tiktokLogo model.weight |> toHtml []
                , Phosphor.tiktokLogo Thin |> toHtml []
                , Phosphor.tiktokLogo Light |> toHtml []
                , Phosphor.tiktokLogo Regular |> toHtml []
                , Phosphor.tiktokLogo Bold |> toHtml []
                , Phosphor.tiktokLogo Fill |> toHtml []
                , Phosphor.tiktokLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.timer model.weight |> toHtml []
                , Phosphor.timer Thin |> toHtml []
                , Phosphor.timer Light |> toHtml []
                , Phosphor.timer Regular |> toHtml []
                , Phosphor.timer Bold |> toHtml []
                , Phosphor.timer Fill |> toHtml []
                , Phosphor.timer Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tipi model.weight |> toHtml []
                , Phosphor.tipi Thin |> toHtml []
                , Phosphor.tipi Light |> toHtml []
                , Phosphor.tipi Regular |> toHtml []
                , Phosphor.tipi Bold |> toHtml []
                , Phosphor.tipi Fill |> toHtml []
                , Phosphor.tipi Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.toggleLeft model.weight |> toHtml []
                , Phosphor.toggleLeft Thin |> toHtml []
                , Phosphor.toggleLeft Light |> toHtml []
                , Phosphor.toggleLeft Regular |> toHtml []
                , Phosphor.toggleLeft Bold |> toHtml []
                , Phosphor.toggleLeft Fill |> toHtml []
                , Phosphor.toggleLeft Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.toggleRight model.weight |> toHtml []
                , Phosphor.toggleRight Thin |> toHtml []
                , Phosphor.toggleRight Light |> toHtml []
                , Phosphor.toggleRight Regular |> toHtml []
                , Phosphor.toggleRight Bold |> toHtml []
                , Phosphor.toggleRight Fill |> toHtml []
                , Phosphor.toggleRight Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.toilet model.weight |> toHtml []
                , Phosphor.toilet Thin |> toHtml []
                , Phosphor.toilet Light |> toHtml []
                , Phosphor.toilet Regular |> toHtml []
                , Phosphor.toilet Bold |> toHtml []
                , Phosphor.toilet Fill |> toHtml []
                , Phosphor.toilet Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.toiletPaper model.weight |> toHtml []
                , Phosphor.toiletPaper Thin |> toHtml []
                , Phosphor.toiletPaper Light |> toHtml []
                , Phosphor.toiletPaper Regular |> toHtml []
                , Phosphor.toiletPaper Bold |> toHtml []
                , Phosphor.toiletPaper Fill |> toHtml []
                , Phosphor.toiletPaper Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.toolbox model.weight |> toHtml []
                , Phosphor.toolbox Thin |> toHtml []
                , Phosphor.toolbox Light |> toHtml []
                , Phosphor.toolbox Regular |> toHtml []
                , Phosphor.toolbox Bold |> toHtml []
                , Phosphor.toolbox Fill |> toHtml []
                , Phosphor.toolbox Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tooth model.weight |> toHtml []
                , Phosphor.tooth Thin |> toHtml []
                , Phosphor.tooth Light |> toHtml []
                , Phosphor.tooth Regular |> toHtml []
                , Phosphor.tooth Bold |> toHtml []
                , Phosphor.tooth Fill |> toHtml []
                , Phosphor.tooth Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tote model.weight |> toHtml []
                , Phosphor.tote Thin |> toHtml []
                , Phosphor.tote Light |> toHtml []
                , Phosphor.tote Regular |> toHtml []
                , Phosphor.tote Bold |> toHtml []
                , Phosphor.tote Fill |> toHtml []
                , Phosphor.tote Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.toteSimple model.weight |> toHtml []
                , Phosphor.toteSimple Thin |> toHtml []
                , Phosphor.toteSimple Light |> toHtml []
                , Phosphor.toteSimple Regular |> toHtml []
                , Phosphor.toteSimple Bold |> toHtml []
                , Phosphor.toteSimple Fill |> toHtml []
                , Phosphor.toteSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trademark model.weight |> toHtml []
                , Phosphor.trademark Thin |> toHtml []
                , Phosphor.trademark Light |> toHtml []
                , Phosphor.trademark Regular |> toHtml []
                , Phosphor.trademark Bold |> toHtml []
                , Phosphor.trademark Fill |> toHtml []
                , Phosphor.trademark Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trademarkRegistered model.weight |> toHtml []
                , Phosphor.trademarkRegistered Thin |> toHtml []
                , Phosphor.trademarkRegistered Light |> toHtml []
                , Phosphor.trademarkRegistered Regular |> toHtml []
                , Phosphor.trademarkRegistered Bold |> toHtml []
                , Phosphor.trademarkRegistered Fill |> toHtml []
                , Phosphor.trademarkRegistered Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trafficCone model.weight |> toHtml []
                , Phosphor.trafficCone Thin |> toHtml []
                , Phosphor.trafficCone Light |> toHtml []
                , Phosphor.trafficCone Regular |> toHtml []
                , Phosphor.trafficCone Bold |> toHtml []
                , Phosphor.trafficCone Fill |> toHtml []
                , Phosphor.trafficCone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trafficSign model.weight |> toHtml []
                , Phosphor.trafficSign Thin |> toHtml []
                , Phosphor.trafficSign Light |> toHtml []
                , Phosphor.trafficSign Regular |> toHtml []
                , Phosphor.trafficSign Bold |> toHtml []
                , Phosphor.trafficSign Fill |> toHtml []
                , Phosphor.trafficSign Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trafficSignal model.weight |> toHtml []
                , Phosphor.trafficSignal Thin |> toHtml []
                , Phosphor.trafficSignal Light |> toHtml []
                , Phosphor.trafficSignal Regular |> toHtml []
                , Phosphor.trafficSignal Bold |> toHtml []
                , Phosphor.trafficSignal Fill |> toHtml []
                , Phosphor.trafficSignal Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.train model.weight |> toHtml []
                , Phosphor.train Thin |> toHtml []
                , Phosphor.train Light |> toHtml []
                , Phosphor.train Regular |> toHtml []
                , Phosphor.train Bold |> toHtml []
                , Phosphor.train Fill |> toHtml []
                , Phosphor.train Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trainRegional model.weight |> toHtml []
                , Phosphor.trainRegional Thin |> toHtml []
                , Phosphor.trainRegional Light |> toHtml []
                , Phosphor.trainRegional Regular |> toHtml []
                , Phosphor.trainRegional Bold |> toHtml []
                , Phosphor.trainRegional Fill |> toHtml []
                , Phosphor.trainRegional Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trainSimple model.weight |> toHtml []
                , Phosphor.trainSimple Thin |> toHtml []
                , Phosphor.trainSimple Light |> toHtml []
                , Phosphor.trainSimple Regular |> toHtml []
                , Phosphor.trainSimple Bold |> toHtml []
                , Phosphor.trainSimple Fill |> toHtml []
                , Phosphor.trainSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tram model.weight |> toHtml []
                , Phosphor.tram Thin |> toHtml []
                , Phosphor.tram Light |> toHtml []
                , Phosphor.tram Regular |> toHtml []
                , Phosphor.tram Bold |> toHtml []
                , Phosphor.tram Fill |> toHtml []
                , Phosphor.tram Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.translate model.weight |> toHtml []
                , Phosphor.translate Thin |> toHtml []
                , Phosphor.translate Light |> toHtml []
                , Phosphor.translate Regular |> toHtml []
                , Phosphor.translate Bold |> toHtml []
                , Phosphor.translate Fill |> toHtml []
                , Phosphor.translate Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trash model.weight |> toHtml []
                , Phosphor.trash Thin |> toHtml []
                , Phosphor.trash Light |> toHtml []
                , Phosphor.trash Regular |> toHtml []
                , Phosphor.trash Bold |> toHtml []
                , Phosphor.trash Fill |> toHtml []
                , Phosphor.trash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trashSimple model.weight |> toHtml []
                , Phosphor.trashSimple Thin |> toHtml []
                , Phosphor.trashSimple Light |> toHtml []
                , Phosphor.trashSimple Regular |> toHtml []
                , Phosphor.trashSimple Bold |> toHtml []
                , Phosphor.trashSimple Fill |> toHtml []
                , Phosphor.trashSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tray model.weight |> toHtml []
                , Phosphor.tray Thin |> toHtml []
                , Phosphor.tray Light |> toHtml []
                , Phosphor.tray Regular |> toHtml []
                , Phosphor.tray Bold |> toHtml []
                , Phosphor.tray Fill |> toHtml []
                , Phosphor.tray Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.tree model.weight |> toHtml []
                , Phosphor.tree Thin |> toHtml []
                , Phosphor.tree Light |> toHtml []
                , Phosphor.tree Regular |> toHtml []
                , Phosphor.tree Bold |> toHtml []
                , Phosphor.tree Fill |> toHtml []
                , Phosphor.tree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.treeEvergreen model.weight |> toHtml []
                , Phosphor.treeEvergreen Thin |> toHtml []
                , Phosphor.treeEvergreen Light |> toHtml []
                , Phosphor.treeEvergreen Regular |> toHtml []
                , Phosphor.treeEvergreen Bold |> toHtml []
                , Phosphor.treeEvergreen Fill |> toHtml []
                , Phosphor.treeEvergreen Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.treePalm model.weight |> toHtml []
                , Phosphor.treePalm Thin |> toHtml []
                , Phosphor.treePalm Light |> toHtml []
                , Phosphor.treePalm Regular |> toHtml []
                , Phosphor.treePalm Bold |> toHtml []
                , Phosphor.treePalm Fill |> toHtml []
                , Phosphor.treePalm Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.treeStructure model.weight |> toHtml []
                , Phosphor.treeStructure Thin |> toHtml []
                , Phosphor.treeStructure Light |> toHtml []
                , Phosphor.treeStructure Regular |> toHtml []
                , Phosphor.treeStructure Bold |> toHtml []
                , Phosphor.treeStructure Fill |> toHtml []
                , Phosphor.treeStructure Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trendDown model.weight |> toHtml []
                , Phosphor.trendDown Thin |> toHtml []
                , Phosphor.trendDown Light |> toHtml []
                , Phosphor.trendDown Regular |> toHtml []
                , Phosphor.trendDown Bold |> toHtml []
                , Phosphor.trendDown Fill |> toHtml []
                , Phosphor.trendDown Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trendUp model.weight |> toHtml []
                , Phosphor.trendUp Thin |> toHtml []
                , Phosphor.trendUp Light |> toHtml []
                , Phosphor.trendUp Regular |> toHtml []
                , Phosphor.trendUp Bold |> toHtml []
                , Phosphor.trendUp Fill |> toHtml []
                , Phosphor.trendUp Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.triangle model.weight |> toHtml []
                , Phosphor.triangle Thin |> toHtml []
                , Phosphor.triangle Light |> toHtml []
                , Phosphor.triangle Regular |> toHtml []
                , Phosphor.triangle Bold |> toHtml []
                , Phosphor.triangle Fill |> toHtml []
                , Phosphor.triangle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.trophy model.weight |> toHtml []
                , Phosphor.trophy Thin |> toHtml []
                , Phosphor.trophy Light |> toHtml []
                , Phosphor.trophy Regular |> toHtml []
                , Phosphor.trophy Bold |> toHtml []
                , Phosphor.trophy Fill |> toHtml []
                , Phosphor.trophy Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.truck model.weight |> toHtml []
                , Phosphor.truck Thin |> toHtml []
                , Phosphor.truck Light |> toHtml []
                , Phosphor.truck Regular |> toHtml []
                , Phosphor.truck Bold |> toHtml []
                , Phosphor.truck Fill |> toHtml []
                , Phosphor.truck Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.twitchLogo model.weight |> toHtml []
                , Phosphor.twitchLogo Thin |> toHtml []
                , Phosphor.twitchLogo Light |> toHtml []
                , Phosphor.twitchLogo Regular |> toHtml []
                , Phosphor.twitchLogo Bold |> toHtml []
                , Phosphor.twitchLogo Fill |> toHtml []
                , Phosphor.twitchLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.twitterLogo model.weight |> toHtml []
                , Phosphor.twitterLogo Thin |> toHtml []
                , Phosphor.twitterLogo Light |> toHtml []
                , Phosphor.twitterLogo Regular |> toHtml []
                , Phosphor.twitterLogo Bold |> toHtml []
                , Phosphor.twitterLogo Fill |> toHtml []
                , Phosphor.twitterLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.umbrella model.weight |> toHtml []
                , Phosphor.umbrella Thin |> toHtml []
                , Phosphor.umbrella Light |> toHtml []
                , Phosphor.umbrella Regular |> toHtml []
                , Phosphor.umbrella Bold |> toHtml []
                , Phosphor.umbrella Fill |> toHtml []
                , Phosphor.umbrella Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.umbrellaSimple model.weight |> toHtml []
                , Phosphor.umbrellaSimple Thin |> toHtml []
                , Phosphor.umbrellaSimple Light |> toHtml []
                , Phosphor.umbrellaSimple Regular |> toHtml []
                , Phosphor.umbrellaSimple Bold |> toHtml []
                , Phosphor.umbrellaSimple Fill |> toHtml []
                , Phosphor.umbrellaSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.unite model.weight |> toHtml []
                , Phosphor.unite Thin |> toHtml []
                , Phosphor.unite Light |> toHtml []
                , Phosphor.unite Regular |> toHtml []
                , Phosphor.unite Bold |> toHtml []
                , Phosphor.unite Fill |> toHtml []
                , Phosphor.unite Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.uniteSquare model.weight |> toHtml []
                , Phosphor.uniteSquare Thin |> toHtml []
                , Phosphor.uniteSquare Light |> toHtml []
                , Phosphor.uniteSquare Regular |> toHtml []
                , Phosphor.uniteSquare Bold |> toHtml []
                , Phosphor.uniteSquare Fill |> toHtml []
                , Phosphor.uniteSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.upload model.weight |> toHtml []
                , Phosphor.upload Thin |> toHtml []
                , Phosphor.upload Light |> toHtml []
                , Phosphor.upload Regular |> toHtml []
                , Phosphor.upload Bold |> toHtml []
                , Phosphor.upload Fill |> toHtml []
                , Phosphor.upload Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.uploadSimple model.weight |> toHtml []
                , Phosphor.uploadSimple Thin |> toHtml []
                , Phosphor.uploadSimple Light |> toHtml []
                , Phosphor.uploadSimple Regular |> toHtml []
                , Phosphor.uploadSimple Bold |> toHtml []
                , Phosphor.uploadSimple Fill |> toHtml []
                , Phosphor.uploadSimple Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.usb model.weight |> toHtml []
                , Phosphor.usb Thin |> toHtml []
                , Phosphor.usb Light |> toHtml []
                , Phosphor.usb Regular |> toHtml []
                , Phosphor.usb Bold |> toHtml []
                , Phosphor.usb Fill |> toHtml []
                , Phosphor.usb Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.user model.weight |> toHtml []
                , Phosphor.user Thin |> toHtml []
                , Phosphor.user Light |> toHtml []
                , Phosphor.user Regular |> toHtml []
                , Phosphor.user Bold |> toHtml []
                , Phosphor.user Fill |> toHtml []
                , Phosphor.user Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userCircle model.weight |> toHtml []
                , Phosphor.userCircle Thin |> toHtml []
                , Phosphor.userCircle Light |> toHtml []
                , Phosphor.userCircle Regular |> toHtml []
                , Phosphor.userCircle Bold |> toHtml []
                , Phosphor.userCircle Fill |> toHtml []
                , Phosphor.userCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userCircleGear model.weight |> toHtml []
                , Phosphor.userCircleGear Thin |> toHtml []
                , Phosphor.userCircleGear Light |> toHtml []
                , Phosphor.userCircleGear Regular |> toHtml []
                , Phosphor.userCircleGear Bold |> toHtml []
                , Phosphor.userCircleGear Fill |> toHtml []
                , Phosphor.userCircleGear Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userCircleMinus model.weight |> toHtml []
                , Phosphor.userCircleMinus Thin |> toHtml []
                , Phosphor.userCircleMinus Light |> toHtml []
                , Phosphor.userCircleMinus Regular |> toHtml []
                , Phosphor.userCircleMinus Bold |> toHtml []
                , Phosphor.userCircleMinus Fill |> toHtml []
                , Phosphor.userCircleMinus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userCirclePlus model.weight |> toHtml []
                , Phosphor.userCirclePlus Thin |> toHtml []
                , Phosphor.userCirclePlus Light |> toHtml []
                , Phosphor.userCirclePlus Regular |> toHtml []
                , Phosphor.userCirclePlus Bold |> toHtml []
                , Phosphor.userCirclePlus Fill |> toHtml []
                , Phosphor.userCirclePlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userFocus model.weight |> toHtml []
                , Phosphor.userFocus Thin |> toHtml []
                , Phosphor.userFocus Light |> toHtml []
                , Phosphor.userFocus Regular |> toHtml []
                , Phosphor.userFocus Bold |> toHtml []
                , Phosphor.userFocus Fill |> toHtml []
                , Phosphor.userFocus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userGear model.weight |> toHtml []
                , Phosphor.userGear Thin |> toHtml []
                , Phosphor.userGear Light |> toHtml []
                , Phosphor.userGear Regular |> toHtml []
                , Phosphor.userGear Bold |> toHtml []
                , Phosphor.userGear Fill |> toHtml []
                , Phosphor.userGear Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userList model.weight |> toHtml []
                , Phosphor.userList Thin |> toHtml []
                , Phosphor.userList Light |> toHtml []
                , Phosphor.userList Regular |> toHtml []
                , Phosphor.userList Bold |> toHtml []
                , Phosphor.userList Fill |> toHtml []
                , Phosphor.userList Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userMinus model.weight |> toHtml []
                , Phosphor.userMinus Thin |> toHtml []
                , Phosphor.userMinus Light |> toHtml []
                , Phosphor.userMinus Regular |> toHtml []
                , Phosphor.userMinus Bold |> toHtml []
                , Phosphor.userMinus Fill |> toHtml []
                , Phosphor.userMinus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userPlus model.weight |> toHtml []
                , Phosphor.userPlus Thin |> toHtml []
                , Phosphor.userPlus Light |> toHtml []
                , Phosphor.userPlus Regular |> toHtml []
                , Phosphor.userPlus Bold |> toHtml []
                , Phosphor.userPlus Fill |> toHtml []
                , Phosphor.userPlus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userRectangle model.weight |> toHtml []
                , Phosphor.userRectangle Thin |> toHtml []
                , Phosphor.userRectangle Light |> toHtml []
                , Phosphor.userRectangle Regular |> toHtml []
                , Phosphor.userRectangle Bold |> toHtml []
                , Phosphor.userRectangle Fill |> toHtml []
                , Phosphor.userRectangle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userSquare model.weight |> toHtml []
                , Phosphor.userSquare Thin |> toHtml []
                , Phosphor.userSquare Light |> toHtml []
                , Phosphor.userSquare Regular |> toHtml []
                , Phosphor.userSquare Bold |> toHtml []
                , Phosphor.userSquare Fill |> toHtml []
                , Phosphor.userSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.userSwitch model.weight |> toHtml []
                , Phosphor.userSwitch Thin |> toHtml []
                , Phosphor.userSwitch Light |> toHtml []
                , Phosphor.userSwitch Regular |> toHtml []
                , Phosphor.userSwitch Bold |> toHtml []
                , Phosphor.userSwitch Fill |> toHtml []
                , Phosphor.userSwitch Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.users model.weight |> toHtml []
                , Phosphor.users Thin |> toHtml []
                , Phosphor.users Light |> toHtml []
                , Phosphor.users Regular |> toHtml []
                , Phosphor.users Bold |> toHtml []
                , Phosphor.users Fill |> toHtml []
                , Phosphor.users Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.usersFour model.weight |> toHtml []
                , Phosphor.usersFour Thin |> toHtml []
                , Phosphor.usersFour Light |> toHtml []
                , Phosphor.usersFour Regular |> toHtml []
                , Phosphor.usersFour Bold |> toHtml []
                , Phosphor.usersFour Fill |> toHtml []
                , Phosphor.usersFour Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.usersThree model.weight |> toHtml []
                , Phosphor.usersThree Thin |> toHtml []
                , Phosphor.usersThree Light |> toHtml []
                , Phosphor.usersThree Regular |> toHtml []
                , Phosphor.usersThree Bold |> toHtml []
                , Phosphor.usersThree Fill |> toHtml []
                , Phosphor.usersThree Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.van model.weight |> toHtml []
                , Phosphor.van Thin |> toHtml []
                , Phosphor.van Light |> toHtml []
                , Phosphor.van Regular |> toHtml []
                , Phosphor.van Bold |> toHtml []
                , Phosphor.van Fill |> toHtml []
                , Phosphor.van Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.vault model.weight |> toHtml []
                , Phosphor.vault Thin |> toHtml []
                , Phosphor.vault Light |> toHtml []
                , Phosphor.vault Regular |> toHtml []
                , Phosphor.vault Bold |> toHtml []
                , Phosphor.vault Fill |> toHtml []
                , Phosphor.vault Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.vibrate model.weight |> toHtml []
                , Phosphor.vibrate Thin |> toHtml []
                , Phosphor.vibrate Light |> toHtml []
                , Phosphor.vibrate Regular |> toHtml []
                , Phosphor.vibrate Bold |> toHtml []
                , Phosphor.vibrate Fill |> toHtml []
                , Phosphor.vibrate Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.video model.weight |> toHtml []
                , Phosphor.video Thin |> toHtml []
                , Phosphor.video Light |> toHtml []
                , Phosphor.video Regular |> toHtml []
                , Phosphor.video Bold |> toHtml []
                , Phosphor.video Fill |> toHtml []
                , Phosphor.video Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.videoCamera model.weight |> toHtml []
                , Phosphor.videoCamera Thin |> toHtml []
                , Phosphor.videoCamera Light |> toHtml []
                , Phosphor.videoCamera Regular |> toHtml []
                , Phosphor.videoCamera Bold |> toHtml []
                , Phosphor.videoCamera Fill |> toHtml []
                , Phosphor.videoCamera Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.videoCameraSlash model.weight |> toHtml []
                , Phosphor.videoCameraSlash Thin |> toHtml []
                , Phosphor.videoCameraSlash Light |> toHtml []
                , Phosphor.videoCameraSlash Regular |> toHtml []
                , Phosphor.videoCameraSlash Bold |> toHtml []
                , Phosphor.videoCameraSlash Fill |> toHtml []
                , Phosphor.videoCameraSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.vignette model.weight |> toHtml []
                , Phosphor.vignette Thin |> toHtml []
                , Phosphor.vignette Light |> toHtml []
                , Phosphor.vignette Regular |> toHtml []
                , Phosphor.vignette Bold |> toHtml []
                , Phosphor.vignette Fill |> toHtml []
                , Phosphor.vignette Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.vinylRecord model.weight |> toHtml []
                , Phosphor.vinylRecord Thin |> toHtml []
                , Phosphor.vinylRecord Light |> toHtml []
                , Phosphor.vinylRecord Regular |> toHtml []
                , Phosphor.vinylRecord Bold |> toHtml []
                , Phosphor.vinylRecord Fill |> toHtml []
                , Phosphor.vinylRecord Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.virtualReality model.weight |> toHtml []
                , Phosphor.virtualReality Thin |> toHtml []
                , Phosphor.virtualReality Light |> toHtml []
                , Phosphor.virtualReality Regular |> toHtml []
                , Phosphor.virtualReality Bold |> toHtml []
                , Phosphor.virtualReality Fill |> toHtml []
                , Phosphor.virtualReality Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.virus model.weight |> toHtml []
                , Phosphor.virus Thin |> toHtml []
                , Phosphor.virus Light |> toHtml []
                , Phosphor.virus Regular |> toHtml []
                , Phosphor.virus Bold |> toHtml []
                , Phosphor.virus Fill |> toHtml []
                , Phosphor.virus Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.voicemail model.weight |> toHtml []
                , Phosphor.voicemail Thin |> toHtml []
                , Phosphor.voicemail Light |> toHtml []
                , Phosphor.voicemail Regular |> toHtml []
                , Phosphor.voicemail Bold |> toHtml []
                , Phosphor.voicemail Fill |> toHtml []
                , Phosphor.voicemail Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.volleyball model.weight |> toHtml []
                , Phosphor.volleyball Thin |> toHtml []
                , Phosphor.volleyball Light |> toHtml []
                , Phosphor.volleyball Regular |> toHtml []
                , Phosphor.volleyball Bold |> toHtml []
                , Phosphor.volleyball Fill |> toHtml []
                , Phosphor.volleyball Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wall model.weight |> toHtml []
                , Phosphor.wall Thin |> toHtml []
                , Phosphor.wall Light |> toHtml []
                , Phosphor.wall Regular |> toHtml []
                , Phosphor.wall Bold |> toHtml []
                , Phosphor.wall Fill |> toHtml []
                , Phosphor.wall Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wallet model.weight |> toHtml []
                , Phosphor.wallet Thin |> toHtml []
                , Phosphor.wallet Light |> toHtml []
                , Phosphor.wallet Regular |> toHtml []
                , Phosphor.wallet Bold |> toHtml []
                , Phosphor.wallet Fill |> toHtml []
                , Phosphor.wallet Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.warehouse model.weight |> toHtml []
                , Phosphor.warehouse Thin |> toHtml []
                , Phosphor.warehouse Light |> toHtml []
                , Phosphor.warehouse Regular |> toHtml []
                , Phosphor.warehouse Bold |> toHtml []
                , Phosphor.warehouse Fill |> toHtml []
                , Phosphor.warehouse Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.warning model.weight |> toHtml []
                , Phosphor.warning Thin |> toHtml []
                , Phosphor.warning Light |> toHtml []
                , Phosphor.warning Regular |> toHtml []
                , Phosphor.warning Bold |> toHtml []
                , Phosphor.warning Fill |> toHtml []
                , Phosphor.warning Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.warningCircle model.weight |> toHtml []
                , Phosphor.warningCircle Thin |> toHtml []
                , Phosphor.warningCircle Light |> toHtml []
                , Phosphor.warningCircle Regular |> toHtml []
                , Phosphor.warningCircle Bold |> toHtml []
                , Phosphor.warningCircle Fill |> toHtml []
                , Phosphor.warningCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.warningDiamond model.weight |> toHtml []
                , Phosphor.warningDiamond Thin |> toHtml []
                , Phosphor.warningDiamond Light |> toHtml []
                , Phosphor.warningDiamond Regular |> toHtml []
                , Phosphor.warningDiamond Bold |> toHtml []
                , Phosphor.warningDiamond Fill |> toHtml []
                , Phosphor.warningDiamond Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.warningOctagon model.weight |> toHtml []
                , Phosphor.warningOctagon Thin |> toHtml []
                , Phosphor.warningOctagon Light |> toHtml []
                , Phosphor.warningOctagon Regular |> toHtml []
                , Phosphor.warningOctagon Bold |> toHtml []
                , Phosphor.warningOctagon Fill |> toHtml []
                , Phosphor.warningOctagon Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.watch model.weight |> toHtml []
                , Phosphor.watch Thin |> toHtml []
                , Phosphor.watch Light |> toHtml []
                , Phosphor.watch Regular |> toHtml []
                , Phosphor.watch Bold |> toHtml []
                , Phosphor.watch Fill |> toHtml []
                , Phosphor.watch Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.waveSawtooth model.weight |> toHtml []
                , Phosphor.waveSawtooth Thin |> toHtml []
                , Phosphor.waveSawtooth Light |> toHtml []
                , Phosphor.waveSawtooth Regular |> toHtml []
                , Phosphor.waveSawtooth Bold |> toHtml []
                , Phosphor.waveSawtooth Fill |> toHtml []
                , Phosphor.waveSawtooth Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.waveSine model.weight |> toHtml []
                , Phosphor.waveSine Thin |> toHtml []
                , Phosphor.waveSine Light |> toHtml []
                , Phosphor.waveSine Regular |> toHtml []
                , Phosphor.waveSine Bold |> toHtml []
                , Phosphor.waveSine Fill |> toHtml []
                , Phosphor.waveSine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.waveSquare model.weight |> toHtml []
                , Phosphor.waveSquare Thin |> toHtml []
                , Phosphor.waveSquare Light |> toHtml []
                , Phosphor.waveSquare Regular |> toHtml []
                , Phosphor.waveSquare Bold |> toHtml []
                , Phosphor.waveSquare Fill |> toHtml []
                , Phosphor.waveSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.waveTriangle model.weight |> toHtml []
                , Phosphor.waveTriangle Thin |> toHtml []
                , Phosphor.waveTriangle Light |> toHtml []
                , Phosphor.waveTriangle Regular |> toHtml []
                , Phosphor.waveTriangle Bold |> toHtml []
                , Phosphor.waveTriangle Fill |> toHtml []
                , Phosphor.waveTriangle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.waveform model.weight |> toHtml []
                , Phosphor.waveform Thin |> toHtml []
                , Phosphor.waveform Light |> toHtml []
                , Phosphor.waveform Regular |> toHtml []
                , Phosphor.waveform Bold |> toHtml []
                , Phosphor.waveform Fill |> toHtml []
                , Phosphor.waveform Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.waves model.weight |> toHtml []
                , Phosphor.waves Thin |> toHtml []
                , Phosphor.waves Light |> toHtml []
                , Phosphor.waves Regular |> toHtml []
                , Phosphor.waves Bold |> toHtml []
                , Phosphor.waves Fill |> toHtml []
                , Phosphor.waves Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.webcam model.weight |> toHtml []
                , Phosphor.webcam Thin |> toHtml []
                , Phosphor.webcam Light |> toHtml []
                , Phosphor.webcam Regular |> toHtml []
                , Phosphor.webcam Bold |> toHtml []
                , Phosphor.webcam Fill |> toHtml []
                , Phosphor.webcam Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.webcamSlash model.weight |> toHtml []
                , Phosphor.webcamSlash Thin |> toHtml []
                , Phosphor.webcamSlash Light |> toHtml []
                , Phosphor.webcamSlash Regular |> toHtml []
                , Phosphor.webcamSlash Bold |> toHtml []
                , Phosphor.webcamSlash Fill |> toHtml []
                , Phosphor.webcamSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.webhooksLogo model.weight |> toHtml []
                , Phosphor.webhooksLogo Thin |> toHtml []
                , Phosphor.webhooksLogo Light |> toHtml []
                , Phosphor.webhooksLogo Regular |> toHtml []
                , Phosphor.webhooksLogo Bold |> toHtml []
                , Phosphor.webhooksLogo Fill |> toHtml []
                , Phosphor.webhooksLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wechatLogo model.weight |> toHtml []
                , Phosphor.wechatLogo Thin |> toHtml []
                , Phosphor.wechatLogo Light |> toHtml []
                , Phosphor.wechatLogo Regular |> toHtml []
                , Phosphor.wechatLogo Bold |> toHtml []
                , Phosphor.wechatLogo Fill |> toHtml []
                , Phosphor.wechatLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.whatsappLogo model.weight |> toHtml []
                , Phosphor.whatsappLogo Thin |> toHtml []
                , Phosphor.whatsappLogo Light |> toHtml []
                , Phosphor.whatsappLogo Regular |> toHtml []
                , Phosphor.whatsappLogo Bold |> toHtml []
                , Phosphor.whatsappLogo Fill |> toHtml []
                , Phosphor.whatsappLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wheelchair model.weight |> toHtml []
                , Phosphor.wheelchair Thin |> toHtml []
                , Phosphor.wheelchair Light |> toHtml []
                , Phosphor.wheelchair Regular |> toHtml []
                , Phosphor.wheelchair Bold |> toHtml []
                , Phosphor.wheelchair Fill |> toHtml []
                , Phosphor.wheelchair Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wheelchairMotion model.weight |> toHtml []
                , Phosphor.wheelchairMotion Thin |> toHtml []
                , Phosphor.wheelchairMotion Light |> toHtml []
                , Phosphor.wheelchairMotion Regular |> toHtml []
                , Phosphor.wheelchairMotion Bold |> toHtml []
                , Phosphor.wheelchairMotion Fill |> toHtml []
                , Phosphor.wheelchairMotion Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wifiHigh model.weight |> toHtml []
                , Phosphor.wifiHigh Thin |> toHtml []
                , Phosphor.wifiHigh Light |> toHtml []
                , Phosphor.wifiHigh Regular |> toHtml []
                , Phosphor.wifiHigh Bold |> toHtml []
                , Phosphor.wifiHigh Fill |> toHtml []
                , Phosphor.wifiHigh Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wifiLow model.weight |> toHtml []
                , Phosphor.wifiLow Thin |> toHtml []
                , Phosphor.wifiLow Light |> toHtml []
                , Phosphor.wifiLow Regular |> toHtml []
                , Phosphor.wifiLow Bold |> toHtml []
                , Phosphor.wifiLow Fill |> toHtml []
                , Phosphor.wifiLow Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wifiMedium model.weight |> toHtml []
                , Phosphor.wifiMedium Thin |> toHtml []
                , Phosphor.wifiMedium Light |> toHtml []
                , Phosphor.wifiMedium Regular |> toHtml []
                , Phosphor.wifiMedium Bold |> toHtml []
                , Phosphor.wifiMedium Fill |> toHtml []
                , Phosphor.wifiMedium Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wifiNone model.weight |> toHtml []
                , Phosphor.wifiNone Thin |> toHtml []
                , Phosphor.wifiNone Light |> toHtml []
                , Phosphor.wifiNone Regular |> toHtml []
                , Phosphor.wifiNone Bold |> toHtml []
                , Phosphor.wifiNone Fill |> toHtml []
                , Phosphor.wifiNone Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wifiSlash model.weight |> toHtml []
                , Phosphor.wifiSlash Thin |> toHtml []
                , Phosphor.wifiSlash Light |> toHtml []
                , Phosphor.wifiSlash Regular |> toHtml []
                , Phosphor.wifiSlash Bold |> toHtml []
                , Phosphor.wifiSlash Fill |> toHtml []
                , Phosphor.wifiSlash Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wifiX model.weight |> toHtml []
                , Phosphor.wifiX Thin |> toHtml []
                , Phosphor.wifiX Light |> toHtml []
                , Phosphor.wifiX Regular |> toHtml []
                , Phosphor.wifiX Bold |> toHtml []
                , Phosphor.wifiX Fill |> toHtml []
                , Phosphor.wifiX Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wind model.weight |> toHtml []
                , Phosphor.wind Thin |> toHtml []
                , Phosphor.wind Light |> toHtml []
                , Phosphor.wind Regular |> toHtml []
                , Phosphor.wind Bold |> toHtml []
                , Phosphor.wind Fill |> toHtml []
                , Phosphor.wind Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.windowsLogo model.weight |> toHtml []
                , Phosphor.windowsLogo Thin |> toHtml []
                , Phosphor.windowsLogo Light |> toHtml []
                , Phosphor.windowsLogo Regular |> toHtml []
                , Phosphor.windowsLogo Bold |> toHtml []
                , Phosphor.windowsLogo Fill |> toHtml []
                , Phosphor.windowsLogo Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wine model.weight |> toHtml []
                , Phosphor.wine Thin |> toHtml []
                , Phosphor.wine Light |> toHtml []
                , Phosphor.wine Regular |> toHtml []
                , Phosphor.wine Bold |> toHtml []
                , Phosphor.wine Fill |> toHtml []
                , Phosphor.wine Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.wrench model.weight |> toHtml []
                , Phosphor.wrench Thin |> toHtml []
                , Phosphor.wrench Light |> toHtml []
                , Phosphor.wrench Regular |> toHtml []
                , Phosphor.wrench Bold |> toHtml []
                , Phosphor.wrench Fill |> toHtml []
                , Phosphor.wrench Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.x model.weight |> toHtml []
                , Phosphor.x Thin |> toHtml []
                , Phosphor.x Light |> toHtml []
                , Phosphor.x Regular |> toHtml []
                , Phosphor.x Bold |> toHtml []
                , Phosphor.x Fill |> toHtml []
                , Phosphor.x Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.xCircle model.weight |> toHtml []
                , Phosphor.xCircle Thin |> toHtml []
                , Phosphor.xCircle Light |> toHtml []
                , Phosphor.xCircle Regular |> toHtml []
                , Phosphor.xCircle Bold |> toHtml []
                , Phosphor.xCircle Fill |> toHtml []
                , Phosphor.xCircle Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.xSquare model.weight |> toHtml []
                , Phosphor.xSquare Thin |> toHtml []
                , Phosphor.xSquare Light |> toHtml []
                , Phosphor.xSquare Regular |> toHtml []
                , Phosphor.xSquare Bold |> toHtml []
                , Phosphor.xSquare Fill |> toHtml []
                , Phosphor.xSquare Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.yinYang model.weight |> toHtml []
                , Phosphor.yinYang Thin |> toHtml []
                , Phosphor.yinYang Light |> toHtml []
                , Phosphor.yinYang Regular |> toHtml []
                , Phosphor.yinYang Bold |> toHtml []
                , Phosphor.yinYang Fill |> toHtml []
                , Phosphor.yinYang Duotone |> toHtml []
                ]
           , div [ class "row" ]
                [ Phosphor.youtubeLogo model.weight |> toHtml []
                , Phosphor.youtubeLogo Thin |> toHtml []
                , Phosphor.youtubeLogo Light |> toHtml []
                , Phosphor.youtubeLogo Regular |> toHtml []
                , Phosphor.youtubeLogo Bold |> toHtml []
                , Phosphor.youtubeLogo Fill |> toHtml []
                , Phosphor.youtubeLogo Duotone |> toHtml []
                ]
            ]
